/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sockios.h>
#include <linux/proc_fs.h>
#ifdef CONFIG_FILTER_UNM
#include <linux/filter.h>
#endif

#include <linux/version.h>
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS) && !defined(__GENKSYMS__)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/modversions.h>
#endif
#endif
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/net.h>
#include <linux/poll.h>
#include <linux/skbuff.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/ip.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/inetdevice.h>

#include <linux/jhash.h>
#include <net/route.h>

#include <net/tcp.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#include <net/inet_hashtables.h>
#endif

#include <linux/inetdevice.h>
#include <linux/if.h>
#include <linux/if_vlan.h>

#include <unm_inc.h>
#include <pegnet_inc.h>
#include "nic_cmn.h"
#include "hostInterface.h"
#include "unm_xport.h"
#include "flow_table.h"
#include "unm_host_pnic_types.h"

#include "driver_info.h"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(TOE_DRIVER_DESCRIPTION);
MODULE_LICENSE("GPL");

/*
 * Function declaration of static functions.
 */
static long unm_wait_for_connect(pegnet_sock_id_t *sid, struct sock *sk,
				 long timeo);
static void nx_sock_cleanup_task(void *v);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
static int check_non_offloaded_conn(struct sock *sk,struct sk_buff *skb,
				    struct iphdr *iph);
static struct open_request *check_tuple_existence(void *tp,
						  struct open_request ***prevp,
						  __u16 rport, __u32 raddr,
						  __u32 laddr);
static u32 tcp_v4_synq_hash_tbl(u32 raddr, u16 rport, u32 rnd);
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9) && LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13)
extern struct sock *tcp_v4_lookup(u32 saddr, u16 sport, u32 daddr, u16 dport,
                                   int dif);
#endif

static pegnet_sock_id_t *nx_search_synq(pegnet_sock_id_t *psid,
					uint32_t local_ip, uint16_t local_port,
					uint32_t remote_ip,
					uint16_t remote_port);
void nx_remove_synq(listener_group_t *grp, pegnet_sock_id_t *sid);					

/*
 * Global variables.
 */
nx_tnic_dbg_cntrs_t	nx_tnic_dbg_cntrs;

/* Whenever possible compress qmsg IOVs before passing it to the card */
extern int			global_use_compressed_iov;
/* Size beyond which use zero copy transmits, if possible */
extern uint32_t                 tx_zero_copy_size_thresh;
/* Connection threshold beyond which use zero copy transmits, if possible */
extern uint32_t                 tx_zero_copy_active_thresh;
/* Number of outstanding tx bytes to trigger zero copy */
extern uint32_t                 tx_bytes_out_thresh;


static unm_card_response_handler_t card_resp_handlers[UNM_P2H_OPCODE_LAST];


/*
 * Pegnet proc handling code. Located under /proc/pegnet
 */
static struct proc_dir_entry	*pegnet_dir = NULL;
static struct proc_dir_entry	*zcopy_dir = NULL;
static struct proc_dir_entry	*compressed_iov_file = NULL;
static struct proc_dir_entry	*tx_size_thresh_file = NULL;
static struct proc_dir_entry	*tx_conn_thresh_file = NULL;
static struct proc_dir_entry	*tx_bytes_out_file = NULL;
extern atomic_t g_devid;
#define UNM_DMA_BUFFER_SIZE	(9 * 1024)

typedef enum { slosh_nocopy, slosh_docopy } unm_slosh;

/* XXX This is required to keep insmod quiet */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
long long __udivdi3(long long  op1, long long op2)
{
	return 0;
}
#endif

/*
 *
 */
int unm_create(struct net_device *netdev, struct socket *sock, int protocol,
	       int type, int zcopy)
{
        pegnet_sock_id_t        *sid = NULL;
        int                     rv;


        if (sock->sk == NULL) {
                return (-EINVAL);
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if (!try_module_get(THIS_MODULE)) {
                return (-EINVAL);
        }
#else
        MOD_INC_USE_COUNT;
#endif

        /* Talk to card and get back sock_id_t */
        sid = host_create(netdev, PF_INET, protocol, &rv);
        if (sid != NULL) {
                sid->parent = sock;
                sid->type = type;

		if(type == CONNECTION_SOCK) {
                        UNM_SOCK_SET_PEGNET_SOCK_ID(sock, sid);
		} else {
			if(sock->sk->SK_PROTINFO == NULL) {
        			listener_group_t	*grp = NULL;

                		grp = (listener_group_t *)
					kmalloc(sizeof(listener_group_t), GFP_KERNEL);
                		if (!grp) {
					host_release(sid, 0);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                			module_put(THIS_MODULE);
#else
                			MOD_DEC_USE_COUNT;
#endif                        		
					return -ENOMEM;
                		}
                		memset(grp, 0, sizeof(listener_group_t));
                		spin_lock_init(&grp->lock);
               	 		INIT_LIST_HEAD(&(grp->listener_list));
                		grp->type = LISTENER_GROUP;
                		UNM_SOCK_SET_PEGNET_SOCK_ID(sock, grp);
			}
                        list_add_tail(&sid->listener_list,
				      &((NX_GET_LISTENER_GROUP_FROM_SOCK(sock))->listener_list));
		}

                sid->zero_copy_enabled = zcopy;
                rv = 0;
        } else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                module_put(THIS_MODULE);
#else
                MOD_DEC_USE_COUNT;
#endif
        }

        return (rv);
}

/*
 *
 */
void nx_set_route_reference(pegnet_sock_id_t *sid, uint32_t saddr,
			    uint16_t sport, uint32_t daddr, uint16_t dport)
{
	sid->lport = sport;
	sid->rport = dport;
	sid->lip_addr.v4 = saddr;
	sid->rip_addr.v4 = daddr;

	sid->route_added = 1;

	nx_print1("saddr = 0x%x, sport = %u, daddr = 0x%x, dport = %u\n",
		  saddr, sport, daddr, dport);
}

/*
 *
 */
static int nx_init_connection(pegnet_sock_id_t *sid)
{
        if (sid->recv_dma.initialized) {
                return (0);
        }

        /* Check if this is a high performance flow, and act accrodingly. */
        if (atomic_read(&sid->adapter->high_perf_cnt) > 0) {
                atomic_dec(&sid->adapter->high_perf_cnt);
                sid->recv_dma.num_pcb_rxbuffers = NUM_PCB_RXBUFFERS_HP;
                sid->send_dma.num_vbuf_pages = UNM_TX_VBUF_MAX_PAGES_HP;
                sid->high_perf_flag = 1;
        } else {
                sid->recv_dma.num_pcb_rxbuffers = NUM_PCB_RXBUFFERS_LP;
                sid->send_dma.num_vbuf_pages = UNM_TX_VBUF_MAX_PAGES_LP;
                sid->high_perf_flag = 0;
        }

	return (nx_init_rxbuffers_pcb(sid));
}

/*
 *
 */
void cleanup_dma_buffers(pegnet_sock_id_t * sid)
{
        int     i;

        for (i = 0; i < pegnet_num_bufs; i++) {
                if (sid->dma[i].buffer != NULL) {
#if 0
                        pci_free_consistent(unm_device_ptr,
                                            UNM_DMA_BUFFER_SIZE,
                                            sid->dma[i].buffer,
                                            sid->dma[i].phys);
#else
                        kfree(sid->dma[i].buffer);
//// Start debug.
//            printk( "cleanup_dma_buffer: freeing buffer=%p\n", sid->dma[i].buffer ) ;
//// End debug.
#endif
                        sid->dma[i].buffer = NULL;
                }
        }

	cleanup_rxbuffers_global(sid);
        cleanup_rxbuffers_pcb(sid);

        spin_lock_bh(&sid->lock);
        cleanup_tx_vbuf_queues(sid);
        spin_unlock_bh(&sid->lock);

	if (sid->high_perf_flag) {
		atomic_inc(&sid->adapter->high_perf_cnt);
	}
}

static int nx_release_listener_sid(pegnet_sock_id_t *sid)
{
	int	rv = 0;

	spin_lock_bh(&sid->lock);
	list_del_init(&sid->listener_list);
	spin_unlock_bh(&sid->lock);

	rv = host_release(sid, 0);	
        if (rv != 0) {
             	printk("%s: host_release failed %d" 
			"sid=%p\n", __FUNCTION__, rv,sid);
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        module_put(THIS_MODULE);
#else
        MOD_DEC_USE_COUNT;
#endif
	return rv;
}
/*
 *
 */
int unm_release(struct socket *sock, int delay_sk_release)
{
        int                     rv = 0;
        pegnet_sock_id_t        *sid= NULL;
        listener_group_t        *grp;
        struct list_head        *node,*_node;

        if (sock->sk->SK_STATE == TCP_LISTEN) {
                grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		grp->released = 1;
		if (grp->accept_passive.lopt) {
			release_accept_queues(grp);
                	kfree(grp->accept_passive.lopt);
                	grp->accept_passive.lopt = NULL;
		}
                list_for_each_safe(node, _node,&grp->listener_list) {
                        sid = list_entry(node, pegnet_sock_id_t,
					 listener_list);
			rv = nx_release_listener_sid(sid);
                }
//		nx_sk_cleanup(sock->sk);
		kfree(grp);
        } else {
                sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
                rv = host_release(sid, delay_sk_release);
        	if (rv != 0) {
                	nx_print("host_release failed %d\n", rv);
        	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        	module_put(THIS_MODULE);
#else
        	MOD_DEC_USE_COUNT;
#endif
        }
	if (sock->sk != NULL) {
		UNM_SOCK_SET_PEGNET_SOCK_ID(sock, NULL);
	}

        return (rv);
}

/*
 *
 */
static long unm_wait_for_connect(pegnet_sock_id_t *sid, struct sock *sk,
                                 long timeo)
{
	long ret = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	DEFINE_WAIT(wait);
#else
        DECLARE_WAITQUEUE(wait, current);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        prepare_to_wait(sk->SK_SLEEP, &wait, TASK_INTERRUPTIBLE);
#else
        add_wait_queue(sk->SK_SLEEP, &wait);
        current->state = TASK_INTERRUPTIBLE;
#endif

        /*
         * If the socket is non blocking then the call still waits for the
         * reply from the card.
         */
        while (SS_CONNECTING == sid->parent->state) {
		prepare_to_wait_exclusive(sk->SK_SLEEP, &wait,
				TASK_INTERRUPTIBLE);
		release_sock(sk);
		timeo = schedule_timeout(timeo);
		lock_sock(sk);

		if (signal_pending(current)) {
			ret = sock_intr_errno(timeo);
			break;
		}
		else if (!timeo)
		{
			ret = -ETIMEDOUT;
			break;
		}
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        finish_wait(sk->SK_SLEEP, &wait);
#else
        current->state = TASK_RUNNING;
        remove_wait_queue(sk->SK_SLEEP, &wait);
#endif
        return (ret);
}

/*
 *
 */
int unm_connect(struct socket *sock, struct sockaddr *saddr, int len,
                int flags, int do_bind)
{
        struct sock             *sk = sock->sk;
        int                     err;
        long                    timeo = 0;
        pegnet_sock_id_t        *sid = NULL;

	if(sk->SK_STATE == TCP_LISTEN) {
		return (-EISCONN);
	}

        lock_sock(sk);

	sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);

        if (saddr->sa_family == AF_UNSPEC) {
                err = unm_disconnect(sock, flags, sock_sndtimeo(sk, 0));
                sock->state = err ? SS_DISCONNECTING : SS_UNCONNECTED;
                goto out;
        }

	switch (sock->state) {
	default:
		err = -EINVAL;
		goto return_err;
	case SS_CONNECTED:
		err = -EISCONN;
		/* First time ? */
		if(sid->conn_compl == 0) {
			sid->conn_compl = 1;
			err = 0;
		}
		goto return_err;
	case SS_CONNECTING:
		err = -EALREADY;
		/* Fall out of switch with err, set for this state */
		goto return_err;
	case SS_UNCONNECTED:
		err = -EISCONN; /*Connect after listen */
		if (sk->SK_STATE != TCP_CLOSE)
			goto return_err; 
		/* Always connect as async and wait for card's initial response */
		err = host_connect(sid, saddr, len, flags | O_NONBLOCK,
				   do_bind);

		if (-EALREADY == err)
			goto return_err; /* Take no action. Specifically
					    don't clean up the socket. */

		if (-EINPROGRESS == err) /* Ignore */
			break;
		else if (0 > err)
			goto sock_error;
	}

	/* In host_connect, we have already waited for the rcv_qid, so we can
	   init rx_buffers now. We will reach this point if host_connect
	   returns 0 or -EINPROGRESS. */

	/*
	 * init_rxbuffers_pcb() sets a flag once it is initialized and
	 * does not initialize again. It is now safe to call multiple
	 * times.
	 */
	nx_init_connection(sid);

	/* That's all for async, in case of sync wait for connection
	   completetion */

	if (flags & O_NONBLOCK)
		goto out;

	/* Only sync gets this far, wait for connection completion,
	   indefinitely. */

	timeo = sock_sndtimeo(sk, 0);

	err = unm_wait_for_connect(sid, sk, timeo);

	if (err < 0)
		goto clobber_error; /* If the wait did not complete,
				       it should be reported as an
				       internal error */

	/* Here only if our wait had completed successfully, otherwise we would
	   have been elsewhere because of unm_wait_for_connect's error code */

	switch (sid->parent->state) {
	case SS_CONNECTED:
		sid->conn_compl = 1;
		err = 0;
		break;
	default:
		err = -1 * sk->SK_ERR;
		goto sock_error;
	}

out:

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        /*
         * TODO: Do we really need this code or do we need it in all versions.
         * Need to check linux code to confirm.
         */
        if (err == 0 || err == -EINPROGRESS) {
                struct sockaddr_in      *sin;

                sin = (struct sockaddr_in *)saddr;
                sock->sk->daddr = sin->sin_addr.s_addr;
                sock->sk->dport = sin->sin_port;
        }
#endif
        release_sock(sk);
        return err;

  clobber_error:
	if (-EINTR != err && -ERESTARTSYS != err)
		err = -ECONNABORTED;
  sock_error:
        sock->state = SS_UNCONNECTED;
        if (unm_disconnect(sock, flags, timeo)) {
                sock->state = SS_DISCONNECTING;
        }
        goto out;

  return_err:
	release_sock(sk);
	return err;
}

/*
 *
 */
int unm_disconnect(struct socket *sock, int flags, long timeo)
{
        pegnet_sock_id_t        *sid;
        int                     rv;

        sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);

        spin_lock_bh(&sid->lock);
        rv = host_conn_disconnect(sid, 1, 0, 0);
        spin_unlock_bh(&sid->lock);

        return (rv);
}

/*
 *
 */
int unm_host_is_accept_queue_empty(struct socket *sock)
{
        //pegnet_sock_id_t        *sid;
        listener_group_t        *grp;
        int                     ret;

        grp = (listener_group_t *) UNM_SOCK_GET_PEGNET_SOCK_ID(sock);

        spin_lock_bh(&grp->lock);
        ret = list_empty(&(grp->accept_passive.accept_queue));
        spin_unlock_bh(&grp->lock);

        return (ret);
}

/*
 *
 */
int unm_host_get_next_accept_stack(struct socket *sock)
{
        listener_group_t        *grp;
        int                     ret;

        grp = (listener_group_t *) NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
        spin_lock_bh(&grp->lock);
        grp->accept_passive.nxt_stack++;
        grp->accept_passive.nxt_stack &= UNM_ACCEPT_STACK_MASK;
        ret = grp->accept_passive.nxt_stack;
        spin_unlock_bh(&grp->lock);

        return (ret);
}

/*
 *
 */
int unm_accept(struct socket *sock, struct socket *newsock, int flags)
{
        pegnet_sock_id_t        *newsid;
        pegnet_sock_id_t        *sid = NULL;
	listener_group_t        *grp;
	struct list_head 	*node;
        int                     rv,found = 0;


	grp =  NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
	if(grp->shutdown & RCV_SHUTDOWN){
                return (-EINVAL);
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        if (!try_module_get(THIS_MODULE)) {
                return (-EINVAL);
        }
#else
        MOD_INC_USE_COUNT;
#endif

        newsid = unm_host_async_accept(grp, flags, &rv);
        /* newsid = host_accept(sid, flags, &err); */

        if (NULL == newsid) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                module_put(THIS_MODULE);
#else
                MOD_DEC_USE_COUNT;
#endif
                return (rv);
        }

        newsid->parent = newsock;
        UNM_SOCK_SET_PEGNET_SOCK_ID(newsock, newsid);

        nx_init_connection(newsid);
        list_for_each(node, &grp->listener_list) {
                sid = list_entry(node, pegnet_sock_id_t, listener_list);
		if(sid->adapter == newsid->adapter) {
			found = 1;
			break;
		}
        }
	if (found == 0) {
		nx_print("No listener sid is found\n");
		return -1;
	}
	/* Replenish the accept just used up */
	nx_host_post_accept(sid, 1);

        return (0);
}   /* unm_accept */

/*
 *
 */
int unm_getname(struct socket *sock, struct sockaddr *uaddr, int *addr_len,
                int peer)
{
	int ret =0;
	pegnet_sock_id_t *sid;

	if(sock->sk->SK_STATE == TCP_LISTEN){
                listener_group_t *grp =
                        NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		sid = list_entry(grp->listener_list.next,pegnet_sock_id_t, listener_list);
	}
	else {
		sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
	}
        ret = host_getname(sid, uaddr,addr_len, peer);
	return ret;
}   /*  unm_getname */

/*
 *
 */
unsigned int unm_poll(struct socket *sock)
{
        unsigned int            rv = 0;

        if (sock->sk->SK_STATE == TCP_LISTEN) {
                listener_group_t *grp =
                        NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		spin_lock_bh(&grp->lock);
                if (!list_empty(&grp->accept_passive.accept_queue)) {
                        rv |= POLLIN | POLLRDNORM;
                }
		spin_unlock_bh(&grp->lock);
	} else {
        	pegnet_sock_id_t        *sid;
       	 	sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
        	spin_lock_bh(&sid->lock);
		rv = sid->poll_flags & ~POLLHUP;

		if (sid->shutdown == SHUTDOWN_MASK ||
		    (sid->recv_dma.tmp_rcv_shutdown &&
		     (sid->shutdown & SEND_SHUTDOWN))) {
			rv |= POLLHUP;
		}

		if (0 == list_empty(&(sid->recv_dma.list))) {
			rv |= (POLLIN | POLLRDNORM);
		}

		if (sid->send_dma.last_comp_valid) {
			process_last_tx_completion(sid);
		}

		if (is_tx_vbuf_allocatable(NX_ZERO_AND_REG_TX_VBUF, sid)) {
			rv |= (POLLOUT|POLLWRNORM) ;
		} else {
			sid->send_dma.need_buffer_wakeup = 1;
		}
        	spin_unlock_bh(&sid->lock);
	}


        return (rv);
}   /* unm_poll */


/*
 *
 */
int unm_listen(struct socket *sock, int backlog, int do_bind)
{
	int rv = 0;
	struct list_head *node,*_node;
	pegnet_sock_id_t *sid;
	listener_group_t *grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);

	list_for_each_safe(node, _node, &grp->listener_list) {
		sid = list_entry(node, pegnet_sock_id_t, listener_list);
		if(0 != (rv = host_listen(sid, backlog, do_bind))) {
			nx_print("Listen Error for Listener grp = %p sid:%p "
				"tnic_adapter=%p dev_port=%d err=%d\n",
				grp,sid,sid->adapter,sid->portnum,rv);
			nx_release_listener_sid(sid);	
		}
	}
	rv = 0;
	if (list_empty(&(grp->listener_list))) {
		rv = -1;
	}
	return rv;
}

/*
 *
 */
int unm_shutdown(struct socket *sock, int how)
{
        int                     rv = 0;
	int			rc = 0;
        pegnet_sock_id_t        *sid = NULL;
        struct list_head 	*node, *_node;

        if (how < 0 || how > 2) {
                return (-EINVAL);
        }

	if (sock->sk->SK_STATE == TCP_LISTEN) {
        	listener_group_t *grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
		grp->shutdown |= how + 1;
        	list_for_each_safe(node, _node,&grp->listener_list) {
                	sid = list_entry(node, pegnet_sock_id_t,
					 listener_list);
                	if(0 != (rv = host_shutdown(sid, how))) {
                		nx_print("Shutdown Error for Listener grp = %p"
					 " sid:%p tnic_adapter = %p "
					 "dev_port = %d err=%d\n",
					 grp,sid,sid->adapter,sid->portnum,rv);
//				nx_release_sid(sid);	 
			}
		}
 		if (list_empty(&(grp->listener_list))) {
			return rv;
 		}
		return 0;
	} else {
        	sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
        	/*
         	* Adding 1 to how will make it such that bit0 set will mean RX,
		* bit1 set will mean TX.
         	*
         	* how:     RX:    0 -> 0+1 -> 0x1
         	*          TX:    1 -> 1+1 -> 0x2
         	*          Both:  2 -> 2+1 -> 0x3
         	*
         	* TBD: Need to see what is to be done for RX.
         	*
         	* Both the semaphore as well as the lock have to be taken and
		* in the correct order. This is done to lock out user sendmsg
		* call and the bottom half processing.
         	*/
	        down(&sid->send_dma.lock);
        	spin_lock_bh(&sid->lock);
	        sid->shutdown |= how + 1;	/* It has to be '|' because a
						   previous shutdown could have
						   turned on some bits. */
	        if (sid->shutdown & SEND_SHUTDOWN) {
        	        wake_up_all(&sid->send_dma.wq);
        	}

		if (sid->shutdown & SEND_SHUTDOWN) {

			if (!list_empty(&sid->send_dma.zero_wait_list)) {
				rc = nx_flush_sendpage(sid);
				UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SENDPAGE_FLUSHED2);
				if (rc) {
					nx_print("Need to fix sid = %p\n", sid);
				}
			}
		}

     		if (!list_empty(&sid->send_dma.vec_wait_list) ||
		    !list_empty(&sid->send_dma.failed_list)) {
                	/*
			 * Turn off the TX shutdown condition. It will be sent
			 * when the last TX data is sent to the card.
			 */
                	how = (how + 1) & ~SEND_SHUTDOWN;
                	how -= 1;
        	}

        	spin_unlock_bh(&sid->lock);
        	up(&sid->send_dma.lock);

        	/*
		 * If the original how was 0 (tx shutdown) and we had a failure
		 * or flow control condition 'how' would have gotten
		 * decremented and will be negative in which case don't send
		 * shutdown message.
		 */
       	 	if (how >= 0) {
                	rv = host_shutdown(sid, how);
        	}
        	return (rv);
	}
}

#if (UNM_X_HARDWARE==UNM_X_ASIC)
#include "af_unm_asic.c"
#else
#error "Unexpected hardware specification"
#endif

/* Finds matching sid for given TOE adapter 
 *
 */
pegnet_sock_id_t *get_sid_by_grp(nx_tnic_adapter_t *tnic_adapter, listener_group_t *listener_grp)
{
	struct list_head *node;
	pegnet_sock_id_t *sid;
	list_for_each(node, &listener_grp->listener_list) {
		sid = list_entry(node, pegnet_sock_id_t, listener_list);
		if (sid->adapter == tnic_adapter)
			return sid;
	}
	return NULL;
}

/*
 *
 */
static inline pegnet_sock_id_t *get_sid_by_idx(flow_table_t *tcp_tbl,uint16_t idx)
{
        return ((pegnet_sock_id_t *)host_flowtbl_get_flow_by_idx(tcp_tbl,idx));
}

/*
 *
 */
static void invalid_card_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
		pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        printk("%s: Invalid card response\n", __FUNCTION__);
}

/*
 *
 */
static void proc_create_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
		pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
	host_peg_msg_hdr_t      *hdr = (host_peg_msg_hdr_t *)&(msg->body);
        create_response_t       *resp = (create_response_t *)&msg->body;
        create_response_t       *sptr;
	uint32_t		tmp_ctx;
	int			id;
        int                     replies;

	/* Retrieve the sid */
	tmp_ctx = host_get_uint32_from_peg(hdr->ctx);
	sid = host_map_tmpctx_2_sid(tnic_adapter->p_ftable,tmp_ctx);
	if (!sid) {
		printk("%s: Could not map tmp_ctx %x to sid\n",
				__FUNCTION__, tmp_ctx);
		return;
	}

	/* Retrieve the transaction to wake up */
	id = hdr->op_id;

        spin_lock(&sid->lock);
        for (trans = sid->op; trans != NULL; trans = trans->link) {
                if (trans->op_id == id) {
                        break;
                }
        }
        spin_unlock(&sid->lock);

	if (!trans) {
		printk("%s: Could not retrieve transaction for sid = %p, "
				" tmp_ctx = %x\n", __FUNCTION__, sid, tmp_ctx);
		return;
	}

	atomic_inc(&trans->replied);
        replies = atomic_read(&trans->replied);     
        if (replies != 1) {
                printk("%s: WARNING replies:%d != 1\n", __FUNCTION__, replies);
		return;
        }

        sptr = (create_response_t *)&(trans->msg.body);
        sptr->ret_code = resp->ret_code;
	if (!sptr->ret_code) {
		sptr->qid = resp->qid;
		sptr->new_ctx = resp->new_ctx;
		sptr->connId = resp->connId;

		/* update the connID table */
		host_update_connID_table(tnic_adapter->p_ftable,sid, resp->connId);
	}

        wake_up(&trans->wq);

	return;
}

/*
 *
 */
static void proc_release_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        release_response_t      *resp = (release_response_t *)&msg->body;
        release_response_t      *sptr;

        sptr = (release_response_t *)&(trans->msg.body);
        sptr->ret_code = host_get_uint32_from_peg(resp->ret_code);
        if(sid->accept_active.parent_sid) {
                list_del(&sid->accept_active.list);
        }	
        wake_up(&trans->wq);
}


/*
 *
 */
static void proc_poll_status(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        poll_status_t   *status = (poll_status_t *)&msg->body;

        spin_lock_bh(&sid->lock);
        sid->poll_flags = host_get_uint32_from_peg(status->flags);
        spin_unlock_bh(&sid->lock);

        if (sid->parent && sid->parent->sk &&
            sid->parent->sk->SK_SLEEP &&
            waitqueue_active(sid->parent->sk->SK_SLEEP)) {
                wake_up_interruptible(sid->parent->sk->SK_SLEEP);
        }
}

/*
 *
 */
static void unm_proc_conn_signal(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        conn_signal_t   *sig_msg = (conn_signal_t *)&msg->body;
        struct sock     *sk;

        spin_lock_bh(&sid->lock);

	if (sid && sid->parent && sid->parent->sk) {
                sk = sid->parent->sk;
	    
                if (sk && sk->SK_SOCKET && sk->SK_SOCKET->fasync_list) {
                        sock_wake_async(sk->SK_SOCKET, sig_msg->how,
                                        sig_msg->band);
                }
        }

        spin_unlock_bh(&sid->lock);
}

void nx_proc_sock_cleanup(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg, 
			  pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
	struct nx_sk_cleanup_queue	*sk_cleanup_queue;

	sk_cleanup_queue = &sid->adapter->sk_cleanup_queue;

        spin_lock_bh(&sid->lock);
        host_flowtbl_put_reference(sid->adapter->p_ftable, sid->connID);
	host_reset_connID_table(sid->adapter->p_ftable, sid, sid->connID);
        sid->cleanup_flag = 1;
        if (sid->release_seen && sid->sk_cleanup) {
                sid->cleanup_flag = 2;
                INIT_LIST_HEAD((&sid->tw_list));
                spin_lock_bh(&sk_cleanup_queue->lock);
                list_add_tail(&sid->tw_list, &sk_cleanup_queue->list);
                spin_unlock_bh(&sk_cleanup_queue->lock);
		NX_KC_SCHEDULE_WORK(&sid->adapter->sk_cleanup_task);

        } else if (sid->release_seen && !sid->sk_cleanup) {
        	spin_unlock_bh(&sid->lock);
		free_pegnet_sock_id(sid);
		goto done;
	} else {
//                printk(KERN_ERR "%s: got invalid sid:%p\n", __func__, sid);
	}
        spin_unlock_bh(&sid->lock);
  done:
	return;
}

void nx_sk_cleanup(struct sock *sk)
{
	/* Part of host_release() */
	sk->SK_PROTINFO = NULL;

	/* Part of unm_inet_release() */
	if (sk->SK_DESTRUCT) {
		// We have waited long enough !!
		sk->SK_STATE = TCP_CLOSE;
		//ip_mc_drop_socket(sk);
		sk->SK_PROT->close(sk, 0);
	} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
		inet_put_port(&tcp_hashinfo, sk);
#else
		tcp_put_port(sk);
#endif
		sk_free(sk);
		UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_SK);
	}
}

void nx_init_sk_cleanup_task(nx_tnic_adapter_t *tnic_adapter)
{
        INIT_LIST_HEAD(&tnic_adapter->sk_cleanup_queue.list);
        spin_lock_init(&tnic_adapter->sk_cleanup_queue.lock);

        NX_KC_INIT_WORK(&tnic_adapter->sk_cleanup_task, nx_sock_cleanup_task,
			tnic_adapter);
}

void nx_sock_cleanup_task_internal(pegnet_sock_id_t *sid)
{
//        spin_lock(&sid->lock);
	if(sid->sk_cleanup) {
		nx_sk_cleanup(sid->sk_cleanup);
	}	
//        spin_unlock(&sid->lock);
        free_pegnet_sock_id(sid);
}

/*
 *
 */
static void nx_sock_cleanup_task(void *v)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	struct work_struct	*_work = (struct work_struct *)v;
	nx_tnic_adapter_t	*tnic_adapter = container_of(_work,
						     struct nx_tnic_adapter_s,
						     sk_cleanup_task);
#else
	nx_tnic_adapter_t	*tnic_adapter = (nx_tnic_adapter_t *)v;
#endif
        struct list_head	free_list;
        struct list_head	*node, *n;
        pegnet_sock_id_t	*sid = NULL;

        INIT_LIST_HEAD(&free_list);

        spin_lock_bh(&tnic_adapter->sk_cleanup_queue.lock);
        list_for_each_safe(node, n, &tnic_adapter->sk_cleanup_queue.list) {
                sid = list_entry(node, pegnet_sock_id_t, tw_list);
                list_del(&sid->tw_list);
                list_add_tail(node, &free_list);
        }
        spin_unlock_bh(&tnic_adapter->sk_cleanup_queue.lock);

        list_for_each_safe(node, n, (&free_list)) {
                sid = list_entry(node, pegnet_sock_id_t, tw_list);
                if(!sid) {
                        printk(KERN_ERR "%s: Got empty node\n", __func__);
                        goto done;
                }
                if (!sid->sk_cleanup) {
                        nx_print("Got invalid sid:%p\n", sid);
                        goto done;
                }

		UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_SOCK_CLEANUP_CONNS);
                list_del(&sid->tw_list);
		nx_sock_cleanup_task_internal(sid);
        }

  done:
        return;
}
                                

/*
 *
 */
static void proc_tcp_event(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg, 
			   pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
        unm_tcp_event_msg_t     *event;

        event = (unm_tcp_event_msg_t *)&msg->body;

        switch (event->param.type) {
        case UNM_TCP_EVENT_FIN_ACKED:
                spin_lock_bh(&sid->lock);
                cleanup_tx_vbuf_queues(sid);
                spin_unlock_bh(&sid->lock);
                break;

        case UNM_TCP_EVENT_RST_RECEIVED:
		if (sid->accept_active.parent_sid) {
			/* We are still a child yet to be accepted */

			pegnet_sock_id_t	*parent_sid;
                        listener_group_t	*grp;

			parent_sid = sid->accept_active.parent_sid;
			grp = NX_GET_LISTENER_GROUP_FROM_SID(parent_sid);

			/* Hold the parent's lock */
			spin_lock_bh(&parent_sid->lock);

			if (grp->released) {
				spin_unlock_bh(&parent_sid->lock);
				return; /* we will be cleaned
					   up with the parent */
			}

			/* Silently discard this child */
			list_del(&sid->accept_active.list);

			spin_lock_bh(&grp->lock);
			grp->accept_passive.accepted_len--;
			spin_unlock_bh(&grp->lock);

			/* Post an accept ticket back */
			nx_host_post_accept(parent_sid, 1);
			spin_unlock_bh(&parent_sid->lock);

			host_release_unaccepted_child(sid);

			cleanup_rxbuffers_global(sid);

			/*
			 * There is no way a sock cleanup can come before this
			 * message. If it comes in before this or is processed
			 * before this we have a serious problem.
			 *
			 * No need to free the sid here. The sid will be freed
			 * when the sock cleanup comes down.
			 */
			sid->release_seen = 1;

			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_RST_ON_ACCEPT_QUEUE);

			return; /* Application does not yet know, and
				  we should not have struct socket references
				  at this point. */
		}

        case UNM_TCP_EVENT_ABORT_ON_SYN:
        case UNM_TCP_EVENT_ABORT_ON_DATA:
        case UNM_TCP_EVENT_ABORT_ON_CLOSE:
        case UNM_TCP_EVENT_ABORT_ON_MEMORY:
        case UNM_TCP_EVENT_ABORT_ON_LINGER:
        case UNM_TCP_EVENT_ABORT_DISCONNECT:
        case UNM_TCP_EVENT_KEEPALIVE_EXPIRED:
        case UNM_TCP_EVENT_RETRANSMIT_TIMEOUT:
        case UNM_TCP_EVENT_WND_PROBE_TIMEOUT:
                spin_lock_bh(&sid->lock);

                sid->shutdown = SHUTDOWN_MASK;
		if (sid->parent && sid->parent->sk) {
		    sid->parent->sk->SK_ERR = (int)event->param.reason;
                }
		del_timer_sync(&sid->send_dma.sendpage_timer);
                cleanup_tx_vbuf_queues(sid);
                wake_up_all(&sid->send_dma.wq);
                wake_up_all(&sid->recv_dma.wq);

                spin_unlock_bh(&sid->lock);
                break;
        }
}

/*
 *
 */
static void debug_print_tcp_event_msg(unm_msg_t *msg, pegnet_sock_id_t *sid,
				      pegnet_inprogress_t *trans)
{
        unm_tcp_event_msg_t     *event = (unm_tcp_event_msg_t *)&msg->body;
        static char		*str[] = {
		"Unknown Type",
		"FIN_ACKED",
		"RST_RECEIVED",
		"ABORT_ON_SYN",
		"ABORT_ON_DATA",
		"ABORT_ON_CLOSE",
		"ABORT_ON_MEMORY",
		"ABORT_ON_LINGER",
		"ABORT_DISCONNECT",
		"KEEPALIVE_EXPIRED",
		"RETRANSMIT_TIMEOUT",
		"WND_PROBE_TIMEOUT"
	};
	static int		count = sizeof (str) / sizeof (char *);

        printk("%s: sid is NULL for connID[%u], TCP_EVENT, %s[%u], "
	       "reason[%u/%d]\n", __FUNCTION__, (uint32_t)event->hdr.ctx,
               event->param.type < count ? str[event->param.type] : str[0],
	       event->param.type,
	       (__uint32_t)event->param.reason, (int)event->param.reason);
}

/*
 *
 */
static void proc_ifconfig_reply(nx_tnic_adapter_t *tnic_adapter,
				unm_msg_t *msg, pegnet_sock_id_t *sid,
				pegnet_inprogress_t *trans)
{
        ifconfig_response_t     *reply;

        reply = (ifconfig_response_t *)&msg->body;

	if (!reply->logical && reply->cmd == NX_NETDEV_UP) {
                atomic_inc(&pegnet_init_done);
	}
	switch (reply->cmd) {
	case NX_NETDEV_ADD:
		nx_print("ifconfig %s ADD : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	case NX_NETDEV_REMOVE:
		nx_print("ifconfig %s REMOVE : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	case NX_NETDEV_UP:
		nx_print("ifconfig %s UP : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	case NX_NETDEV_DOWN:
		nx_print("ifconfig %s DOWN : %s\n", reply->ifname,
			 reply->ret_code ? "FAILED" : "SUCCESS");
		break;
	}
}

/*
 *
 */
static void proc_resource_notice(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        netxen_resource_notice_t        *notice;

        notice = (netxen_resource_notice_t *)&msg->body;
        switch (notice->type) {
        case NETXEN_TX_MEM_MS_LOW_WATER:
                printk("%s: Received NETXEN_TX_MEM_MS_LOW_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        case NETXEN_TX_MEM_MS_HIGH_WATER:
                printk("%s: Received NETXEN_TX_MEM_MS_HIGH_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        case NETXEN_TX_MEM_HS_LOW_WATER:
                printk("%s: Received NETXEN_TX_MEM_HS_LOW_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        case NETXEN_TX_MEM_HS_HIGH_WATER:
                printk("%s: Received NETXEN_TX_MEM_HS_HIGH_WATER, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->param);
                break;

        default:
                printk("%s: type = %u, data = %u\n",
                       __FUNCTION__, (uint32_t)notice->type,
		       (uint32_t)notice->param);
                break;
        }
}
/* Update connID at host.
 *
 */
static void nx_proc_update_connID(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg,
                        pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        nx_update_connID_table_t        *resp;
        pegnet_sock_id_t                *syn_sid = NULL;
        resp = (nx_update_connID_table_t *)&msg->body;
        /* get the child sid from listener syn q */
        if (!(syn_sid = nx_search_synq(sid, resp->local_ip, resp->local_port,
                                       resp->remote_ip, resp->remote_port))) {
                printk("Could not find child sid for sid [%p]\n", sid);
                return;
        }

        spin_lock_bh(&syn_sid->lock);
        syn_sid->ctx.context =  host_get_uint32_from_peg(resp->child_ctx);
        syn_sid->connID = host_get_uint32_from_peg(resp->child_connID);
        host_update_connID_table(tnic_adapter->p_ftable,syn_sid, resp->child_connID);
        spin_unlock_bh(&syn_sid->lock);
        return;
}

/* Handle handshake complete msg. 
 *
 */
static void unm_proc_tcp_handshake_response(nx_tnic_adapter_t *tnic_adapter,
					    unm_msg_t *msg,
					    pegnet_sock_id_t *sid,
					    pegnet_inprogress_t *trans)
{
	tcp_handshake_complete_t	*resp;
	pegnet_sock_id_t		*syn_sid = NULL;
	listener_group_t        	*grp = NX_GET_LISTENER_GROUP_FROM_SID(sid);	

	resp = (tcp_handshake_complete_t *)&msg->body;
//	printk("%s: ret_code:%d\n", __func__, resp->ret_code);

	spin_lock_bh(&sid->lock);

	if (sid->released || grp->released) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_CONN_FOR_FREED_LISTENER);
		spin_unlock_bh(&sid->lock);
		return;
	}

	/* get the child sid from listener syn q */
	syn_sid = nx_search_synq(sid, resp->local_ip, resp->local_port,
				 resp->remote_ip, resp->remote_port);
	if (!syn_sid) {
		printk("Could not find child sid for sid [%p]\n", sid);
		spin_unlock_bh(&sid->lock);
		return;
	}

	spin_unlock_bh(&sid->lock);

	spin_lock_bh(&grp->lock);
	/* remove this from the syn queue */
	nx_remove_synq(grp, syn_sid);

	/* free the sid if we got error */
	if (resp->ret_code != 0) {
		spin_unlock_bh(&grp->lock);

		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_ERR_PASSIVE_CONN_COMPLETE);
		if (resp->ret_code != PEGNET_SYN_OFFLOAD_RESP_LISTENER_FAIL) {
			/* processing has gone past the listening peg,
			   we should save some fields for cleanup. */
			host_assign_qid_from_peg(syn_sid->qid, resp->qid);
			host_release_unaccepted_child(syn_sid);
			/*
			 * There is no way a sock cleanup can come before this
			 * message. If it comes in before this or is processed
			 * before this we have a serious problem.
			 *
			 * No need to free the sid here. The sid will be freed
			 * when the sock cleanup comes down.
			 */
			syn_sid->release_seen = 1;
		} else {
			free_pegnet_sock_id(syn_sid);
		}
		return;
	}

	/* make the child sid usable */
	host_assign_qid_from_peg(syn_sid->qid, resp->qid);
	host_assign_qid_from_peg(syn_sid->rcvBuf_qid, resp->rcvBuf_qid);
	syn_sid->ctx.context =  host_get_uint32_from_peg(resp->child_ctx);
	syn_sid->connID = host_get_uint32_from_peg(resp->child_connID);
	host_update_connID_table(tnic_adapter->p_ftable, syn_sid,
				 resp->child_connID);


	/* add to the listener accept q */
	list_add_tail(&syn_sid->accept_active.list,
		      &grp->accept_passive.accept_queue);

	grp->accept_passive.accepted_len++;
	spin_unlock_bh(&grp->lock);

//	printk("%s: About to wake up sk=%p sid=%p grp=%p\n", __func__, sid->parent->sk,sid,grp);

	/* wakeup any pending accepts */
	if (sid->parent->sk) {
		bh_lock_sock(sid->parent->sk);
		if (sid->parent->sk->SK_DATA_READY) {
			sid->parent->sk->SK_DATA_READY(sid->parent->sk, 0);
		}
		bh_unlock_sock(sid->parent->sk);
	}

	return;
}

/*
 * A child sk has declared that it can be pruned
 */
static void unm_proc_prunable_sk(nx_tnic_adapter_t *tnic_adapter,
				 unm_msg_t *msg, pegnet_sock_id_t *sid,
				 pegnet_inprogress_t *trans)
{
	pegnet_sock_id_t	*child;
	tcp_prunable_sk_t	*pr_sk = (tcp_prunable_sk_t *)&msg->body;
        nx_lopt_t		*lopt;
        listener_group_t	*grp = NX_GET_LISTENER_GROUP_FROM_SID(sid);

	spin_lock_bh(&sid->lock);
	if (!(child = nx_search_synq(sid, pr_sk->local_ip,
					pr_sk->local_port,
					pr_sk->remote_ip,
					pr_sk->remote_port)) ) {
		nx_print("Could not find child of %p: rport = %x, rip = %x\n",
			 sid, pr_sk->remote_port, pr_sk->remote_ip);
		spin_unlock_bh(&sid->lock);
		return;
	}

	spin_unlock_bh(&sid->lock);
	host_assign_qid_from_peg(child->qid, pr_sk->child_qid);
	child->ctx.context =  host_get_uint32_from_peg(pr_sk->child_ctx);
	child->connID = host_get_uint32_from_peg(pr_sk->child_connID);
	host_update_connID_table(tnic_adapter->p_ftable, child,
				 pr_sk->child_connID);


	spin_lock_bh(&grp->lock);
	lopt = grp->accept_passive.lopt;
	list_add_tail(&child->accept_active.young_sid, &lopt->youngsters);
	spin_unlock_bh(&grp->lock);

	UNM_HOST_PNIC_STATS_INC(DBG_CNTR_TOTAL_PRUNABLES);

	return;
}

/*
 *
 */
static void nx_proc_routecache_msg(nx_tnic_adapter_t *tnic_adapter,
				   unm_msg_t *msg, pegnet_sock_id_t *sid,
				   pegnet_inprogress_t *trans)
{
	nx_proc_route_invalid(tnic_adapter, (routecache_invalid_t *)msg);
}

/*
 *
 */
void register_card_response_handler(unm_card_response_handler_t *handler)
{
        card_resp_handlers[handler->opcode] = *handler;
}

EXPORT_SYMBOL(register_card_response_handler);

static unm_card_response_handler_t	initial_card_resp_handlers[] =
{
	{UNM_P2H_OPCODE_START, 0, 0,
	 invalid_card_response, NULL, "START"},
        {UNM_P2H_OPCODE_CREATE_RESPONSE, 0, 0,
	 proc_create_response, NULL, "CREATE_RESPONSE"},
        {UNM_P2H_OPCODE_RELEASE_RESPONSE, 1, 1,
	 proc_release_response, NULL, "RELEASE_RESPONSE"},
        {UNM_P2H_OPCODE_CONNECT_RESPONSE, 1, 1,
	 unm_proc_connect_response, NULL, "CONNECT_RESPONSE"},
        {UNM_P2H_OPCODE_CONN_DISCONNECT_RESPONSE, 1, 0,
	 unm_proc_conn_disconnect_response, NULL, "CONN_DISCONNECT_RESPONSE"},

        {UNM_P2H_OPCODE_POLL_STATUS, 1, 0,
	 proc_poll_status, NULL, "POLL_STATUS"},
        {UNM_P2H_OPCODE_RX_URGENT_POINTER, 1, 0,
	 nx_proc_rx_urg_pointer, NULL, "RX_URGENT_POINTER"},
        {UNM_P2H_OPCODE_CONN_SIGNAL, 1, 0,
	 unm_proc_conn_signal, NULL, "CONN_SIGNAL"},
        {UNM_P2H_OPCODE_SENDMSG_CONTROL, 1, 0,
	 proc_sendmsg_control, NULL, "SENDMSG_CONTROL"},

        {UNM_P2H_OPCODE_TCP_EVENT, 1, 0,
	 proc_tcp_event, debug_print_tcp_event_msg, "TCP_EVENT"},

        {UNM_P2H_OPCODE_LISTEN_RESPONSE, 1, 1,
	 unm_host_proc_listen_response, NULL, "LISTEN_RESPONSE"},
        {UNM_P2H_OPCODE_IOCTL_RESPONSE, 1, 1,
	 unm_host_proc_ioctl_response, NULL, "IOCTL_RESPONSE"},
        {UNM_P2H_OPCODE_SETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_setsockopt_response, NULL, "SETSOCKOPT_RESPONSE"},
        {UNM_P2H_OPCODE_GETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_getsockopt_response, NULL, "GETSOCKOPT_RESPONSE"},
        {UNM_P2H_OPCODE_IPTSETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_iptsetsockopt_response, NULL, "IPTSETSOCKOPT_RESPONSE"},
        {UNM_P2H_OPCODE_IPTGETSOCKOPT_RESPONSE, 1, 1,
	 unm_host_proc_iptgetsockopt_response, NULL, "IPTGETSOCKOPT_RESPONSE"},

        {UNM_P2H_OPCODE_CONN_SHUTDOWN_RESPONSE, 1, 1,
	 unm_host_proc_shutdown_response, NULL, "CONN_SHUTDOWN_RESPONSE"},

        {UNM_P2H_OPCODE_IFCONFIG_REPLY, 0, 0,
	 proc_ifconfig_reply, NULL, "IFCONFIG_REPLY"},
	{UNM_P2H_OPCODE_CONNECT_COMPLETE, 1, 0,
	 unm_proc_connect_complete, NULL, "CONNECT_COMPLETE"},
        {UNM_P2H_OPCODE_GET_STATS_REPLY, 0, 0,
	 nx_proc_get_stats_reply, NULL, "GET_STATS_REPLY"},
        {UNM_P2H_OPCODE_RESOURCE_NOTICE, 0, 0,
	 proc_resource_notice, NULL, "RESOURCE_NOTICE"},
        {NX_P2H_OPCODE_SOCK_CLEANUP, 1, 0,
	 nx_proc_sock_cleanup, NULL, "SOCK_CLEANUP"},
	{UNM_P2H_OPCODE_3WAY_HANDSHAKE_COMPLETE, 1, 0,
	 unm_proc_tcp_handshake_response, NULL, "3WAY_HANDSHAKE_COMPLETE"},
	{UNM_P2H_OPCODE_PRUNABLE_SK, 1, 0,
	 unm_proc_prunable_sk, NULL, "PRUNABLE_SK"},
	{UNM_P2H_OPCODE_ROUTECACHE_INVALID, 0, 0,
	 nx_proc_routecache_msg, NULL, "ROUTECACHE_INVALID"},
        {NX_P2H_OPCODE_UPDATE_CONNID, 1, 0,
         nx_proc_update_connID, NULL, "UPDATE_CONNID"}

};

/*
 *
 */
static void init_card_response_handlers(void)
{
	int	i;
	int	count;

        memset(card_resp_handlers, 0, sizeof (card_resp_handlers));

	count = (sizeof (initial_card_resp_handlers) /
		 sizeof (unm_card_response_handler_t));
	for (i = 0; i < count; i++) {
		register_card_response_handler(&initial_card_resp_handlers[i]);
	}
};

/*
 *
 */
void handleCardResponse(struct net_device *netdev,unm_msg_t *msg)
{
	host_peg_msg_hdr_t      *hdr = (host_peg_msg_hdr_t *)&(msg->body);
	int                     id;
	pegnet_sock_id_t        *sid;
	pegnet_inprogress_t     *trans;
        uint8_t                 opcode;
        uint32_t                connID;
        unm_card_response_handler_t     *handler;
        int                     replies;
	nx_tnic_adapter_t 	*tnic_adapter;

        opcode = hdr->opcode;
	tnic_adapter = nx_get_dev_tnic_adapter(netdev);

        if (opcode == 0 || opcode >= UNM_P2H_OPCODE_LAST) {
                printk("%s: Invalid opcode[%u] in message\n", __FUNCTION__,
                       opcode);
                return;
        }

        handler = &card_resp_handlers[opcode];

        if (handler->func == NULL) {
		printk("%s: Unknown opcode 0x%x\n", __FUNCTION__, opcode);
		return;
	}

        if (!handler->retreive_socket) {
                handler->func(tnic_adapter,msg, NULL, NULL);
                return;
        }

        connID = host_get_uint32_from_peg(hdr->ctx);
        sid = retreive_pegnet_sock_id(tnic_adapter->p_ftable,connID);
        if (sid == NULL) {
                if (handler->printmsg) {
                        handler->printmsg(msg, sid, NULL);
                } else {
			printk("%s: NULL sid for connID[%u]: %s[0x%x]\n",
			       __FUNCTION__, connID,
			       (handler->opcode_str != NULL ?
				handler->opcode_str : "Unknown"),
			       opcode);
                        printk("%016llx  %016llx  %016llx  %016llx\n"
                               "%016llx  %016llx  %016llx  %016llx\n",
                               ((uint64_t *)msg)[0], ((uint64_t *)msg)[1],
                               ((uint64_t *)msg)[2], ((uint64_t *)msg)[3],
                               ((uint64_t *)msg)[4], ((uint64_t *)msg)[5],
                               ((uint64_t *)msg)[6], ((uint64_t *)msg)[7]);
                }
                return;
        }

        if (!handler->retreive_transaction) {
                handler->func(tnic_adapter,msg, sid, NULL);
		goto release_reference;
        }

	id = hdr->op_id;

        spin_lock_bh(&sid->lock);
        for (trans = sid->op; trans != NULL; trans = trans->link) {
                if (trans->op_id == id) {
                        break;
                }
        }
        spin_unlock_bh(&sid->lock);

        if (trans == NULL) {
                if (handler->printmsg) {
                        handler->printmsg(msg, sid, NULL);
                } else {
			printk("%s: NULL transaction for connID[%u]: "
			       "%s[0x%x]: op_id[%u]\n",
			       __FUNCTION__, connID,
			       (handler->opcode_str != NULL ?
				handler->opcode_str : "Unknown"),
			       opcode, id);
                        printk("%016llx  %016llx  %016llx  %016llx\n"
                               "%016llx  %016llx  %016llx  %016llx\n",
                               ((uint64_t *)msg)[0], ((uint64_t *)msg)[1],
                               ((uint64_t *)msg)[2], ((uint64_t *)msg)[3],
                               ((uint64_t *)msg)[4], ((uint64_t *)msg)[5],
                               ((uint64_t *)msg)[6], ((uint64_t *)msg)[7]);
                }
		goto release_reference;
        }

        atomic_inc(&trans->replied);
        replies = atomic_read(&trans->replied);     
        if (replies == 1) {
                handler->func(tnic_adapter,msg, sid, trans);
        } else {
                printk("%s: WARNING replies:%d != 1\n", __FUNCTION__, replies);
        }

  release_reference:
	host_flowtbl_put_reference(tnic_adapter->p_ftable,connID);
        return;
}

/*
 * Allocate a receive page.
 */
int alloc_netxen_page(netxen_host_page_t *page)
{
#if	defined(NETXEN_USE_ALLOC_PAGES_HM)
#error	"Has bug in 32 bit kernel"
	page->mm_page = alloc_pages(GFP_HIGHUSER, 0);
	if (page->mm_page == NULL)
#elif	defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
	page->mm_page = alloc_pages(GFP_KERNEL, 0);
	if (page->mm_page == NULL)
#else
	page->buffer = kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (page->buffer == NULL)
#endif
	{
		printk("%s: Memory Allocation failed\n", __FUNCTION__);
		return (-ENOMEM);
	}

#if	defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
	page->buffer = kmap(page->mm_page);
	if (page->buffer == NULL) {
		printk("%s: Memory Mapping failed\n", __FUNCTION__);
		__free_pages(page->mm_page, 0);
		return (-ENOMEM);
	}
#endif

	page->size = PAGE_SIZE;
	return (0);
}

/*
 * Free a receive page.
 */
int free_netxen_page(netxen_host_page_t *page,
		struct pci_dev *unm_device_ptr,int direction)
{
	netxen_page_unmap_pci(page, unm_device_ptr, direction);
	netxen_page_unmap_virt(page);

#if	defined(NETXEN_USE_ALLOC_PAGES_HM) || defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
#if	defined(NETXEN_USE_ALLOC_PAGES_LM)
	if (page->buffer) {
		kunmap(page->mm_page);
		page->buffer = NULL;
	}
#endif	/* NETXEN_USE_ALLOC_PAGES_LM */

	if (page->mm_page) {
		__free_pages(page->mm_page, 0);
	}
#else
	if (page->buffer) {
		kfree(page->buffer);
	}
#endif
	else {
		printk("%s: can't free NULL buffer. %s.\n", __FUNCTION__,
		       (direction == PCI_DMA_FROMDEVICE) ? "RX" : "TX");
		return (1);
	}

	page->mm_page = NULL;
	page->buffer = NULL;
	page->phys = 0;

	return (0);
}


unsigned int max_tcp_conn = UNM_MAX_TCP_CONNS;

static pegnet_sock_id_t *nx_search_synq(pegnet_sock_id_t *psid,
					uint32_t local_ip,
					uint16_t local_port,
					uint32_t remote_ip,
					uint16_t remote_port)
{
	u32 h;
	struct list_head *node;
	pegnet_sock_id_t *tmp;
        listener_group_t *grp = NX_GET_LISTENER_GROUP_FROM_SID(psid);
        nx_lopt_t *lopt = grp->accept_passive.lopt;

	if (!lopt) {
		printk("%s: lopt NULL for sid [%p]\n", __FUNCTION__, psid);
	       	return 0;
	}

       	h = jhash_2words(remote_ip, remote_port, lopt->hash_rnd)
				& (NX_SYNQ_HSIZE - 1);

	local_port = htons(local_port);

	list_for_each(node, &lopt->syn_table[h]) {
		tmp = list_entry(node, pegnet_sock_id_t,
					accept_active.list);

		if (tmp->rip_addr.v4 == remote_ip &&
				tmp->rport == remote_port &&
				tmp->lip_addr.v4 == local_ip &&
				tmp->lport == local_port) {
			return tmp;
		}
	}

	return NULL;
}

void nx_remove_synq(listener_group_t *grp, pegnet_sock_id_t *sid)
{
	list_del(&sid->accept_active.list);

	list_del_init(&sid->accept_active.young_sid); /* We may be attached */
	grp->accept_passive.lopt->qlen--;
}

/* check -
 *   	- ip options
 *	- syn with data
 *	- tcp options
 */
static int check_syn_options(struct sk_buff *skb, struct iphdr *iph)
{
	struct tcphdr	*th;

	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));

	// printk("%s: iph->ihl:%d so(iphdr):%d tot_len:%d th->doff:%d\n", __func__, iph->ihl, sizeof(struct iphdr), ntohs(iph->tot_len), th->doff);
	/* no ip options */
	if ( (iph->ihl * 4) > sizeof(struct iphdr))
		return 1;

	/* no syn with data */
	if ( ntohs(iph->tot_len) > ( (iph->ihl << 2) + (th->doff << 2)) )
		return 1;

	/* Will the TCP header fit it a single host-to-card message */
	if (NX_MAX_TCP_OFFLOAD_HDR_SIZE < (th->doff << 2)) {
		printk("%s: DOFF = %d is too big (max = %d)\n", __FUNCTION__,
				(int)th->doff, (int)NX_MAX_TCP_OFFLOAD_HDR_SIZE);
		return 1;
	}

	return 0;
}

/* offload the syn -
 *   - allocate sid
 *   - add sid to synq
 *   - send UNM_MSGTYPE_RXED_SYN_OFFLOAD to card
 */
static int offload_syn(struct net_device *netdev,pegnet_sock_id_t *parent_sid, 
			struct sock *sk,struct sk_buff *skb, struct iphdr *iph)
{
	pegnet_sock_id_t	*sid;
	listener_group_t        *grp;
	unm_msg_t		msg[NX_NUM_MSGS_PER_SYN];
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)	
	struct tcp_options_received		opt_rx = {0};
#else
	struct tcp_opt		opt_rx = {0};
#endif
	nx_lopt_t		*lopt;
	uint32_t		hash;
	struct list_head	*node;
	struct rtable		*rt;
	int			ifport;
	nx_tnic_adapter_t *tnic_adapter = nx_get_dev_tnic_adapter(netdev);

        grp = NX_GET_LISTENER_GROUP_FROM_SID(parent_sid);
        lopt = grp->accept_passive.lopt;
	skb->h.th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));

	/*
	 * Check if it is a retransmitted SYN.
	 */
	hash = jhash_2words(iph->saddr, skb->h.th->source, lopt->hash_rnd);
	hash &= (NX_SYNQ_HSIZE - 1);

	list_for_each(node, &lopt->syn_table[hash]) {
		sid = list_entry(node, pegnet_sock_id_t, accept_active.list);
		if (sid->rip_addr.v4 == iph->saddr &&
		    sid->lip_addr.v4 == iph->daddr &&
		    sid->rport == skb->h.th->source) {
			printk("%s: Dropping retransmitted SYN. Should "
			       "forward.\n", __FUNCTION__);
			return (0);
		}
	}

	/* ip options and syn with data not allowed */
	if (check_syn_options(skb, iph)) {
		printk("%s: IP options packet forwarded\n", __FUNCTION__);
		return (1);
	}

	/* card resources */
	if (check_card_resources(tnic_adapter->p_ftable)) {
		static unsigned int print_cnt = 0;
		if ((print_cnt % 10000) == 0) {
			nx_print("Num connections maxed out on card: %u\n", print_cnt);
		}
		print_cnt++;

		return (1);
	}

	/*
	 * Check if we can accept it.
	 */
	if ((lopt->qlen > NX_MAX_SYNQ_LEN) ||
	    grp->accept_passive.accepted_len > sk->sk_max_ack_backlog) {

		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYNQ_DROPPED);
		return (0);
	}

	/* We will accept this one, but do we need pruning? */
	if (lopt->qlen >= (NX_MAX_SYNQ_LEN >> 1)) {
		nx_prune_synq(grp, NX_SYNQ_PRUNETYPE_BH);
	}

	sid = alloc_pegnet_sock_id(netdev);
	if (!sid) {
		printk("null sid\n");
		return (1);
	}
        sid->type = CONNECTION_SOCK;

	/* parse tcp options */
	tcp_parse_options(skb, &opt_rx, 0);

	/*
	 * Do the output routing and if required populate the card with the
	 * route.
	 */
	if (nx_route_connection(&rt, iph->daddr, skb->h.th->dest, iph->saddr,
				skb->h.th->source, iph->tos,
				0/* dev->ifindex */)) {
		nx_print("Unable to route connection\n");
		free_pegnet_sock_id(sid);
		return (1);
	}

	ifport = nx_get_device_port(NX_KC_ROUTE_DEVICE(rt));
	if (ifport < 0) {
		nx_print("Not netxen device? Weird\n");
		free_pegnet_sock_id(sid);
		ip_rt_put(rt);
		return (1);
	}

	if (nx_route_add(tnic_adapter, rt, ifport)) {
		nx_print("Error: Route add failed\n");
		free_pegnet_sock_id(sid);
		ip_rt_put(rt);
		return (1);
	}
	ip_rt_put(rt);
	nx_set_route_reference(sid, iph->daddr, skb->h.th->dest, iph->saddr,
			       skb->h.th->source);


	/* initialize some fields */
	sid->lport = skb->h.th->dest;
	sid->rport = skb->h.th->source;
	sid->lip_addr.v4 = iph->daddr;
	sid->rip_addr.v4 = iph->saddr;
	sid->zero_copy_enabled = parent_sid->zero_copy_enabled;
	sid->accept_active.parent_sid = parent_sid;

	/* add to the syn q */
//	printk("%s: raddr:0x%x rport:0x%x rnd:0x%x hash:%d\n", __FUNCTION__, sid->rip_addr.v4, sid->rport, lopt->hash_rnd, h);

	list_add_tail(&sid->accept_active.list, &lopt->syn_table[hash]);
	lopt->qlen++;

	INIT_LIST_HEAD(&sid->accept_active.young_sid); /* not needed */

	/* First, copy out the iphdr */
	UNM_MSGQ_HDR(msg[NX_SYN_IPH_IDX].hdr, parent_sid->qid.major,
			parent_sid->qid.minor,
			RXED_SYN_OFFLOAD);
	memcpy(&msg[NX_SYN_IPH_IDX].body, iph, iph->ihl << 2);

	/* Next, the tcp hdr */
	UNM_MSGQ_HDR(msg[NX_SYN_TCPH_IDX].hdr, parent_sid->qid.major,
			parent_sid->qid.minor,
			RXED_SYN_OFFLOAD);
	memcpy(&msg[NX_SYN_TCPH_IDX].body, skb->h.th, skb->h.th->doff << 2);

#if 0
	printk("%s: lip:0x%x rip:0x%x lport:0x%x rport:0x%x conn:0x%x\n",
	       __func__, syn_req->local_ip.v4, syn_req->remote_ip.v4,
	       syn_req->local_port, syn_req->remote_port, syn_req->connID);
#endif
	/* here we go */
	if (unm_send_h2p_msg_atomic(netdev,&msg[0], NX_NUM_MSGS_PER_SYN)) {
		printk("error sending compressed syn\n");
		nx_remove_synq(grp, sid);
		free_pegnet_sock_id(sid);
	}

	return (0);
}

/* perform some sanity checks on the packet
 */
static int check_syn_packet(struct sk_buff *skb, struct iphdr *iph)
{
	struct tcphdr	*th;
	uint32_t	len;

	/* ip version
	 * hdr length
	 * total length
	 */
	if (iph->ihl < 5 || iph->version != 4 ) {
		return 1;
	}

	len = ntohs(iph->tot_len);
	if (skb->len < len || len < (iph->ihl << 2)) {
		return 1;
	}
	
	/* tcp doff
	 */
	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));
	if (th->doff < sizeof(struct tcphdr) / 4) {
		return 1;
	}

	return 0;

}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)	
static u32 tcp_v4_synq_hash_tbl(u32 raddr, u16 rport, u32 rnd)
{
	return (jhash_2words(raddr, (u32) rport, rnd) & (TCP_SYNQ_HSIZE - 1));
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)	
static struct open_request * check_tuple_existence(void *tp,struct open_request ***prevp,__u16 rport,
						__u32 raddr, __u32 laddr)
{

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
        struct tcp_listen_opt *lopt = ((struct tcp_opt *)tp)->listen_opt;
#else
        struct tcp_listen_opt *lopt = ((struct tcp_sock *)tp)->listen_opt;
#endif
	struct open_request *req, **prev;
	for (prev = &lopt->syn_table[tcp_v4_synq_hash_tbl(raddr, rport, lopt->hash_rnd)];
	    (req = *prev) != NULL;
	    prev = &req->dl_next) {
		if (req->rmt_port == rport &&
		    req->af.v4_req.rmt_addr == raddr &&
		    req->af.v4_req.loc_addr == laddr &&
		    TCP_INET_FAMILY(req->class->family)) {
		    	*prevp = prev;
			break;
		}	
	}	
	return req;
}
#endif

static int check_non_offloaded_conn(struct sock *sk,struct sk_buff *skb,
				    struct iphdr *iph)
{
        struct tcphdr *th;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
        struct request_sock **prev;
        struct request_sock *req;
#else
	struct open_request **prev;
	struct open_request *req;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct tcp_opt *tp = &(sk->tp_pinfo.af_tcp);
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
	struct tcp_opt *tp = tcp_sk(sk);
#else
	struct tcp_sock *tp = tcp_sk(sk);
#endif
	
	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));

	if(sk->SK_STATE==TCP_LISTEN){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)	
		req = (struct request_sock *)inet_csk_search_req((struct sock *)tp, &prev, th->source,iph->saddr,
					  iph->daddr);
#else
		req = check_tuple_existence(tp, &prev, th->source,iph->saddr,
					    iph->daddr);
#endif		
		if(req){
			return 1;
		} else {
			return 0;
		}	
	} else if (TCP_TIME_WAIT == sk->SK_STATE) {
		/* protinfo is not usable here, because the host stack uses it
		   differently. tcp_rcv_state_process -> tcp_time_wait ->
		   __tcp_hashdance will replace sk with a tcp_tw_bucket, and
		   only explicitly initialised fields can be relied upon after
		   that. Luckily, if we are in TIME_WAIT on the host, it could
		   only be because we do not need protinfo! */
#if 0
		printk("%s: Time-wait, not relying on protinfo\n",
		       __FUNCTION__);
		printk("%s: Not offloaded sport = %x, saddr %x, dport = %x, "
		       "daddr = %x\n", __FUNCTION__, th->source,
		       iph->saddr, th->dest, iph->daddr);
#endif
		return 1;
    } else {
		if(sk->SK_PROTINFO){
			return 0;
                }
		else {
			return 1;
            }
        }
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
static __inline__ int nx_tcp_hashfn(__u32 laddr, __u16 lport,
                                  __u32 faddr, __u16 fport)
 {
         int h = (laddr ^ lport) ^ (faddr ^ fport);
         h ^= h >> 16;
         h ^= h >> 8;
         return h & (tcp_ehash_size - 1);
 }
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
static inline struct sock *nx_tcp_v4_lookup_established(u32 saddr, u16 sport,
                                                        u32 daddr, u16 hnum,
                                                        int dif)
 {
         struct tcp_ehash_bucket *head;
         TCP_V4_ADDR_COOKIE(acookie, saddr, daddr)
         __u32 ports = TCP_COMBINED_PORTS(sport, hnum);
         struct sock *sk;
         struct hlist_node *node;
         /* Optimize here for direct hit, only listening connections can
          * have wildcards anyways.
          */
         int hash = nx_tcp_hashfn(daddr, hnum, saddr, sport);
         head = &tcp_ehash[hash];
         read_lock(&head->lock);
         sk_for_each(sk, node, &head->chain) {
                 if (TCP_IPV4_MATCH(sk, acookie, saddr, daddr, ports, dif))
                         goto hit; /* You sunk my battleship! */
         }

         /* Must check for a TIME_WAIT'er before going to listener hash. */
         sk_for_each(sk, node, &(head + tcp_ehash_size)->chain) {
                 if (TCP_IPV4_TW_MATCH(sk, acookie, saddr, daddr, ports, dif))
                         goto hit;
         }
         sk = NULL;
 out:
         read_unlock(&head->lock);
         return sk;
 hit:
         sock_hold(sk);
         goto out;
 }

#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
static inline struct sock *__nx_tcp_v4_lookup(u32 saddr, u16 sport,
                                            u32 daddr, u16 hnum, int dif)
 {
         struct sock *sk = nx_tcp_v4_lookup_established(saddr, sport,
                                                       daddr, hnum, dif);

         return sk ? : tcp_v4_lookup_listener(daddr, hnum, dif);
 }
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
static inline struct sock *nx_tcp_v4_lookup(u32 saddr, u16 sport, u32 daddr,
                                   u16 dport, int dif)
 {
         struct sock *sk;

         local_bh_disable();
         sk = __nx_tcp_v4_lookup(saddr, sport, daddr, ntohs(dport), dif);
         local_bh_enable();

         return sk;
 }
#endif

/* This will handle SYNs for offloaded connections
 *
 * Return - 0: pegnet handled the SYN, nic should drop it
 *          1: nic should handle it
 */
int nx_handle_offloaded_syn(struct net_device *netdev,
			    uint16_t listener_id, struct sk_buff *skb)
{
	struct sock	*sk;
	struct iphdr	*iph;
	struct tcphdr	*th;
	uint32_t	saddr;
	uint32_t	daddr;
	uint16_t	sport;
	uint16_t	dport;
	pegnet_sock_id_t	*sid = NULL;
	listener_group_t *grp;
	int		rv;
	nx_tnic_adapter_t *tnic_adapter = nx_get_dev_tnic_adapter(netdev); 

	if (!skb || !skb->data) {
		printk("skb or skb data\n");
		return 1;
	}

	if (skb->protocol == htons(ETH_P_8021Q)) {
		iph = (struct iphdr *)(skb->data + VLAN_HLEN);
	} else {
		iph = (struct iphdr *)skb->data;
	}
	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));
    
/* 	if (!th->syn || th->ack) { */
	if (!(tcp_flag_word(th) & TCP_FLAG_SYN) ||
	    (tcp_flag_word(th) & (TCP_FLAG_CWR | TCP_FLAG_ECE | TCP_FLAG_URG |
				  TCP_FLAG_ACK | TCP_FLAG_RST |
				  TCP_FLAG_FIN))) {
		printk("%s: Not a packet that can be processed\n",
		       __FUNCTION__);
		return (1);
	}

	saddr = iph->saddr;
	daddr = iph->daddr;
	sport = th->source;
	dport = th->dest;

#if 0
	printk("%s: SYN saddr:0x%x daddr:0x%x sport:0x%x dport:0x%x\n",
	       __func__, saddr, daddr, sport, dport);
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
	sk = nx_tcp_v4_lookup(saddr, sport, daddr, dport, 0);
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13)
	sk = tcp_v4_lookup(saddr, sport, daddr, dport, 0);
#else
	sk = inet_lookup(&tcp_hashinfo, saddr, sport, daddr, dport,0);
#endif    

	if (!sk) {
		return 1;
	}

	/* lookups do a hold */
	sock_put(sk);

#if 0
	printk("%s: Found sk [%p]\n", __func__, sk);
#endif

	if (check_non_offloaded_conn(sk, skb, iph)) {
		return 1;
	}

	grp = sk->SK_PROTINFO;
	if (grp == NULL) {
		printk("%s: Listener group not found for sk [%p]\n", __FUNCTION__, sk);
		return 1;
	}
	if (grp->released) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYN_FOR_FREED_LISTENER);
		return 0;
	}
	sid = get_sid_by_grp(tnic_adapter, grp);
	if (sid == NULL) {
		printk("%s: sid not found for sk [%p] listener group [%p] adapter [%p]\n", __FUNCTION__, sk, grp, tnic_adapter);
		return 1;
	}

	host_flowtbl_get_reference(tnic_adapter->p_ftable,sid->connID);
	rv = 0;
	spin_lock_bh(&sid->lock);
	if (!sid->listener || grp->type != LISTENER_GROUP) {
		printk("%s: sid for listener_id[%u] is not a listener\n",
		       __FUNCTION__, listener_id);
		rv = 1;
		goto done;
	}

	if (sid->released) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYN_FOR_FREED_LISTENER);
		rv = 0;
		goto done;
	}

	/* do some sanity tests on the packet */
	if (check_syn_packet(skb, iph)) {
		printk("%s: Invalid packet, drop it\n", __FUNCTION__);
		rv = 0;
		goto done;
	}

	/* Ok, everything's fine. Let's offload this */
	if (offload_syn(netdev,sid, sk, skb, iph)) {
		rv = 1;
		goto done;
	}

  done:
	spin_unlock_bh(&sid->lock);

	host_flowtbl_put_reference(tnic_adapter->p_ftable,sid->connID);
	return (rv);
}

/*
 * END OF ROUTINES TO HANDLE RECEIVED DATA FROM THE CARD.
 */

extern void proc_tx_completion(struct net_device *netdev,unm_msg_t *msg);
extern void proc_tx_switch_completion(struct net_device *netdev,unm_msg_t *msg);

/*
 * The tasklet to wake up sleeping processes when we get a reply
 * from our board
 */
void unm_process_pegnet_msg(struct net_device *netdev, unm_msg_t *msg)
{
        if (msg->hdr.type == UNM_MSGTYPE_CARD_RESPONSE) {
		/*
		 * We got a bunch of responses from the card. Extract the
		 * info and dispatch accordingly.
		 */
		handleCardResponse(netdev,msg);
		return;
        }

        if (msg->hdr.type == UNM_MSGTYPE_RXBUFFER) {
                /* 
                 * Received data from the card.
                 */
                nx_proc_rx_buffer(netdev,msg);
                return;
        }

        if (msg->hdr.type == UNM_MSGTYPE_SENDMSG) {
                /* 
                 * Used send buffer from the card.
                 */
                proc_tx_completion(netdev,msg);
                return;
        }

        if (msg->hdr.type == UNM_MSGTYPE_TXBUFFER_SWITCH) {
                /*
                 * Completion of switch of zcopy buffer from the card.
                 */
                proc_tx_switch_completion(netdev,msg);
                return;
        }

        /* Now check what we have to do */
	printk(KERN_ERR "%s: UNM_ISR: #5 Unexpected msg\n"
	       "0x%llx 0x%llx 0x%llx 0x%llx\n"
	       "0x%llx 0x%llx 0x%llx 0x%llx\n", __FUNCTION__,
	       msg->hdr.word, msg->body.values[0],
	       msg->body.values[1], msg->body.values[2],
	       msg->body.values[3], msg->body.values[4],
	       msg->body.values[5], msg->body.values[6]);
}

/*
 * This function is called for each flow when the table is destroyed. This
 * allows us to free the flows in the table and also notify the card to kill
 * i.e. reset each flow.
 *
 * Parameters:
 *      id      - The ID of the flow being freed.
 *      flow    - The specified flow pointer being freed.
 *
 * Return:
 *      void.
 */
void flow_tbl_destroy_flow_free(flow_id_t id, void *flow)
{
        pegnet_sock_id_t        *sid = (pegnet_sock_id_t *)flow;

        if (sid->sk_cleanup) {
		nx_sk_cleanup(sid->sk_cleanup);
		/* kfree(sid); */
        } else {
		printk("%s: Should not be here for sid = %p\n", __FUNCTION__,
		       sid);
	}
}

static int proc_read_tx_zcopy_iov_type(char *page, char **start, off_t off,
				       int count, int *eof, void *data)
{
	int len = 0;

	len = sprintf(page, "Compressed IOVs are  %s\n",
		      global_use_compressed_iov ? "ENABLED" : "DISABLED");
	return len;

}

static int proc_write_tx_zcopy_iov_type(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
	int len = 0;
	char str[2];
	

	if (count == 0) {
		return 0;
	}

	if(count > 2)
		len = 2;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';
	
	if (!strcmp(str, "1")) {
		global_use_compressed_iov = 1;

	} else if (!strcmp(str,"0")) {
		global_use_compressed_iov = 0;
	}

	return len;
}


static int proc_read_tx_zcopy_size_thresh(char *page, char **start, off_t off,
					  int count, int *eof, void *data)
{
	int len = 0;
	
	len = sprintf(page, "Zero copy Tx size threshold = %u bytes [default = 0, max = 2097152] \n",
		      tx_zero_copy_size_thresh);
	return len;

}

static int proc_write_tx_zcopy_size_thresh(struct file *file,
					   const char *buffer,
					   unsigned long count, void *data)
{
	int len = 0;
	char str[8];
	uint32_t size = 0;
	char *end;
	
	
	if (count == 0) {
		return 0;
	}

	if(count > 8)
		len = 8;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';

	size = simple_strtoul(str, &end, 0);
	if (end == str) {
		printk("Invalid size\n");
		return len;

	}

	if (size <= 2*1024*1024) {
		tx_zero_copy_size_thresh = size;
	}


	return len;
}


static int proc_read_tx_zcopy_conn_thresh(char *page, char **start, off_t off,
					  int count, int *eof, void *data)
{
	int len = 0;
	
	len = sprintf(page, "Zero copy Tx connection threshold = %u [default = 0, max = 1000000]\n",
		      tx_zero_copy_active_thresh);
	return len;
}

static int proc_write_tx_zcopy_conn_thresh(struct file *file,
					   const char *buffer,
					   unsigned long count,
					   void *data)
{
	int len = 0;
	char str[8];
	uint32_t conn = 0;
	char *end;
	
	
	if (count == 0) {
		return 0;
	}

	if(count > 8)
		len = 8;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';

	conn = simple_strtoul(str, &end, 0);
	if (end == str) {
		printk("Invalid size\n");
		return len;

	}

	if (conn <= 1000000) {
		tx_zero_copy_active_thresh = conn;
	}


	return len;
}


static 
int proc_read_tx_bytes_out_thresh(char *page, char **start,
				       off_t off, int count, 
				       int *eof, void *data)
{
	int len = 0;
	
	len = sprintf(page, "Zero copy Tx outstanding bytes threshold = %u "
                      "[default = 1000000, max = 4294967295 (2^32-1)]\n",
		      tx_bytes_out_thresh);
	return len;
}

static 
int proc_write_tx_bytes_out_thresh(struct file *file,
					const char *buffer,
					unsigned long count, 
					void *data)
{
	int len = 0;
	char str[8];
	uint32_t out_bytes = 0;
	char *end;
	
	
	if (count == 0) {
		return 0;
	}

	if(count > 11)
		len = 11;
	else
		len = count;
	
	if(copy_from_user(str, buffer, len))
		return -EFAULT;
	

	str[len-1] = '\0';

	out_bytes = simple_strtoul(str, &end, 0);
	if (end == str) {
		printk("Invalid size\n");
		return len;

	}

	tx_bytes_out_thresh = out_bytes;


	return len;
}

static int proc_read_max_tcp_conn(char *page, char **start,
					off_t off, int count,
					int *eof, void *data)
{
	int len = 0;

	len = sprintf(page, "%u\n", max_tcp_conn);
	return len;
}

static int proc_write_max_tcp_conn(struct file *file,
					const char *buffer,
					unsigned long count,
					void *data)
{
	int len = 0;
	char str[9];
	unsigned int max_conn;

	if (count == 0)
		return 0;

	if (count > 8)
		len = 8;
	else
		len = count;

	if (copy_from_user(str, buffer, len))
		return -EFAULT;

	str[len] = 0;

	max_conn = simple_strtoul(str, NULL, 10);
	
	if (max_conn > UNM_MAX_TCP_CONNS)
		return -EINVAL;

	printk("%s: max_conn:%d\n", __FUNCTION__, max_conn);
	max_tcp_conn = max_conn;

	return len;
}

static int proc_read_dbg_counters(char *page, char **start, off_t off,
				  int count, int *eof, void *data)
{
	int	len = 0;

#define DBG_CNTR_PRINT(BUF, LEN, STR, VARIABLE)			\
	(LEN) += sprintf((BUF) + (LEN), "    %44s\t = %d\n",	\
			 STR, UNM_HOST_PNIC_STATS_GET(VARIABLE))

        len += sprintf(page, "NetXen TOE Debug Counters\n");
        len += sprintf(page + len, "\n");

        DBG_CNTR_PRINT(page, len, "Active Connections", DBG_CNTR_ACTIVE_CONNS);
        DBG_CNTR_PRINT(page, len, "Connection Table Overflows",
		       DBG_CNTR_CONN_TBL_OVERFLOWS);
        DBG_CNTR_PRINT(page, len, "'struct sock' in use", DBG_CNTR_INUSE_SK);
        DBG_CNTR_PRINT(page, len, "TX Pages in use", DBG_CNTR_INUSE_TX_PAGES);
        DBG_CNTR_PRINT(page, len, "RX Pages in use", DBG_CNTR_INUSE_RX_PAGES);
        DBG_CNTR_PRINT(page, len, "Global RX Pages in use",
		       DBG_CNTR_INUSE_GLOBAL_RX_PAGES);
        DBG_CNTR_PRINT(page, len, "Dropped SYNs", DBG_CNTR_SYNQ_DROPPED);
        DBG_CNTR_PRINT(page, len, "SYN received for freed listener",
		       DBG_CNTR_SYN_FOR_FREED_LISTENER);
        DBG_CNTR_PRINT(page, len, "Conn complete received for freed listener",
		       DBG_CNTR_CONN_FOR_FREED_LISTENER);
        DBG_CNTR_PRINT(page, len, "Error passive connection complete",
		       DBG_CNTR_ERR_PASSIVE_CONN_COMPLETE);
        DBG_CNTR_PRINT(page, len, "Released from SYN queue",
		       DBG_CNTR_FREED_FROM_SYNQ);
        DBG_CNTR_PRINT(page, len, "Released from accept queue",
		       DBG_CNTR_FREED_FROM_ACCEPTQ);
        DBG_CNTR_PRINT(page, len, "TX Freeing In BH",
		       DBG_CNTR_TX_FREEING_IN_BH);
        DBG_CNTR_PRINT(page, len, "TX Delayed Freeing",
		       DBG_CNTR_TX_DELAY_FREEING);
        DBG_CNTR_PRINT(page, len, "Sendpage Timer Fired",
		       DBG_CNTR_SENDPAGE_TIMER_FIRED);
        DBG_CNTR_PRINT(page, len, "Sendpage Flushed 1",
		       DBG_CNTR_SENDPAGE_FLUSHED1);
        DBG_CNTR_PRINT(page, len, "Sendpage Flushed 2",
		       DBG_CNTR_SENDPAGE_FLUSHED2);
        DBG_CNTR_PRINT(page, len, "Sendpage Flushed 3",
		       DBG_CNTR_SENDPAGE_FLUSHED3);
        DBG_CNTR_PRINT(page, len, "Total Pruned", DBG_CNTR_TOTAL_PRUNED);
        DBG_CNTR_PRINT(page, len, "Total Prunables", DBG_CNTR_TOTAL_PRUNABLES);
        DBG_CNTR_PRINT(page, len, "Rst to child on accept queue",
		       DBG_CNTR_RST_ON_ACCEPT_QUEUE);
        len += sprintf(page + len, "\n");

/*         len += sprintf(page + len, "%-60s: %d\n", "RX read timeout count", */
/* 		       val); */
        *eof = 1;
        return len;
}

static char *ip_stats_str[] = {
	"IpInReceives",
	"IpInHdrErrors",
	"IpInAddrErrors",
	"IpForwDatagrams",
	"IpInUnknownProtos",
	"IpInDiscards",
	"IpInDelivers",
	"IpOutRequests",
	"IpOutDiscards",
	"IpOutNoRoutes",
	"IpReasmTimeout",
	"IpReasmReqds",
	"IpReasmOKs",
	"IpReasmFails",
	"IpFragOKs",
	"IpFragFails",
	"IpFragCreates"
};

static char *tcp_stats_str[] = {
	"TcpRtoAlgorithm",
	"TcpRtoMin",
	"TcpRtoMax",
	"TcpMaxConn",
	"TcpActiveOpens",
	"TcpPassiveOpens",
	"TcpAttemptFails",
	"TcpEstabResets",
	"TcpCurrEstab",
	"TcpInSegs",
	"TcpOutSegs",
	"TcpRetransSegs",
	"TcpInErrs",
	"TcpOutRsts"
};

/*
 *
 */
static int procfs_print_ip_statistics(char *buffer, netxen_pstats_t *pstat,
				      int port)
{
	int		len = 0;
	int		ii;
	int		cnt1;
	int		cnt2;
	int		jj;
	__uint64_t	value;

	cnt1 = sizeof (ip_stats_str) / sizeof (char *);
	cnt2 = sizeof (nx_ip_mib_t) / sizeof (__uint64_t);

	if (cnt1 != cnt2) {
		len += sprintf(buffer, "Mismatch in number of elements in IP "
			       "string and statistics: strings[%d] != "
			       "statistics[%d]\n", cnt1, cnt2);
		return (len);
	}

	for (jj = 0; jj < cnt1; jj++) {
		len += sprintf(buffer + len, "%-20s", ip_stats_str[jj]);

		value = 0;
		for (ii = 0; ii < NETXEN_NUM_PEGS; ++ii) {
			value += ((__uint64_t *)&pstat->peg[ii].port[port].ip_statistics)[jj];
			len += sprintf(buffer + len, "%10llu  ",
				       ((__uint64_t *)&pstat->peg[ii].port[port].ip_statistics)[jj]);
		}

		len += sprintf(buffer + len, "%10llu\n", value);
	}

	return (len);
}

/*
 *
 */
static int procfs_print_tcp_statistics(char *buffer, netxen_pstats_t *pstat,
				       int port)
{
	int		len = 0;
	int		ii;
	int		cnt1;
	int		cnt2;
	int		jj;
	__uint64_t	value;

	cnt1 = sizeof (tcp_stats_str) / sizeof (char *);
	cnt2 = sizeof (nx_tcp_mib_t) / sizeof (__uint64_t);

	if (cnt1 != cnt2) {
		len += sprintf(buffer, "Mismatch in number of elements in TCP "
			       "string and statistics: strings[%d] != "
			       "statistics[%d]\n", cnt1, cnt2);
		return (len);
	}

	for (jj = 0; jj < cnt1; jj++) {
		len += sprintf(buffer + len, "%-20s", tcp_stats_str[jj]);

		value = 0;
		for (ii = 0; ii < NETXEN_NUM_PEGS; ++ii) {
			value += ((__uint64_t *)&pstat->peg[ii].port[port].tcp_statistics)[jj];
			len+= sprintf(buffer + len, "%10llu  ",
				      ((__uint64_t *)&pstat->peg[ii].port[port].tcp_statistics)[jj]);
		}

		len += sprintf(buffer + len, "%10llu\n", value);
	}

	return (len);
}

/*
 *
 */
static int proc_read_stats(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{
	int			len = 0;
	nx_tnic_adapter_t	*tnic = (nx_tnic_adapter_t *)data;
        uint32_t		i;
	int			rv;


        len += sprintf(page + len, "TCP/IP Statistics\n");
        len += sprintf(page + len, "=================\n");
        len += sprintf(page + len, "\n");
/*         len += sprintf(page + len, "%-60s: %d\n", "RX read timeout count", */
/* 		       val); */

        for (i = 0; i < NX_TNIC_MAX_PORTS; i++) {
                if (NX_GET_DEV_FROM_TNIC_ADAPTER(tnic, i)) {
			break;
		}
        }

	if (i == NX_TNIC_MAX_PORTS) {
		len += sprintf(page + len, "No device active for getting "
			       "statistics\n");
		goto done;
	}

	rv = host_get_stats(tnic, i);
	if (rv) {
		len += sprintf(page + len, "Error[%d]: Getting statistics "
			       "from card failed\n", rv);
		goto done;
	}

	len += sprintf(page + len, "%20s%10s  %10s  %10s  %10s  %10s\n",
		       " ", "Peg0", "Peg1", "Peg2", "Peg3", "Total");

	len += procfs_print_ip_statistics(page + len, tnic->stats.data, 0);
	len += procfs_print_tcp_statistics(page + len, tnic->stats.data, 0);

  done:
        *eof = 1;
        return len;
}

/*
 *
 */
static int proc_write_stats(struct file *file, const char *buffer,
			    unsigned long count, void *data)
{
	int			len = 0;
	char			str[9];
	unsigned int		value;
	nx_tnic_adapter_t	*tnic = (nx_tnic_adapter_t *)data;
        uint32_t		i;
	int			rv;

	nx_print("adapter = %p\n", tnic);

	if (count == 0) {
		return 0;
	}

	len = (count > 8) ? 8 : count;

	if (copy_from_user(str, buffer, len)) {
		return -EFAULT;
	}

	str[len] = 0;

	value = simple_strtoul(str, NULL, 10);

	if (value != 0) {
		return -EINVAL;
	}


        for (i = 0; i < NX_TNIC_MAX_PORTS; i++) {
                if (!NX_GET_DEV_FROM_TNIC_ADAPTER(tnic,i)) {
			continue;
		}

		nx_print("Resetting stats for port <%u>.\n", i);

		rv = host_reset_stats(NX_GET_DEV_FROM_TNIC_ADAPTER(tnic,i), i);
		if (rv) {
			nx_print("Error[%d]: Reset of statistics for "
				 "port <%u> failed\n", rv, i);
		}
        }

	return (len);
}


/*
 *
 */
static void nx_tnic_procfs_destroy_parent(void)
{
	if (tx_bytes_out_file) {
		remove_proc_entry("zcopy_bytes_out_thresh", zcopy_dir);
		tx_bytes_out_file = NULL;
	}

	if (tx_conn_thresh_file) {
		remove_proc_entry("zcopy_conn_thresh", zcopy_dir);
		tx_conn_thresh_file = NULL;
	}

	if (tx_size_thresh_file) {
		remove_proc_entry("zcopy_size_thresh", zcopy_dir);
		tx_size_thresh_file = NULL;
	}

	if (compressed_iov_file) {
		remove_proc_entry("use_compressed_IOV", zcopy_dir);
		compressed_iov_file = NULL;
	}

	if (zcopy_dir) {
		remove_proc_entry("zcopytx", pegnet_dir);
		zcopy_dir = NULL;
	}

	if (pegnet_dir) {
		remove_proc_entry("pegnet", NULL);
		pegnet_dir = NULL;
	}
}

/*
 * Create all the TOE related proc entries.
 */
static int nx_tnic_procfs_create_parent(void)
{
	pegnet_dir = proc_mkdir("pegnet", NULL);
	if (pegnet_dir == NULL) {
		goto error;
	}

	pegnet_dir->owner = THIS_MODULE;

	/*
	 * Create zero copy related directory and entries.
	 */
	zcopy_dir = proc_mkdir("zcopytx", pegnet_dir);
	if (zcopy_dir == NULL) {
		goto error;
	}

	zcopy_dir->owner = THIS_MODULE;

	/*
	 *
	 */
	compressed_iov_file = create_proc_entry("use_compressed_IOV", 0644,
						zcopy_dir);
	if (compressed_iov_file == NULL) {
		goto error;
	}
	compressed_iov_file->data = NULL;
	compressed_iov_file->read_proc = proc_read_tx_zcopy_iov_type;
	compressed_iov_file->write_proc = proc_write_tx_zcopy_iov_type;
	compressed_iov_file->owner = THIS_MODULE;

	/*
	 *
	 */
	tx_size_thresh_file = create_proc_entry("zcopy_size_thresh", 0644,
						zcopy_dir);
	if (tx_size_thresh_file == NULL) {
		goto error;
	}
	tx_size_thresh_file->data = NULL;
	tx_size_thresh_file->read_proc = proc_read_tx_zcopy_size_thresh;
	tx_size_thresh_file->write_proc = proc_write_tx_zcopy_size_thresh;
	tx_size_thresh_file->owner = THIS_MODULE;	

	/*
	 *
	 */
	tx_conn_thresh_file = create_proc_entry("zcopy_conn_thresh", 0644,
						zcopy_dir);
	if (tx_conn_thresh_file == NULL) {
		goto error;
	}
	tx_conn_thresh_file->data = NULL;
	tx_conn_thresh_file->read_proc = proc_read_tx_zcopy_conn_thresh;
	tx_conn_thresh_file->write_proc = proc_write_tx_zcopy_conn_thresh;
	tx_conn_thresh_file->owner = THIS_MODULE;	

	/*
	 *
	 */
	tx_bytes_out_file = create_proc_entry("zcopy_bytes_out_thresh", 0644,
					      zcopy_dir);
	if (tx_bytes_out_file == NULL) {
		goto error;
	}
	tx_bytes_out_file->data = NULL;
	tx_bytes_out_file->read_proc = proc_read_tx_bytes_out_thresh;
	tx_bytes_out_file->write_proc = proc_write_tx_bytes_out_thresh;
	tx_bytes_out_file->owner = THIS_MODULE;

	return (0);

  error:
	nx_tnic_procfs_destroy_parent();
	return (-ENOMEM);
}

/*
 *
 */
int nx_tnic_adapter_create_procfs(nx_tnic_adapter_t *tnic,
				  struct proc_dir_entry *adapter_dir)
{
	char tnic_dir[20];
	if (adapter_dir == NULL) {
		nx_print("Specified adapter_dir is NULL using pegnet_dir\n");
		adapter_dir = pegnet_dir;
	}

//	tnic->procfs.base_dir = proc_mkdir("tnic", adapter_dir);
        atomic_inc(&g_devid);
        atomic_set(&tnic->dev_id,atomic_read(&g_devid));
        sprintf(tnic_dir,"tnic_%d",atomic_read(&tnic->dev_id));
        tnic->procfs.base_dir = proc_mkdir(tnic_dir, adapter_dir);

	if (tnic->procfs.base_dir == NULL) {
		goto error;
	}

	tnic->procfs.base_dir->owner = THIS_MODULE;

	/*
	 * Create all the common variables.
	 */
	tnic->procfs.max_tcp_conn_file =
		create_proc_entry("max_offloaded_tcp_conn", 0644,
				  tnic->procfs.base_dir);
	if (tnic->procfs.max_tcp_conn_file == NULL) {
		goto error;
	}
	tnic->procfs.max_tcp_conn_file->owner = THIS_MODULE;
	tnic->procfs.max_tcp_conn_file->read_proc = proc_read_max_tcp_conn;
	tnic->procfs.max_tcp_conn_file->write_proc = proc_write_max_tcp_conn;
	tnic->procfs.max_tcp_conn_file->data = tnic;

	/*
	 *
	 */
	tnic->procfs.dbg_counters_file =
		create_proc_entry("dbg_counters", 0644, tnic->procfs.base_dir);
	if (tnic->procfs.dbg_counters_file == NULL) {
		goto error;
	}
	tnic->procfs.dbg_counters_file->owner = THIS_MODULE;
	tnic->procfs.dbg_counters_file->read_proc = proc_read_dbg_counters;
	tnic->procfs.dbg_counters_file->write_proc = NULL;
	tnic->procfs.dbg_counters_file->data = tnic;

	/*
	 *
	 */
	tnic->procfs.stats_file = create_proc_entry("stats", 0644,
						    tnic->procfs.base_dir);
	if (tnic->procfs.stats_file == NULL) {
		goto error;
	}
	tnic->procfs.stats_file->owner = THIS_MODULE;
	tnic->procfs.stats_file->read_proc = proc_read_stats;
	tnic->procfs.stats_file->write_proc = proc_write_stats;
	tnic->procfs.stats_file->data = tnic;

	return (0);

  error:
	nx_tnic_adapter_destroy_procfs(tnic, adapter_dir);
	return (-ENOMEM);
}

/*
 *
 */
void nx_tnic_adapter_destroy_procfs(nx_tnic_adapter_t *tnic,
				    struct proc_dir_entry *adapter_dir)
{
	char tnic_dir[20];
	if (adapter_dir == NULL) {
		nx_print("Specified adapter_dir is NULL using pegnet_dir\n");
		adapter_dir = pegnet_dir;
	}

	if (tnic->procfs.max_tcp_conn_file) {
		remove_proc_entry("max_offloaded_tcp_conn",
				  tnic->procfs.base_dir);
		tnic->procfs.max_tcp_conn_file = NULL;
	}

	if (tnic->procfs.dbg_counters_file) {
		remove_proc_entry("dbg_counters", tnic->procfs.base_dir);
		tnic->procfs.dbg_counters_file = NULL;
	}

	if (tnic->procfs.stats_file) {
		remove_proc_entry("stats", tnic->procfs.base_dir);
		tnic->procfs.stats_file = NULL;
	}

	if (tnic->procfs.base_dir) {
//		remove_proc_entry("tnic", adapter_dir);
                sprintf(tnic_dir,"tnic_%d",atomic_read(&tnic->dev_id));
                remove_proc_entry(tnic_dir, adapter_dir);
		tnic->procfs.base_dir = NULL;
	}
}

/*
 *
 */
static int __init unm_init(void)
{
        int     rv;


        UNM_HOST_PNIC_STATS_INIT();
        init_card_response_handlers();

	if ((rv = nx_tnic_procfs_create_parent()) != 0) {
		printk("unm_init: Procfs initialization failed %d\n", rv);
		return rv;
	}

	pstack_init();

#if 0
	{
		pegnet_sock_id_t	u;

		nx_print1("sizeof pegnet_sock_id_t            = %u\n",
			  sizeof (pegnet_sock_id_t));
		nx_print1("sizeof hostRxBuffer_t              = %u\n",
			  sizeof (hostRxBuffer_t));
		nx_print1("sizeof netxen_host_page_t          = %u\n",
			  sizeof (netxen_host_page_t));
		nx_print1("sizeof recv_dma                    = %u\n",
			  sizeof (u.recv_dma));
		nx_print1("sizeof skb_shared_info             = %u\n",
			  sizeof (struct skb_shared_info));
		nx_print1("sizeof sk_buff                     = %u\n",
			  sizeof (struct sk_buff));
		nx_print1("sizeof SMP_CACHE_BYTES             = %u\n",
			  SMP_CACHE_BYTES);
	}
#endif

	/* nx_init_rx_global_buffers(); */

#ifdef CONFIG_UNM_NETFILTER
	unm_iptable_init();
#endif

        printk("ENABLING LINUX TOE\n") ;

	return 0;
}

void nx_cleanup_all_tnics(void);
static void __exit unm_done(void)
{
	/*
	 * First make sure all our scheduled tasks have completed.
	 */
	nx_print("In unm done");
	nx_cleanup_all_tnics();
	pstack_exit();

	flush_scheduled_work();

	nx_tnic_procfs_destroy_parent();
}

module_init(unm_init);
module_exit(unm_done);
/*** SYN-queue pruning implementation ***/

#define MAX_SIDS_PER_PRUNING_CYCLE 8
#define MAX_PRUNED_SIDS_BH         8
#define MAX_PRUNED_SIDS            64

static nx_kc_work_queue_t	g_synq_pruning_task;
static nx_synq_prun_requests_t g_prune_reqs;



static void nx_synq_pruning_worker(void *request_q);
static void nx_schedule_pruning(listener_group_t  *grp);
static listener_group_t *nx_get_pruning_request(nx_synq_prun_requests_t *q);

void nx_init_pruning_task(void)
{
	INIT_LIST_HEAD(&g_prune_reqs.sid_list);
	spin_lock_init(&g_prune_reqs.lock);

	NX_KC_INIT_WORK(&g_synq_pruning_task,
			nx_synq_pruning_worker, &g_prune_reqs);
	return;
}

/* sid spin lock should be held when we enter this function */
void nx_prune_synq(listener_group_t   *grp, unsigned int type)
{
	unm_msg_t                       msg;
	nx_send_immediate_conn_compl_t  *req =
		(nx_send_immediate_conn_compl_t *)&msg.body;

	unsigned int			i = 0;
	unsigned int			max = ((type == NX_SYNQ_PRUNETYPE_BH) ?
					       MAX_PRUNED_SIDS_BH :
					       MAX_PRUNED_SIDS);

	pegnet_sock_id_t                *child;
	struct list_head                *node = NULL, *tmp = NULL;
        nx_lopt_t			*lopt = grp->accept_passive.lopt;


	if (NX_SYNQ_PRUNETYPE_BH == type) {
		if (grp->accept_passive.pruning_active) {
			return;
		}
	}

	req->hdr.opcode = UNM_H2P_OPCODE_FORCE_3WAY_HSHAKE_COMPL;

	list_for_each_safe(node, tmp, &lopt->youngsters) {
		child = list_entry(node, pegnet_sock_id_t,
				accept_active.young_sid);

		UNM_MSGQ_HDR(msg.hdr,
			     child->accept_active.parent_sid->qid.major,
			     child->accept_active.parent_sid->qid.minor,
			     HOST_REQUEST);
		msg.hdr.dst_major 	= child->qid.major;
		msg.hdr.dst_minor 	= child->qid.minor;

		req->local_ip.v4        = child->lip_addr.v4;
		req->remote_ip.v4       = child->rip_addr.v4;
		req->local_port         = child->lport;
		req->remote_port        = child->rport;

		/* You can be young only once! */
		list_del_init(&child->accept_active.young_sid);

		if (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(child->adapter,child->portnum),
				     &msg, sizeof(*req))) {
			nx_print("Unable to send pruning request\n");
		} else {
			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_TOTAL_PRUNED);
		}

		if (++i == max) {
			break;
		}
	}

	if (NX_SYNQ_PRUNETYPE_BH == type) {
                grp->accept_passive.pruning_active = 1;
                nx_schedule_pruning(grp);
	}
	else {
		grp->accept_passive.pruning_active = 0;
	}

	return;
}

static void nx_schedule_pruning(listener_group_t  *grp)
{
	/* Update the task queue and request scheduling */

	/* this should not be needed here */
	INIT_LIST_HEAD(&grp->accept_passive.pruning);

	spin_lock(&g_prune_reqs.lock);
        list_add_tail(&grp->accept_passive.pruning,
                        &g_prune_reqs.sid_list);
        spin_unlock(&g_prune_reqs.lock);
        grp->accept_passive.pruning_active = 1;

	NX_KC_SCHEDULE_WORK(&g_synq_pruning_task);
	return;
}

static void nx_synq_pruning_worker(void *request_q)
{
	nx_synq_prun_requests_t *q = &g_prune_reqs;
        listener_group_t *next_grp = NULL;
        unsigned int    i = MAX_SIDS_PER_PRUNING_CYCLE;

        while (i-- && (next_grp = nx_get_pruning_request(q))) {
                if (next_grp->released) {
                        i++;
                        continue;
                }

                spin_lock_bh(&next_grp->lock);
                nx_prune_synq(next_grp, NX_SYNQ_PRUNETYPE_NORMAL);
                spin_unlock_bh(&next_grp->lock);
        }

	spin_lock_bh(&q->lock);
	if (!list_empty(&q->sid_list)) {
		spin_unlock_bh(&q->lock);
		NX_KC_SCHEDULE_WORK(&g_synq_pruning_task);
	} else {
		spin_unlock_bh(&q->lock);
	}

	return;
}

listener_group_t *nx_get_pruning_request(nx_synq_prun_requests_t *q)
{
        listener_group_t *grp = NULL;
        struct list_head *node;

        spin_lock_bh(&q->lock);

        list_for_each(node, &q->sid_list) {
                grp = list_entry(node, listener_group_t,
                                accept_passive.pruning);

                list_del_init(&grp->accept_passive.pruning);
                break;
        }

        spin_unlock_bh(&q->lock);

        return grp;

}

#if defined(KGDB) && !defined(DEBUG)
void NOINLINE inline prefetch(const void *x) { }        /* Might be compiled without inlines */
#endif
