/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <unm.h>
#include <unm_xport.h>
#include <linux/if_pppox.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if.h>
#include <unm_inc.h>
#include <pegnet_inc.h>
#include <asm/ioctls.h>
#include <linux/route.h>

extern int unm_ipt_getsockopt(pegnet_sock_id_t *sid, int level, int optname,
                              char *optval, int *optlen);
extern int unm_ipt_setsockopt(pegnet_sock_id_t *sid, int level, int optname,
                              char *optval, int optlen);

atomic_t	pegnet_init_done = ATOMIC_INIT(0);



int get_dma_buffer(pegnet_sock_id_t *sid, pegnet_buf_dir dir, void **buf)
{
	assert((pegnet_send_buf == dir) || (pegnet_sopt_buf == dir));

	if (down_interruptible(&(sid->dma[dir].lock))) return -ERESTARTSYS;

	if (sid->released) {
		up(&(sid->dma[dir].lock));
		return -ENOTSOCK;
	}

	sid->dma[dir].buffer = kmalloc(UNM_DMA_BUFFER_SIZE, GFP_KERNEL);

	if (sid->dma[dir].buffer == NULL) {
		up(&(sid->dma[dir].lock));
		return -ENOMEM;
	}

	sid->dma[dir].phys = virt_to_bus(sid->dma[dir].buffer);

	assert(virt_to_bus(sid->dma[dir].buffer) == sid->dma[dir].phys);
	*buf = sid->dma[dir].buffer;

	return 0;
}

void put_dma_buffer(pegnet_sock_id_t * sid, pegnet_buf_dir dir, void *buf)
{
	assert((pegnet_send_buf == dir) || (pegnet_sopt_buf == dir));
	assert(buf == sid->dma[dir].buffer);

	kfree(sid->dma[dir].buffer);
	sid->dma[dir].phys = 0;
	sid->dma[dir].buffer = NULL;

	up(&(sid->dma[dir].lock));
}

int smart_memcpy(void *dest, void *src, int size)
{
	int ret = 0;
    
	if ((unsigned long) dest < TASK_SIZE) {
		/* dest is in user space */
		//assert((unsigned long) src >= TASK_SIZE);
		ret = copy_to_user(dest, src, size);
#if defined(PEGNET_DEBUG_PRINT) && (PEGNET_DEBUG_LEVEL>1)
		debug_print_payload(pegnet_debug_mask, src, size);
#endif
	} else {
		/* dest is in kernel space */
		if ((unsigned long) src < TASK_SIZE) {
			/* src is in user space */
			ret = copy_from_user(dest, src, size);
#if defined(PEGNET_DEBUG_PRINT) && (PEGNET_DEBUG_LEVEL>1)
			debug_print_payload(pegnet_debug_mask, dest, size);
#endif
		} else {
			/* src is in kernel space */
			memcpy(dest, src, size);
		}
	}

	return ret;
}

/*
 *
 */
int unm_sendmsg(struct socket *sock, struct msghdr *m, size_t size)
{
        int                     rv;
        pegnet_sock_id_t        *sid;


        if (sock->type == SOCK_STREAM && sock->state != SS_CONNECTED) {
                return (-EPIPE);
        }

        sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
        if (sid->released) {
                return (-ENOTSOCK);
        }

        rv = host_sendmsg_fast(sid, m);
        return (rv);
    
}  /* unm_sendmsg */

/*
 *
 */
int unm_recvmsg(struct socket *sock, struct msghdr *m, int total_len,
                int flags)
{
        int                     rv;
        pegnet_sock_id_t        *sid;

        sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);

//// Start debug.
//    printk( "Entering unm_recvmsg: sock=%p\n", sock ) ;
//// End debug.
        if (sock->type == SOCK_STREAM && sock->state != SS_CONNECTED &&
	    sock->state != SS_CONNECTING) {
                return (-EPIPE);
        }

        rv = host_recvmsg_fast(sid, m, total_len, flags, NULL);
        if (rv >= 0) {
                // This is always zero in case of tcp
                m->msg_namelen = 0;
        }

        /*
         * We assume intel arch, where I/O is to uncached memory
         * and thus doesn't have to be flushed.
         */
        return rv;
}  /* unm_recvmsg */


#define IPT_BASE_CTL	0x40
int unm_setsockopt(pegnet_sock_id_t *sid, int level, int optname, char *optval,
                   int optlen)
{
        int                     rv;
        void                    *buf=NULL;

#if defined(CONFIG_UNM_NETFILTER)
        if (optname >= IPT_BASE_CTL) 
                return (unm_ipt_setsockopt(sid, level, optname, optval,
                                           optlen));
#else
        if (optname >= IPT_BASE_CTL) return(-ENOPROTOOPT);
#endif
        rv = get_dma_buffer(sid, pegnet_sopt_buf, &buf);
        if (0 != rv) return rv;
        if (optlen){
                if(smart_memcpy(buf, optval, optlen)){
                        put_dma_buffer(sid, pegnet_sopt_buf, buf);
                        return -EFAULT;
                }
        }
        rv = host_setsockopt(sid, level, optname, 
                             (unsigned long)virt_to_bus(buf), optlen, 0);
        put_dma_buffer(sid, pegnet_sopt_buf, buf);
        return(rv);
}

/*
 *
 */
int unm_getsockopt(pegnet_sock_id_t *sid, int level, int optname, char *optval,
                   int *optlen)
{
        int                     rv;
        void                    *buf/* , *data */;
        int                     len/* , alen */;
        uint64_t                addr1/* , addr2 */;
        struct pci_dev *pdev;


#if defined(CONFIG_UNM_NETFILTER)
        if (optname >= IPT_BASE_CTL) 
                return(unm_ipt_getsockopt(sid, level, optname, optval,
                                          optlen));
#else
        if (optname >= IPT_BASE_CTL) return(-ENOPROTOOPT);
#endif
        rv = get_dma_buffer(sid, pegnet_sopt_buf, &buf);
        if (0 != rv) return rv;
        if(copy_from_user(&len, optlen, sizeof(int))){
                put_dma_buffer(sid, pegnet_sopt_buf, buf);
                return -EFAULT;
        }
        if (len) {
                if(copy_from_user(buf, optval, len)){
                        put_dma_buffer(sid, pegnet_sopt_buf, buf);
                        return -EFAULT;
                }
        }
        /* alen = ((len +8) >> 3) << 3; */
        /* data = (void *)(buf+ alen); */
        /* memcpy(data, &len, sizeof(int)); */
        pdev = get_pci_dev_from_sid(sid);
        addr1 = ((len)?pci_map_single(pdev,buf, len, 0):0);
        if(addr1)addr1 = PEGNET_MARK_NONLOCAL(addr1);
        /* addr2 = pci_map_single(NULL, data, sizeof(int),0); */
        /* if(addr2)addr2 = PEGNET_MARK_NONLOCAL(addr2); */
        /* rv = host_getsockopt(sid, level, optname, addr1, addr2); */
        rv = host_getsockopt(sid, level, optname, addr1, &len);
        if(rv) goto finish;
        /* len = *(int *)data; */

        if( copy_to_user(optlen, &len/* data */, sizeof(int))){
                put_dma_buffer(sid, pegnet_sopt_buf, buf);
                return -EFAULT;
        }
        if(copy_to_user(optval, buf, len)){
                put_dma_buffer(sid, pegnet_sopt_buf, buf);
                return -EFAULT;
        }
  finish:    
        put_dma_buffer(sid, pegnet_sopt_buf, buf);
        return(rv);
}

/*
 *
 */
int unm_ioctl(pegnet_sock_id_t *sid, unsigned int cmd, unsigned long arg)
{
        int                     rv;
        void                    *kbuf;
        size_t                  len;
        unm_slosh               op;

        switch (cmd) {
        case SIOCSIFADDR:           /* set pa address               */
        case SIOCSIFLINK:           /* set iface channel            */
        case SIOCSIFFLAGS:          /* set flags                    */
        case SIOCSIFMEM:            /* set memory address (bsd)     */
        case SIOCSIFMTU:            /* set mtu size                 */
        case SIOCSIFENCAP:          /* set encapsulations           */
        case SIOCSIFSLAVE:          /* set driver slaving support   */
        case SIOCADDMULTI:          /* set multicast address lists  */
        case SIOCDELMULTI:          /* del multicast address lists  */
        case SIOCSIFPFLAGS:         /* set extended flags set       */
        case SIOCSIFBR:             /* set bridging options         */
        case SIOCETHTOOL:           /* ethtool interface            */
        case SIOCGIFFLAGS:          /* get flags                    */
        case SIOCGIFADDR:           /* get pa address               */
        case SIOCGIFDSTADDR:        /* get remote pa address        */
        case SIOCSIFDSTADDR:        /* set remote pa address        */
        case SIOCGIFBRDADDR:        /* get broadcast pa address     */
        case SIOCGIFNETMASK:        /* get network pa mask          */
        case SIOCGIFMETRIC:         /* get metric                   */
        case SIOCGIFMEM:            /* get memory address (bsd)     */
        case SIOCGIFMTU:            /* get mtu size                 */
        case SIOCGIFHWADDR:         /* get hardware address         */
        case SIOCGIFSLAVE:          /* driver slaving support       */
        case SIOCGIFINDEX:          /* name -> if_index mapping     */
        case SIOCGIFPFLAGS:         /* get extended flags set       */
        case SIOCGIFCOUNT:          /* get number of devices */
        case SIOCGIFTXQLEN:         /* get the tx queue length      */
        case SIOCSIFTXQLEN:         /* set the tx queue length      */
        case SIOCSIFMAP:            /* set device parameters        */
        case SIOCGIFMAP:            /* get device parameters        */
        case SIOCADDRT:             /* add routing table entry  */
        case SIOCDELRT:             /* delete routing table entry   */
        case SIOCRTMSG:             /* call to routing system   */
		return (0);
	}


        //printk("%s: cmd = 0x%x, arg:0x%p\n", __FUNCTION__, cmd, (void *)arg);

        rv = get_dma_buffer(sid, pegnet_send_buf, &kbuf);
        if (rv != 0) return rv;


        switch (cmd)
        {
/****************************************************************
 ** SET OPERATIONS
 *****************************************************************/
/* Routing table calls                                          */
        case SIOCGSTAMP:
                len = sizeof(struct timeval);
                op = slosh_nocopy;
                break;
        
        case SIOCSIFNAME:           /* set interface name */
        case SIOCSIFHWADDR:         /* set hardware address         */
                /* In integrated build, the nic driver calls these ioctls from 
                 * kernel space. So a memcpy is required instead of copy_from_user.
                 */
#ifdef PEGNET_NIC
                memcpy(kbuf, (void*)arg, sizeof(struct ifreq));
                rv = host_ioctl(sid, cmd, virt_to_bus(kbuf), 0);
                goto ioctl_finish;
#endif

/* Socket configuration controls                                */
        case SIOCSIFBRDADDR:        /* set broadcast pa address     */
        case SIOCSIFNETMASK:        /* set network pa mask          */
        case SIOCSIFMETRIC:         /* set metric                   */
        case SIOCDIFADDR:           /* delete pa address            */
        case SIOCSIFHWBROADCAST:    /* set hardware broadcast addr  */

                op = slosh_docopy;
                len = sizeof(struct ifreq);
                break;

/* ARP cache control calls.                                     */
/*  0x8950 - 0x8952             ** obsolete calls, don't re-use */
        case SIOCDARP:              /* delete arp table entry       */
        case SIOCSARP:              /* set arp table entry          */
                op = slosh_docopy;
                len = sizeof(struct arpreq);
                break;


/****************************************************************
 ** GET OPERATIONS
 *****************************************************************/
/* Socket configuration controls                                */

#if 0
        case SIOCPROTOPRIVATE + 3:
                /* Useful for debugging. */
                rv = sid->qid.word;
                goto ioctl_finish;
#endif
        case SIOCINQ:
	{
                rv=host_recv_available(sid,arg);
		goto ioctl_finish;
	}	
				
#if 0
                /****************************************************************
                 ** NOT SUPPORTED
                 *****************************************************************/
                /* Linux-specific socket ioctls */
        case SIOCOUTQ:
#endif

        case SIOCATMARK:
                op = slosh_nocopy;
                len = sizeof(int);
                break;

        case SIOCSFASYNC:
                memcpy(kbuf, (char *)arg, sizeof(int));
                rv = host_ioctl(sid, cmd, virt_to_bus(kbuf), 1);
                goto ioctl_finish;

        default:
                rv = -EOPNOTSUPP;  /* command not supported */
                goto ioctl_finish;
        }

        /*
        ** For ioctls we assume that the data is an IN/OUT parameter.
        ** Therefore, we always copy the data to the card and then
        ** optionally copy it back to the user if indicated.
        */
#if defined(PEGNET_DEBUG_PRINT) && (PEGNET_DEBUG_LEVEL>4)
        printk("AF_UNM: ioctl(#0x%x) copying %u bytes\n", cmd, len);
#endif

        rv = copy_from_user(kbuf, (void*)arg, len);
        if (rv) {
                rv = -EFAULT;
                goto ioctl_finish;
        }
        rv = host_ioctl(sid, cmd, virt_to_bus(kbuf), 0);
        if (slosh_nocopy == op) {
                if (copy_to_user((void*)arg, kbuf, len)) {
                        rv = -EFAULT;
                }
        }

  ioctl_finish:
        put_dma_buffer(sid, pegnet_send_buf, kbuf);

        return rv;
}


