/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Copyright and CVS version tags.
 */

/*
 * This has the functions for managing a host table of flows. Currently only
 * TCP.
 */
#include "flow_table.h"
#include "unm_host_pnic_types.h"

#include <asm/string.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>




/*
 * This function is used to allocate space for the flow table. It allocates the
 * maximum IDs and flow pointers. Does not allocate the flow itself.
 *
 * Parameters:
 *	None.
 *
 * Return:
 *	0		- If allocated successfully.
 *	ENOMEM		- If out of memory.
 */
int host_flowtbl_init(flow_table_t *tcp_tbl)
{
	int	i;
	int	alloc_size;

        tcp_tbl->lock = SPIN_LOCK_UNLOCKED;

	alloc_size = sizeof (flow_tbl_entry_t) * FLOW_TABLE_SIZE;
//	tcp_tbl->flows = (flow_tbl_entry_t *)kmalloc(alloc_size, GFP_KERNEL);
	tcp_tbl->flows = (flow_tbl_entry_t *)vmalloc(alloc_size);
	if (tcp_tbl->flows == NULL) {
		printk("host_flowtbl_init: Flows alloc failed\n");
		return (-ENOMEM);
	}

/* 	printk("#####################host_flowtbl_init: 1 = %u\n", */
/* 	       (sizeof (flow_tbl_entry_t) * UNM_HOST_MAX_TCP_CONNECTIONS)); */

	tcp_tbl->free_list.ids = (flow_id_t *)kmalloc((sizeof (flow_id_t) *
						      FLOW_TABLE_SIZE),
						     GFP_KERNEL);

/* 	printk("#####################host_flowtbl_init: 2 = %u\n", */
/* 	       (sizeof (flow_id_t) * UNM_HOST_MAX_TCP_CONNECTIONS)); */

	if (tcp_tbl->free_list.ids == NULL) {
		printk("host_flowtbl_init: IDs alloc failed\n");
		vfree(tcp_tbl->flows);
		return (-ENOMEM);
	}

	tcp_tbl->connID_table = (int32_t *)kmalloc((sizeof(int32_t) *
						   FLOW_TABLE_SIZE),
						  GFP_KERNEL);

	if (!tcp_tbl->connID_table) {
		printk("host_flowtbl_init: connID_table alloc failed\n");
		kfree(tcp_tbl->free_list.ids);
		vfree(tcp_tbl->flows);
		return (-ENOMEM);
	}

	tcp_tbl->max = FLOW_TABLE_SIZE;
	tcp_tbl->free_list.max = FLOW_TABLE_SIZE;
	tcp_tbl->free_list.tail = FLOW_TABLE_SIZE - 1;


	for (i = 0; i < FLOW_TABLE_SIZE; i++) {
		tcp_tbl->flows[i].id.idx = UNM_INVALID_FLOW_ID;
		tcp_tbl->flows[i].flow = kmalloc(sizeof (pegnet_sock_id_t),
						GFP_KERNEL);
		if (!tcp_tbl->flows[i].flow) {
			printk("%s: Unable to allocate flows\n", __FUNCTION__);
			goto free_and_out;
		}
		tcp_tbl->flows[i].gnrtn = 0;
		atomic_set(&tcp_tbl->flows[i].refcnt, 0);

		tcp_tbl->free_list.ids[i].idx = i;

		tcp_tbl->connID_table[i] = UNM_INVALID_CONNID;

	}

        printk("host_flowtbl_init: Initialized\n");

	/* Add everything to the free list */

	return (0);

free_and_out:
	for (i = 0; i < FLOW_TABLE_SIZE; i++) {
		if (tcp_tbl->flows[i].flow)
			kfree(tcp_tbl->flows[i].flow);
	}
	kfree(tcp_tbl->free_list.ids);
	kfree(tcp_tbl->connID_table);
	vfree(tcp_tbl->flows);
	return -ENOMEM;
}

/*
 * This function is used to free space allocated for the flow table.
 *
 * Parameters:
 *	None.
 *
 * Return:
 *	None.
 */
void host_flowtbl_destroy(flow_table_t *tcp_tbl,void (*free_flow_cb)(flow_id_t id, void *flow))
{
	int	i;
	/*
	 * WARNING: Need to check whether following lock needs to be held
	 * with spin_lock_irq(). If that is required, problem is that
	 * unm_proc_twkill() runs into problem when doing spin_unlock_bh()
	 * since bh should not be enabled at elevated irq.
	 */
	spin_lock(&tcp_tbl->lock);

	for (i = 0; i < FLOW_TABLE_SIZE; i++) {
		if (tcp_tbl->flows[i].flow != NULL &&
		    tcp_tbl->flows[i].id.idx != UNM_INVALID_FLOW_ID) {

			free_flow_cb(tcp_tbl->flows[i].id,
				     tcp_tbl->flows[i].flow);
		}
		if (tcp_tbl->flows[i].flow)
			kfree(tcp_tbl->flows[i].flow);
	}

	spin_unlock(&tcp_tbl->lock);

	kfree(tcp_tbl->connID_table);
	kfree(tcp_tbl->free_list.ids);
	vfree(tcp_tbl->flows);

	return;
}

/*
 * This function is used to allocate a flow from the flow table.
 * Returns a flow if allocated else returns NULL.
 *
 * Parameters:
 *	flow_idx	- The pointer to fill in the index of allocated
 *			  flow
 *
 * Return:
 *	The flow that was allocated.
 *	NULL	- If allocation fails.
 */
void *host_flowtbl_allocate_flow(flow_table_t *tcp_tbl,uint32_t *flow_idx)
{
	flow_id_t	flow_id;
        flow_free_ids_t *free_list;
	void		*flow = NULL;
	unsigned int 	allocated_index;

	/* make sure we don't exceed any limits */
	if (check_card_resources(tcp_tbl)) {
		return NULL;
	}

	spin_lock_bh(&tcp_tbl->lock);

        free_list = &tcp_tbl->free_list;

	if (free_list->head != free_list->tail) {
		/*
		 * First pick the ID. Don't let generation hit 0. That will
		 * cause problems on the Pegasus. A context of 0 is a special
                 * case in the Peg code.
		 */
		flow_id.idx = allocated_index =
			free_list->ids[free_list->head].idx;

		tcp_tbl->flows[allocated_index].gnrtn++;

		if (!tcp_tbl->flows[allocated_index].gnrtn) {
			tcp_tbl->flows[allocated_index].gnrtn++;
		}

		free_list->head++;
		free_list->head %= free_list->max;

		/*
		 * Now add the flow and exit.
		 */
		//tcp_tbl->flows[flow_id.bits.idx].flow = flow;
		tcp_tbl->flows[allocated_index].id.idx = flow_id.idx;
		flow = tcp_tbl->flows[allocated_index].flow;
		atomic_set(&tcp_tbl->flows[allocated_index].refcnt, 1);

		tcp_tbl->nr_flows++;

/*                 printk("%s: id = %u, flow = %p\n", */
/*                        __FUNCTION__, flow_id.id, flow); */
	}

	spin_unlock_bh(&tcp_tbl->lock);

	if (flow_idx) {
		*flow_idx = flow_id.idx;
	} else {
		printk("%s: flow_idx pointer is NULL\n", __FUNCTION__);
	}

	return (flow);
}

/*
 * This function is used to free a flow from the table.
 *
 * Parameters:
 *	connId		- The connId of the flow to free.
 *
 * Return:
 *	0		- If released successfully.
 *	-1		- If the flow is not found.
 */
int host_flowtbl_free_flow(flow_table_t *tcp_tbl,uint32_t connId)
{
	int		rv = -1;
	flow_id_t	flow_id = {idx:0};
	uint32_t	old_tail;
	uint32_t	flow_idx;
        flow_free_ids_t *free_list;
        /* printk("%s: id = %u\n", __FUNCTION__, id); */

	if (UNM_INVALID_CONNID == connId) {
		printk("%s: Attempt to free invalid connID\n",
				__FUNCTION__);
		return rv;
	}

	spin_lock_bh(&tcp_tbl->lock);

	flow_idx = tcp_tbl->connID_table[connId];

	if (UNM_INVALID_FLOW_ID == flow_idx) {
		printk("%s: Attempt to free invalid flow-id\n",
				__FUNCTION__);
		spin_unlock_bh(&tcp_tbl->lock);
		return rv;
	}

	flow_id.idx = flow_idx;

        free_list = &tcp_tbl->free_list;

	if (flow_id.idx < tcp_tbl->max) {

		if (!atomic_dec_and_test(&tcp_tbl->flows[flow_idx].refcnt)) {
			printk("%s: Reference not zero\n", __FUNCTION__);
		}

		/*
		 * Clear the table.
		 */
		tcp_tbl->flows[flow_idx].id.idx = UNM_INVALID_FLOW_ID;
		tcp_tbl->connID_table[connId] = UNM_INVALID_CONNID;

		//tcp_tbl->flows[flow_id.bits.idx].flow = NULL;

		/*
		 * Add to tail of free list.
		 */
		old_tail = free_list->tail;
		free_list->tail++;
		free_list->tail %= free_list->max;

		if (free_list->tail != free_list->head) {
			free_list->ids[free_list->tail].idx = flow_idx;
			rv = 0;
		} else {
			printk("host_flowtbl_free_tcp_flow: tail reached head "
			       "this should never happen\n");
			tcp_tbl->free_list.tail = old_tail;
		}
		tcp_tbl->nr_flows--;
	}

	spin_unlock_bh(&tcp_tbl->lock);

	if (rv != 0) {
		printk("host_flowtbl_free_tcp_flow: Flow table free flow "
		       "failed. ID %u\n", connId);
	}
	return (rv);
}

int host_flowtbl_free_flow_by_flowidx(flow_table_t *tcp_tbl,uint32_t flow_idx)
{
        flow_free_ids_t *free_list;
	uint32_t	old_tail;
	int 		rv = -1;

        free_list = &tcp_tbl->free_list;

	spin_lock_bh(&tcp_tbl->lock);

	if (flow_idx < tcp_tbl->max) {

		if (!atomic_dec_and_test(&tcp_tbl->flows[flow_idx].refcnt)) {
			printk("%s: Reference not zero\n", __FUNCTION__);
		}

		/*
		 * Clear the table.
		 */
		tcp_tbl->flows[flow_idx].id.idx = UNM_INVALID_FLOW_ID;

		//tcp_tbl->flows[flow_id.bits.idx].flow = NULL;

		/*
		 * Add to tail of free list.
		 */
		old_tail = tcp_tbl->free_list.tail;
		free_list->tail++;
		free_list->tail %= free_list->max;

		if (free_list->tail != free_list->head) {
			free_list->ids[free_list->tail].idx = flow_idx;
			rv = 0;
		} else {
			printk("host_flowtbl_free_tcp_flow: tail reached head "
			       "this should never happen\n");
			tcp_tbl->free_list.tail = old_tail;
		}
		tcp_tbl->nr_flows--;
	}

	spin_unlock_bh(&tcp_tbl->lock);
	return rv;

}

/*
 * This function is used to get the table index of a connection ID.
 *
 * Parameters:
 *	id		- The ID to convert to index.
 *
 * Return:
 *	The index into the table.
 */
uint16_t host_flowtbl_get_index(flow_table_t *tcp_tbl,uint32_t id)
{
	/* there should be no race in this read access */
	return ((uint16_t)tcp_tbl->connID_table[id]);
}

/*
 * This function is used to get a flow from the flow table.
 *
 * Parameters:
 *	idx		- The index of the flow to get.
 *
 * Return:
 *	The flow that was found.
 *	NULL		- If there is no flow for the index.
 */
void *host_flowtbl_get_flow_by_idx(flow_table_t *tcp_tbl,uint16_t idx)
{
        void	*flow = NULL;
        /* printk("%s: id = %u\n", __FUNCTION__, id); */

	spin_lock_bh(&tcp_tbl->lock);

	if (idx < tcp_tbl->max &&
	    tcp_tbl->flows[idx].id.idx != UNM_INVALID_FLOW_ID) {

		flow = tcp_tbl->flows[idx].flow;
		atomic_inc(&tcp_tbl->flows[idx].refcnt);
	}

	spin_unlock_bh(&tcp_tbl->lock);

        /* printk("%s: id = %u, flow = %p\n", __FUNCTION__, id, flow); */

	return (flow);
}

/*
 * This function is used to get a flow from the flow table.
 *
 * Parameters:
 *	id		- The ID of the flow to get.
 *
 * Return:
 *	The flow that was found.
 *	NULL		- If there is no flow for the ID.
 */
void *host_flowtbl_get_flow(flow_table_t *tcp_tbl,uint32_t connId)
{
        flow_id_t	flow_id;
        void		*flow = NULL;
        /* printk("%s: id = %u\n", __FUNCTION__, id); */

	if (UNM_INVALID_CONNID == connId) {
		printk("%s: Attempt to retrive flow on invalid connId\n",
				__FUNCTION__);
		return flow;
	}

	spin_lock_bh(&tcp_tbl->lock);

	flow_id.idx = tcp_tbl->connID_table[connId];

	if (flow_id.idx < tcp_tbl->max &&
	    tcp_tbl->flows[flow_id.idx].id.idx == flow_id.idx) {

		flow = tcp_tbl->flows[flow_id.idx].flow;
		atomic_inc(&tcp_tbl->flows[flow_id.idx].refcnt);
	}
	spin_unlock_bh(&tcp_tbl->lock);

        /* printk("%s: id = %u, flow = %p\n", __FUNCTION__, id, flow); */

	return (flow);
}

void host_flowtbl_get_reference(flow_table_t *tcp_tbl,uint32_t id)
{
        flow_id_t	flow_id;
	if (UNM_INVALID_CONNID == id) {
		printk("%s: Attempt to put reference on invalid connId\n",
				__FUNCTION__);
		return;
	}

	spin_lock_bh(&tcp_tbl->lock);

	flow_id.idx = tcp_tbl->connID_table[id];

	if (flow_id.idx < tcp_tbl->max &&
	    tcp_tbl->flows[flow_id.idx].id.idx == flow_id.idx) {

		atomic_inc(&tcp_tbl->flows[flow_id.idx].refcnt);
	}

	spin_unlock_bh(&tcp_tbl->lock);

	return;
}
/*
 * This function is used to release reference to a flow from the flow table.
 *
 * Parameters:
 *	id		- The ID of the flow to release a reference
 *
 * Return:
 */
void host_flowtbl_put_reference(flow_table_t *tcp_tbl,uint32_t id)
{
        flow_id_t	flow_id;
	if (UNM_INVALID_CONNID == id) {
		printk("%s: Attempt to put reference on invalid connId\n",
				__FUNCTION__);
		return;
	}

	spin_lock_bh(&tcp_tbl->lock);

	flow_id.idx = tcp_tbl->connID_table[id];

	if (flow_id.idx < tcp_tbl->max &&
	    tcp_tbl->flows[flow_id.idx].id.idx == flow_id.idx) {

		atomic_dec(&tcp_tbl->flows[flow_id.idx].refcnt);
	}

	spin_unlock_bh(&tcp_tbl->lock);

	return;
}

int host_flowtbl_get_nr_flows(flow_table_t *tcp_tbl)
{
	int flows;

	spin_lock_bh(&tcp_tbl->lock);
	flows = tcp_tbl->nr_flows;
	spin_unlock_bh(&tcp_tbl->lock);

	return flows;
}

void host_update_connID_table(flow_table_t *tcp_tbl,
				struct pegnet_sock_id_s *sid, uint32_t connId)
{
	if (connId > FLOW_TABLE_SIZE) {
		nx_print("connID %x allocated by card cannot be handled\n",
			 connId);
		return;
	}

	spin_lock_bh(&tcp_tbl->lock);

	if (tcp_tbl->connID_table[connId] != UNM_INVALID_CONNID) {
		if (sid->flow_id != tcp_tbl->connID_table[connId]) {
			nx_print("WARNING: connID[%d] already mapped to "
				 "flow[%d], sid[%p], remapping to sid[%p], "
				 "flow[%d]\n",
				 connId, tcp_tbl->connID_table[connId],
				 tcp_tbl->flows[tcp_tbl->connID_table[connId]].flow,
				 sid, sid->flow_id);
		}
	}

	tcp_tbl->connID_table[connId] = sid->flow_id;
	spin_unlock_bh(&tcp_tbl->lock);

	return;
}

void host_reset_connID_table(flow_table_t *tcp_tbl,
				struct pegnet_sock_id_s *sid,uint32_t connId)
{
        if (connId > FLOW_TABLE_SIZE) {
                nx_print("connID %x allocated is out of range\n",
                         connId);
                return;
        }
	spin_lock_bh(&tcp_tbl->lock);
	if (tcp_tbl->connID_table[connId] != UNM_INVALID_CONNID) {
		if (sid->flow_id != tcp_tbl->connID_table[connId]) {
			nx_print("WARNING: connID[%d] already mapped to "
				 "flow[%d], sid[%p], not matches to sid[%p], "
				 "flow[%d]\n",
				 connId, tcp_tbl->connID_table[connId],
				 tcp_tbl->flows[tcp_tbl->connID_table[connId]].flow,
				 sid, sid->flow_id);
		}
	}
	else {
		nx_print("Warning:release called on Invalid connID=%d\n",connId);
	}
	tcp_tbl->connID_table[connId] = UNM_INVALID_CONNID;
	sid->connID = UNM_INVALID_CONNID;
	spin_unlock_bh(&tcp_tbl->lock);
}

pegnet_sock_id_t *host_map_tmpctx_2_sid(flow_table_t *tcp_tbl,uint32_t tmp_ctx)
{
	pegnet_sock_id_t	*sid = NULL;
        /* printk("%s: id = %u\n", __FUNCTION__, id); */

        spin_lock_bh(&tcp_tbl->lock);

        if (tmp_ctx < tcp_tbl->max &&
            tcp_tbl->flows[tmp_ctx].id.idx != UNM_INVALID_FLOW_ID) {

                sid = (pegnet_sock_id_t *)(tcp_tbl->flows[tmp_ctx].flow);
        }

        spin_unlock_bh(&tcp_tbl->lock);
	return sid;
}

