/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 *
 * 20 May 2004 JR
 *     - IP Table structure mapping 
 */
#include <unm_inc.h>
#include <pegnet_inc.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <net/sock.h>          /* To deref sock->sk */
#include <linux/skbuff.h>
#include <asm/uaccess.h>

/*  Host's definitions */
#define ASSERT_READ_LOCK(head)      /* No locking needed */
#define ASSERT_WRITE_LOCK(head)
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/listhelp.h>
/*  Card's definitions */
#include <ipt_unm.h>
#include "unm_host_pnic_types.h"

#define IPT_LEN_ALIGN(ptr, ofs, s) ((void *)(ptr) + (((ofs) + sizeof(s) -1) & ~(sizeof(s)-1)))

#define IPT_UNM_ALIGN_OFFSET(l,o) (( (l + sizeof(uint64_t)- 1) & ~(sizeof(uint64_t)-1)) + o)
#define IPT_ERR_DELTA (sizeof(struct ipt_error_target) - sizeof(struct ipt_standard_target))

#ifdef UNM_HWBUG_5_WORKAROUND
# define IPT_HOSTBUFF_SIZE	((9*1024) - UNM_HWBUG_5_WORKAROUND_MIN_DMA_BYTES)
#else
# define IPT_HOSTBUFF_SIZE	(9*1024)
#endif

static ipt_unm_gateway_t ipt_sentry;
static int enable_gateway = 0;
extern int get_dma_buffer(pegnet_sock_id_t * sid, pegnet_buf_dir dir, void **buf);
extern void put_dma_buffer(pegnet_sock_id_t * sid, pegnet_buf_dir dir, void *buf);
extern void smart_memcpy(void *dest, void *src, int size);

extern int host_ipt_setsockopt( pegnet_sock_id_t* sock, int level, 
                                int optname, uint64_t optval, int optlen);

extern int host_ipt_getsockopt( pegnet_sock_id_t* sock, int level,
                         int optname, uint64_t optval, int *optlen);

extern void print_ipt_entry(struct ipt_entry *e);
extern void print_ipt_getinfo(struct ipt_getinfo *info);
extern void print_ipt_replace(struct ipt_replace *info);

static int check_entry(struct ipt_entry *e, const char *name, unsigned int size);

static void ipt_update_sentry_with_getinfo(struct ipt_getinfo *info, int side)
{
    struct ipt_getinfo_unm *gw;
    if (side == IPT_UNM_HOST) gw = &ipt_sentry.host;
    else gw = &ipt_sentry.card;

    memcpy(gw, info, sizeof(struct ipt_getinfo));
    ipt_sentry.flags |= side;
}

static void ipt_update_sentry_with_replace(void *info, int side)
{
    struct ipt_getinfo *gw = (struct ipt_getinfo *)info;
    ipt_update_sentry_with_getinfo(gw, side);
    ipt_sentry.flags &= (~IPT_UNM_PASS);
 
}

static int ipt_patch_getinfo(struct ipt_getinfo_trailer *trailer,
    struct ipt_getinfo *info)
{
    int hook;

    /* Update before patch */
    ipt_update_sentry_with_getinfo(info, IPT_UNM_CARD); 
#ifdef	CONFIG_UNM_NETFILTER_DEBUG
    print_ipt_trailer(trailer);
#endif
    /* If gateway cache is enabled, and table index is fresh, serve */
    if ((ipt_sentry.flags & IPT_UNM_HOST) && enable_gateway) {
        memcpy(info->hook_entry, ipt_sentry.host.hook_entry, 
                sizeof(uint32_t)*NF_IP_NUMHOOKS);
        memcpy(info->underflow, ipt_sentry.host.underflow, 
                sizeof(uint32_t)*NF_IP_NUMHOOKS);
        goto update_size;
    }
    for (hook = 0; hook < NF_IP_NUMHOOKS; hook++) {
        if (info->hook_entry[hook]) {
            info->hook_entry[hook] -= (trailer->hook_count[hook]* (IPT_HOOK_TRANSLATE));
            //info->hook_entry[hook] -= trailer->match_align[hook];
        }
        if (info->underflow[hook]){
            info->underflow[hook] -= (trailer->underflow_count[hook]*(IPT_HOOK_TRANSLATE));
            //info->underflow[hook] -= trailer->match_align[hook];
        }
    }
update_size:
    info->size -= (info->num_entries * (IPT_HOOK_TRANSLATE));
    info->size += IPT_ERR_DELTA; 
    ipt_update_sentry_with_getinfo(info, IPT_UNM_HOST); 
    return(sizeof(struct ipt_getinfo));
}

uint16_t ipt_normalize(uint16_t val, int type)
{
    if (!val) return(val);
    switch(type){
    case IPT_STDTARGET_TO_32_HOST:
        return(sizeof(struct ipt_standard_target));
    case IPT_ERRTARGET_TO_32_HOST:
        return(sizeof(struct ipt_error_target));
    case IPT_STDTARGET_TO_64_PEG:
        return(IPT_ALIGN(sizeof(struct ipt_standard_target_unm)));
    case IPT_ERRTARGET_TO_64_PEG:
        return(IPT_ALIGN(sizeof(struct ipt_error_target_unm)));
    default :
        return(val);
    }
}
                                                                                                          
static int
copy_entries_from_host(const char *name, struct ipt_entry_unm *e,
             struct ipt_entry *ebar,
             uint32_t *hooks, uint32_t *underflow,
             unsigned int len, unsigned int entries )
{
    unsigned int off, num, koff=0;
    unsigned int set_hooks=0, set_underflow=0, hk, kroff=0;
    struct ipt_entry_match_unm *m;
    struct ipt_entry_match  *mbar;
    struct ipt_entry_target_unm *t=NULL;
    struct ipt_entry_target *tbar;
    void *kernelptr = (void *)e;
    void *userptr = (void *)ebar;
    int ret = 0;
   
         /* copy entry; matches + target , fix hooks*/
    for (off = 0,  num = 0;
         off < len; off += ebar->next_offset, num++){
        unsigned int i,j;
        ebar = (struct ipt_entry *)((void *)userptr + off);
        memcpy((void *)e, (void *)ebar, sizeof(struct ipt_entry));
        koff = sizeof(struct ipt_entry_unm);
        for (i = sizeof(struct ipt_entry),
            j = sizeof(struct ipt_entry_unm);
            i < ebar->target_offset;
            i += mbar->u.match_size, j += m->u.match_size) {

            m = (struct ipt_entry_match_unm *)((void *)e + j);
            mbar = (struct ipt_entry_match *)((void *)ebar + i);
            memcpy((void *)m, (void *)mbar, mbar->u.match_size);
            m->u.match_size = IPT_UNM_ALIGN_OFFSET(mbar->u.match_size, 0);
            koff += m->u.match_size;
 
       }
       e->target_offset = (uint16_t )koff;
       t = (struct ipt_entry_target_unm *)((void *)e + e->target_offset);
       tbar = (struct ipt_entry_target *) ((void *)ebar + ebar->target_offset);
       memcpy((void *)t, (void *)tbar, tbar->u.target_size);
       if (strcmp(t->u.user.name, IPT_ERROR_TARGET))
           t->u.target_size = ipt_normalize(tbar->u.target_size, 
                                        IPT_STDTARGET_TO_64_PEG);
       else 
           t->u.target_size = ipt_normalize(tbar->u.target_size,
                  IPT_ERRTARGET_TO_64_PEG );
       e->next_offset = koff + t->u.target_size;

       kroff += e->next_offset;  /* Cumulative offset */
       for (hk=0; hk < NF_IP_NUMHOOKS; ++hk){
           if (underflow[hk] == (off+ebar->next_offset)) {
               if(!(set_underflow & (1 << hk))) {
                   underflow[hk] = kroff;
                   set_underflow |= (1 << hk);
               }
           }
           if (hooks[hk] == (off+ebar->next_offset)) {
               if(!(set_hooks & (1 << hk))) {
                   hooks[hk] = kroff;
                   set_hooks |= (1 << hk);
               }
          }
      }

      if (ebar->next_offset == 0 || num >= entries) {
          off = len;
      }
#ifdef	CONFIG_UNM_NETFILTER_DEBUG
        print_ipt_entry(ebar);
#endif
     if (check_entry(ebar, name, ebar->next_offset)) return(-EINVAL);
      e = (void *)e + e->next_offset;
  }

  ret = (int)((void *)e - kernelptr);
  return ret;
}

static int copy_entries_to_host(struct ipt_entry *ebar,
       struct ipt_entry_unm *e, unsigned int total_size )
{
    unsigned int off, koff=0;
    struct ipt_entry_target_unm *t=NULL;
    struct ipt_entry_target *tbar=NULL;
    void *kernelptr = (void *)e;
    int ret = 0, tsize=0;


    /* copy matches + target and then fix counters and names */
    for (off = 0; off < total_size; off += e->next_offset){
        unsigned int i;
        struct ipt_entry_match_unm *m;
        struct ipt_entry_match  *mbar;

        e = (struct ipt_entry_unm *)((void *)kernelptr + off);
        ebar = (struct ipt_entry *)((void *)ebar + koff + tsize);
       
        koff = sizeof(struct ipt_entry);
        memcpy((void *)ebar, (void *)e, koff);

        for (i = sizeof(struct ipt_entry_unm);
            i < e->target_offset; i += m->u.match_size) {

            m = (struct ipt_entry_match_unm *)((void *)e + i);
            mbar = (struct ipt_entry_match *)((void *)ebar + koff);
            memcpy((void *)mbar, (void *)m, m->u.match_size);

            memcpy((void *)mbar + offsetof(struct ipt_entry_match, u.user.name),
                        m->u.user.name,
                        strlen(m->u.user.name)+1);
            koff += mbar->u.match_size;
        }
        ebar->target_offset = (uint16_t )koff;
        t = ipt_get_target_unm(e);
        tbar = (struct ipt_entry_target *) ((void *)ebar +koff);
        if (strcmp(t->u.user.name, IPT_ERROR_TARGET))
           tsize =  ipt_normalize(t->u.target_size, IPT_STDTARGET_TO_32_HOST);
        else
           tsize =  ipt_normalize(t->u.target_size, IPT_ERRTARGET_TO_32_HOST);

        memcpy((void *)tbar, (void *)t, tsize);
        tbar->u.target_size = tsize;

        ebar->next_offset = koff + tsize;
#ifdef	CONFIG_UNM_NETFILTER_DEBUG
        print_ipt_entry(ebar);
#endif
        ret += (koff + tsize);
    }

    return ret;
}

static int get_entries_from_unm(struct ipt_get_entries *host,
              struct ipt_get_entries_unm *unm)
{
    int ret;
    ret = copy_entries_to_host((void *) host + IPT_ENTRYHEADER,
             (void *)unm + IPT_ENTRYHEADER_UNM, unm->size);
    return(ret + IPT_ENTRYHEADER);
}

static int copy_iptreplace_from_host(void *user, void *card, unsigned int len)
{
    struct ipt_replace_unm *tmp=(struct ipt_replace_unm *)card;
    struct ipt_entry *host_entries;
    struct ipt_entry_unm *card_entries;
		int ret=-EINVAL;


    /* 
     * Copy ipt_replace 
     */
    memcpy(card, user, sizeof(struct ipt_replace_unm));
    card_entries = (struct ipt_entry_unm *)(card + sizeof(struct ipt_replace_unm));
    host_entries = (struct ipt_entry *)(user + sizeof(struct ipt_replace));
   
    ipt_update_sentry_with_replace(user, IPT_UNM_HOST);
#ifdef	CONFIG_UNM_NETFILTER_DEBUG
    print_ipt_replace((struct ipt_replace *)user);
#endif

    /*
     * Copy entries
     */
    ret = copy_entries_from_host(tmp->name, card_entries,
                host_entries, tmp->hook_entry, tmp->underflow,
                tmp->size, tmp->num_entries);
		if(ret < 0) return ret;
		tmp->size = ret;
#ifdef	CONFIG_UNM_NETFILTER_DEBUG
     printk("Adjusted Entry Size = %d\n", tmp->size);
#endif
    /* set tmp->counters <-- end of buffer leaving space for len */
    tmp->counters = PEGNET_MARK_NONLOCAL(pci_map_single(NULL,(void *)card_entries +tmp->size, tmp->num_counters * sizeof(struct ipt_counters_unm), 0));

    return (tmp->size + sizeof(struct ipt_replace_unm));
}

static int copy_out_counters(void *user, void *card)
{
    void *src;
    struct ipt_replace_unm *card_info=(struct ipt_replace_unm *)card;
    struct ipt_replace *host_info = (struct ipt_replace *)user;

    src = IPT_LEN_ALIGN(card, sizeof(struct ipt_replace_unm) + card_info->size,           uint64_t);

    copy_to_user(host_info->counters, src,
                    sizeof(struct ipt_counters)*host_info->num_counters);
    return 0;
}

static int copy_counters_from_host(void *user, void *card, unsigned int len)
{
    struct ipt_counters_info *host_info = (struct ipt_counters_info *)user;

    memcpy(card,user, sizeof(struct ipt_counters_info_unm));
    memcpy(card+sizeof(struct ipt_counters_info_unm),
              user+sizeof(struct ipt_counters_info),
              host_info->num_counters*sizeof(struct ipt_counters_info));

    len = sizeof(struct ipt_counters_info_unm) +
                 host_info->num_counters*sizeof(struct ipt_counters);
    return len;
 
}

static int do_ipt_set_ctl(int cmd, ipt_op_t op, void *user,
                   void *card, unsigned int len)
{
    int ret=-EINVAL;

    switch (cmd) {
    case IPT_SO_SET_REPLACE:
        switch(op){
        case ipt_op_copyin:
            ret = copy_iptreplace_from_host(user, card, len);
            break;
        case ipt_op_copyout:
            ret = copy_out_counters(user, card);
            break;
        }
        break;
    case IPT_SO_SET_ADD_COUNTERS:
        switch(op){
        case ipt_op_copyin:
            ret = copy_counters_from_host(user, card, len);
            break;
        default: return 0;
        }
        break;
    }
    return ret;
}

static int do_ipt_get_ctl(int cmd, ipt_op_t op, void *user, 
                  void *card, int len)
{
    int ret=-EINVAL;
    struct ipt_getinfo *infouser;
    struct ipt_getinfo_unm *infocard;
    struct ipt_get_entries *get_user;
    struct ipt_get_entries_unm *get_card;
    struct ipt_getinfo_trailer *trailer;

    switch (cmd) {
    case IPT_SO_GET_INFO:
        switch(op){
        case ipt_op_copyin :
            /* Just to simplify code flow; expensive though  */
            memcpy(card, user, len);
            ret = IPT_UNM_ALIGN_OFFSET(sizeof(struct ipt_getinfo_unm), 
                      sizeof(struct ipt_getinfo_trailer));
            break;
        case ipt_op_copyout :
            infouser = (struct ipt_getinfo *)user;
            infocard = (struct ipt_getinfo_unm *)card;
            memcpy(infouser, infocard, sizeof(struct ipt_getinfo));
            trailer = (struct ipt_getinfo_trailer *) IPT_LEN_ALIGN(card , 
                            sizeof(struct ipt_getinfo_unm), uint64_t);
            ret = ipt_patch_getinfo(trailer, infouser);
#ifdef	CONFIG_UNM_NETFILTER_DEBUG
            print_ipt_getinfo(infouser);
#endif
        }
        break;

    case IPT_SO_GET_ENTRIES:
        get_card = (struct ipt_get_entries_unm *)card;
        get_user = (struct ipt_get_entries *)user;
        if (strcmp(get_user->name, "filter")) return (-EINVAL);
        switch(op){
        case ipt_op_copyin:
            memcpy(get_card, get_user, sizeof(struct ipt_get_entries));
            get_card->size = ipt_sentry.card.size; /* From socket */
            ret = get_card->size + sizeof(struct ipt_get_entries_unm);
            break;
        case ipt_op_copyout:
            ret = get_entries_from_unm(get_user, get_card);
            break;
        }
        break;
    }
    return ret;
}


int unm_ipt_setsockopt(pegnet_sock_id_t *sid, int level, int optname,
                       char *optval, int optlen)
{
        int rv;
        void *card_buf, *host_buf;

        if (!capable(CAP_NET_ADMIN)) return -EPERM;
        if (!optlen) return -EINVAL;
        rv = get_dma_buffer(sid, pegnet_sopt_buf, &card_buf);
        if (0 != rv) return rv;

        if ((host_buf = kmalloc(optlen, GFP_KERNEL)) == NULL) goto finish0;
        smart_memcpy(host_buf, optval, optlen); /* Get the user copy */
        optlen = do_ipt_set_ctl(optname, ipt_op_copyin, host_buf, 
                                card_buf, optlen); /* transform */
        if(optlen <= 0) {
                rv=optlen;
                goto finish;
        }

        /* change to host_ipt_setsockopt() */
        rv = host_ipt_setsockopt(sid, level, optname,
                                 PEGNET_MARK_NONLOCAL(virt_to_bus(card_buf)),
                                 optlen);
        if (!rv)
                do_ipt_set_ctl(optname, ipt_op_copyout, host_buf, card_buf,
                               optlen);
  finish:
        kfree(host_buf);
  finish0:
        put_dma_buffer(sid, pegnet_sopt_buf, card_buf);
        return(rv);
}

/*
 * tcp_setsockopt() will call copy_from_user() 
 * and DMA if INDIRECT_IOCTL_PARMS isn't set.
 */
int unm_ipt_getsockopt(pegnet_sock_id_t *sid, int level, int optname,
                       char *optval, int *optlen)
{
        void *card_buf, *host_buf; 
        uint64_t addr1;
        int len,  rv;
        struct pci_dev *pdev;

        if (!capable(CAP_NET_ADMIN)) return -EPERM;
        /* Get the transfer buffers */
        rv = get_dma_buffer(sid, pegnet_sopt_buf, &card_buf);
        if (0 != rv) return rv;
        smart_memcpy(&len, optlen, sizeof(int));
        if (!len) goto finish0;
        if ( (host_buf = kmalloc(IPT_HOSTBUFF_SIZE, GFP_KERNEL)) == NULL) goto finish0;

        smart_memcpy(host_buf, optval, len);  /* copy user buffer */
        /* fit to peg */
        len = do_ipt_get_ctl(optname, ipt_op_copyin, host_buf, card_buf, len);
        if (len <= 0) {
                rv = len;	
                goto finish;
        }

        /* Prepare optval */
        pdev = get_pci_dev_from_sid(sid);
        addr1 = PEGNET_MARK_NONLOCAL(pci_map_single(pdev,card_buf, len, 0));  

        /* change to host_ipt_getsockopt() */
        if((rv = host_ipt_getsockopt(sid, level, optname, addr1, &len)))
                goto finish;
        len = do_ipt_get_ctl(optname, ipt_op_copyout, host_buf, card_buf, len);
        if(len > 0){
                smart_memcpy(optlen, &len, sizeof(int));
                smart_memcpy(optval, host_buf, len);
        }else {
                rv = len;
                printk("UNM_IPT: Failed to map content\n");
        }

  finish:
        kfree(host_buf);
  finish0:
        put_dma_buffer(sid, pegnet_sopt_buf, card_buf);
        return(rv);
}

/*
 *   Entry validation
 */
#ifdef CONFIG_UNM_NETFILTER_DEBUG
#define duprintf(format, args...)  printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

struct ipt_table_info
{
    /* Size per table */
    uint32_t size;
    /* Number of entries: FIXME. --RR */
    uint32_t number;
    /* Initial number of entries. Needed for module usage count */
    uint32_t initial_entries;

    /* Entry points and underflows */
    uint32_t hook_entry[NF_IP_NUMHOOKS];
    uint32_t underflow[NF_IP_NUMHOOKS];
    char entries[0] __attribute__((__aligned__(SMP_CACHE_BYTES)));
};

static LIST_HEAD(ipt_target);
static LIST_HEAD(ipt_match);
static int tcp_checkentry(const char *tablename, const struct ipt_ip *ip,
           void *matchinfo, unsigned int matchsize, unsigned int hook_mask);
static int udp_checkentry(const char *tablename, const struct ipt_ip *ip,
           void *matchinfo, unsigned int matchsize, unsigned int hook_mask);
static int icmp_checkentry(const char *tablename, const struct ipt_ip *ip,
           void *matchinfo, unsigned int matchsize, unsigned int hook_mask);

static struct ipt_target ipt_standard_target
= { { NULL, NULL }, IPT_STANDARD_TARGET, NULL, NULL, NULL };

static struct ipt_target ipt_error_target
= { { NULL, NULL }, IPT_ERROR_TARGET, NULL, NULL, NULL };

static struct ipt_target ipt_reject_target
= { { NULL, NULL }, "REJECT", NULL, NULL, NULL };

static struct ipt_match tcp_matchstruct
= { { NULL, NULL }, "tcp", NULL, &tcp_checkentry, NULL};
static struct ipt_match udp_matchstruct
= { { NULL, NULL }, "udp", NULL, &udp_checkentry, NULL};
static struct ipt_match icmp_matchstruct
= { { NULL, NULL }, "icmp", NULL, &icmp_checkentry, NULL};

static inline struct ipt_match *
find_match(const char *name, int *error)
{
    struct ipt_match *ret;
    if ((ret = list_named_find(&ipt_match, name)) == NULL) *error = -ENOENT;
    return ret; 
}

static inline struct ipt_target *
find_target(const char *name, int *error)
{
    struct ipt_target *ret;
    if ((ret = list_named_find(&ipt_target, name))==NULL) *error = -ENOENT;
    return ret;
}


static inline int
ip_checkentry(const struct ipt_ip *ip)
{
    if (ip->flags & ~IPT_F_MASK) {
        duprintf("Unknown flag bits set: %08X\n",
             ip->flags & ~IPT_F_MASK);
        return 0;
    }
    if (ip->invflags & ~IPT_INV_MASK) {
        duprintf("Unknown invflag bits set: %08X\n",
             ip->invflags & ~IPT_INV_MASK);
        return 0;
    }
    return 1;
}

static inline int
check_match(struct ipt_entry_match *m,
        const char *name,
        const struct ipt_ip *ip,
        unsigned int hookmask,
        unsigned int *i)
{
    int ret;
    struct ipt_match *match;

    match = find_match(m->u.user.name, &ret);
    if (!match) {
        duprintf("check_match: `%s' not found\n", m->u.user.name);
        return ret;
    }
    m->u.kernel.match = match;

    if (m->u.kernel.match->checkentry
        && !m->u.kernel.match->checkentry(name, ip, m->data,
                          m->u.match_size - sizeof(*m),
                          hookmask)) {
        duprintf("ip_tables: check failed for `%s'.\n",
             m->u.kernel.match->name);
        return -EINVAL;
    }

    (*i)++;
    return 0;
}

static inline int
standard_check(const struct ipt_entry_target *t,
           unsigned int max_offset)
{
    struct ipt_standard_target *targ = (void *)t;

    /* Check standard info. */
    if (t->u.target_size
        != IPT_ALIGN(sizeof(struct ipt_standard_target))) {
        duprintf("standard_check: target size %u != %u\n",
             t->u.target_size,
             IPT_ALIGN(sizeof(struct ipt_standard_target)));
        return 0;
    }
    if (!strcmp(t->u.user.name,IPT_ERROR_TARGET)) return(1);


    if (targ->verdict >= 0
        && targ->verdict > max_offset - sizeof(struct ipt_entry)) {
        duprintf("ipt_standard_check: bad verdict (%d)\n",
             targ->verdict);
        return 0;
    }

    if (targ->verdict < -NF_MAX_VERDICT - 1) {
        duprintf("ipt_standard_check: bad negative verdict (%d)\n",
             targ->verdict);
        return 0;
    }
    return 1;
}


static inline int
check_entry(struct ipt_entry *e, const char *name, unsigned int size)
{
    struct ipt_entry_target *t;
    struct ipt_target *target;
    int ret;
    unsigned int j;

    if (!ip_checkentry(&e->ip)) {
        duprintf("ip_tables: ip check failed %p %s.\n", e, name);
        return -EINVAL;
    }

    j = 0;
    ret = IPT_MATCH_ITERATE(e, check_match, name, &e->ip, e->comefrom, &j);
    if (ret != 0)
        goto cleanup_matches;

    t = ipt_get_target(e);
    target = find_target(t->u.user.name, &ret);
    if (!target) {
        duprintf("check_entry: `%s' not found\n", t->u.user.name);
        goto cleanup_matches;
    }
    t->u.kernel.target = target;

    if (t->u.kernel.target == &ipt_standard_target) {
        if (!standard_check(t, size)) {
            ret = -EINVAL;
            goto cleanup_matches;
        }
    } else if (t->u.kernel.target->checkentry
           && !t->u.kernel.target->checkentry(name, e, t->data,
                              t->u.target_size
                              - sizeof(*t),
                              e->comefrom)) {
        duprintf("ip_tables: check failed for `%s'.\n",
             t->u.kernel.target->name);
        ret = -EINVAL;
        goto cleanup_matches;
    }

    return 0;

 cleanup_matches:
    /* 
     * We are only sniffing. We dont have kernel copy, 
     * so cleanup is to throw error 
     */
    return ret;
}

static inline int
check_entry_size_and_hooks(struct ipt_entry *e,
               struct ipt_table_info *newinfo,
               unsigned char *base,
               unsigned char *limit,
               const unsigned int *hook_entries,
               const unsigned int *underflows,
               unsigned int *i)
{
    unsigned int h;

    if ((unsigned long)e % __alignof__(struct ipt_entry) != 0
        || (unsigned char *)e + sizeof(struct ipt_entry) >= limit) {
        duprintf("Bad offset %p Limit %p Size %d \n", e, limit, 
                sizeof(struct ipt_entry));
        return -EINVAL;
    }

    if (e->next_offset
        < sizeof(struct ipt_entry) + sizeof(struct ipt_entry_target)) {
        duprintf("checking: element %p size %u\n",
             e, e->next_offset);
        return -EINVAL;
    }

    /* Check hooks & underflows */
    h = (unsigned char *)e - base;
    for (h = 0; h < NF_IP_NUMHOOKS; h++) {
        if ((unsigned char *)e - base == hook_entries[h])
            newinfo->hook_entry[h] = hook_entries[h];
        if ((unsigned char *)e - base == underflows[h])
            newinfo->underflow[h] = underflows[h];
    }

    /* FIXME: underflows must be unconditional, standard verdicts
           < 0 (not IPT_RETURN). --RR */

    /* Clear counters and comefrom */
    e->counters = ((struct ipt_counters) { 0, 0 });
    e->comefrom = 0;

    (*i)++;
    return 0;
}

static int
tcp_checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{
    const struct ipt_tcp *tcpinfo = matchinfo;

    /* Must specify proto == TCP, and no unknown invflags */
    return ip->proto == IPPROTO_TCP
        && !(ip->invflags & IPT_INV_PROTO)
//        && matchsize == IPT_ALIGN(sizeof(struct ipt_tcp))
        && !(tcpinfo->invflags & ~IPT_TCP_INV_MASK);
}

static int
udp_checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchinfosize,
           unsigned int hook_mask)
{
    const struct ipt_udp *udpinfo = matchinfo;

    /* Must specify proto == UDP, and no unknown invflags */
    if (ip->proto != IPPROTO_UDP || (ip->invflags & IPT_INV_PROTO)) {
        duprintf("ipt_udp: Protocol %u != %u\n", ip->proto,
             IPPROTO_UDP);
        return 0;
    }
    /*
    if (matchinfosize != IPT_ALIGN(sizeof(struct ipt_udp))) {
        duprintf("ipt_udp: matchsize %u != %u\n",
             matchinfosize, IPT_ALIGN(sizeof(struct ipt_udp)));
        return 0;
    }
    */
    if (udpinfo->invflags & ~IPT_UDP_INV_MASK) {
        duprintf("ipt_udp: unknown flags %X\n",
             udpinfo->invflags);
        return 0;
    }

    return 1;
}

static int icmp_checkentry(const char *tablename,
       const struct ipt_ip *ip, void *matchinfo,
       unsigned int matchsize, unsigned int hook_mask)
{
    const struct ipt_icmp *icmpinfo = matchinfo;

    /* Must specify proto == ICMP, and no unknown invflags */
    return ip->proto == IPPROTO_ICMP
        && !(ip->invflags & IPT_INV_PROTO)
   /* We re-align it anyway */
//        && matchsize == IPT_ALIGN(sizeof(struct ipt_icmp))
        && !(icmpinfo->invflags & ~IPT_ICMP_INV);
}


/* Done at module init time */
int unm_iptable_init(void)
{
    list_append(&ipt_target, &ipt_standard_target);
    list_append(&ipt_target, &ipt_error_target);
    list_append(&ipt_target, &ipt_reject_target);
    list_append(&ipt_match, &tcp_matchstruct);
    list_append(&ipt_match, &udp_matchstruct);
    list_append(&ipt_match, &icmp_matchstruct);

    return 0;
}

