/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * NetXen:
 *
 * Hash table implemented on top of a link list.
 */
#ifndef _NX_HASH_TABLE_H
#define _NX_HASH_TABLE_H

#include <linux/list.h>
#include <linux/kernel.h>

typedef struct {
	spinlock_t		lock;
	struct list_head	head;
} nx_hbucket_head_t;

#define INIT_NX_HBUCKET_HEAD(PTR)			\
	do {					\
		spin_lock_init(&(PTR)->lock);	\
		INIT_LIST_HEAD(&(PTR)->head);	\
	} while (0)

/*
 *
 */
static inline void nx_hbucket_add(struct list_head *new,
				  nx_hbucket_head_t *head)
{
	spin_lock_bh(&head->lock);
	list_add(new, &head->head);
	spin_unlock_bh(&head->lock);
}

/*
 *
 */
static inline void nx_hbucket_add_tail(struct list_head *new,
				       nx_hbucket_head_t *head)
{
	spin_lock_bh(&head->lock);
	list_add_tail(new, &head->head);
	spin_unlock_bh(&head->lock);
}

/*
 *
 */
static inline void nx_hbucket_del(struct list_head *node,
				  nx_hbucket_head_t *head)
{
	spin_lock_bh(&head->lock);
	list_del(node);
	spin_unlock_bh(&head->lock);
}
/*
 *
 */
static inline void nx_hbucket_del_init(struct list_head *node,
				  nx_hbucket_head_t *head)
{
	spin_lock_bh(&head->lock);
	list_del_init(node);
	spin_unlock_bh(&head->lock);
}

#endif /* _NX_HASH_TABLE_H */
