/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * NetXen:
 *
 * Provides a cache for destination lookup.
 */
#ifndef _NX_ROUTE_TABLE_H
#define _NX_ROUTE_TABLE_H


#include <unm_inc.h>
#include <pegnet_inc.h>

#include <linux/unistd.h>
#include <linux/types.h>
#include <linux/autoconf.h>
#include <net/route.h>
#include <linux/ctype.h>
#include <hostInterface.h>
#include "nx_hash_table.h"

#define	RC_HASH_TBL_SIZE	(16 * 1024)
#define	RC_MAX_ENTRIES		( 8 * 1024)
#define	RC_MAX_NHS		1


typedef struct {
	uint32_t		gw;
	uint8_t			port;
	uint8_t			tos;
	uint16_t		vlanid;
	uint8_t			qos;
	uint8_t			vlan_set;
#ifdef CONFIG_IP_ROUTE_MULTIPATH
	int			weight;
	int			power;
#endif
} nx_host_rt_nh_t;

typedef struct {
	uint32_t	daddr;
	uint32_t	saddr;
} nx_host_rt_key_t;

/*
 * This structure contains data shared by many routes.
 */
typedef struct {
	struct list_head	list;
	nx_host_rt_key_t	key;
	uint32_t		mtu;
	uint32_t		window;
	uint32_t		irtt;
	uint32_t		advmss;
	int			refcnt;
	int			invalid;

	int			nh_cnt;
	nx_host_rt_nh_t		nhs[RC_MAX_NHS];
} nx_host_rt_entry_t;

typedef struct {
	struct list_head	free_list;
	struct {
		spinlock_t		lock;
		wait_queue_head_t	wq;
		int			inuse;
	} lock;
	int			cnt;
	nx_hbucket_head_t	routes[RC_HASH_TBL_SIZE];
} nx_host_route_tbl_t;

typedef struct {
	nx_hbucket_head_t		list;
	nx_kc_work_queue_t		task;
} nx_host_rt_sync_work_queue_t;

#endif /* _NX_ROUTE_TABLE_H */
