/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <linux/sockios.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include <linux/version.h>
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS) && !defined(__GENKSYMS__)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/modversions.h>
#endif
#endif

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/net.h>
#include <linux/poll.h>
#include <linux/skbuff.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/ip.h>
#include <net/tcp.h>

#include <linux/inetdevice.h>
#include <linux/if.h>
#include <linux/if_vlan.h>

#include <unm_inc.h>
#include <pegnet_inc.h>
#include "nic_cmn.h"
#include "hostInterface.h"
#include "flow_table.h"
#include "unm_host_pnic_types.h"


/*
 * Function declaration of static functions.
 */
static int host_send_rxbuffer(struct net_device *netdev, hostRxBuffer_t *rbuf, 
			      pegnet_sock_id_t* sid);


/*
 *
 */
void nx_proc_rx_urg_pointer(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
		pegnet_sock_id_t *sid,pegnet_inprogress_t *trans)
{
        rcv_urg_pointer_t       *urg_msg = (rcv_urg_pointer_t *)&msg->body;
        struct sock             *sk;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	int 			pid;
#endif

        spin_lock_bh(&sid->lock);

        if (sid && sid->parent && sid->parent->sk) {
                sk = sid->parent->sk;
                if(sid->urg_flg == URG_WITH_APP) {
                        /* If we havent handled earlier urgent data completely
                           ignore this. (Compliant with protocol, if multiple
                           URG datas are stacked, protocol does not gurentee 
                           if all of them would be delivered to app) */
                        goto done;
                }
                sid->urg_flg = URG_WITH_HOST;
                sid->urg_data = urg_msg->byte_0;
                sid->urg_ptr = urg_msg->position - 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                sk_send_sigurg(sk);
#else
                // This code (for 2.4.xx) not tested yet
                if (sk->socket && sk->socket->file) {
                        pid = sk->socket->file->f_owner.pid;
                        if (pid > 0) {
                                kill_proc(pid, SIGURG, 1);
                        } else {
                                kill_pg(-pid, SIGURG, 1);
                        }
                }
#endif
        }
  done:
        spin_unlock_bh(&sid->lock);
}


/*
 *
 */
static inline int rx_buffer_map_pci(struct net_device *dev,hostRxBuffer_t *rbuf)
{
	int	ii;

	for (ii = 0; ii < rbuf->numPages; ii++) {
		netxen_page_unmap_virt(&rbuf->page[ii]);

		if(netxen_page_map_pci(&rbuf->page[ii],nx_get_pcidev(dev),
				       PCI_DMA_FROMDEVICE)) {
			printk("%s: PCI map call failed\n", __FUNCTION__);
			ii--;
			for ( ; ii >= 0; ii--) {
				netxen_page_unmap_virt(&rbuf->page[ii]);
			}
			return (-1);
		}
	}

	return (0);
}

/*
 *
 */
static inline int rx_buffer_unmap_pci(struct net_device *dev,hostRxBuffer_t *rbuf)
{
	int	ii;

	for (ii = 0; ii < rbuf->numPages; ii++) {
		netxen_page_unmap_pci(&rbuf->page[ii],nx_get_pcidev(dev),
				      PCI_DMA_FROMDEVICE);
	}

	return (0);
}

/* Initialize the per-connection receive buffers and post
 * them to the card. */
/* Typically, called during connect or accept. */
int nx_init_rxbuffers_pcb(pegnet_sock_id_t *sid)
{
        int		ii, jj;
        hostRxBuffer_t	*rbuf;
        int		totalPages = 0;
	int		rv = 0;


	sid->recv_dma.remoteBuf = kmalloc(sizeof (hostRxBuffer_t) *
					  sid->recv_dma.num_pcb_rxbuffers,
					  GFP_KERNEL);
	if (sid->recv_dma.remoteBuf == NULL) {
		nx_print("Unable to allocate receive buffers\n");
		return (-ENOMEM);
	}
        rbuf = sid->recv_dma.remoteBuf;

        for (ii = 0; ii < sid->recv_dma.num_pcb_rxbuffers; ii++, rbuf++) {
                memset(rbuf, 0, sizeof(*rbuf));
		rbuf->hostCtrlFlag = UNM_HT_RXBUFF_NOTVALID;

                for (jj = 0; jj < HOST_PCB_RX_PAGES; jj++) {
			rv = alloc_netxen_page(&rbuf->page[jj]);
			if (rv) {
                                printk("%s: sid=%p, allocate memory failed\n",
				       __FUNCTION__, sid);
                                break;
                        }
                        rbuf->numPages++;
                        totalPages++;
			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_INUSE_RX_PAGES);
                }
                if (jj != 0) {
                        rbuf->hostCtrlFlag = UNM_HT_RXBUFF_EMPTY;
                        INIT_LIST_HEAD(&rbuf->list);
                        rbuf->index = ii ;
                        rbuf->bufferType = PCB_RXBUFFER ;
                        if( 0 == host_send_rxbuffer(NX_GET_DEV_FROM_TNIC_ADAPTER(
				sid->adapter,sid->portnum),rbuf,sid) ) {
                                int kk ;
                                printk("Failed sending pcb buffer to card\n");
                                /*
                                 * TODO : Handle failures more gracefully.
                                 * Free the remaining allocated buffers, and
                                 * let the card know that the buffers are
                                 * not usable.
                                 */
                                for( kk=0; kk<jj; kk++ ) {
					free_netxen_page(&rbuf->page[kk],
						nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
						sid->adapter,sid->portnum)),PCI_DMA_FROMDEVICE);
                                        UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_RX_PAGES);
                                }
                                return (-ENOMEM);
                        }
                }
                if (jj != HOST_PCB_RX_PAGES) {
                        break;
                }
        }
//// Start debug.
//    printk( "init_rxbuffers_pcb: sid=%p, allocated %d pages\n", sid, totalPages ) ;
//// End debug.
        sid->recv_dma.initialized = 1;

        return( ii ) ;
}

/*
 * Allocate the global RX buffer array and the elements.
 */
static int nx_rx_global_buf_array_alloc(nx_tnic_adapter_t *tnic)
{
	int		ii;
	hostRxBuffer_t	*rbuf;


	if (tnic->total_global_rx_bufs || tnic->global_rx_bufs) {
		nx_print("BUG: global buffers already allocated\n");
		return (-EINVAL);
	}

	tnic->global_rx_bufs =
		(hostRxBuffer_t **)kmalloc((sizeof (hostRxBuffer_t *) *
					    NUM_GLOBAL_RX_BUFFERS),
					   GFP_KERNEL);
	if (tnic->global_rx_bufs == NULL) {
		nx_print("Global RX buffer array allocation failed\n");
		return (-ENOMEM);
	}

	memset(tnic->global_rx_bufs, 0, (sizeof (hostRxBuffer_t *) *
					 NUM_GLOBAL_RX_BUFFERS));

	for (ii = 0; ii < NUM_GLOBAL_RX_BUFFERS; ii++) {
		rbuf = (hostRxBuffer_t *)kmalloc(sizeof (hostRxBuffer_t),
						 GFP_KERNEL);
		if (rbuf == NULL) {
			nx_print("%d Global RX buffer alloc failed\n", ii);

			for (ii--; ii >= 0; ii--) {
				kfree(tnic->global_rx_bufs[ii]);
			}
			kfree(tnic->global_rx_bufs);
			tnic->global_rx_bufs = NULL;

			return (-ENOMEM);
		}

		tnic->global_rx_bufs[ii] = rbuf;
	}

	return (0);
}

/*
 * Allocate the global RX buffer array and the elements.
 */
static void nx_rx_global_buf_array_free(nx_tnic_adapter_t *tnic)
{
	int		ii;

	for (ii = 0; ii < NUM_GLOBAL_RX_BUFFERS; ii++) {
		kfree(tnic->global_rx_bufs[ii]);
	}

	kfree(tnic->global_rx_bufs);
	tnic->global_rx_bufs = NULL;

	return;
}

/*
 * Initialize the global receive buffers and post
 * them to the card. Called during startup.
 */
int nx_init_rx_global_buffers(nx_tnic_adapter_t *tnic)
{
	int		ii;
	int		jj;
	int		pages_allocated = 0;
	hostRxBuffer_t	*rbuf;
	int		rv;


	rv = nx_rx_global_buf_array_alloc(tnic);
	if (rv) {
		return (rv);
	}

	for (ii = 0; ii < NUM_GLOBAL_RX_BUFFERS; ii++) {
		rbuf = tnic->global_rx_bufs[ii];

		memset(rbuf, 0, sizeof (*rbuf));

		for (jj = 0; jj < HOST_GLOBAL_RX_PAGES; jj++) {
			rv = alloc_netxen_page(&rbuf->page[jj]);
			if (rv) {
				nx_print("Failed to allocate global buffer\n");
				break;
			}
			rbuf->numPages++;
			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_INUSE_GLOBAL_RX_PAGES);
//// Start debug.
//            memset( rbuf->page[jj].buffer, 0xa5, rbuf->page[jj].size ) ;
//// End debug.
		}

		if (jj != 0) {
			rbuf->hostCtrlFlag = UNM_HT_RXBUFF_EMPTY;
			INIT_LIST_HEAD(&(rbuf->list));
			rbuf->index = ii;
			rbuf->bufferType = GLOBAL_RXBUFFER;

			if (0 == host_send_rxbuffer(tnic->nx_dev, rbuf,
						    NULL)) {
				int	kk;

				nx_print("Failed sending a global buffer to "
					 "card\n");
				/*
				 * TODO : Handle failures more gracefully.
				 * Note that this might not be a fatal state.
				 * For instance, the global buffer queue might
				 * just be full.
				 */
				for (kk = 0; kk < jj; kk++) {
					free_netxen_page(&rbuf->page[kk],
							 nx_get_pcidev(tnic->nx_dev),
							 PCI_DMA_FROMDEVICE);
					UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_GLOBAL_RX_PAGES);
				}
				ii--;
				break;
			}
			pages_allocated += jj;
		} else {
			ii--;
		}

		if (jj != HOST_GLOBAL_RX_PAGES) {
			break;
		}
	}

	tnic->total_global_rx_bufs = ii;

//// Start debug.
	nx_print("Allocated %d buffers, containing a total of %d pages\n",
		 ii, pages_allocated);
//// End debug.

	return (0);
}

/*
 * Free all global receive data dma buffers.
 */
void nx_destroy_rx_global_buffers(nx_tnic_adapter_t *tnic)
{
	int		ii;
	int		jj;
	int		pages_freed = 0;
        hostRxBuffer_t	*rbuf;

	if (tnic->global_rx_bufs == NULL) {
		nx_print("BUG: No globals allocated to free\n");
		return;
	}

        for (ii = 0; ii < tnic->total_global_rx_bufs; ii++) {
		rbuf = tnic->global_rx_bufs[ii];

		if (rbuf->hostCtrlFlag != UNM_HT_RXBUFF_EMPTY &&
		    rbuf->hostCtrlFlag != UNM_HT_RXBUFF_CARD) {
			nx_print("Invalid state = %u\n", rbuf->hostCtrlFlag);
		}

                for (jj = 0; jj < rbuf->numPages; jj++) {
			free_netxen_page(&rbuf->page[jj],
					 nx_get_pcidev(tnic->nx_dev),
					 PCI_DMA_FROMDEVICE);
			UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_GLOBAL_RX_PAGES);
		}
                rbuf->hostCtrlFlag = UNM_HT_RXBUFF_NOTVALID;

                pages_freed += rbuf->numPages;
	}

	nx_print("Freed %d Global buffers and %d pages\n",
		 tnic->total_global_rx_bufs, pages_freed);

	tnic->total_global_rx_bufs = 0;

	nx_rx_global_buf_array_free(tnic);
}

/* Post empty receive buffers to the card */
static int host_send_rxbuffer(struct net_device *netdev, hostRxBuffer_t *rbuf, 
			      pegnet_sock_id_t *sid)
{
	rxbuf_msg_t		msg_arr[CMD_DESC_PER_RXBUFF];
	rxbuf_msg_t		*msg;
	int			numPages;
	int			numPagesRemaining;
	int			ii;
	netxen_host_page_t	*page;
	pegnet_queue_addr_t	dst;
	int 			nr_elements = 0;
	int			rv;
	uint8_t			counter = 0;
	int 			bufferBytesRemaining;
	int			rcv_shutdown;
	int			hostCtrlFlag;

	bufferBytesRemaining 	= rbuf->bufferBytesRemaining;
	rcv_shutdown		= rbuf->rcv_shutdown;
	hostCtrlFlag		= rbuf->hostCtrlFlag;

	/* Pick our destination queue. */
	if (GLOBAL_RXBUFFER == rbuf->bufferType) {
		if (sid != NULL) {
			/* printk("%s: sid not NULL for global\n", __FUNCTION__); */
			sid = NULL;
		}
		dst.major = UNM_MSGQ_GLOBAL_RXBUFF_MAJ;
		dst.minor = UNM_MSGQ_GLOBAL_RXBUFF_MIN;
		dst.side = 0;
	} else {
		if (sid == NULL) {
			printk("%s: sid is NULL for PCB\n", __FUNCTION__);
			return (0);
		}
		dst = sid->rcvBuf_qid;

		/* Is our sid worthy to receive empty receive buffers. */
		spin_lock_bh(&sid->lock);
		if (sid->released) {
			spin_unlock_bh(&sid->lock);
//// Start debug.
			printk("%s: sid=%p, socket is released\n",
			       __FUNCTION__, sid);
//// End debug.
			return (0);
		}
		spin_unlock_bh(&sid->lock);
	}

	if (null_queue(dst)) {
		return (0);
	}
	msg = &msg_arr[0];

//// Start debug.
	if (UNM_HT_RXBUFF_EMPTY != rbuf->hostCtrlFlag) {
		printk("%s: Invalid buffer state: rbuf=%p\n",
		       __FUNCTION__, rbuf);
		return (0);
	}
//// End debug.

//// Start debug.
//    printk( "host_send_rxbuffer: sending empty buffer to card: index=%d\n",
//            rbuf->index ) ;
//    for( ii=0; ii<rbuf->numPages; ii++ ) {
//        printk( "    rx buffer[%d]=%p, phys[%d]=0x%x\n",
//                ii, rbuf->page[ii].buffer, ii, rbuf->page[ii].phys ) ;
//    }
//// End debug.

	if (rx_buffer_map_pci(netdev,rbuf)) {
		printk("%s: rx_buffer_map_pci failed\n", __FUNCTION__);
		return (0);
	}
	numPages = numPagesRemaining = rbuf->numPages;
	if (sid) {
		counter = sid->recv_dma.post_cntr;
	}
	while (numPagesRemaining > 0) {
		int numPages_qmsg;
		int jj ;

		memset(&msg->hdr, 0, sizeof (msg->hdr));
		msg->dst_subq = 1;
		msg->dst_side  = dst.side;
		msg->dst_major = dst.major;
		msg->dst_minor = dst.minor;
		msg->type = UNM_MSGTYPE_RXBUFFER;

		if (numPagesRemaining > QMSG_RX_PAGES) {
			numPages_qmsg = QMSG_RX_PAGES;
		} else {
			numPages_qmsg = numPagesRemaining;
		}
		msg->numBufs = numPages_qmsg;
		for (ii = 0, jj = numPages-numPagesRemaining;
		     ii < numPages_qmsg; ii++, jj++ ) {

			page = &rbuf->page[jj];
			msg->iov[ii].addr = page->phys;
			msg->iov[ii].len = page->size;
		}
		msg->bufferStart = 0;
		msg->bufferComplete = 0;
		msg->pushSet = 1;
		msg->pushTimer = 1;
		if (numPages == numPagesRemaining) {
			msg->bufferStart = 1;
		}
		numPagesRemaining -= numPages_qmsg;
		if (0 == numPagesRemaining) {
			msg->bufferComplete = 1;
		}

                msg->cookie = rbuf->index;
		if (sid) {
			msg->counter = sid->recv_dma.post_cntr;
			sid->recv_dma.post_cntr++;
		} else {
			msg->counter = 0;
		}

		nr_elements++;
		msg = &msg_arr[nr_elements];
	}
	rbuf->bufferBytesRemaining = 0;
	rbuf->hostCtrlFlag = UNM_HT_RXBUFF_CARD;
	rbuf->rcv_shutdown = 0;

	rv = unm_send_h2p_msg_atomic(netdev,(unm_msg_t *)msg_arr, nr_elements);

	if (rv) {
		/* rollback */
		rbuf->bufferBytesRemaining = bufferBytesRemaining;
		rbuf->hostCtrlFlag = hostCtrlFlag;
		rbuf->rcv_shutdown = rcv_shutdown;

		printk("%s: failed\n", __FUNCTION__);
		if (sid) {
			sid->recv_dma.post_cntr = counter;
		}
		rv = 0;
	} else 
		rv = 1;

	return rv;
}

/* Routine to handle filled data buffers from the card. */
/* Invoked from the tasklet initiated by the card isr. */
void nx_proc_rx_buffer(struct net_device *netdev, unm_msg_t *msg)
{
        rxbuff_completion_t     *body = (rxbuff_completion_t *)(&msg->body);
        pegnet_sock_id_t        *sid;
        int                     index = body->cookie;
        hostRxBuffer_t          *rbuf;
        uint64_t		len = body->length;
        int                     size;
        rxbuff_flags_t          flags;
	uint32_t		connID;
	nx_tnic_adapter_t 	*tnic;

	tnic = nx_get_dev_tnic_adapter(netdev);	

        flags.word = body->flags;
	if (flags.surrender) {
		if (!flags.global_buffer) {
			nx_print("Cannot surrender a PCB: 0x%llx  0x%llx\n",
				 body->body[0], body->body[1]);
			return;
		}

		if (index < 0 || index >= tnic->total_global_rx_bufs) {
			nx_print("Surrender: Invalid GLOBAL index [%d]\n",
				 index);
                        return;
		}

		rbuf = tnic->global_rx_bufs[index];
		if (rbuf->bufferType != GLOBAL_RXBUFFER) {
                        nx_print("Surrender: flags.global_buffer = 0x%x "
				 "rbuf->bufferType = 0x%x\n",
				 flags.global_buffer, rbuf->bufferType);
                }
		rbuf->hostCtrlFlag = UNM_HT_RXBUFF_EMPTY;
		if (0 == host_send_rxbuffer(netdev,rbuf, NULL)) {
			nx_print("Surrender: Reposting failed\n");
		}
		return;
	}

	connID = host_get_uint32_from_peg(body->ctx);
        sid = retreive_pegnet_sock_id(tnic->p_ftable, connID);
        if (sid == NULL) {
                nx_print("sid not found for %lu\n", (unsigned long)body->ctx);
                return;
        }

        /* Either pcb or global buffer. */
        if (flags.global_buffer) {
		if (index < 0 || index >= tnic->total_global_rx_bufs) {
			nx_print("Invalid GLOBAL index [%d]\n", index);
                        goto release_reference;
		}

		rbuf = tnic->global_rx_bufs[index];
		if (rbuf->bufferType != GLOBAL_RXBUFFER) {
                        nx_print("sid %p, flags.global_buffer = 0x%x "
				 "rbuf->bufferType = 0x%x\n",
				 sid, flags.global_buffer, rbuf->bufferType);
                }

        } else {
                if (index < 0 || index >= sid->recv_dma.num_pcb_rxbuffers) {
                        printk("%s: Invalid PCB index [%d]\n", __FUNCTION__,
                               index);
                        goto release_reference;
                }
		if (sid->recv_dma.remoteBuf == NULL) {
                        nx_print("remoteBuf == NULL\n");
                        goto release_reference;
		}
                rbuf = &(sid->recv_dma.remoteBuf[index]);
                /* Make sure that we are in sync. */
                if (rbuf->bufferType != PCB_RXBUFFER) {
                        printk("%s: sid %p, flags.global_buffer = 0x%x "
                               "rbuf->bufferType = 0x%x\n", __FUNCTION__,
                               sid, flags.global_buffer, rbuf->bufferType);
                }
        }
//// Start debug

        if( index != rbuf->index ) {
                printk("%s: sid=%p, index=0x%x !=  rbuf->index=0x%x\n",
		       __FUNCTION__, sid, index, rbuf->index ) ;
        }

        /* Make sure that we are not updating a buffer currently in use */
        if (UNM_HT_RXBUFF_CARD != rbuf->hostCtrlFlag) {
                nx_print("rbuf->hostCtrlFlag[%u] != UNM_HT_RXBUFF_CARD[%u], "
			 "index = %u, sid=%p\n",
			 rbuf->hostCtrlFlag, UNM_HT_RXBUFF_CARD, index, sid);
        }
//// End debug.

        size = rbuf->page[0].size;
        rbuf->pageBytesRemaining = (size < len) ? size : len;
        rbuf->pageBytesRead = 0;
        rbuf->bufferBytesRemaining = len - rbuf->pageBytesRemaining;
        rbuf->currPage = rbuf->page;
        rbuf->hostCtrlFlag = UNM_HT_RXBUFF_READY;
        rbuf->rcv_shutdown = flags.fin_rxed;

	rx_buffer_unmap_pci(netdev,rbuf);

//// Start debug.
//    printk( "host_get_rxbuffer: index=%d, bufferBytesRemainig=%d, pageBytesRemaining=%d, currPage=%p\n",
//            index, rbuf->bufferBytesRemaining, rbuf->pageBytesRemaining, rbuf->currPage ) ;
//// End debug.

        spin_lock_bh(&sid->lock);
	sid->recv_dma.tmp_rcv_shutdown = flags.fin_rxed ? 1 : 0;
        list_add_tail(&rbuf->list, &sid->recv_dma.list);
        spin_unlock_bh(&sid->lock);

        /* Got data. */
        if (sid->parent && sid->parent->sk) {
		if (sid->parent->sk->SK_SLEEP &&
		    waitqueue_active(sid->parent->sk->SK_SLEEP)) {
			wake_up_interruptible(sid->parent->sk->SK_SLEEP);
		}

		/*
		 * Notify parent of the available received data.
		 */
                bh_lock_sock(sid->parent->sk);
                if (sid->parent->sk->SK_DATA_READY) {
                        sid->parent->sk->SK_DATA_READY(sid->parent->sk, 0);
                }
                bh_unlock_sock(sid->parent->sk);
        }
        /* We need to wake up host_recvmesg, if its sleeping */
        wake_up_interruptible(&sid->recv_dma.wq);

//// Start debug.
//    printk( "    xGotit: sid=%p, index=%d\n", sid, index ) ;
//    printk( "   xGotit, rbuf->page[0].phys=0x%x, rbuf->bufferBytesRemaining=%d, rbuf->rcv_shutdown=%d\n",
//        rbuf->page[0].phys, rbuf->bufferBytesRemaining, rbuf->rcv_shutdown ) ;
//// End debug.

  release_reference:
	host_flowtbl_put_reference(tnic->p_ftable,connID);
}

/*
 * Transfer received data to application.
 */
int host_recvmsg_fast(pegnet_sock_id_t *sid, struct msghdr *msg,
		int requested_len, int flags, struct scm_cookie *scm)
{
	hostRxBuffer_t		*rbuf;
	int			iovlen;
	struct iovec		*iov;
	int			copied = 0;
	int			pageBytesRemaining;
	int			pageBytesRead;
	int			bufferBytesRemaining;
	netxen_host_page_t	*currPage;
	char			*from;
	hostRxBuffer_t		*arr_rxbuf[NUM_PCB_RXBUFFERS_HP];
	int			cnt_rxbuf=0,access_flag=0;
	long            	timeo;
	struct list_head	*node;
			

        nx_print1("Entering\n");

	/* Before grabbing any locks, let the card know if RCVBUF has changed */
	if (sid->reported_rcvbuf != sid->parent->sk->SK_RCVBUF) {
		int sockopt_ret;

		if (!(sockopt_ret =
			host_setsockopt(sid, SOL_SOCKET, SO_RCVBUF,
				sid->parent->sk->SK_RCVBUF/2,
				sizeof(sid->parent->sk->SK_RCVBUF), 1)) ) {
			sid->reported_rcvbuf = sid->parent->sk->SK_RCVBUF;
		}
		else
			printk("%s: Unable to set RCVBUF, ret = %d\n",
					__FUNCTION__, sockopt_ret);
	}

	if (down_interruptible(&(sid->recv_dma.lock))) {
		return -ERESTARTSYS;
	}
	spin_lock_bh(&(sid->lock));

	nx_print1("Entering proper, sid=%p\n", sid);

	/* Check if the socket is still able to receive data. */
	if (0 != sid->recv_dma.rcv_shutdown) {
		spin_unlock_bh(&(sid->lock));
		nx_print1("No more data to receive\n" );
		goto fault;
	}

	/* Urgent data needs to be handled specially. */
	if ((flags & MSG_OOB) && !(sid->shutdown & RCV_SHUTDOWN)) {
		goto recv_urg;
	}

	timeo = sock_rcvtimeo(sid->parent->sk, flags & MSG_DONTWAIT);

	/* Check if we have a buffer available. */
	if (list_empty(&(sid->recv_dma.list))) {

		/* No buffer available */
		/* TODO: need to handle non-blocking */

		DECLARE_WAITQUEUE(wait, current);

		/* In case the list is empty, and caller has requested MSG_DONTWAIT,
		   then return error. It's okay to place this check after
		   the MSG_OOB check, because the user will never block when requesting MSG_OOB. */
        if ( (MSG_DONTWAIT & flags) && !(sid->shutdown & RCV_SHUTDOWN))
		{
			spin_unlock_bh( &(sid->lock) );
			copied = -EAGAIN;
			goto fault; /* recv_dma lock will be upped there, no need to do it here */
		}

		add_wait_queue(&(sid->recv_dma.wq), &wait);
		while (list_empty(&(sid->recv_dma.list)) &&
		    !(sid->shutdown & RCV_SHUTDOWN)) {
			set_current_state( TASK_INTERRUPTIBLE );
			spin_unlock_bh( &(sid->lock) );
			if( signal_pending(current) ) {

				spin_lock_bh(&(sid->lock));
				set_current_state(TASK_RUNNING);
				remove_wait_queue(&(sid->recv_dma.wq), &wait);
				spin_unlock_bh(&(sid->lock));

				nx_print1("Signal received: sid=%p, "
					"sid->qid=0x%x\n",
					sid, *((__uint32_t *)&(sid->qid)));
				nx_print1("list_empty(recv_dma.lst)=%d\n",
					list_empty(&sid->recv_dma.list));

				copied = -ERESTARTSYS;
				goto fault;
			}

			if (!timeo) {
				spin_lock_bh(&(sid->lock));
				set_current_state(TASK_RUNNING);
				remove_wait_queue(&(sid->recv_dma.wq), &wait);
				spin_unlock_bh( &(sid->lock) );

				copied = -EAGAIN;
				goto fault;
			}

			nx_print1("sid=%p, No buffer available: sleeping\n",
				sid);

			timeo = schedule_timeout(timeo);
			spin_lock_bh(&(sid->lock));
		}
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&(sid->recv_dma.wq), &wait);

		nx_print1("sid=%p, revived!!\n", sid);
	}

    /* In case of any error RST etc ... */
	if (sid->parent->sk->SK_ERR != 0) {
		spin_unlock_bh(&(sid->lock));
		copied = -(sid->parent->sk->SK_ERR);
		goto fault;
	}

	if (sid->shutdown & RCV_SHUTDOWN) {
		spin_unlock_bh(&(sid->lock));
		copied = 0;
		goto fault;
	}
	node = sid->recv_dma.list.next;
	rbuf = list_entry(node, hostRxBuffer_t, list);
	spin_unlock_bh(&(sid->lock));
	arr_rxbuf[cnt_rxbuf++] = rbuf;

//// Start debug.
//	if(rbuf->rcv_shutdown
//	    && ((rbuf->bufferBytesRemaining)+(rbuf->pageBytesRemaining)) ) {
//		printk( "host_recvmsg_fast: got a fin with data: sid=%p, datalen=0x%x\n",
//	        	(rbuf->bufferBytesRemaining)+(rbuf->pageBytesRemaining) );
//	}
//	printk( "host_recvmsg_fast: hooray got a buffer\n" );
//// End debug.

	/* We got a buffer. Move the puppy out of kernel space. */
	pageBytesRemaining = rbuf->pageBytesRemaining;
	pageBytesRead = rbuf->pageBytesRead;
	bufferBytesRemaining = rbuf->bufferBytesRemaining;
	currPage = rbuf->currPage;
	from = netxen_page_map_virt(currPage) + pageBytesRead;
	copied = 0;
	iovlen = msg->msg_iovlen;
	iov = msg->msg_iov;
	while (--iovlen >= 0) {
		int seglen = iov->iov_len;
		char *to = iov->iov_base;
		while (seglen > 0) {

			int chunksize ;
			if (sid->urg_flg && sid->urg_ptr == sid->abs_posn)
			{
				/* Are we at urgent data? Stop if we have read anything. */
				if (copied && URG_WITH_HOST == sid->urg_flg)
					break;
				/* If we have already eaten urgent data, we need to skip it */
				if (URG_WITH_APP == sid->urg_flg)
				{
					sid->urg_flg = URG_IGN;
					/* skip copy for 1 byte*/
					/* If NULL has been passed to app (due to 
					   tcp_stdurg=1), don't do any re-adjustment */
					chunksize = !(!(sid->urg_data));
					goto size_adjusting;
				}
				/* If first byte is urgent data, treat it like normal data */
				if (!copied)
					sid->urg_flg = URG_IGN;
			}

			if( 0 != pageBytesRemaining ) {
				/* Here, if we got a receive page. */
				chunksize = ((seglen < pageBytesRemaining) ?
						seglen : pageBytesRemaining);
				chunksize = ((requested_len < chunksize) ?
						requested_len : chunksize);
				if(sid->urg_flg)
				{
					if ((sid->urg_ptr - sid->abs_posn) < chunksize)
						chunksize = sid->urg_ptr - sid->abs_posn;
				}

				if (!access_ok(VERIFY_WRITE, to, chunksize)) {
					if (copied == 0) {
						copied = -EFAULT;
						goto fault;
					}
					access_flag=1;
					goto finish_check;
				}
#if 1
				smart_memcpy(to, from, chunksize);
#else // XXX: Use this to bypass memcpy for iperf performance test
				if (chunksize < 100)
					smart_memcpy(to, from, chunksize);
#endif
//// Start debug.
//				printk( "host_recvmsg_fast: copied %d bytes from %p to %p\n",
//				    chunksize, from, to );
//				printk( "    pageBytesRemaining=0x%x, pageBytesRead=0x%x\n",
//				    pageBytesRemaining, pageBytesRead );
//				printk( "    bufferBytesRemaining=0x%x, currPage=%p, requested_len=0x%x\n",
//				    bufferBytesRemaining, currPage, requested_len );
//// End debug.
				copied += chunksize ;
				requested_len -= chunksize ;
				to += chunksize ;
size_adjusting :
				seglen -= chunksize;
				pageBytesRemaining -= chunksize;
				pageBytesRead += chunksize;
				from += chunksize;
				sid->abs_posn += chunksize ;
			}
finish_check:
			if ((0 == requested_len) || (0==(pageBytesRemaining +  bufferBytesRemaining)) 
			    || (1==access_flag)) {
				if (!(flags & MSG_PEEK)) {
					rbuf->bufferBytesRemaining = bufferBytesRemaining;
					rbuf->pageBytesRemaining = pageBytesRemaining;
					rbuf->pageBytesRead = pageBytesRead;
					rbuf->currPage = currPage;
				}
				if((0 == requested_len) || (1==access_flag)){
					goto out;
				}
			}

			if (0 == pageBytesRemaining) {
				/* Here if we need to grab another page. */
				if (0 == bufferBytesRemaining) {
					/* Here we are done with our buffer. */
					spin_lock_bh( &(sid->lock) );
					/* Check for list head */
					if ((node = node->next) != &(sid->recv_dma.list)) { 
						/*Retrieve another buffer */
						rbuf = list_entry(node, hostRxBuffer_t, list);
						arr_rxbuf[cnt_rxbuf++]=rbuf;
						spin_unlock_bh( &(sid->lock) );
						pageBytesRemaining = rbuf->pageBytesRemaining;
						pageBytesRead = rbuf->pageBytesRead;
						bufferBytesRemaining = rbuf->bufferBytesRemaining;
						currPage = rbuf->currPage;
						from = netxen_page_map_virt(currPage) + pageBytesRead;
					}
					else {
						spin_unlock_bh( &(sid->lock) );
						goto out;
					}

				}
				else {
					currPage++;
					from = netxen_page_map_virt(currPage);
					pageBytesRemaining = ((currPage->size <
						    bufferBytesRemaining) ?  
						    currPage->size :
						    bufferBytesRemaining);
					pageBytesRead = 0;
					bufferBytesRemaining -= pageBytesRemaining;
				}
			}
		}
		iov++;
	}

out:
	/* We have transferred data.  If we have emptied our buffer,
	 * give it back to the card. */
	while((--cnt_rxbuf) >= 0) {
		rbuf=arr_rxbuf[cnt_rxbuf];

		if( 0 == ((rbuf->bufferBytesRemaining)+(rbuf->pageBytesRemaining)) ) {
			/* Here if we're done with the buffer. Normally we just take
			 * the puppy off the link list. However, we have to keep the
			 * buffer on the list if rcv_shutdown is set for a buffer
			 * that contains data. This ensures that poll still works
			 * properly.
			 */
			if( !(rbuf->rcv_shutdown) || (0==copied) ) {

				/* Buffer is read. Remove it from receive queue. */
				spin_lock_bh( &(sid->lock) );
				list_del_init( &(rbuf->list) );
				spin_unlock_bh( &(sid->lock) );

				/* Handle receive shutdown. */
				sid->recv_dma.rcv_shutdown = rbuf->rcv_shutdown;
//// Start debug.
//				if( rbuf->rcv_shutdown ) {
//					printk( "host_recvmsg_fast: sid=%p, setting a "
//					    "rec_dma.rcv_shutdown\n", sid );
//				}
//// End debug.
				rbuf->hostCtrlFlag = UNM_HT_RXBUFF_EMPTY;
				host_send_rxbuffer(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,
					sid->portnum),rbuf, sid );
			}
		}
	}

fault:
	up(&(sid->recv_dma.lock));

	nx_print1("Exiting, sid=%p, rv=%d\n", sid, copied);
	return (copied);

recv_urg :
	spin_unlock_bh( &(sid->lock) ) ;
	if(URG_WITH_HOST == sid->urg_flg)
	{
		iov = msg->msg_iov ;
		smart_memcpy(iov->iov_base, &(sid->urg_data), 1);
		sid->urg_flg = URG_WITH_APP;
		up(&(sid->recv_dma.lock));
		return 1;
	}
	up(&(sid->recv_dma.lock));
	return -EINVAL; /* recv(..., MSG_OOB) should never block */
}

/* Checks for unread data in recv buffer */
int host_recv_available(pegnet_sock_id_t *sid,unsigned long arg)
{
	hostRxBuffer_t  *rbuf;
	struct list_head *p;
	int             bufferBytesRemaining,pageBytesRemaining;
	int 		answ=0;
	spin_lock_bh(&(sid->lock));
	if((sid->parent) && (sid->parent->sk) &&
           (sid->parent->sk->SK_STATE == TCP_LISTEN)){
		answ=-EINVAL; /* checking for listen state */
	}
	else if (!list_empty(&(sid->recv_dma.list))){
		if(URG_WITH_HOST == sid->urg_flg){
			answ = sid->urg_ptr - sid->abs_posn;                 
		}
		else {
			list_for_each(p, &(sid->recv_dma.list)){
				rbuf = list_entry(p, hostRxBuffer_t, list);
				bufferBytesRemaining = rbuf->bufferBytesRemaining;
				pageBytesRemaining = rbuf->pageBytesRemaining;
				answ += pageBytesRemaining + bufferBytesRemaining ;
			}	
		 	if(URG_WITH_APP == sid->urg_flg){
				answ = answ -1;
			}
		}
	}		
	spin_unlock_bh(&(sid->lock));
	return put_user(answ, (int __user *)arg);
}	

/*
 * This function traverses a list of already received buffers and reposts the
 * global buffers to the global buffer queue.
 */
void cleanup_rxbuffers_global(pegnet_sock_id_t *sid)
{
        hostRxBuffer_t		*rbuf;
        struct list_head	*node;
	struct list_head	*tmp_node;

        list_for_each_safe(node, tmp_node, &sid->recv_dma.list) {
                rbuf = list_entry(node, hostRxBuffer_t, list);

		if (rbuf->bufferType == GLOBAL_RXBUFFER) {
			list_del_init(&rbuf->list);
			rbuf->hostCtrlFlag = UNM_HT_RXBUFF_EMPTY;
			if (0 == host_send_rxbuffer(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,
					sid->portnum),rbuf, NULL)) {
				printk("%s: Reposting Global buffer failed\n",
				       __FUNCTION__);
			}
                }
        }
}

/* Free the pcb receive data dma buffers. */
void cleanup_rxbuffers_pcb( pegnet_sock_id_t *sid )
{
	int	i;
	int	jj;
	int	totalPages = 0 ;

	if (sid->recv_dma.remoteBuf == NULL) {
		/* nx_print("No receive buffers allocated\n"); */
		return;
	}
//// Start debug.
	if (waitqueue_active(&(sid->recv_dma.wq))) {
		nx_print("sid=%p: stuff on the wait q!!!\n", sid);
	}
//// End debug.

	for( i=0; i<sid->recv_dma.num_pcb_rxbuffers; i++ ) {
		hostRxBuffer_t *rbuf = &(sid->recv_dma.remoteBuf[i]) ;
		int numPages = rbuf->numPages ;

		for( jj=0; jj<numPages; jj++ ) {
			free_netxen_page(&rbuf->page[jj], 
			    nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum)),
			    PCI_DMA_FROMDEVICE);
			UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_RX_PAGES);
//// Start debug.
//            printk( "cleanup_rxbuffers_pcb: freeing rx buffer=%p, phys=0x%x, hostCtrlFlag=%d\n",
//                rbuf->page[jj].buffer, rbuf->page[jj].phys, rbuf->hostCtrlFlag ) ;
//// End debug.
			totalPages++ ;
		}
	}
//// Start debug.
//    printk( "cleanup_rxbuffers_pcb: sid=%p, freed %d pages\n", sid, totalPages ) ;
//// End debug.
	kfree(sid->recv_dma.remoteBuf);
	sid->recv_dma.remoteBuf = NULL;
}
