/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#include <linux/sockios.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#ifdef CONFIG_FILTER_UNM
#include <linux/filter.h>
#endif

#include <linux/version.h>
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if defined(MODVERSIONS) && !defined(__GENKSYMS__)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/modversions.h>
#endif
#endif

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/net.h>
#include <linux/poll.h>
#include <linux/skbuff.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/ip.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/inetdevice.h>

#include <linux/jhash.h>

#include <net/tcp.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#include <net/inet_hashtables.h>
#endif

#include <linux/inetdevice.h>
#include <linux/if.h>

#include <unm_inc.h>
#include <pegnet_inc.h>
#include "nic_cmn.h"
#include "hostInterface.h"
#include "flow_table.h"
#include "unm_host_pnic_types.h"


#define GLOBAL_TX_BUFFER_SIZE	4096
#define MAX_BYTES_PER_TX_QMSG	(GLOBAL_TX_BUFFER_SIZE * QMSG_TX_PAGES)


typedef struct {
	struct msghdr	*msg;
	int		iovlen;
	struct iovec	*iov;
	char		*from;
	int		seglen;
	int		copied;
	int		vec_cnt;
	int		try_reuse;
	int		compress_iovs;
	int		entries_per_qmsg;
	long		timeo;
} nx_msghdr_t;

typedef enum {NX_ZERO_COPY_TX_MODE, NX_COPY_TX_MODE } tx_mode_t;

#define SET_RB_POINT(VBUF, RB)                                          \
{                                                                       \
        (RB).page_bytes_remaining = (VBUF)->page_bytes_remaining;       \
        (RB).tot_bytes_sent = (VBUF)->tot_bytes_sent;                   \
        (RB).index = (VBUF)->index;                                     \
        (RB).refid = (VBUF)->refid;                                     \
        (RB).vec_cnt = (VBUF)->vec_cnt;                                 \
}
#define REVERT_TO_RB_POINT(VBUF, RB)                                    \
{                                                                       \
        (VBUF)->page_bytes_remaining = (RB).page_bytes_remaining;       \
        (VBUF)->tot_bytes_sent = (RB).tot_bytes_sent;                   \
        (VBUF)->index = (RB).index;                                     \
        (VBUF)->refid = (RB).refid;                                     \
        (VBUF)->vec_cnt = (RB).vec_cnt;                                 \
}


static void zero_copy_tx_callback(host_tx_vbuf_t *vbuf);
static void host_sendmsg_send_backlog(pegnet_sock_id_t *sid);

int host_sendpage(pegnet_sock_id_t *sid,struct page **pages, int poffset,
		  size_t psize, int flags);


/* Whenever possible compress qmsg IOVs before passing it to the card */
int		global_use_compressed_iov  = 0;
/* Size beyond which use zero copy transmits, if possible */
uint32_t	tx_zero_copy_size_thresh   = 0;
/* Connection threshold beyond which use zero copy transmits, if possible */
uint32_t	tx_zero_copy_active_thresh = 0;
/* Number of outstanding tx bytes to trigger zero copy */
uint32_t	tx_bytes_out_thresh = 1000000;


/*
 * We currently only enable TX zero-copy on systems that don't have an IO-MMU
 * The dma_map_page and pci_map_page APIs currently don't deal correctly with
 * user virtual addresses. Almost all of them assume the address to be mapped
 * was allocated using pci_alloc_consistent or came from the kernel kmalloc
 * pool. On systems that don't have an IO-MMU we can get away by NOT using the 
 * pci_map_* API. 
 * Furthermore on X86 and X86_64 systems we don't need to worry about flushing
 * dma buffers as the cache coherence hardware ensures all parties(CPU and
 * peripherals) will see consistent memory.
 */
static inline int use_zero_copy(struct msghdr *msg, pegnet_sock_id_t *sid)
{
#if defined(CONFIG_X86) || (defined(CONFIG_X86_64) && defined(PCI_DMA_BUS_IS_PHYS))
	if (sid->zero_copy_enabled) {

		if (msg->msg_iovlen > 1) {
			return 0;
		}

                if (msg->msg_iov->iov_len >= tx_zero_copy_size_thresh &&
		    atomic_read(&sid->adapter->num_tx_active_flows) >= tx_zero_copy_active_thresh &&
		    atomic_read(&sid->adapter->tx_bytes_out) >= tx_bytes_out_thresh) {

			return 1;
		}
	}
	return 0;
#else
	return 0;
#endif
}

/*
 * Allocates a virtual buffer containing multiple pages.
 *
 * Parameters:
 *      num_pages       - The total number of pages in the virtual buffer. This
 *                        is clamped to a maximum of UNM_HOST_TX_PAGES (14)
 *                        pages.
 *
 * Return:
 *      The allocated virtual buffer if successful or NULL in case of failure.
 */
static host_tx_vbuf_t *alloc_tx_vbuf(struct net_device *dev,vbuf_type_t type, int num_pages)
{
        int			ii;
	host_tx_vbuf_t		*vbuf;
        netxen_host_page_t	*page;


        if (num_pages > UNM_HOST_TX_PAGES) {
                printk("%s: num_pages[%d] > UNM_HOST_TX_PAGES[%d]. "
                       "Clamping to %d\n", __FUNCTION__, num_pages,
                       UNM_HOST_TX_PAGES, UNM_HOST_TX_PAGES);
                num_pages = UNM_HOST_TX_PAGES;
        }

        vbuf = (host_tx_vbuf_t *)kmalloc(sizeof (host_tx_vbuf_t), GFP_KERNEL);
        if (vbuf == NULL) {
                printk("%s: VBUF alloc failed\n", __FUNCTION__);
                return (NULL);
        }

        memset(vbuf, 0, sizeof (host_tx_vbuf_t));

	vbuf->type = type;

	if (vbuf->type == NX_COPY_TX_VBUF) {
		for (ii = 0; ii < num_pages; ii++) {
			page = &vbuf->page[ii];
			if (alloc_netxen_page(page)) {
				nx_print("Failed to allocate page buffer\n");
				break;
			}
			vbuf->tot_capacity += page->size;
			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_INUSE_TX_PAGES);
		}
		
		if (ii == num_pages) {
			vbuf->num_pages = ii;
			return (vbuf);
		}
		
		for (ii--; ii >= 0; ii--) {
			free_netxen_page(&vbuf->page[ii],nx_get_pcidev(dev),PCI_DMA_TODEVICE);
			UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_TX_PAGES);
		}
		
		kfree(vbuf);
		
		return (NULL);
	} else {
		// zero copy vbuf
		for (ii = 0; ii < num_pages; ii++) {
			page = &vbuf->page[ii];
			page->buffer = NULL;
			page->size = 0;
		}
		vbuf->num_pages = num_pages;
	}

	return vbuf;
}

/*
 *
 */
static void free_tx_vbuf(struct net_device *dev,host_tx_vbuf_t *vbuf)
{
        int	ii;


        if (vbuf->allocated) {
                printk("%s: vbuf[%p] is currently allocated and should not be "
                       "freed.\n", __FUNCTION__, vbuf);
                return;
        }

	if (vbuf->type == NX_COPY_TX_VBUF) {
		for (ii = 0; ii < vbuf->num_pages; ii++) {
			free_netxen_page(&vbuf->page[ii],nx_get_pcidev(dev),PCI_DMA_TODEVICE);
			UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_TX_PAGES);
		}
	} else {
		// zero copy vbuf
		ZERO_COPY_DBG("%s: vbuf = %p, tot_bytes_sent = %u\n",
			      __FUNCTION__, vbuf,
			      vbuf->tot_bytes_sent);
		for (ii = 0; ii < UNM_HOST_TX_PAGES; ii++) {
			if (vbuf->page[ii].phys) {
				ZERO_COPY_DBG("ii = %u, (vbuf = %p) Unmapping "
					      "%p, size = %u\n",
					      ii, vbuf,
					      (void *)(vbuf->page[ii].phys),
					      vbuf->page[ii].size);
				
				page_cache_release(vbuf->page[ii].mm_page);
				vbuf->page[ii].phys = 0;
				vbuf->page[ii].size = 0;
				vbuf->page[ii].buffer = NULL;
			}
		}
		if(vbuf->txcb) {  /* sendpage check */
			zero_copy_tx_callback(vbuf);
		}	
	}

	kfree(vbuf);
}

/*
 *
 */
static void free_tx_vbuf_queue(struct net_device *dev,struct list_head *list)
{
        host_tx_vbuf_t          *vbuf;

        while (!list_empty(list)) {
                vbuf = list_entry(list->next, host_tx_vbuf_t, list);
                list_del_init(&vbuf->list);
                vbuf->allocated = 0;
                free_tx_vbuf(dev,vbuf);
        }
}

/*
 *
 */
void cleanup_tx_vbuf_queues(pegnet_sock_id_t *sid)
{
        free_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.failed_list);
        free_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.vec_wait_list);
        free_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.sent_list);
        free_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.alloc_list);
        free_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.zero_wait_list);
        free_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.zcopy_switch_list);
}

/*
 * Allocates a list of free pages.
 */
static host_tx_vbuf_t *get_tx_vbuf(vbuf_type_t type, pegnet_sock_id_t *sid)
{
	host_tx_vbuf_t  *vbuf = NULL;


	if (type == NX_COPY_TX_VBUF) {
		spin_lock_bh(&sid->lock);
		
		if (!list_empty(&sid->send_dma.alloc_list)) {
			vbuf = list_entry(sid->send_dma.alloc_list.next,
					  host_tx_vbuf_t, list);
			list_del_init(&vbuf->list);
			atomic_dec(&sid->send_dma.alloc_q_len);
		}
		
		spin_unlock_bh(&sid->lock);
	}

        if (vbuf == NULL) {
                vbuf = alloc_tx_vbuf(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
			    type, sid->send_dma.num_vbuf_pages);
                if (vbuf == NULL) {
                        printk("%s: Alloc failed for vbuf\n", __FUNCTION__);
                        return (NULL);
                }
		if (type == NX_COPY_TX_VBUF) {
			atomic_inc(&sid->send_dma.allocated);
		}
		if (type == NX_ZERO_COPY_TX_VBUF) {
			atomic_inc(&sid->send_dma.zerocopy_vbuf_cnt);
		}
        }

        vbuf->allocated = 1;
        vbuf->full = 0;
        vbuf->tx_flags = 0;

        vbuf->failed_ptr = 0;
        vbuf->vec_cnt = 0;
        vbuf->penalty_vecs = 0;
        vbuf->index = 0;
        vbuf->refid = 0;
        vbuf->tot_bytes_sent = 0;
        vbuf->page_bytes_remaining = vbuf->page[0].size;
        vbuf->vec_wait_idx = 0;
        vbuf->failed_idx = 0;
	vbuf->type = type;
	vbuf->txcb = NULL;
	vbuf->try_reuse = (type == NX_COPY_TX_VBUF) ? 1 : 0;

        return (vbuf);
}

/*
 * prepare the buffer for DMA(using pci_map_single) and store
 * the paddrs into the host_tx_page structure
 */
static int add_tx_page_to_vbuf(host_tx_vbuf_t *vbuf, netxen_host_page_t *page,
			       char *buffer, int chunksize)
{
	int		ret;
	unsigned long	uaddr = (unsigned long)buffer; 
	dma_addr_t	temp;

	
	ZERO_COPY_DBG("%s:vbuf = %p, page = %p, buffer = %p, chunksize = %d\n",
		      __FUNCTION__, vbuf, page, buffer, chunksize);
	
	down_read(&current->mm->mmap_sem);
	ret = get_user_pages(current, current->mm, uaddr & PAGE_MASK,
			     1, 0, 0, &(page->mm_page), NULL);
	
	up_read(&current->mm->mmap_sem);
	
	if (ret < 1) {
		printk("%s: get_user_pages failed\n", __FUNCTION__);
		return -1; // get_user_pages failed
	}
	
	temp = page_to_pfn(page->mm_page) << PAGE_SHIFT;
	page->phys = temp;
	page->phys += offset_in_page(uaddr);
	if (page->phys == 0) {
		printk("%s: pci_map_single failed on buffer = %p, size = %d\n",
		       __FUNCTION__, buffer, chunksize);

		page_cache_release(page->mm_page);
		return -1;
	}

	page->size = chunksize;
	page->zcopy_user_addr = buffer;
	vbuf->page_bytes_remaining = 0;

	ZERO_COPY_DBG("%s: physaddr = %p\n", __FUNCTION__,
		      (void *)(page->phys));
	return 0;
}


/*
 * Frees a list of free pages.
 */
static inline void put_tx_vbuf(pegnet_sock_id_t *sid, host_tx_vbuf_t *vbuf)
{
	int	i;

        vbuf->allocated = 0;


        if (vbuf->type == NX_ZERO_COPY_TX_VBUF ||
	    atomic_read(&sid->send_dma.allocated) > UNM_TX_FREE_VBUF_THRESHOLD) {
                /* printk("%s: releasing to kernel\n", __FUNCTION__); */
                free_tx_vbuf(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),vbuf);
		if (vbuf->type == NX_COPY_TX_VBUF) {
			atomic_dec(&sid->send_dma.allocated);
		}
		if (vbuf->type == NX_ZERO_COPY_TX_VBUF) {
			atomic_dec(&sid->send_dma.zerocopy_vbuf_cnt);
		}
                return;
        }

	for (i = 0; i < vbuf->num_pages; i++) {
		netxen_page_unmap_pci(&vbuf->page[i],
			nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum))
			,PCI_DMA_TODEVICE);
	}

        list_add_tail(&vbuf->list, &sid->send_dma.alloc_list);
        atomic_inc(&sid->send_dma.alloc_q_len);
}

static inline void set_tx_iov(struct pci_dev *unm_device_ptr,
			      txbuf_msg_t *txmsg, dma_addr_t addr,
			      uint32_t len, int compress_iovs)
{
	uint32_t index = txmsg->numBufs;

	pci_dma_sync_single_for_device(unm_device_ptr, addr, len,
				       PCI_DMA_TODEVICE);

	// nx_print("addr = %u, len = %u\n", addr, len);
	if (compress_iovs) {
		txmsg->compressed_iov = 1;
		txmsg->iov40.end_len  = len;
		txmsg->iov40.addr32[index] =  (__uint32_t) addr;
		txmsg->iov40.addr_hi[index] = (__uint8_t) (addr >> 31);
	} else {
		txmsg->compressed_iov = 0;
		txmsg->iov[index].len = len;
		txmsg->iov[index].addr = addr;
	}
	txmsg->numBufs++;
}

/*
 * This function resends buffers that could not be accepted by the card before.
 * It only sends one vbuf at a time. The main reason is that we don't flood the
 * card and go back into failure.
 */
static void host_sendmsg_resend_bufs(pegnet_sock_id_t *sid)
{
        host_tx_vbuf_t          *vbuf;
	netxen_host_page_t	*page;
	txbuf_msg_t		msg;
	dma_addr_t		addr;
	uint32_t		len;
        int                     failed_idx;

        printk("%s: Resend after failure\n", __FUNCTION__);

        if (list_empty(&sid->send_dma.failed_list)) {
                printk("%s: Error: Nothing to resend\n", __FUNCTION__);
                goto done;
        }

	/* Queue message initialization. */
	memset(&msg.hdr, 0, sizeof (msg.hdr));
	msg.bufferStart = 0;
	msg.bufferComplete = 0;
	msg.cookie = 0;
	msg.numBufs = 0;
	msg.hdr.dst_subq = 1;
	msg.hdr.dst_side  = sid->qid.side;
	msg.hdr.dst_major = sid->qid.major;
	msg.hdr.dst_minor = sid->qid.minor;
	msg.hdr.type = UNM_MSGTYPE_SENDMSG;

        vbuf = list_entry(sid->send_dma.failed_list.next,
                          host_tx_vbuf_t, list);
        list_del_init(&vbuf->list);
        sid->send_dma.fl_cnt--;
        failed_idx = vbuf->failed_idx;

        while (vbuf) {
                page = &vbuf->page[failed_idx];
                addr = page->phys;
                len = page->size;
                if (failed_idx == vbuf->index) {
                        len -= vbuf->page_bytes_remaining;
                }

                if (vbuf->failed_ptr && vbuf->failed_idx == failed_idx) {
                        len -= vbuf->failed_ptr - page->phys;
                        addr = vbuf->failed_ptr;
                }

                nx_print("vbuf->failed_ptr = 0x%llx, vbuf->failed_idx = %u\n",
			 (unsigned long long)vbuf->failed_ptr, (uint32_t)vbuf->failed_idx);
                nx_print("addr = 0x%llx, len = %u\n", (unsigned long long)addr, len);
                {
                        int             jj;
                        uint8_t         *data;

                        data = ((uint8_t *)netxen_page_map_virt(page)) +
				addr - page->phys;
                        printk("%s: bytes = ", __FUNCTION__);
                        for (jj = 0; jj < 16; jj++) {
                                printk("%02x ", data[jj]);
                        }
                        printk("\n");
                }
                /*
                 * There is a chance that the page pointed to by index does
                 * not have any data in it. But index cannot be 0 in this case.
                 * So skip over the page.
                 */
                if (len) {
                        /*
                         * Add the current page to the queue message.
                         */
			set_tx_iov(nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
			    sid->adapter,sid->portnum)),&msg, addr, len, 0);
                } else {
                        if (failed_idx == 0) {
                                printk("%s:%d: ERROR: This should not "
                                       "happen\n", __FUNCTION__, __LINE__);
                        }
                }

                failed_idx++;
                if (failed_idx == vbuf->index &&
                    vbuf->page[failed_idx].size == vbuf->page_bytes_remaining) {
                        printk("%s:%d: Skipping unused page. numBufs = %u\n",
                               __FUNCTION__, __LINE__, msg.numBufs);
                        failed_idx++;
                }

                if (msg.numBufs && (QMSG_TX_PAGES == msg.numBufs ||
                                    failed_idx > vbuf->index)) {
                        /*
                         * Take a completion when vbuf is done.
                         */
                        msg.cookie = sid->send_dma.refid;
                        if (failed_idx > vbuf->index) {

                                msg.bufferComplete = 1;
                                vbuf->refid = sid->send_dma.refid;
                                sid->send_dma.refid++;
                                if (sid->send_dma.refid == 0) {
                                        sid->send_dma.refid++;
                                }
                        } else {
                                msg.bufferComplete = 0;
                        }

                        if (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(
			    sid->adapter,sid->portnum),(unm_msg_t *)&msg,
			    sizeof (msg.body))) {
                                printk("%s: Failed queueing\n", __FUNCTION__);
                                list_add(&vbuf->list,
                                         &sid->send_dma.failed_list);
                                sid->send_dma.fl_cnt++;
                                break;
                        }
                        msg.numBufs = 0;
                        vbuf->failed_idx = failed_idx;
                        vbuf->failed_ptr = 0;
		}

                if (failed_idx > vbuf->index) {
                        vbuf->failed_idx = 0;
                        vbuf->failed_ptr = 0;
                        list_add_tail(&vbuf->list, &sid->send_dma.sent_list);
                        sid->send_dma.sl_cnt++;
                        vbuf = NULL;
                }
	}

        /*
         * If the last data was sent to the card and the SEND_SHUTDOWN is set
         * then notify the card of the shutdown.
         */
        if ((sid->shutdown & SEND_SHUTDOWN) &&
            list_empty(&sid->send_dma.failed_list) &&
            list_empty(&sid->send_dma.vec_wait_list) &&
	    !sid->released) {
                int     rv;

                rv = host_conn_disconnect(sid, 0, 1, 0);
                if (rv != 0) {
                        printk("%s: sending tx_shutdown to card FAILED[%d]\n",
                               __FUNCTION__, rv);
                }
        }

  done:
	/* If we have user context waiting for resource wake them up. */
	if (list_empty(&sid->send_dma.failed_list) &&
            sid->send_dma.need_failure_wakeup) {

		if (sid->parent && sid->parent->sk &&
		    sid->parent->sk->SK_SLEEP &&
		    waitqueue_active(sid->parent->sk->SK_SLEEP)) {
			wake_up_interruptible(sid->parent->sk->SK_SLEEP);
		}

		wake_up_all(&sid->send_dma.wq);
	}
}

/*
 *
 */
static void host_sendmsg_send_veclist(pegnet_sock_id_t *sid)
{
        host_tx_vbuf_t          *vbuf;
	netxen_host_page_t	*page;
	txbuf_msg_t		msg;
	dma_addr_t		addr;
	uint32_t		len;
	int16_t			vecs;
        int16_t                 vec_cnt;
        int                     mss = 1460;     /* sid->mss */
        int                     vec_wait_idx;
        int                     vbuf_vec_cnt;


	/* Queue message initialization. */
	memset(&msg.hdr, 0, sizeof (msg.hdr));
	msg.bufferStart = 0;
	msg.bufferComplete = 0;
	msg.cookie = 0;
	msg.numBufs = 0;
	msg.hdr.dst_subq = 1;
	msg.hdr.dst_side  = sid->qid.side;
	msg.hdr.dst_major = sid->qid.major;
	msg.hdr.dst_minor = sid->qid.minor;
	msg.hdr.type = UNM_MSGTYPE_SENDMSG;

        vbuf = NULL;
	vec_wait_idx = 0;
        vbuf_vec_cnt = 0;
        vec_cnt = sid->send_dma.vec_cnt;
        while (vec_cnt > 0) {
                if (vbuf == NULL) {
                        if (list_empty(&sid->send_dma.vec_wait_list)) {
                                vbuf = NULL;
                                break;
                        }

                        vbuf = list_entry(sid->send_dma.vec_wait_list.next,
                                          host_tx_vbuf_t, list);
                        list_del_init(&vbuf->list);
                        sid->send_dma.vl_cnt--;
                        vec_wait_idx = vbuf->vec_wait_idx;
                        vbuf_vec_cnt = 0;
                }

                page = &vbuf->page[vec_wait_idx];
                addr = page->phys;
                len = page->size;
                if (vec_wait_idx == vbuf->index) {
                        len -= vbuf->page_bytes_remaining;
                }

                /*
                 * There is a chance that the page pointed to by index does
                 * not have any data in it. But index cannot be 0 in this case.
                 * So skip over the page.
                 */
                if (len) {
                        /*
                         * Add the current page to the queue message.
                         */
			set_tx_iov(nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
			    sid->adapter,sid->portnum)),&msg, addr, len, 0);
                        vecs = (len + mss - 1) / mss;
                        vec_cnt -= vecs;
                        vbuf_vec_cnt += vecs;
                } else {
                        if (vec_wait_idx == 0) {
                                printk("%s:%d: ERROR: This should not "
                                       "happen\n", __FUNCTION__, __LINE__);
                        }
                }

                vec_wait_idx++;
                if (vec_wait_idx == vbuf->index &&
                    vbuf->page[vec_wait_idx].size == vbuf->page_bytes_remaining) {
                        /* Skip unused page */
                        vec_wait_idx++;
                }

                if (QMSG_TX_PAGES == msg.numBufs || vec_cnt <= 0 ||
                    vec_wait_idx > vbuf->index) {

                        /*
                         * Take a completion when vbuf is done.
                         */
                        msg.cookie = sid->send_dma.refid;
                        if (vec_wait_idx > vbuf->index) {
				if (vbuf->penalty_vecs) {
					vec_cnt -= UNM_SEND_BUF_QUEUED_PENALTY;
					vbuf_vec_cnt += UNM_SEND_BUF_QUEUED_PENALTY;
				}
				vbuf->penalty_vecs = 0;

                                msg.bufferComplete = 1;
                                vbuf->refid = sid->send_dma.refid;
                                sid->send_dma.refid++;
                                if (sid->send_dma.refid == 0) {
                                        sid->send_dma.refid++;
                                }
                        } else {
                                msg.bufferComplete = 0;
                        }

                        if (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(
			    sid->adapter,sid->portnum),(unm_msg_t *)&msg,
			    sizeof (msg.body))) {
                                nx_print("Failed queueing\n");
                                list_add(&vbuf->list,
                                         &sid->send_dma.vec_wait_list);
                                sid->send_dma.vl_cnt++;
                                break;
                        }
                        msg.numBufs = 0;

                        sid->send_dma.vec_cnt = vec_cnt;
                        vbuf->vec_wait_idx = vec_wait_idx;
                        vbuf->vec_cnt += vbuf_vec_cnt;
                        vbuf_vec_cnt = 0;

                        if (vec_wait_idx > vbuf->index) {
                                vbuf->vec_wait_idx = 0;
                                list_add_tail(&vbuf->list,
                                              &sid->send_dma.sent_list);
                                sid->send_dma.sl_cnt++;
                                vbuf = NULL;

                        } else if (vec_cnt <= 0) {
                                /*
                                 * Insert back in the wait list.
                                 */
                                list_add(&vbuf->list,
                                         &sid->send_dma.vec_wait_list);
                                sid->send_dma.vl_cnt++;
                                vbuf = NULL;
                        }
		}
	}

        /*
         * If the last data was sent to the card and the SEND_SHUTDOWN is set
         * then notify the card of the shutdown.
         */
        if ((sid->shutdown & SEND_SHUTDOWN) &&
            list_empty(&sid->send_dma.failed_list) &&
            list_empty(&sid->send_dma.vec_wait_list) &&
	    !sid->released) {
                int     rv;

                rv = host_conn_disconnect(sid, 0, 1, 0);
                if (rv != 0) {
                        printk("%s: sending tx_shutdown to card FAILED[%d]\n",
                               __FUNCTION__, rv);
                }
        }
}

void process_last_tx_completion(pegnet_sock_id_t *sid)
{
	uint32_t		refid;
	host_tx_vbuf_t		*vbuf;
	struct list_head	*node;
	host_tx_vbuf_t		*refid_vbuf;


	if (!sid->send_dma.last_comp_valid) {
		nx_print("Called when invalid\n");
		return;
	}
	sid->send_dma.last_comp_valid = 0;
	refid = sid->send_dma.last_comp_cookie;

	/*
	 * Traverse the nodes to be released.
	 */
	refid_vbuf = NULL;

	list_for_each(node, &sid->send_dma.sent_list) {
		vbuf = (host_tx_vbuf_t *)node;
		if (refid == vbuf->refid) {
			refid_vbuf = vbuf;
		} else if (refid_vbuf) {
                        break;
                }

		if (!vbuf->allocated) {
			nx_print("vbuf = %p, No marked as allocated\n", vbuf);
		}
	}
	if (refid_vbuf) {
                struct list_head	*end_node;
		/*
		 * delink the free nodes from the current list and hook it to
		 * a new head
		 */
                node = sid->send_dma.sent_list.next;
                end_node = refid_vbuf->list.next;
                while (node != end_node) {
                        vbuf = (host_tx_vbuf_t *)node;
                        atomic_sub(vbuf->tot_bytes_sent, &sid->adapter->tx_bytes_out);
                        sid->send_dma.vec_cnt += vbuf->vec_cnt;
                        if (sid->send_dma.vec_cnt > 512) {
                                nx_print("vec_cnt[%d] > 512 +_+_+_+_+_+_+\n",
					 sid->send_dma.vec_cnt);
                        }

                        node = node->next;
                        list_del_init(&vbuf->list);
                        sid->send_dma.sl_cnt--;
                        put_tx_vbuf(sid, vbuf);
                }
        }

        if (sid->send_dma.need_buffer_wakeup || sid->released) {

                /* If we have send buffers available, tell everyone. */
                if (sid->parent && sid->parent->sk &&
                    sid->parent->sk->SK_SLEEP &&
                    waitqueue_active(sid->parent->sk->SK_SLEEP)) {
                        wake_up_interruptible(sid->parent->sk->SK_SLEEP);
                }

		sid->send_dma.need_buffer_wakeup = 0;
                wake_up_all(&sid->send_dma.wq);
        }
}

/*
 * Handle empty send buffers from the card.
 */
void proc_tx_completion(struct net_device *netdev,unm_msg_t *msg)
{
	txbuff_completion_t	*body;
	pegnet_sock_id_t	*sid;
	uint32_t		connID;
	nx_tnic_adapter_t 	*tnic_adapter = nx_get_dev_tnic_adapter(netdev);	


	body = (txbuff_completion_t *)(&msg->body);
	connID = host_get_uint32_from_peg(body->ctx);
	sid = retreive_pegnet_sock_id(tnic_adapter->p_ftable,connID);
        if (sid == NULL) {
		if (body->comp_type != NX_TX_COMP_SHUTDOWN) {
			nx_print("Socket not found for context %lu\n",
				 (unsigned long)body->ctx);
		}
                return;
        }

	spin_lock(&sid->lock);

	if (body->comp_type & NX_TX_COMP_SHUTDOWN) {
		nx_print1("FIN Complete. Cleanup TX buffers: %u\n", body->ctx);
                cleanup_tx_vbuf_queues(sid);
                wake_up_all(&sid->send_dma.wq);
		goto done;
	}

	sid->send_dma.last_comp_cookie = body->cookie;
	sid->send_dma.last_comp_valid = 1;

        if (!sid->send_dma.in_sendmsg) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_TX_FREEING_IN_BH);
		process_last_tx_completion(sid);

                if (!list_empty(&sid->send_dma.failed_list)) {

                        host_sendmsg_resend_bufs(sid);

                } else if (sid->send_dma.vec_cnt > 0 &&
                           !list_empty(&sid->send_dma.vec_wait_list)) {

                        host_sendmsg_send_veclist(sid);

                } else {
			if (waitqueue_active(&sid->send_dma.wq) &&
			    list_empty(&sid->send_dma.sent_list) &&
			    list_empty(&sid->send_dma.vec_wait_list))
				wake_up(&sid->send_dma.wq);
		}
        } else {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_TX_DELAY_FREEING);
	}

        if (sid->send_dma.need_buffer_wakeup || sid->released) {

		/* If we have send buffers available, tell everyone. */
		if (sid->parent && sid->parent->sk &&
		    sid->parent->sk->SK_SLEEP &&
		    waitqueue_active(sid->parent->sk->SK_SLEEP)) {
			wake_up_interruptible(sid->parent->sk->SK_SLEEP);
		}

		sid->send_dma.need_buffer_wakeup = 0;
		wake_up_all(&sid->send_dma.wq);
        }

  done:
	spin_unlock(&sid->lock);

	host_flowtbl_put_reference(tnic_adapter->p_ftable,connID);
}

void proc_tx_switch_completion(struct net_device *netdev,unm_msg_t *msg)
{
	txbuf_switch_completion_t	*body;
	pegnet_sock_id_t		*sid;
	uint32_t			connID;
	host_tx_vbuf_t			*vbuf;
	struct list_head		*node;
	struct list_head		*tmp_node;
	int				found = 0;
	nx_tnic_adapter_t 		*tnic_adapter = nx_get_dev_tnic_adapter(netdev);	


	body = (txbuf_switch_completion_t *)(&msg->body);

	nx_print1("Received txbuf_switch_completion_t: refid = %u\n",
		  (unsigned int)body->cookie);

	connID = host_get_uint32_from_peg(body->ctx);
	sid = retreive_pegnet_sock_id(tnic_adapter->p_ftable,connID);
        if (sid == NULL) {
                nx_print("Socket not found for context %lu\n",
			 (unsigned long)body->ctx);
                return;
        }

	spin_lock(&sid->lock);

	/*
	 * Traverse the nodes to be released.
	 */
        list_for_each_safe(node, tmp_node, &sid->send_dma.zcopy_switch_list) {
		vbuf = list_entry(node, host_tx_vbuf_t, list);
		if (body->cookie == vbuf->refid) {
			nx_print1("Found and released vbuf for refid %u\n",
				  (unsigned int)body->cookie);
                        list_del_init(&vbuf->list);
                        put_tx_vbuf(sid, vbuf);
			found = 1;
                        break;
                } else {
			nx_print1("Not the vbuf[%u], refid %u\n", vbuf->refid,
				  (unsigned int)body->cookie);
		}
	}

	if (!found) {
		nx_print("VBUF not found for cookie[%u]\n",
			 (unsigned int)body->cookie);
	}
	spin_unlock(&sid->lock);

	host_flowtbl_put_reference(tnic_adapter->p_ftable,connID);
}

/*
 * Splits starting from loc from list and inserts in the head of new_list.
 */
static __inline__ void unm_list_split(struct list_head *list,
                                      struct list_head *new_list,
                                      struct list_head *loc)
{
        new_list->next->prev = list->prev;
        list->prev->next = new_list->next;
        new_list->next = loc;

        loc->prev->next = list;
        list->prev = loc->prev;
        loc->prev = new_list;
}

/*
 *
 */
static void requeue_sendmsg_failed_ptr(pegnet_sock_id_t *sid)
{
	host_tx_vbuf_t          *vbuf;
	netxen_host_page_t	*page;
	struct list_head	*node;
	int			found;
	int			i;


        if (!list_empty(&sid->send_dma.failed_list)) {
                printk("%s: Got another failure when clearing previous "
                       "failure: sid = %p, pointer = 0x%llx\n", __FUNCTION__, sid,
                       (unsigned long long)sid->send_dma.failed_ptr);
                vbuf = list_entry(sid->send_dma.failed_list.next,
                                  host_tx_vbuf_t, list);
                vbuf->failed_ptr = 0;
                vbuf->failed_idx = 0;
        }

	/*
	 * Traverse the vbufs and find the page with the failure and
         * split the list and create the failed_list.
	 */
	found = 0;
	list_for_each(node, &sid->send_dma.sent_list) {
		vbuf = (host_tx_vbuf_t *)node;
                for (i = 0; i <= vbuf->index; i++) {
                        page = &vbuf->page[i];
                        if ((sid->send_dma.failed_ptr &
                             (~(page->size - 1))) == (uint64_t)page->phys) {

                                printk("%s: found page: refid[%u], page[%u], "
                                       "pointer[0x%llx], phys[0x%llx]\n",
                                       __FUNCTION__, vbuf->refid, i,
                                       (unsigned long long)sid->send_dma.failed_ptr,
                                       (unsigned long long)page->phys);
                                found = 1;
                                vbuf->failed_ptr = sid->send_dma.failed_ptr;
                                vbuf->failed_idx = i;
                                if (page->size == vbuf->page_bytes_remaining) {
                                        printk("%s: Failed page is invalid\n",
                                               __FUNCTION__);
                                }
                                unm_list_split(&sid->send_dma.sent_list,
                                               &sid->send_dma.failed_list,
                                               &vbuf->list);
                                break;
                        }
                }
                if (found) {
                        break;
                }
	}

	if (!found) {
		printk("%s: No Page Found for: sid = %p, pointer = 0x%llx\n",
		       __FUNCTION__, sid, (unsigned long long)sid->send_dma.failed_ptr);
	}

        sid->send_dma.failed_ptr = 0;
}

/*
 * Failure happens when the card fails to allocate memory and will notify the
 * host the pointer it could not accept.
 *
 * Even before a failure is cleared a new failure can happen and has to be
 * handled.
 */
void proc_sendmsg_control(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg,
			  pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
        peg_sendmsg_control_t	*cntl_msg;

        cntl_msg = (peg_sendmsg_control_t *)&msg->body;


	printk("%s: sid = %p, pointer = %llx\n", __FUNCTION__, sid,
               cntl_msg->pointer);

	spin_lock_bh(&sid->lock);

        if (sid->send_dma.failed_ptr) {
                printk("%s: Got another failure when first has not "
                       "been marked: sid = %p, pointer = %llx\n",
                       __FUNCTION__, sid, cntl_msg->pointer);
        }

        sid->send_dma.failed_ptr = (dma_addr_t)cntl_msg->pointer;
        if (!sid->send_dma.in_sendmsg) {
                requeue_sendmsg_failed_ptr(sid);

		/* Try to resend the first buffer */
		if (!list_empty(&sid->send_dma.failed_list)) {
			host_sendmsg_resend_bufs(sid);
		}
        }

	spin_unlock_bh(&sid->lock);
}

/*
 *
 */
static int wait_for_returned_vbufs(vbuf_type_t type, pegnet_sock_id_t *sid,
				   long *timeo_p)
{
        int             rc = 0;
        DECLARE_WAITQUEUE(wait, current);

	/* nx_print("Waiting for returned buffer\n"); */

        spin_lock_bh(&sid->lock);

        add_wait_queue(&sid->send_dma.wq, &wait);

        while (!is_tx_vbuf_allocatable(type, sid) &&
               !(sid->shutdown & SEND_SHUTDOWN)) {

                set_current_state(TASK_INTERRUPTIBLE);
		if (!*timeo_p) {
			rc = -EAGAIN;
			goto wait_done;
		}

                sid->send_dma.need_buffer_wakeup = 1;
                spin_unlock_bh(&sid->lock);
                if (signal_pending(current)) {
                        spin_lock_bh(&sid->lock);
			sid->send_dma.need_buffer_wakeup = 0;
			nx_print("got signal\n");
                        rc = sock_intr_errno(*timeo_p);
                        goto wait_done;
                }

                *timeo_p = schedule_timeout(*timeo_p);

                spin_lock_bh(&sid->lock);

		if (sid->send_dma.last_comp_valid) {
			sid->send_dma.vec_cnt -= 2 * UNM_SEND_BUF_MAX_VECS;
			process_last_tx_completion(sid);
			sid->send_dma.vec_cnt += 2 * UNM_SEND_BUF_MAX_VECS;
			if (sid->send_dma.vec_cnt > 0 &&
			    !list_empty(&sid->send_dma.vec_wait_list)) {
				host_sendmsg_send_veclist(sid);
			}
		}

                sid->send_dma.need_buffer_wakeup = 0;
        }

  wait_done:
        set_current_state(TASK_RUNNING);
        remove_wait_queue(&sid->send_dma.wq, &wait);

        if (sid->shutdown & SEND_SHUTDOWN) {
                rc = -EPIPE;
        }

        spin_unlock_bh(&sid->lock);

        return (rc);
}

/*
 *
 */
static inline int tx_sent_list_add_tail(pegnet_sock_id_t *sid,
                                        host_tx_vbuf_t *vbuf)
{
        if (sid->shutdown & SEND_SHUTDOWN) {
                put_tx_vbuf(sid, vbuf);
                return (-EPIPE);
        }

        list_add_tail(&vbuf->list, &sid->send_dma.sent_list);
        sid->send_dma.sl_cnt++;
        return (0);
}

/*
 *
 */
static inline int tx_vec_wait_list_add_tail(pegnet_sock_id_t *sid,
                                            host_tx_vbuf_t *vbuf)
{
        if (sid->shutdown & SEND_SHUTDOWN) {
                put_tx_vbuf(sid, vbuf);
                return (-EPIPE);
        }

        list_add_tail(&vbuf->list, &sid->send_dma.vec_wait_list);
        sid->send_dma.vl_cnt++;
        return (0);
}

/*
 *
 */
static inline int tx_zero_wait_list_add_tail(pegnet_sock_id_t *sid,
					     host_tx_vbuf_t *vbuf)
{
        if (sid->shutdown & SEND_SHUTDOWN) {
                put_tx_vbuf(sid, vbuf);
                return (-EPIPE);
        }

        list_add_tail(&vbuf->list, &sid->send_dma.zero_wait_list);
        return (0);
}


static void init_zero_copy_tx_cb(zero_copy_tx_cb_t *txcb)
{
	init_waitqueue_head( &(txcb->block_wq));  
	txcb->total_bytes = 0;
	txcb->bytes_acked = 0;
}

/*
 * Test if the all the bytes sent have been acknowledged by
 * the peer.
 */
static int zero_copy_tx_is_complete(zero_copy_tx_cb_t *txcb)
{
	return (txcb->total_bytes >= 0 && 
		txcb->total_bytes == txcb->bytes_acked);
}

static unsigned long getTSC(void)
{
#ifdef ZERO_COPY_TX_SAMPLE
	unsigned long temp;
	
	rdtscll(temp);
	
	return temp;
#else
	return 0;
#endif // ZERO_COPY_TX_SAMPLE
}

/*
 * If the blocked write has completed(i.e acknowledged
 * by the peer) then release the blocked send thread
 *
 */
static void zero_copy_tx_callback(host_tx_vbuf_t *vbuf)
{
	zero_copy_tx_cb_t	*txcb = vbuf->txcb;

	if (txcb->bytes_acked == 0) {
		txcb->tstamp_end = getTSC();
	}
	txcb->bytes_acked += vbuf->tot_bytes_sent;

	ZERO_COPY_DBG("%s:  txcb = %p, bytes_acked = %u, total_bytes = %u\n", 
		      __FUNCTION__, txcb, txcb->bytes_acked,
		      txcb->total_bytes);
	
	if (zero_copy_tx_is_complete(txcb)) {
		ZERO_COPY_DBG("waking up thread\n");
		wake_up(&(txcb->block_wq));
	}

	vbuf->txcb = NULL;
}

/*
 *
 */
static void zcopy_clear_tx_vbuf_queue(struct net_device *dev,
				      struct list_head *list,
				      zero_copy_tx_cb_t *txcb)
{
        host_tx_vbuf_t          *vbuf;
        struct list_head	*node;
	struct list_head	*tmp_node;

        list_for_each_safe(node, tmp_node, list) {
                vbuf = list_entry(node, host_tx_vbuf_t, list);
		if (vbuf->txcb == txcb) {
			list_del_init(&vbuf->list);
			vbuf->allocated = 0;
#if 0
			printk("%s:  txcb = %p, bytes_acked = %u, "
			       "total_bytes = %u\n", 
			       __FUNCTION__, txcb, txcb->bytes_acked,
			       txcb->total_bytes);
#endif
			free_tx_vbuf(dev,vbuf);
		}
        }
}

/*
 *
 */
static void zcopy_clear_tx_vbuf_queues(pegnet_sock_id_t *sid,
				       zero_copy_tx_cb_t *txcb)
{
        zcopy_clear_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.failed_list, txcb);
        zcopy_clear_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.vec_wait_list, txcb);
        zcopy_clear_tx_vbuf_queue(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum),
		&sid->send_dma.sent_list, txcb);
}

/*
 *
 */
int is_tx_vbuf_allocatable(vbuf_type_t type, pegnet_sock_id_t *sid)
{
        int	max_zcopy;
	int	max_rcopy;

	if (type == NX_ZERO_COPY_TX_VBUF) {
        	max_zcopy = (sid->high_perf_flag ?
			     UNM_TX_VBUF_MAX_HP : UNM_ZCOPY_TX_VBUF_MAX_LP);
        	return (atomic_read(&sid->send_dma.zerocopy_vbuf_cnt) < max_zcopy);
	} else if (type == NX_COPY_TX_VBUF){
        	max_rcopy = (sid->high_perf_flag ?
			     UNM_TX_VBUF_MAX_HP : UNM_TX_VBUF_MAX_LP);
        	return ((atomic_read(&sid->send_dma.alloc_q_len) != 0) ||
                	atomic_read(&sid->send_dma.allocated) < max_rcopy);

	} else if (type == NX_ZERO_AND_REG_TX_VBUF){
        	max_zcopy = (sid->high_perf_flag ?
			     UNM_TX_VBUF_MAX_HP : UNM_ZCOPY_TX_VBUF_MAX_LP);
        	max_rcopy = (sid->high_perf_flag ?
			     UNM_TX_VBUF_MAX_HP : UNM_TX_VBUF_MAX_LP);
        	return ((atomic_read(&sid->send_dma.zerocopy_vbuf_cnt) < max_zcopy)
			&&((atomic_read(&sid->send_dma.alloc_q_len) != 0) ||
           		atomic_read(&sid->send_dma.allocated) < max_rcopy));
	}
	nx_print("invalid type:0x%x\n", type);
	return 0;
}

/*
 *
 */
static inline char *nx_tx_get_copy_addr(struct pci_dev *unm_device_ptr,
			host_tx_vbuf_t *vbuf, dma_addr_t *phys)
{
	netxen_host_page_t	*page;
	char			*to;
	int			idx;


	page = &vbuf->page[vbuf->index];
	idx = page->size - vbuf->page_bytes_remaining;
	to = netxen_page_map_virt(page);
	if (to == NULL || netxen_page_map_pci(page, unm_device_ptr,
					      PCI_DMA_TODEVICE)) {
		nx_print("virtual/physical address mapping failed\n");
		return (NULL);
	}
	to += idx;
	if (idx) {
		pci_dma_sync_single_for_cpu(unm_device_ptr, page->phys + idx,
					    vbuf->page_bytes_remaining,
					    PCI_DMA_TODEVICE);
	}
	*phys = page->phys + idx;

	return (to);
}

static inline void nx_tx_inc_page(host_tx_vbuf_t *vbuf)
{
	netxen_page_unmap_virt(&vbuf->page[vbuf->index]);

	vbuf->index++;
	if (vbuf->index == vbuf->num_pages) {
		vbuf->full = 1;
		vbuf->index--;          /* Keep it in valid range */
	} else {
		vbuf->page_bytes_remaining = vbuf->page[vbuf->index].size;
	}
}

static void zero_copy_print_stats(nx_tnic_adapter_t *tnic_adapter,
				  unsigned long before, unsigned long after,
				  int blocked, zero_copy_tx_cb_t *txcb)
{
        ZERO_COPY_DBG("%s: Enter\n", __FUNCTION__);
        if (blocked) {
                tnic_adapter->zcopy_stats.num_blocks++;
                tnic_adapter->zcopy_stats.delta += (after-before)/3000;

                if (time_after(txcb->tstamp_end,
                               txcb->tstamp_start)) {
                        tnic_adapter->zcopy_stats.ack_delta +=
                                (txcb->tstamp_end - txcb->tstamp_start) / 3000;
                }
        }
        tnic_adapter->zcopy_stats.num_completions++;
        tnic_adapter->zcopy_stats.write_delta += ((txcb->write_end) -
						  (txcb->write_start)) / 3000;

        if (tnic_adapter->zcopy_stats.num_completions % 10000 == 0) {
                printk("completions = %u, blocks = %u delta = %lu usec "
                       "ack_delta = %lu usec write_latency = %lu usec\n",
                       tnic_adapter->zcopy_stats.num_completions,
                       tnic_adapter->zcopy_stats.num_blocks,
                       tnic_adapter->zcopy_stats.delta/10000,
                       tnic_adapter->zcopy_stats.ack_delta/10000,
                       tnic_adapter->zcopy_stats.write_delta/10000);
                tnic_adapter->zcopy_stats.delta = 0;
                tnic_adapter->zcopy_stats.num_completions = 0;
                tnic_adapter->zcopy_stats.num_blocks = 0;
                tnic_adapter->zcopy_stats.ack_delta = 0;
                tnic_adapter->zcopy_stats.write_delta = 0;
        }
        ZERO_COPY_DBG("%s: Exit\n", __FUNCTION__);
}

/*
 *
 */
static host_tx_vbuf_t *zcopy_get_vbuf(struct list_head *list,
				      zero_copy_tx_cb_t *txcb)
{
        host_tx_vbuf_t          *zcopy_vbuf;
        struct list_head	*node;
	struct list_head	*tmp_node;


        list_for_each_safe(node, tmp_node, list) {
                zcopy_vbuf = list_entry(node, host_tx_vbuf_t, list);
		if (zcopy_vbuf->txcb == txcb) {
			return (zcopy_vbuf);
		}
	}

	return (NULL);
}

static inline void nx_txswitch_qmsg_init(pegnet_sock_id_t *sid,
					 txbuf_switch_t *qmsg)
{
	/* Queue message initialization. */
	qmsg->hdr.word = 0;
	qmsg->hdr.dst_subq = 1;
	qmsg->hdr.dst_side = sid->qid.side;
	qmsg->hdr.dst_major = sid->qid.major;
	qmsg->hdr.dst_minor = sid->qid.minor;
	qmsg->hdr.type = UNM_MSGTYPE_TXBUFFER_SWITCH;
}

static inline void set_tx_switch_iov(struct pci_dev *unm_device_ptr,
				txbuf_switch_t *txswitch, dma_addr_t old,
				dma_addr_t new, uint32_t len)
{
	pci_dma_sync_single_for_device(unm_device_ptr, new, len,
				       PCI_DMA_TODEVICE);

	nx_print1("old = %p, new = %p, len = %u\n",
		  (void *)old, (void *)new, len);

	txswitch->old[txswitch->numBufs].addr = old;
	txswitch->new[txswitch->numBufs].addr = new;
	txswitch->old[txswitch->numBufs].len = len;
	txswitch->new[txswitch->numBufs].len = len;
	txswitch->numBufs++;
}

/*
 *
 */
static int zcopy_copy_vbuf_data(pegnet_sock_id_t *sid,
				host_tx_vbuf_t *from_vbuf,
				host_tx_vbuf_t *to_vbuf, txbuf_switch_t *qmsg)
{
	netxen_host_page_t	*from_page;
	netxen_host_page_t	*to_page;
	int			ii;
	dma_addr_t		phys;
	int			qmsg_idx;
	int			num_pages;
	char			*to;
	int			rv;


	if (from_vbuf->num_pages != to_vbuf->num_pages) {
		nx_print("Number of pages are different\n");
		return (-1);
	}

	qmsg_idx = 0;
	nx_txswitch_qmsg_init(sid, &qmsg[qmsg_idx]);

	num_pages = from_vbuf->index;
	if (from_vbuf->full) {
		num_pages++;
	}
        to_vbuf->tot_bytes_sent = from_vbuf->tot_bytes_sent;
	to_vbuf->refid = from_vbuf->refid;
	to_vbuf->vec_cnt = from_vbuf->vec_cnt;
	to_vbuf->failed_idx = from_vbuf->failed_idx;
	to_vbuf->failed_ptr = from_vbuf->failed_ptr;
	to_vbuf->full = from_vbuf->full;
	to_vbuf->tx_flags = from_vbuf->tx_flags;
	to_vbuf->try_reuse = 0;

	for (ii = 0; ii < num_pages; ii++) {

		from_page = &from_vbuf->page[ii];
		to_page = &to_vbuf->page[ii];
		to_vbuf->page_bytes_remaining = from_page->size;

		if (to_vbuf->page_bytes_remaining > to_page->size) {
			nx_print("Mismatch in page sizes.\n");
			return (-1);
		}

		to = nx_tx_get_copy_addr(nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
			sid->adapter,sid->portnum)),to_vbuf, &phys);
		if (to == NULL) {
			nx_print("FAILED mapping page\n");
			return (-1);
		}

		rv = smart_memcpy(to, from_page->zcopy_user_addr,
				  from_page->size);
		if (rv != 0) {
			nx_print("copyin of %u bytes FAILED\n",
				 from_page->size);
			return (-1);
		}

                to_vbuf->page_bytes_remaining -= from_page->size;
		nx_tx_inc_page(to_vbuf);

                if (QMSG_MAX_TX_SWITCH_IOV == qmsg[qmsg_idx].numBufs) {
			qmsg_idx++;
			nx_txswitch_qmsg_init(sid, &qmsg[qmsg_idx]);
		}
		set_tx_switch_iov(nx_get_pcidev(
		NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum)),
		&qmsg[qmsg_idx], from_page->phys, phys,from_page->size);
	}

	if (to_vbuf->index != from_vbuf->index) {
		nx_print("to_vbuf->index[%u] != from_vbuf->index[%u]\n",
			 to_vbuf->index, from_vbuf->index);
	}

	qmsg[qmsg_idx].bufferComplete = 1;
	qmsg[qmsg_idx].cookie = from_vbuf->refid;

	return (qmsg_idx + 1);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
/*
 * list_replace_rcu - replace old entry by new one
 * @old : the element to be replaced
 * @new : the new element to insert
 *
 * The old entry will be replaced with the new entry atomically.
 * Note: 'old' should not be empty.
 */
static inline void list_replace_rcu(struct list_head *old,
				struct list_head *new)
{
	new->next = old->next;
	new->prev = old->prev;
	smp_wmb();
	new->next->prev = new;
	new->prev->next = new;
	old->prev = LIST_POISON2;
}
#endif

/*
 *
 */
static int zcopy_replace_vbuf(struct list_head *list, host_tx_vbuf_t *old,
			      host_tx_vbuf_t *new)
{
        host_tx_vbuf_t          *vbuf;
        struct list_head	*node;
	struct list_head	*tmp_node;

        list_for_each_safe(node, tmp_node, list) {
                vbuf = list_entry(node, host_tx_vbuf_t, list);
		if (vbuf == old) {
			list_replace_rcu(&old->list, &new->list);
			return (0);
		}
	}

	return (1);
}

/*
 *
 */
static int zcopy_switch_list_vbufs(pegnet_sock_id_t *sid,
				   struct list_head *list,
				   zero_copy_tx_cb_t *txcb, int bytes_replaced)
{
        host_tx_vbuf_t	*zcopy_vbuf;
        host_tx_vbuf_t	*copy_vbuf;
        host_tx_vbuf_t	orig_vbuf;
	txbuf_switch_t	qmsg[NX_TX_MAX_SWITCH_QMSG_PER_VBUF];
	int		qmsg_cnt;
	int		rv;


	while ((bytes_replaced + txcb->bytes_acked) < txcb->total_bytes) {
		/*
		 * Get the first available vbuf.
		 */
		spin_lock_bh(&sid->lock);

		zcopy_vbuf = zcopy_get_vbuf(list, txcb);
		if (zcopy_vbuf != NULL) {
			memcpy(&orig_vbuf, zcopy_vbuf, sizeof (orig_vbuf));
		}

		spin_unlock_bh(&sid->lock);

		if (zcopy_vbuf == NULL) {
			nx_print("VBUF not FOUND\n");
			break;
		}

		/*
		 * Allocate a copy buffer.
		 */
		copy_vbuf = get_tx_vbuf(NX_COPY_TX_VBUF, sid);
		if (copy_vbuf == NULL) {
			nx_print("Alloc FAILED\n");
			break;
		}

		/*
		 * Copyin the data.
		 */
		qmsg_cnt = zcopy_copy_vbuf_data(sid, &orig_vbuf, copy_vbuf,
						qmsg);
		if (qmsg_cnt < 0) {
			nx_print("Copy FAILED\n");
			spin_lock_bh(&sid->lock);
			put_tx_vbuf(sid, copy_vbuf);
			spin_unlock_bh(&sid->lock);
			break;
		}

		/*
		 * Search and replace vbuf in the specified list.
		 */
		spin_lock_bh(&sid->lock);
		rv = zcopy_replace_vbuf(list, zcopy_vbuf, copy_vbuf);
		if (rv) {
			nx_print("Went away before the copy completed\n");
			put_tx_vbuf(sid, copy_vbuf);
			spin_unlock_bh(&sid->lock);
			continue;
		}
		spin_unlock_bh(&sid->lock);

		/*
		 * Send the switch message to the card.
		 */
		if (unm_send_h2p_msg_atomic(NX_GET_DEV_FROM_TNIC_ADAPTER(
		    sid->adapter,sid->portnum),(unm_msg_t *)qmsg, qmsg_cnt)) {
			nx_print("Failed to send %u messages\n", qmsg_cnt);
			spin_lock_bh(&sid->lock);
			rv = zcopy_replace_vbuf(list, copy_vbuf, zcopy_vbuf);
			if (rv) {
				nx_print("No vbuf after the msg send fail\n");
				put_tx_vbuf(sid, zcopy_vbuf);
			}
			spin_unlock_bh(&sid->lock);
			break;
		}

		/*
		 * Add to the wait for completion list.
		 */
		spin_lock_bh(&sid->lock);
		list_add_tail(&zcopy_vbuf->list,
			      &sid->send_dma.zcopy_switch_list);
		bytes_replaced += zcopy_vbuf->tot_bytes_sent;
		spin_unlock_bh(&sid->lock);
	}

	return (bytes_replaced);
}

/*
 *
 */
static int zcopy_switch_vbufs(pegnet_sock_id_t *sid, zero_copy_tx_cb_t *txcb)
{
	int	bytes = 0;

	if (down_interruptible(&sid->send_dma.lock)) {
		nx_print("down failed\n");
		return (0);
	}

	/*
	 * Mark in_sendmsg. This way nothing will go into failure list. Things
	 * will be a little stable when we move things around.
	 */
	spin_lock_bh(&sid->lock);
	sid->send_dma.in_sendmsg = 1;
	spin_unlock_bh(&sid->lock);

	/*
	 * First check the sent_list.
	 */
	if ((bytes + txcb->bytes_acked) < txcb->total_bytes) {
		//nx_print("Switching sent_list\n");
		bytes = zcopy_switch_list_vbufs(sid, &sid->send_dma.sent_list,
						txcb, bytes);
	}

	/*
	 * Check the failure_list. Truely if any buffer enters failed list it
	 * should go from zero copy to copy type. TBD. TODO. *&*&*&*&*.
	 */
	if ((bytes + txcb->bytes_acked) < txcb->total_bytes) {
		//nx_print("Switching failure_list\n");
		bytes += zcopy_switch_list_vbufs(sid,
						 &sid->send_dma.failed_list,
						 txcb, bytes);
	}

	nx_print1("Switched %u bytes of %u. %u acked\n",
		  bytes, txcb->total_bytes, txcb->bytes_acked);

	/*
	 * Since we had locked down everything now try cleaning up everything.
	 */
	spin_lock_bh(&sid->lock);

	if (sid->send_dma.last_comp_valid) {
		process_last_tx_completion(sid);
	}

        /*
         * Check for failure case.
         */
	host_sendmsg_send_backlog(sid);

	sid->send_dma.in_sendmsg = 0;

	spin_unlock_bh(&sid->lock);

	up(&sid->send_dma.lock);

	return (bytes);
}

/*
 *
 */
static int zcopy_switch_wait_for_completion(pegnet_sock_id_t *sid,
					    zero_copy_tx_cb_t *txcb)
{
	int		rv;
	int		bytes;
	unsigned long	timeo;
       	DECLARE_WAITQUEUE(wait, current);

	nx_print1("Enter\n");

	timeo = 3 * HZ;
	bytes = zcopy_switch_vbufs(sid, txcb);

	add_wait_queue(&txcb->block_wq, &wait);

	while (!zero_copy_tx_is_complete(txcb)) {
		set_current_state(TASK_UNINTERRUPTIBLE);

		timeo = schedule_timeout(timeo);
		if (timeo) {
			continue;
		}

		set_current_state(TASK_RUNNING);
	}

	remove_wait_queue(&(txcb->block_wq), &wait);

	if (zero_copy_tx_is_complete(txcb)) {
		nx_print1("Found everything\n");
		return (0);
	}

	/*
	 * Kill the connection.
	 */
	set_current_state(TASK_UNINTERRUPTIBLE);

	nx_print1("Trying to kill connection\n");

	rv = host_conn_disconnect(sid, 1, 0, 0);
	if (rv != 0) {
		nx_print("host_conn_disconnect FAILED[%d]\n", rv);
	}
	schedule_timeout(3 * HZ);

	spin_lock_bh(&sid->lock);
	zcopy_clear_tx_vbuf_queues(sid, txcb);
	spin_unlock_bh(&sid->lock);

	set_current_state(TASK_RUNNING);

	return (0);
}

/*
 * wait_for_send_completion --
 *
 *   Wait until all the send data has been acknowledged by the
 * peer.
 *
 *
 */
static int wait_for_send_completion(pegnet_sock_id_t *sid,
				    zero_copy_tx_cb_t *txcb)
{
	int		blocked = 0;
	unsigned long	before;
	unsigned long	after;
	
	
	DECLARE_WAITQUEUE(wait, current);

/* 	nx_print("Force a switch for TESTING\n"); */
/* 	zcopy_switch_wait_for_completion(sid, txcb); */

	add_wait_queue(&(txcb->block_wq), &wait);

	ZERO_COPY_DBG("%s: Enter\n", __FUNCTION__);
	before = getTSC();
	after  = before;

	while (!zero_copy_tx_is_complete(txcb)) {
		set_current_state(TASK_INTERRUPTIBLE);
		
		if (signal_pending(current)) {
			nx_print("Try switching buffers: txcb[%p], "
				 "bytes_acked[%u],total_bytes[%u]\n",
				 txcb, txcb->bytes_acked, txcb->total_bytes);

			set_current_state(TASK_RUNNING);
			remove_wait_queue(&(txcb->block_wq), &wait);
			zcopy_switch_wait_for_completion(sid, txcb);
			return (-ERESTARTSYS);
		}
		blocked = 1;
		schedule();
		after = getTSC();

		ZERO_COPY_DBG("after schedule\n");
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&(txcb->block_wq), &wait);

	ZERO_COPY_PRINT_STATS(sid->adapter,before, after, blocked, txcb);
	ZERO_COPY_DBG("%s: Exit\n", __FUNCTION__);
	return (0);
}


static host_tx_vbuf_t *wait_get_tx_vbuf(pegnet_sock_id_t *sid,
					vbuf_type_t type, long *timeo, int *rc)
{
        host_tx_vbuf_t	*vbuf;

	if (!is_tx_vbuf_allocatable(type, sid)) {
		*rc = wait_for_returned_vbufs(type, sid, timeo);
		if (*rc) {
			return (NULL);
		}
	}
	vbuf = get_tx_vbuf(type, sid);
	if (vbuf == NULL) {
		*rc = -ENOMEM;
	}
	return (vbuf);
}

static inline int nx_forward_iov(nx_msghdr_t *msg)
{
  top:
	if (msg->seglen == 0) {
		if (msg->iovlen == 0) {
			return (1);
		}
		msg->seglen = msg->iov->iov_len;
		msg->from = msg->iov->iov_base;
		msg->iov++;
		msg->iovlen--;
		goto top;
	}

	return (0);
}

static inline int nx_tx_copy_data(nx_msghdr_t *msg, char *to, int bytes)
{
	int	len;
	int	copied;
	int	rc;


	copied = 0;
	rc = 0;
        while (bytes) {
		if (nx_forward_iov(msg)) {
			break;	/* nothing more */
		}

		/* Copy to kernel buffer, update state. */
		len = (msg->seglen < bytes) ? msg->seglen : bytes;

#if 1
		rc = smart_memcpy(to, msg->from, len);
#else // XXX: Use this to bypass memcpy for iperf performance test
		rc = (len > 100)? 0 : smart_memcpy(to, msg->from, len);
#endif
		if (rc != 0) {
			nx_print("copyin of %u bytes FAILED\n", len);
			break;
		}

		copied += len;
                msg->copied += len;
                msg->seglen -= len;
		bytes -= len;

                msg->from += len;
                to += len;
	}

	if (copied) {
		return (copied);
	}

	return (rc);
}

static inline void nx_txbuf_qmsg_init(pegnet_sock_id_t *sid, txbuf_msg_t *qmsg,
				      int flags)
{
	/* Queue message initialization. */
	qmsg->hdr.word = 0;
	qmsg->urgFlag = (flags & MSG_OOB) ? 1 : 0;
	qmsg->hdr.dst_subq = 1;
	qmsg->hdr.dst_side = sid->qid.side;
	qmsg->hdr.dst_major = sid->qid.major;
	qmsg->hdr.dst_minor = sid->qid.minor;
	qmsg->hdr.type = UNM_MSGTYPE_SENDMSG;
}

/*
 *
 */
static void host_sendmsg_send_backlog(pegnet_sock_id_t *sid)
{
        if (sid->send_dma.failed_ptr) {
                requeue_sendmsg_failed_ptr(sid);
        }

	if (!list_empty(&sid->send_dma.failed_list)) {
		host_sendmsg_resend_bufs(sid);
	}

	/*
	 * Vec wait list now.
	 */
	if (!list_empty(&sid->send_dma.vec_wait_list) &&
            list_empty(&sid->send_dma.failed_list) &&
            sid->send_dma.vec_cnt > 0) {
                host_sendmsg_send_veclist(sid);
	}
}

/*
 * Send filled send buffers to the card.
 */
static int nx_sendmsg_queue_wait_list(pegnet_sock_id_t *sid, nx_msghdr_t *msg)
{
        host_tx_vbuf_t	*vbuf;
	char		*to;
	int		rc = 0;
	dma_addr_t	phys;


	spin_lock_bh(&sid->lock);

        vbuf = NULL;

        if (!list_empty(&sid->send_dma.vec_wait_list)) {
                vbuf = list_entry(sid->send_dma.vec_wait_list.prev,
				  host_tx_vbuf_t, list);
                if (!vbuf->full) {
                        list_del_init(&vbuf->list);
			sid->send_dma.vl_cnt--;
                } else {
                        vbuf = NULL;
                }
        }

	spin_unlock_bh(&sid->lock);

	if (vbuf == NULL) {
		vbuf = wait_get_tx_vbuf(sid, NX_COPY_TX_VBUF,
					&msg->timeo, &rc);
		if (vbuf == NULL) {
			return (rc);
		}
	}

        to = NULL;

        while ((msg->iovlen != 0 || msg->seglen != 0) && !vbuf->full) {

		if (to == NULL) {
			to = nx_tx_get_copy_addr(nx_get_pcidev(
			NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum)),
			vbuf, &phys);
			if (to == NULL) {
				rc = -ENOMEM;
				break;
			}
		}

		rc = nx_tx_copy_data(msg, to, vbuf->page_bytes_remaining);
		if (rc < 0) {
			break;
		} else if (rc == 0) {
			nx_print("Should not happen.\n");
		}

                vbuf->page_bytes_remaining -= rc;
                vbuf->tot_bytes_sent += rc;
		to += rc;
		rc = 0;

		ZERO_COPY_DBG("%s: vbuf = %p, vbuf->tot_bytes_sent = %u "
			      "chunksize = %u\n", __FUNCTION__, 
			      vbuf, vbuf->tot_bytes_sent, rc);

                if (vbuf->page_bytes_remaining > 0) {
                        continue;
                }

		nx_tx_inc_page(vbuf);
                to = NULL;
        }

        /*
         * Check if the last vbuf was used. If not then free it.
         */
	spin_lock_bh(&sid->lock);

	if (vbuf->tot_bytes_sent == 0) {
		put_tx_vbuf(sid, vbuf);
	} else {
		int     tmp_rc = 0;

		vbuf->refid = 0;
		vbuf->penalty_vecs++;
		tmp_rc = tx_vec_wait_list_add_tail(sid, vbuf);
		if (tmp_rc) {
			rc = tmp_rc;
		}
	}

	spin_unlock_bh(&sid->lock);

	return (rc);
}

static void nx_sent_list_remove_vbuf(pegnet_sock_id_t *sid,
				     host_tx_vbuf_t *requested_vbuf)
{
	host_tx_vbuf_t	*vbuf = NULL;

	spin_lock_bh(&sid->lock);

        if (!list_empty(&sid->send_dma.sent_list)) {
                vbuf = list_entry(sid->send_dma.sent_list.prev,
				  host_tx_vbuf_t, list);
                if (vbuf == requested_vbuf) {
                        list_del_init(&vbuf->list);
			sid->send_dma.sl_cnt--;
			put_tx_vbuf(sid, vbuf);
                } else {
                        vbuf = NULL;
                }
        }

	spin_unlock_bh(&sid->lock);

	if (vbuf == NULL) {
		nx_print("vbuf was already removed\n");
	}
}

static inline int switch_to_regular_copy(int seglen, char *from, int mss ) 
{ 
        int available_space = PAGE_SIZE - ((unsigned long)from & ~PAGE_MASK);
        
        if( (seglen-available_space)  > mss ) {
                return 0;
        } else {
                return 1;
        }
}

static int nx_sendmsg_zcopy(pegnet_sock_id_t *sid, nx_msghdr_t *msg,
			    zero_copy_tx_cb_t *txcb)
{
        host_tx_vbuf_t		*vbuf;
	netxen_host_page_t	*page;
	int			rc = 0;
	txbuf_msg_t		qmsg[NX_TX_MAX_QMSG_PER_VBUF];
	int			qmsg_idx;
	int			chunksize;
	int			vecs;


	vbuf = wait_get_tx_vbuf(sid, NX_ZERO_COPY_TX_VBUF, &msg->timeo, &rc);
	if (vbuf == NULL) {
		return (rc);
	}

	vbuf->txcb = txcb;

	qmsg_idx = 0;
	nx_txbuf_qmsg_init(sid, &qmsg[qmsg_idx],msg->msg->msg_flags);

	/* msg->iovlen != 0 ||  */
	while (msg->seglen != 0 && !vbuf->full && msg->vec_cnt > 0 &&
	       !switch_to_regular_copy(msg->seglen, msg->from, sid->mss)) {

                if (msg->entries_per_qmsg == qmsg[qmsg_idx].numBufs) {
			qmsg_idx++;
			nx_txbuf_qmsg_init(sid, &qmsg[qmsg_idx], msg->msg->msg_flags);
		}

		chunksize = PAGE_SIZE - ((unsigned long)msg->from & ~PAGE_MASK);

		if (chunksize > msg->seglen) {
			chunksize = msg->seglen;
		}

		page = &vbuf->page[vbuf->index];
		if (add_tx_page_to_vbuf(vbuf, page, msg->from, chunksize)) {
			rc = -ENOMEM;
			break;
		}

                msg->seglen -= chunksize;
                vbuf->tot_bytes_sent += chunksize;
                msg->from += chunksize;

                vecs = (chunksize + sid->mss - 1) / sid->mss;
                msg->vec_cnt -= vecs;
                vbuf->vec_cnt += vecs;

		set_tx_iov(nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
		    sid->adapter,sid->portnum)),&qmsg[qmsg_idx], page->phys, 
		    chunksize,msg->compress_iovs);

		nx_tx_inc_page(vbuf);
                page = NULL;

		ZERO_COPY_DBG("%s: vbuf = %p, vbuf->tot_bytes_sent = %u "
			      "chunksize = %u\n", __FUNCTION__, 
			      vbuf, vbuf->tot_bytes_sent, chunksize);
        }

	if (vbuf->tot_bytes_sent == 0) {
		spin_lock_bh(&sid->lock);
		put_tx_vbuf(sid, vbuf);
		spin_unlock_bh(&sid->lock);
	} else {
		int	copied;
		int	tmp_rc;

		copied = vbuf->tot_bytes_sent;
                if (0 == qmsg[qmsg_idx].numBufs) {
			qmsg_idx--;
		}

		qmsg[qmsg_idx].cookie = sid->send_dma.refid;
		qmsg[qmsg_idx].bufferComplete = 1;
		vbuf->refid = sid->send_dma.refid;

		spin_lock_bh(&sid->lock);
		sid->send_dma.refid++;
		if (sid->send_dma.refid == 0) {
			sid->send_dma.refid++;
		}

		txcb->total_bytes += copied;

		tmp_rc = tx_sent_list_add_tail(sid, vbuf);
		spin_unlock_bh(&sid->lock);

		if (tmp_rc) {
			return (tmp_rc);
		}

		nx_print1("qmsg.numBufs = %u, refid = %u\n",
			  qmsg[qmsg_idx].numBufs, qmsg[qmsg_idx].cookie);

		if (unm_send_h2p_msg_atomic(NX_GET_DEV_FROM_TNIC_ADAPTER(
		    sid->adapter,sid->portnum),(unm_msg_t *)qmsg, 
		    qmsg_idx + 1)) {
			nx_print("Failed to send %u messages\n", qmsg_idx+1);
			nx_sent_list_remove_vbuf(sid, vbuf);
			return (-ENOMEM);
		}
		msg->copied += copied;
	}

	return (rc);
}

static int nx_sendmsg_copy(pegnet_sock_id_t *sid, nx_msghdr_t *msg)
{
        host_tx_vbuf_t		*vbuf;
        host_tx_vbuf_rb_info_t  rb;
	char			*to;
	int			rc = 0;
	int			page_bytes;
	dma_addr_t		phys;
	txbuf_msg_t		qmsg[NX_TX_MAX_QMSG_PER_VBUF];
	int			qmsg_idx;
	int			vecs;
	int			penalty_vecs = 0;
	struct {
		int	copied;
		int	vec_cnt;
	} msg_rb;

	memset(&phys,0,sizeof(phys));
        vbuf = NULL;
	if (msg->try_reuse) {
		spin_lock_bh(&sid->lock);

		if (!list_empty(&sid->send_dma.sent_list)) {
			vbuf = list_entry(sid->send_dma.sent_list.prev,
					  host_tx_vbuf_t, list);
			if (vbuf->try_reuse && !vbuf->full &&
			    vbuf->type == NX_COPY_TX_VBUF) {
				list_del_init(&vbuf->list);
				sid->send_dma.sl_cnt--;
			} else {
				vbuf = NULL;
			}
		}
		spin_unlock_bh(&sid->lock);

		penalty_vecs = UNM_SEND_BUF_REUSE_PENALTY;
		msg->try_reuse = 0;
	}

	if (vbuf == NULL) {
		vbuf = wait_get_tx_vbuf(sid, NX_COPY_TX_VBUF,
					&msg->timeo, &rc);
		if (vbuf == NULL) {
			return (rc);
		}
	}

	SET_RB_POINT(vbuf, rb);
        to = NULL;
	qmsg_idx = 0;
	nx_txbuf_qmsg_init(sid, &qmsg[qmsg_idx], msg->msg->msg_flags);
	msg_rb.vec_cnt = msg->vec_cnt;
	msg_rb.copied = msg->copied;
	page_bytes = 0;

        while ((msg->iovlen != 0 || msg->seglen != 0) && !vbuf->full &&
		msg->vec_cnt > 0) {

                if (msg->entries_per_qmsg == qmsg[qmsg_idx].numBufs) {
			qmsg_idx++;
			nx_txbuf_qmsg_init(sid, &qmsg[qmsg_idx], msg->msg->msg_flags);
		}

		if (to == NULL) {
			to = nx_tx_get_copy_addr(nx_get_pcidev(
			NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum)),
			vbuf, &phys);
			if (to == NULL) {
				rc = -ENOMEM;
				break;
			}
		}

		rc = nx_tx_copy_data(msg, to, vbuf->page_bytes_remaining);
		if (rc < 0) {
			break;
		} else if (rc == 0) {
			nx_print("Should not happen.\n");
		}

                vbuf->page_bytes_remaining -= rc;
                vbuf->tot_bytes_sent += rc;
		page_bytes += rc;
		to += rc;
		rc = 0;

		ZERO_COPY_DBG("%s: vbuf = %p, vbuf->tot_bytes_sent = %u "
			      "chunksize = %u\n", __FUNCTION__, 
			      vbuf, vbuf->tot_bytes_sent, rc);

                if (vbuf->page_bytes_remaining == 0 ||
		    (msg->iovlen == 0 && msg->seglen == 0)) {
			vecs = (page_bytes + sid->mss - 1) / sid->mss;
			msg->vec_cnt -= vecs;
			vbuf->vec_cnt += vecs;

			set_tx_iov(nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
			    sid->adapter,sid->portnum)),&qmsg[qmsg_idx], phys, 
			    page_bytes,msg->compress_iovs);

			if (vbuf->page_bytes_remaining == 0) {
				nx_tx_inc_page(vbuf);

				to = NULL;
				page_bytes = 0;
			}
                }
	}

	if (vbuf->tot_bytes_sent == 0) {
		spin_lock_bh(&sid->lock);
		put_tx_vbuf(sid, vbuf);
		spin_unlock_bh(&sid->lock);
	} else if (vbuf->tot_bytes_sent == rb.tot_bytes_sent) {
		int	tmp_rc;

		spin_lock_bh(&sid->lock);
		tmp_rc = tx_sent_list_add_tail(sid, vbuf);
		spin_unlock_bh(&sid->lock);
		if (tmp_rc) {
			return (tmp_rc);
		}
	} else {
		int	tmp_rc;

                if (0 == qmsg[qmsg_idx].numBufs) {
			qmsg_idx--;
		}

		spin_lock_bh(&sid->lock);

		vbuf->refid = sid->send_dma.refid;
		msg->vec_cnt -= penalty_vecs;
		vbuf->vec_cnt += penalty_vecs;

		if (1 || vbuf->full || msg->vec_cnt <= 0 ||
		    !is_tx_vbuf_allocatable(NX_COPY_TX_VBUF, sid)) {
			qmsg[qmsg_idx].cookie = sid->send_dma.refid;
			qmsg[qmsg_idx].bufferComplete = 1;
			sid->send_dma.refid++;
			if (sid->send_dma.refid == 0) {
				sid->send_dma.refid++;
			}
		}

		tmp_rc = tx_sent_list_add_tail(sid, vbuf);
		spin_unlock_bh(&sid->lock);

		if (tmp_rc) {
			return (tmp_rc);
		}

		nx_print1("qmsg.numBufs = %u, refid = %u\n",
			  qmsg[qmsg_idx].numBufs, qmsg[qmsg_idx].cookie);

		if (unm_send_h2p_msg_atomic(NX_GET_DEV_FROM_TNIC_ADAPTER(
		    sid->adapter,sid->portnum),(unm_msg_t *)qmsg,qmsg_idx + 1)) {
			nx_print("Failed to send %u messages\n", qmsg_idx+1);

			spin_lock_bh(&sid->lock);

			if (!list_empty(&sid->send_dma.sent_list)) {
				host_tx_vbuf_t	*tmp = NULL;
				tmp = list_entry(sid->send_dma.sent_list.prev,
						 host_tx_vbuf_t, list);
				if (tmp == vbuf) {
					REVERT_TO_RB_POINT(vbuf, rb);
				}
			}
			spin_unlock_bh(&sid->lock);

			msg->copied = msg_rb.copied;
			msg->vec_cnt = msg_rb.vec_cnt;
			return (-ENOMEM);
		}
	}

	return (rc);
}

static int wait_for_failure_recovery(pegnet_sock_id_t *sid, long *timeo)
{
	int	rc;

	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue(&sid->send_dma.wq, &wait);
	rc = 0;
	while (!list_empty(&sid->send_dma.failed_list) &&
	       !(sid->shutdown & SEND_SHUTDOWN)) {

		set_current_state(TASK_INTERRUPTIBLE);
		if (!*timeo) {
			rc = -EAGAIN;
			break;
		}

		sid->send_dma.need_failure_wakeup = 1;
		spin_unlock_bh(&sid->lock);

		if (signal_pending(current)) {
			spin_lock_bh(&sid->lock);
			sid->send_dma.need_failure_wakeup = 0;
			nx_print("got signal\n");
			rc = sock_intr_errno(*timeo);
			break;
		}
		*timeo = schedule_timeout(*timeo);

		spin_lock_bh(&sid->lock);
		sid->send_dma.need_failure_wakeup = 0;
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&sid->send_dma.wq, &wait);

	return (rc);
}

/*
 * Send filled send buffers to the card.
 */
int nx_host_tcp_sendmsg(pegnet_sock_id_t *sid, struct msghdr *msg,
			zero_copy_tx_cb_t *txcb)
{
	nx_msghdr_t	nxmsg;
	tx_mode_t	tx_mode;
	int		alloc_len;
	int		i;
	int		rc = 0;
	int		min_zcopy_vecs;

	tx_mode = txcb ? NX_ZERO_COPY_TX_MODE : NX_COPY_TX_MODE;

	nxmsg.msg = msg;
	nxmsg.iovlen = msg->msg_iovlen;
	nxmsg.iov = msg->msg_iov;
	nxmsg.seglen = 0;
	nxmsg.copied = 0;
        nxmsg.timeo = sock_sndtimeo(sid->parent->sk,
				    msg->msg_flags & MSG_DONTWAIT);
	if (global_use_compressed_iov && msg->msg_iovlen == 1) {
		nxmsg.compress_iovs = 1;
		nxmsg.entries_per_qmsg = QMSG_TX_COMPRESSED_PAGES;
	} else {
		nxmsg.compress_iovs = 0;
		nxmsg.entries_per_qmsg = QMSG_TX_PAGES;
	}

	alloc_len = 0;
	for (i = 0; i < nxmsg.iovlen; i++) {
		if (!access_ok(VERIFY_READ, nxmsg.iov[i].iov_base,
			       nxmsg.iov[i].iov_len)) {
			nx_print("No access\n");
			return (-EFAULT);
		}

		alloc_len += nxmsg.iov[i].iov_len;
	}

        if (alloc_len <= 0) {
		return (-EINVAL);
        }

	if (down_interruptible(&sid->send_dma.lock)) {
		nx_print("down failed\n");
		return (-ERESTARTSYS);
	}

	spin_lock_bh(&sid->lock);

	/* This flow needs a queue. */
	if (null_queue(sid->qid)) {
		nx_print("No dest queue\n");
		spin_unlock_bh(&sid->lock);
		rc = -ENOTSOCK;
		goto cleanup_return;
	}

	/*
         * Check if the send dma is in failed state and if we need to
         * wait for card to recover from memory shortage.
	 */
	if (!list_empty(&sid->send_dma.failed_list)) {
		rc = wait_for_failure_recovery(sid, &nxmsg.timeo);
		if (rc) {
			nx_print("Error wait_for_failure_recovery\n");
			spin_unlock_bh(&sid->lock);
			goto cleanup_return;
		}
	}

	if (sid->shutdown & SEND_SHUTDOWN) {
		nx_print("Send is shutdown\n");
		spin_unlock_bh(&sid->lock);
		rc = -EPIPE;
		goto cleanup_return;
	}

	/* To handle sendmsg call in between two sendpage */
 	if (!list_empty(&sid->send_dma.zero_wait_list)) { 
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SENDPAGE_FLUSHED1);
 		rc = nx_flush_sendpage(sid);
 		if (rc) { 
 			nx_print("Flush sendpage failed\n"); 
 			spin_unlock_bh(&sid->lock); 
 			goto cleanup_return; 
 		} 
 	}

	/*
	 * The in_sendmsg flag is set and other contexts check this flag.
	 * We have to do this because when we copy in data and we cannot hold
	 * any spinlocks and so we will give up the spinlock.
	 * And in interrupt context we cannot hold any semaphores.
	 */
	nxmsg.vec_cnt = sid->send_dma.vec_cnt;
	sid->send_dma.vec_cnt = 0;
	sid->send_dma.in_sendmsg = 1;

	nxmsg.try_reuse = 0;
	if (!list_empty(&sid->send_dma.sent_list) &&
	    nxmsg.vec_cnt > 0 && tx_mode == NX_COPY_TX_MODE) {
		host_tx_vbuf_t	*vbuf;
		int		len;

		len = alloc_len % (2 * MAX_BYTES_PER_TX_QMSG);
		vbuf = list_entry(sid->send_dma.sent_list.prev,
				  host_tx_vbuf_t, list);
		if (alloc_len < (2 * MAX_BYTES_PER_TX_QMSG) &&
		    vbuf->try_reuse && vbuf->type == NX_COPY_TX_VBUF && len) {
			nxmsg.try_reuse = 1;
		}
	}

	spin_unlock_bh(&sid->lock);

        sid->mss = 1460/* sid->mss */;

	min_zcopy_vecs = 2 * ((PAGE_SIZE + sid->mss - 1) / sid->mss);

        while (1) {

		if (sid->send_dma.last_comp_valid) {
			spin_lock_bh(&sid->lock);
			sid->send_dma.vec_cnt += nxmsg.vec_cnt;
			process_last_tx_completion(sid);
			host_sendmsg_send_backlog(sid);
			nxmsg.vec_cnt = sid->send_dma.vec_cnt;
			sid->send_dma.vec_cnt = 0;
			spin_unlock_bh(&sid->lock);
		}

		if (nx_forward_iov(&nxmsg)) {
			break;	/* nothing more */
		}

		if (nxmsg.vec_cnt <= 0) {
			/* nx_print("Vec count <= 0\n"); */
			rc = nx_sendmsg_queue_wait_list(sid, &nxmsg);
			if (rc != -EPIPE) {
				spin_lock_bh(&sid->lock);
				sid->send_dma.vec_cnt += nxmsg.vec_cnt;
				if (sid->send_dma.last_comp_valid) {
					process_last_tx_completion(sid);
				}
				host_sendmsg_send_backlog(sid);
				nxmsg.vec_cnt = sid->send_dma.vec_cnt;
				sid->send_dma.vec_cnt = 0;
				spin_unlock_bh(&sid->lock);
			}
			if (rc) {
				goto out;
			}

			continue;
		}

		if (tx_mode == NX_ZERO_COPY_TX_MODE) {
			if (nxmsg.vec_cnt > min_zcopy_vecs) {
				nxmsg.vec_cnt -= min_zcopy_vecs;
				rc = nx_sendmsg_zcopy(sid, &nxmsg, txcb);
				nxmsg.vec_cnt += min_zcopy_vecs;
				if (rc) {
					goto out;
				}
				if (switch_to_regular_copy(nxmsg.seglen,
							   nxmsg.from,
							   sid->mss)) {
					tx_mode = NX_COPY_TX_MODE;
				}
				continue;
			}
			tx_mode = NX_COPY_TX_MODE;
		}

		rc = nx_sendmsg_copy(sid, &nxmsg);
		if (rc) {
			goto out;
		}
        }

  out:
	/*
	 * Now update everything in send_dma structure after taking the lock.
	 */
	spin_lock_bh(&sid->lock);

	/*
	 * Give the remaining vecs back. vec_cnt can also be negative.
	 * It has to be within the sid->lock section.
	 */
	sid->send_dma.vec_cnt += nxmsg.vec_cnt;

	if (sid->send_dma.last_comp_valid) {
		process_last_tx_completion(sid);
	}
        /*
         * Check for failure case.
         */
	host_sendmsg_send_backlog(sid);

	sid->send_dma.in_sendmsg = 0;

	spin_unlock_bh(&sid->lock);

	if (nxmsg.copied) {
		rc = nxmsg.copied;
	}

  cleanup_return:
	up(&sid->send_dma.lock);
	return (rc);
}


int host_sendmsg_fast(pegnet_sock_id_t *sid, struct msghdr *msg)
{
	int rc = 0;
	zero_copy_tx_cb_t txcb;
	
	txcb.tstamp_start = 0;
	txcb.tstamp_end = 0;
	
	atomic_inc(&sid->adapter->num_tx_active_flows);
	atomic_add(msg->msg_iov->iov_len, &sid->adapter->tx_bytes_out);

	if (use_zero_copy(msg, sid)) {
		init_zero_copy_tx_cb(&txcb);
		
		ZERO_COPY_DBG("%s: txcb = %p\n", __FUNCTION__, &txcb);
		
		txcb.write_start = getTSC();
		rc = nx_host_tcp_sendmsg(sid, msg, &txcb);
		
		txcb.tstamp_start = getTSC();
		txcb.write_end = txcb.tstamp_start;
		
		ZERO_COPY_DBG("%s: rc = %d\n", __FUNCTION__, rc);

		if (rc > 0) {
			ZERO_COPY_DBG("%s: txcb.total_bytes = %u\n", 
				      __FUNCTION__, txcb.total_bytes);
			wait_for_send_completion(sid, &txcb);
		}
	} else {   // regular copy
		rc = nx_host_tcp_sendmsg(sid, msg, NULL);
	}
	
	atomic_dec(&sid->adapter->num_tx_active_flows);
	return rc;
}



/**************************************************************************
 *
 *			SEND PAGE
 *
 **************************************************************************/



/*
 *
 */
static int sendpage_send_vbuf(pegnet_sock_id_t *sid, host_tx_vbuf_t *vbuf,
			      int flags, int entries_per_qmsg,
			      int compress_iovs)
{
	txbuf_msg_t		qmsg[NX_TX_MAX_QMSG_PER_VBUF];
	int			qmsg_idx = 0;
	int			cnt = 0;
	netxen_host_page_t	*page;
	int			mss;
	int16_t			vecs;

	mss = 1460/* sid->mss */;

	nx_txbuf_qmsg_init(sid, &qmsg[qmsg_idx], flags);
	vbuf->index = vbuf->index - 1;
	for (cnt = 0; cnt <= vbuf->index; cnt++) {
		page = &vbuf->page[cnt];
		set_tx_iov(nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(
		    sid->adapter,sid->portnum)),&qmsg[qmsg_idx], 
		    page->phys, page->size,compress_iovs);

		vecs = (page->size + mss - 1) / mss;
		sid->send_dma.vec_cnt -= vecs;
		vbuf->vec_cnt += vecs;

		if (entries_per_qmsg == qmsg[qmsg_idx].numBufs) {
			if (cnt < vbuf->index) {
				qmsg_idx++;
				nx_txbuf_qmsg_init(sid, &qmsg[qmsg_idx],
						   flags);
			}
		}
	}
	qmsg[qmsg_idx].cookie = sid->send_dma.refid;
	qmsg[qmsg_idx].bufferComplete = 1;
	vbuf->refid = sid->send_dma.refid;
	sid->send_dma.refid++;
	if (sid->send_dma.refid == 0) {
		sid->send_dma.refid++;
	}

	nx_print1("qmsg.numBufs = %u, refid = %u\n",
		  qmsg[qmsg_idx].numBufs, qmsg[qmsg_idx].cookie);

	if (unm_send_h2p_msg_atomic(NX_GET_DEV_FROM_TNIC_ADAPTER(
	    sid->adapter,sid->portnum),(unm_msg_t *)qmsg, qmsg_idx + 1)) {
		nx_print("Failed to send %u messages\n", qmsg_idx + 1);
		return (-ENOMEM);
	}

	/*
	 * We hold the lock when we come here and should never come here if
	 * the SEND_SHUTDOWN flag is set.
	 */
	list_add_tail(&vbuf->list, &sid->send_dma.sent_list);
	sid->send_dma.sl_cnt++;

	return (0);
}

/*
 *
 */
int nx_flush_sendpage(pegnet_sock_id_t *sid)
{
        host_tx_vbuf_t	*vbuf;
	int		rc = 0;
	int		entries_per_qmsg;
	int		compress_iovs;

	if (sid->send_dma.tstamp_flag) {
		del_timer_sync(&sid->send_dma.sendpage_timer);
	}

	if (global_use_compressed_iov) {
		compress_iovs = 1;
		entries_per_qmsg = QMSG_TX_COMPRESSED_PAGES;
	} else {
		compress_iovs = 0;
		entries_per_qmsg = QMSG_TX_PAGES;
	}

	while (!list_empty(&sid->send_dma.zero_wait_list)) {
		vbuf = list_entry(sid->send_dma.zero_wait_list.next,
				  host_tx_vbuf_t, list);
		list_del_init(&vbuf->list);

		if (sid->send_dma.vec_cnt <= 0) {
			vbuf->refid = 0;
			list_add_tail(&vbuf->list,
				      &sid->send_dma.vec_wait_list);
			sid->send_dma.vl_cnt++;
		} else {
			rc = sendpage_send_vbuf(sid, vbuf, 0, entries_per_qmsg,
						compress_iovs);
			if (rc) {
				list_add(&vbuf->list,
					 &sid->send_dma.zero_wait_list);
				return (rc);
			}
		}
	}
	return (0);
}

/*
 *
 */
static void sendpage_write_timer(unsigned long data)
{
	pegnet_sock_id_t	*sid = (pegnet_sock_id_t*)data;
	int			rc;
	unsigned long		timeout = 10 * HZ; /* optimum value needs to
						      be found */

	UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SENDPAGE_TIMER_FIRED);

	nx_print1("Timer expired: sid = %p\n", sid);

	spin_lock(&sid->lock);
	sid->send_dma.tstamp_flag = 0;
	rc = nx_flush_sendpage(sid);
	if (!list_empty(&sid->send_dma.zero_wait_list)) {
		/*
		 * We failed to send the message. Just reset the timer to try
		 * again. Need to add a retry counter and then fail.
		 */
		mod_timer(&sid->send_dma.sendpage_timer, (timeout + jiffies));
	}

	spin_unlock(&sid->lock);
}

/*
 *
 */
int unm_sendpage(struct socket *sock,struct page **pages, int poffset,
		 size_t psize, int flags)
{
	int			rc = 0;
        pegnet_sock_id_t        *sid;
	unsigned long		expires;
	unsigned long		timeout = 10 * HZ; /* optimum value needs to
						      be found */
	int			queue_empty;

        if (sock->type == SOCK_STREAM && sock->state != SS_CONNECTED) {
                return (-EPIPE);
        }

        sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
	expires = timeout + jiffies;
	rc = host_sendpage(sid, pages, poffset, psize, flags);
	if (rc <= 0) {
		return (rc);
	}

	spin_lock_bh(&sid->lock);
	queue_empty = list_empty(&sid->send_dma.zero_wait_list);
	spin_unlock_bh(&sid->lock);

	if (sid->send_dma.tstamp_flag == 0 && !queue_empty) {
		spin_lock_bh(&sid->lock);
		sid->send_dma.sendpage_timer.expires = expires;
		sid->send_dma.sendpage_timer.function = &sendpage_write_timer;
		sid->send_dma.sendpage_timer.data = (unsigned long)sid;
		sid->send_dma.tstamp_flag = 1;
		spin_unlock_bh(&sid->lock);
		add_timer(&sid->send_dma.sendpage_timer);

	} else if (sid->send_dma.tstamp_flag == 1 && (flags & MSG_MORE) &&
		   !queue_empty) {
		mod_timer(&sid->send_dma.sendpage_timer, expires);

	} else if (sid->send_dma.tstamp_flag == 1 && queue_empty) {
		del_timer_sync(&sid->send_dma.sendpage_timer);
		spin_lock_bh(&sid->lock);
		sid->send_dma.tstamp_flag = 0;
		spin_unlock_bh(&sid->lock);
	}

	return (rc);
}

/*
 *
 */
static inline int nx_map_sendpage_to_phys(struct pci_dev *unm_device_ptr,
				netxen_host_page_t *page, int offset,
				int chunksize)
{
        dma_addr_t	temp;

        page_cache_get(page->mm_page);
        temp = pci_map_page(unm_device_ptr, page->mm_page, 0,
			    (1 << PAGE_SHIFT), PCI_DMA_TODEVICE);
        page->phys = temp + offset;
        if (page->phys == 0) {
                nx_print("pci_map_single failed on page = %p, size = %d\n",
			 page, chunksize);

                page_cache_release(page->mm_page);
                return (-1);
        }

        nx_print1("physaddr = %p\n", (void *)(page->phys));

        return (0);
}

/*
 *
 */
int host_sendpage(pegnet_sock_id_t *sid, struct page **pages, int poffset,
		  size_t psize, int flags)
{
	host_tx_vbuf_t		*vbuf;
	netxen_host_page_t	*page;
	int			copied;
	int			chunksize;
	int			rc = 0;
	struct sock		*sk;
	long			timeo;
	vbuf_type_t		type;
	tx_mode_t		tx_mode;
	int			compress_iovs;
	int			entries_per_qmsg;
	int			offset;


	tx_mode	 = NX_ZERO_COPY_TX_MODE;
	type	 = NX_ZERO_COPY_TX_VBUF;

	if (global_use_compressed_iov) {
		compress_iovs = 1;
		entries_per_qmsg = QMSG_TX_COMPRESSED_PAGES;
	} else {
		compress_iovs = 0;
		entries_per_qmsg = QMSG_TX_PAGES;
	}

	sk = sid->parent->sk;

	if (down_interruptible(&sid->send_dma.lock)) {
		printk("%s: down failed\n", __FUNCTION__);
		return (-ERESTARTSYS);
	}

	spin_lock_bh(&sid->lock);

	/* This flow needs a queue. */
	if (null_queue(sid->qid)) {
		printk("%s: no dest queue\n", __FUNCTION__);
		rc = -ENOTSOCK;
		goto cleanup_return;
	}
	timeo = sock_sndtimeo(sid->parent->sk, flags & MSG_DONTWAIT);

	/*
	 * Check if the send dma is in failed state and if we need to
	 * wait for card to recover from memory shortage.
	 */
	if (!list_empty(&sid->send_dma.failed_list)) {
		rc = wait_for_failure_recovery(sid, &timeo);
		if(rc) {
			nx_print("Error wait_for_failure_recovery\n");
			goto cleanup_return;
		}
	}

	if (sid->shutdown & SEND_SHUTDOWN) {
		//nx_print("Send is shutdown\n");
		rc = -EPIPE;
		goto cleanup_return;
	}

	/*
	 * The in_sendmsg flag is set and other contexts check this flag.
	 */
	sid->send_dma.in_sendmsg = 1;

	vbuf = NULL;

	if (!list_empty(&sid->send_dma.zero_wait_list)) {
		vbuf = list_entry(sid->send_dma.zero_wait_list.prev,
				  host_tx_vbuf_t, list);
		if (!vbuf->full) {
			list_del_init(&vbuf->list);
			vbuf->refid = 0;
		} else {
			vbuf = NULL;
		}
	}

	copied = 0;
	while (psize != 0) {

		if (vbuf == NULL) {
			spin_unlock_bh(&sid->lock);
			vbuf = wait_get_tx_vbuf(sid, NX_ZERO_COPY_TX_VBUF,
						&timeo, &rc);
			spin_lock_bh(&sid->lock);

			if (sid->shutdown & SEND_SHUTDOWN) {
				rc = -EPIPE;
				goto cleanup_return;
			}
			host_sendmsg_send_backlog(sid);

			if (vbuf == NULL) {
				break;
			}
		}

		vbuf->txcb = NULL;
		page = &vbuf->page[vbuf->index];
		page->mm_page = pages[poffset / PAGE_SIZE];
		offset = poffset % PAGE_SIZE;
		chunksize = min_t(size_t, psize, PAGE_SIZE - offset);

		if (nx_map_sendpage_to_phys(nx_get_pcidev(
			NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum)),
			page, offset, chunksize)) {
			/*
			 * If the last vbuf was used then queue it.
			 */
			if (vbuf->tot_bytes_sent == 0) {
				put_tx_vbuf(sid, vbuf);
			} else {
				list_add_tail(&vbuf->list,
					      &sid->send_dma.zero_wait_list);
			}
			break;
		}
		page->size = chunksize;
		vbuf->tot_bytes_sent += chunksize;
		vbuf->page_bytes_remaining = 0;

		poffset = poffset + chunksize;
		psize -= chunksize;
		copied += chunksize;

		vbuf->index++;
		if (vbuf->index == vbuf->num_pages) {
			vbuf->full = 1;
		}

		if (!vbuf->full && psize) {
			continue;
		}

		if (!list_empty(&sid->send_dma.zero_wait_list) ||
		    (!vbuf->full && (flags & MSG_MORE))) {

			list_add_tail(&vbuf->list,
				      &sid->send_dma.zero_wait_list);

		} else if (sid->send_dma.vec_cnt <= 0) {
			/*
			 * Since we hold the lock continously just
			 * queue it, SEND_SHUTDOWN cannot be set when
			 * we have the lock.
			 */
			vbuf->refid = 0;
			list_add_tail(&vbuf->list,
				      &sid->send_dma.vec_wait_list);
			sid->send_dma.vl_cnt++;

		} else {
			rc = sendpage_send_vbuf(sid, vbuf, flags,
						entries_per_qmsg,
						compress_iovs);
			if (rc) {
				/*
				 * Queue it up when the timer fires it can
				 * try to handle it.
				 */
				list_add_tail(&vbuf->list,
					      &sid->send_dma.zero_wait_list);
			}
		}
		vbuf = NULL;
	}

	if (copied) {
		rc = copied;
	}

	/*
	 * Now update everything in send_dma structure after taking the
	 * lock.
	 */
	if (sid->send_dma.last_comp_valid) {
		process_last_tx_completion(sid);
	}

	/*
	 * Check for failure case.
	 */
	host_sendmsg_send_backlog(sid);

	sid->send_dma.in_sendmsg = 0;

  cleanup_return:
	spin_unlock_bh(&sid->lock);
	up(&sid->send_dma.lock);
	return (rc);
}


/*
 * Waits until all the pending data processed. This function is invoked by
 * rdma.
 */
int check_pending_data(struct socket *sock)
{
	int ret = 0;
	pegnet_sock_id_t *sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
	DECLARE_WAITQUEUE(wait, current);

	if (down_interruptible(&sid->send_dma.lock))
		return (-ERESTARTSYS);

	spin_lock_bh(&sid->lock);
	add_wait_queue(&sid->send_dma.wq, &wait);

	while ((!list_empty(&sid->send_dma.failed_list)) ||
	       (!list_empty(&sid->send_dma.sent_list)) ||
	       (!list_empty(&sid->send_dma.vec_wait_list)) ||
	       (!list_empty(&sid->send_dma.zero_wait_list))) {

		set_current_state(TASK_INTERRUPTIBLE);
		spin_unlock_bh(&sid->lock);
		if (signal_pending(current)) {
			spin_lock_bh(&sid->lock);
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&sid->send_dma.wq, &wait);
			spin_unlock_bh(&sid->lock);
			ret = -ERESTARTSYS;
			goto earlyret;
		}
		schedule();
		spin_lock_bh(&sid->lock);
	}

	set_current_state(TASK_RUNNING) ;
	remove_wait_queue(&sid->send_dma.wq, &wait);
	spin_unlock_bh(&sid->lock);
earlyret:
	up(&sid->send_dma.lock);
	return(ret);
}

EXPORT_SYMBOL(check_pending_data);
