/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Copyright and CVS version tags.
 */
#ifndef	_UNM_HOST_TYPES_H
#define	_UNM_HOST_TYPES_H

#include <linux/types.h>
#include <linux/wait.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/if_vlan.h>
#include <linux/inetdevice.h>
#include <linux/if.h>
#include <linux/in.h>
#include <linux/proc_fs.h>

#include <unm_inc.h>
#include <pegnet_inc.h>
#include "nx_nic_linux_tnic_api.h"
#include "kernel_compatibility.h"
#include "nic_cmn.h"
#include "hostInterface.h"
#include "netxen_linux_ioctl.h"
#include "flow_table.h"
#include "nx_route_table.h"
#include "unm_pstats.h"

struct unm_adapter_s;   /* It is defined in unm_nic.h */
struct in_ifaddr;
extern unsigned int max_tcp_conn;

#if defined(DEBUG) && defined(PEGNET_DEBUG_PRINT) && PEGNET_DEBUG_LEVEL > 3
#define HOST_TIMEOUT   (5 * HZ)        /* Five Seconds -
					Pegs might be printing a lot */
#else
#define HOST_TIMEOUT   (3 * HZ)        /* Three Seconds */
#endif

//#define	TURN_ON_DEBUG_PRINTS
#undef	TURN_ON_DEBUG_PRINTS

#define	nx_print(format, ...)						 \
        printk("%s[%d]: " format, __FUNCTION__, __LINE__, ##__VA_ARGS__)
#if	defined(TURN_ON_DEBUG_PRINTS)
#define nx_print1		nx_print
#else
#define nx_print1(format, ...)
#endif	/* defined(TURN_ON_DEBUG_PRINTS) */

/*
 * Some assert functionality. Needs to be killed. We can't just assert in the
 * driver and every failure has to be clean.
 */
#if     defined(NDEBUG)
#define assert(A)
#else   /* NDEBUG */
static inline void __assert(const char *txt, const char *file,
                            unsigned int line)
{
        printk(KERN_CRIT "Assertion %s in file %s:line%d failed\n",
               txt, file, line);
        panic("Aiiee\n");
}
#define assert(A)                                               \
        do {                                                    \
                if(!(A)) {                                      \
                        __assert(#A, __FILE__, __LINE__);       \
                }                                               \
        } while(0)
#endif  /* NDEBUG */


// #define DEBUG_ZERO_COPY_TX     // debug information
#define ZERO_COPY_TX_SAMPLE    // sample zero copy tx latency


#ifdef DEBUG_ZERO_COPY_TX
#define ZERO_COPY_DBG( args...) printk( args)
#else
#define ZERO_COPY_DBG( arg...)
#endif // DEBUG_ZERO_COPY_TX


#ifdef ZERO_COPY_TX_SAMPLE
#define ZERO_COPY_PRINT_STATS(_adapter,_before, _after, _blocked, _txcb) zero_copy_print_stats(_adapter,_before, _after, _blocked, _txcb)
#else
#define ZERO_COPY_PRINT_STATS(_a, _b, _c, _d,_e)
#endif

typedef struct zero_copy_stats {
	unsigned long   delta;
	unsigned long   ack_delta;
	unsigned long   write_delta;
	uint32_t        num_blocks;
	uint32_t        num_completions;
} zero_copy_stats;


#define	UNM_SOCK_SET_PEGNET_SOCK_ID(SOCK, SID)  \
	((SOCK)->sk->SK_PROTINFO = SID)

#define UNM_SOCK_GET_PEGNET_SOCK_ID(SOCK)               \
        ((pegnet_sock_id_t *)((SOCK)->sk->SK_PROTINFO))

#define	UNM_SOCK_IS_SOCK_OFFLOADED(SOCK)                        \
	((SOCK)->sk && (SOCK)->sk->SK_FAMILY == AF_INET &&      \
	 (SOCK)->sk->SK_PROTINFO)

#define NX_GET_LISTENER_GROUP_FROM_SID(SID) \
       ((listener_group_t *)((SID)->parent->sk->SK_PROTINFO))

#define NX_GET_LISTENER_GROUP_FROM_SOCK(SOCK)               \
        ((listener_group_t *)((SOCK)->sk->SK_PROTINFO))
	 
#define NX_TNIC_MAX_PORTS       4

#define NX_GET_DEV_FROM_TNIC_ADAPTER(TNIC_ADAPTER,PORTNUM)               \
        ((TNIC_ADAPTER)->nx_dev)

typedef struct pegnet_inprogress        pegnet_inprogress_t;
typedef struct pegnet_sock_id_s         pegnet_sock_id_t;

typedef enum host_op_type {
        op_release = 0,
        op_create,
        op_connect,
        op_ioctl,
        op_shutdown,
        op_setsockopt,
        op_getsockopt,
        op_ipt_setsockopt,
        op_ipt_getsockopt,
        op_listen,
        op_bind,
        op_getstats,
} host_op_type_t;

typedef enum {
	NX_COPY_TX_VBUF		= 0,
	NX_ZERO_COPY_TX_VBUF	= 1,
	NX_ZERO_AND_REG_TX_VBUF = 2
} vbuf_type_t;


/*
** A pegnet_inprogress is a recording of the interesting facts about
** a socket operation that is currently being performed by the card.
** This is what the kernel (linux) used to wait on, and be notified
** about, completion of such operations.
** These are allocated on the stack of the thread that initiated
** an operation and linked together by the runtime. This allows for
** multiple operations on a single socket without malloc'ng structures.
** The elememts are linked and paired up using the context pointer and
** transaction identifier carried in every rpc message.
*/
struct pegnet_inprogress {
        unm_msg_t               msg;    /* the msg used for communication */
        pegnet_inprogress_t     *link; 	/* linked list of these things */
        wait_queue_head_t       wq;     /* user blocks on until reply comes */
        host_op_type_t          type;   /* type of operation */
        atomic_t                replied;/* request has been replied to */
        uint8_t                 op_id;  /* numeric id for this operation */
};


typedef enum {
        pegnet_send_buf = 0,
        pegnet_sopt_buf = 1,
        pegnet_num_bufs = 2
} pegnet_buf_dir;

/* Values for the variable bufferType (see below). */
/* Can't be zero. */
#define PCB_RXBUFFER            1
#define GLOBAL_RXBUFFER         2
/* Receive data buffer stuff. */
// #define USE_GLOBAL_BUFFERS         /* All flows use global buffers. */
#define NUM_HIGH_PERF_FLOWS     32
#define NUM_PCB_RXBUFFERS_HP    16    /* For 'high performance' flows. */
#define NUM_PCB_RXBUFFERS_LP    2
#if NUM_PCB_RXBUFFERS_HP > NUM_PCB_RXBUFFERS_LP
#define NUM_PCB_RXBUFFERS       NUM_PCB_RXBUFFERS_HP
#else
#define NUM_PCB_RXBUFFERS       NUM_PCB_RXBUFFERS_LP
#endif

/* Buffer states for the the variable hostCtrFlag below. */
#define UNM_HT_RXBUFF_NOTVALID  0  /* Buffer is not valid                 */
#define UNM_HT_RXBUFF_READY     1  /* Buffer can be read from.            */
#define UNM_HT_RXBUFF_EMPTY     2  /* Buffer is empty.                    */
#define UNM_HT_RXBUFF_CARD      3  /* Buffer is under card control        */

/* Do not offload a SYN if the TCP header exceeds this size */
#define NX_MAX_TCP_OFFLOAD_HDR_SIZE	(sizeof(unm_msg_body_t))

/* Number of host to card messages per SYN */
#define NX_NUM_MSGS_PER_SYN		2

#define NX_SYN_IPH_IDX	0
#define NX_SYN_TCPH_IDX	1

typedef struct {
        char            *buffer;
	struct page	*mm_page;
        dma_addr_t      phys;          /* Physical address of buffer.      */
        uint32_t        size;          /* Size of buffer.                  */
	uint8_t		pci_mapped;
	char		*zcopy_user_addr;
} netxen_host_page_t;

/* Structures to maintain state of host receive buffers. */
#define HOST_GLOBAL_RX_PAGES    1   /* Pages in a global rx buffer.        */
#define HOST_PCB_RX_PAGES       14  /* Pages in a pcb rx buffer.           */

/* Sundry error checking. */
#if (7 < QMSG_RX_PAGES)
#error "rxbuf_msg_body_t pages won't fit in queue message."
#endif
#if (HOST_GLOBAL_RX_PAGES > QMSG_RX_PAGES)
#error "Global buffer pages must fit in a queue message."
#endif

#if (HOST_GLOBAL_RX_PAGES > HOST_PCB_RX_PAGES)
#define HOST_RX_PAGES    HOST_GLOBAL_RX_PAGES
#else
#define HOST_RX_PAGES    HOST_PCB_RX_PAGES
#endif

#define CMD_DESC_PER_RXBUFF	(HOST_RX_PAGES / QMSG_RX_PAGES)

#define	NUM_GLOBAL_RX_BUFFERS		512

typedef struct {
        struct list_head        list;
        netxen_host_page_t      page[HOST_RX_PAGES];
        uint32_t                numPages;
        uint32_t                bufferBytesRemaining;
        uint32_t                pageBytesRemaining;
        uint32_t                pageBytesRead;
        netxen_host_page_t	*currPage;
        uint32_t                index;
        uint8_t                 hostCtrlFlag: 2,
                                rcv_shutdown:1,
                                bufferType:2,               /* Buffer Type */
                                rsvd0: 3;
} hostRxBuffer_t;

/*
 * Used to rollback some of the vbuf info in case of error in sendmsg.
 */
typedef struct {
        uint32_t                page_bytes_remaining;
        uint32_t                tot_bytes_sent;
        uint32_t                index;

	uint16_t		refid;
	int16_t                 vec_cnt;
        int                     rb_required;
} host_tx_vbuf_rb_info_t;
#define	NX_TX_MAX_QMSG_PER_VBUF		4
#define	NX_TX_MAX_SWITCH_QMSG_PER_VBUF	6

typedef struct {
	// wait queue to block on while socket write completes
	wait_queue_head_t	block_wq;
	uint32_t		total_bytes;
	uint32_t		bytes_acked; 
	unsigned long		tstamp_start;
	unsigned long		tstamp_end;
	unsigned long		write_start;
	unsigned long		write_end;
} zero_copy_tx_cb_t;

/*
 *
 */
#define UNM_HOST_TX_PAGES       14

typedef struct {
        struct list_head        list;
        netxen_host_page_t	page[UNM_HOST_TX_PAGES];
        uint16_t                num_pages;
        uint16_t                index;
        uint32_t                page_bytes_remaining;
        uint32_t                tot_bytes_sent;
	uint32_t		tot_capacity;
        vbuf_type_t		type;
	zero_copy_tx_cb_t	*txcb;
	uint16_t		refid;
	int16_t                 vec_cnt;
	int16_t			penalty_vecs;

        uint16_t                vec_wait_idx;
        uint16_t                failed_idx;
	dma_addr_t		failed_ptr;	/* Phy addr of failure point */

	uint8_t			allocated;
        uint8_t                 full;
        uint8_t                 tx_flags;       /* Shutdown, etc */
	uint8_t			try_reuse;	/* Allow copying at the
						   tail page */
} host_tx_vbuf_t;

#define NX_SYNQ_HSIZE	512
#define NX_MAX_SYNQ_LEN	1024
typedef struct {
	int			qlen;
	uint32_t		hash_rnd;
	struct list_head	youngsters;
	struct list_head	syn_table[NX_SYNQ_HSIZE];
} nx_lopt_t;

struct nx_sk_cleanup_queue {
        struct list_head	list;
        spinlock_t		lock;
};

typedef struct {
	struct list_head        sid_list;
	spinlock_t              lock;
} nx_synq_prun_requests_t;


/*
 *
 */
struct pegnet_sock_id_s {
	uint32_t 		type;
        struct list_head	tw_list;	/* sid list, killed in
						   scheduled cleanup task.
						   Should be first element */
        struct {
                __uint32_t      context;        /* encoded cookie */
        } ctx;                                  /* execution associated
                                                   context handle */
        unsigned int            poll_flags;     /* Poll flags from the card */
        pegnet_inprogress_t     *op;            /* list of in-progress
                                                   operations | NULL */
        uint8_t                 op_id;          /* per socket operation id */
        uint8_t                 cleanup_flag;	/* 1 if host_release()
						   needs to cleanup sid->tw */
	uint8_t                 conn_compl;	/* Set to 1 first time
						   SS_CONNECTED is seen in
						   unm_connect() */

	uint8_t			release_seen;

	uint16_t		mss;
        struct socket           *parent;        /* pointer to parent socket */
        struct sock             *sk_cleanup;	/* ptr to sk which is going to be cleaned up*/
        spinlock_t              lock;           /* a little protection */

	uint32_t                lport : 16,
                                rport : 16;
	ip_addr_t               lip_addr;
	ip_addr_t               rip_addr;

	uint8_t			route_added;

        /* DMA buffer stuff in redirector - kernel memory */
        struct {
                struct semaphore        lock;
                void                    *buffer;        /* pointer to kernel
                                                           buffer | NULL */
                dma_addr_t              phys;           /* physical address of
                                                           DMA buffer */
        } dma[pegnet_num_bufs];

        /* Structures to get async recieve data buffers from the card. */
        struct {
                struct semaphore        lock;
                wait_queue_head_t       wq;
                struct list_head        list;
                uint32_t                readOffset;     /* Next byte to read */
                uint16_t                rcv_shutdown : 1,
                                        initialized : 1,
					tmp_rcv_shutdown : 1,
                                        rsvd0 : 9,
					post_cntr:NETXEN_RXBUF_COUNTER_BITS;
                /* Used only for per connection buffers. */
                uint16_t                num_pcb_rxbuffers;
                hostRxBuffer_t          *remoteBuf;
        } recv_dma;

#define	UNM_SEND_BUF_MAX_VECS		300
#define	UNM_SEND_BUF_REUSE_PENALTY	(UNM_SEND_BUF_MAX_VECS / 10)
#define	UNM_SEND_BUF_QUEUED_PENALTY	(UNM_SEND_BUF_MAX_VECS / 5)
#define UNM_TX_FREE_VBUF_THRESHOLD      6
#define UNM_TX_VBUF_MAX_HP              6
#define UNM_TX_VBUF_MAX_LP              1
#define UNM_ZCOPY_TX_VBUF_MAX_LP        4
#define UNM_TX_VBUF_MAX_PAGES_HP        14
#define UNM_TX_VBUF_MAX_PAGES_LP        7
	struct {
		struct semaphore	lock;
		wait_queue_head_t	wq;
                struct list_head        sent_list;
		struct list_head        vec_wait_list;
		struct list_head        failed_list;
                struct list_head        alloc_list;
                struct list_head        zcopy_switch_list;
                struct list_head        zero_wait_list;

                int                     sl_cnt;
                int                     vl_cnt;
                int                     fl_cnt;

                atomic_t                alloc_q_len;
                atomic_t                allocated;
                uint32_t                num_vbuf_pages;
		uint16_t		refid;
		int16_t			vec_cnt;
                dma_addr_t              failed_ptr;
		uint16_t		last_comp_cookie;
		uint8_t			last_comp_valid;
		uint8_t			need_failure_wakeup : 1,
					need_buffer_wakeup : 1,
					rsvd : 6;
		uint8_t			in_sendmsg;
		uint8_t                 tstamp_flag;
		struct timer_list       sendpage_timer;
		atomic_t		zerocopy_vbuf_cnt;
	} send_dma;

	/*
	 * Async accept requests are queued in posted_queue. The maximum
	 * depth of this queue is the socket backlog value.
	 *
	 * When new connections are accepted they are removed from the
	 * posted_queue and added to the accepted_queue.
	 *
	 * When the application makes an accept call then the connection in
	 * accepted_queue is removed and handed over to the application. Then
	 * a new accept request is posted to the posted_queue.
	 */

	/*
	* This structure is only used by the active connection socket
	* and not by the passive listening socket.
	*/
	struct {
		struct list_head	list;
		uint8_t			accept_id;
		struct list_head        young_sid;
		struct pegnet_sock_id_s *parent_sid;
	} accept_active;
        pegnet_queue_addr_t     free_accept_qid;

        /*
         * Shutdown flags.
         */
        uint8_t                 shutdown;
        char                    urg_data;       /* actual urgent data */
        unsigned char           urg_flg;        /*  See enums bewlow */

        uint8_t			released : 1,	/* When close is called */
				listener : 1,
				zero_copy_enabled : 1,
				high_perf_flag : 1,	/* Indicates an high
							 * performance socket*/
                                rsvd_0 : 4;
        pegnet_queue_addr_t     qid;	/* qid of this socket instance */
        pegnet_queue_addr_t     rcvBuf_qid;	/* Where to post empty receive
						   buffers. */
	uint32_t                connID; /* Connection ID given by card */
	uint32_t                flow_id; /* flow_id used by host */
	uint32_t                tmp_ctx; /* ephemeral ctx */
        uint32_t                abs_posn;       /* amount of data read till
                                                   now */
        int                     urg_ptr;        /* Ptr to absolute posn of
                                                   urgent variable */
	uint32_t                reported_rcvbuf; /* what the card knows */
	nx_tnic_adapter_t 	*adapter;
	int			portnum;
	struct list_head        listener_list;
};


/* Urgent data flags */
enum {
        URG_IGN=0,          /* 0 - No urg data */
        URG_WITH_HOST,      /* 1 - urg data received but not given to host */
        URG_WITH_APP        /* 2 - urg data received and given to host but not
                               yet deleted from stream */
        /* state transition would be
           0->1->2->0 or 0->1->1...1->2->0 or 0>1->0 */
};

enum {
        LISTENER_SOCK = 0,
        LISTENER_GROUP,
        CONNECTION_SOCK
};

struct listener_group_s {
        uint32_t type;
        struct list_head listener_list;
        /*
         * Async accept requests are queued in posted_queue. The maximum
         * depth of this queue is the socket backlog value.
         *
         * When new connections are accepted they are removed from the
         * posted_queue and added to the accepted_queue.
         *
         * When the application makes an accept call then the connection in
         * accepted_queue is removed and handed over to the application. Then
         * a new accept request is posted to the posted_queue.
         */
#define UNM_POSTED_ACCEPT_MAX           PEGNET_MAX_SAVED_ACCEPTS
#define UNM_ACCEPT_STACK_MASK           0x01
        /*
         * This structure is only used by the passive listening
         * socket and not by the active connection socket.
         */
        struct {
                nx_lopt_t               *lopt;
                struct list_head        accept_queue;
                struct list_head        pruning;
                uint32_t                pruning_active;
                /*
                 * For now just alternate between host and offload
                 * stack when accepting connections. nxt_stack is
                 * either 0 or 1.
                 */
                uint8_t                 nxt_stack;
                uint32_t                accepted_len;
        } accept_passive;
        spinlock_t lock;
	uint8_t released;
	uint8_t shutdown;
};

typedef struct listener_group_s listener_group_t;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
extern struct inet_hashinfo	tcp_hashinfo;
#endif

/*
 * Handler registry structure for CARD_RESPONSE messages.
 */
typedef struct {
        uint32_t        opcode;
        uint32_t        retreive_socket;
        uint32_t        retreive_transaction;
        void (*func)(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans);
        void (*printmsg)(unm_msg_t *msg, pegnet_sock_id_t *sid,
			 pegnet_inprogress_t *trans); /* For printing debugs */
	char		*opcode_str;
} unm_card_response_handler_t;


/*
 * Different offload debug counters for the driver.
 */
enum {
	DBG_CNTR_ACTIVE_CONNS,
	DBG_CNTR_CONN_TBL_OVERFLOWS,
	DBG_CNTR_INUSE_SK,
	DBG_CNTR_INUSE_TX_PAGES,
	DBG_CNTR_INUSE_RX_PAGES,
	DBG_CNTR_INUSE_GLOBAL_RX_PAGES,
	DBG_CNTR_SYNQ_DROPPED,
	DBG_CNTR_SYN_FOR_FREED_LISTENER,
	DBG_CNTR_CONN_FOR_FREED_LISTENER,
	DBG_CNTR_ERR_PASSIVE_CONN_COMPLETE,
	DBG_CNTR_FREED_FROM_SYNQ,
	DBG_CNTR_FREED_FROM_ACCEPTQ,
	DBG_CNTR_SOCK_CLEANUP_CONNS,
	DBG_CNTR_TX_FREEING_IN_BH,
	DBG_CNTR_TX_DELAY_FREEING,
	DBG_CNTR_SENDPAGE_TIMER_FIRED,
	DBG_CNTR_SENDPAGE_FLUSHED1,
	DBG_CNTR_SENDPAGE_FLUSHED2,
	DBG_CNTR_SENDPAGE_FLUSHED3,
	DBG_CNTR_TOTAL_PRUNED,
	DBG_CNTR_TOTAL_PRUNABLES,
	DBG_CNTR_RST_ON_ACCEPT_QUEUE,
	DBG_CNTR_MAX,
};

/*
 * Structure to hold the global debug counters.
 */
typedef struct {
	atomic_t	counters[DBG_CNTR_MAX];
} nx_tnic_dbg_cntrs_t;

extern nx_tnic_dbg_cntrs_t	nx_tnic_dbg_cntrs;
#define UNM_HOST_PNIC_STATS_INIT()					\
        memset(&nx_tnic_dbg_cntrs, 0, sizeof (nx_tnic_dbg_cntrs))
#define UNM_HOST_PNIC_STATS_INC(VARIABLE)			\
        atomic_inc(&nx_tnic_dbg_cntrs.counters[VARIABLE])
#define UNM_HOST_PNIC_STATS_DEC(VARIABLE)			\
        atomic_dec(&nx_tnic_dbg_cntrs.counters[VARIABLE])
#define UNM_HOST_PNIC_STATS_GET(VARIABLE)			\
        atomic_read(&nx_tnic_dbg_cntrs.counters[VARIABLE])

typedef struct nx_tnic_procfs_s {
	struct proc_dir_entry	*base_dir;
	struct proc_dir_entry	*max_tcp_conn_file;
	struct proc_dir_entry	*dbg_counters_file;
	struct proc_dir_entry	*stats_file;
} nx_tnic_procfs_t;

/*
 *
 */
struct nx_tnic_adapter_s {
        spinlock_t		lock;
	nx_host_route_tbl_t	*rt_table;
	struct list_head	rt_sync_node;
	int			rt_sync_scheduled;
	flow_table_t		*p_ftable;

	nx_kc_work_queue_t      sk_cleanup_task;

	struct nx_sk_cleanup_queue	sk_cleanup_queue;
	zero_copy_stats         zcopy_stats;
	/* High performance stuff. */
	atomic_t		high_perf_cnt;
	/* Number of active writes being processed */
	atomic_t		num_tx_active_flows;
	/* Number of outstanding tx bytes */
	atomic_t		tx_bytes_out;
	struct net_device	*nx_port[NX_TNIC_MAX_PORTS];
	struct net_device       *nx_dev;

	struct list_head   	list;

	hostRxBuffer_t		**global_rx_bufs;
	int			total_global_rx_bufs;

        struct {
                netxen_pstats_t		*data;
                dma_addr_t		phys;
		wait_queue_head_t	wq;
		int			ret_code;
		uint8_t			pending;
        } stats;

	nx_tnic_procfs_t	procfs;
	atomic_t		dev_id;
	uint16_t                ring_ctx;
};




/***************************************************************************
 *
 *              Inline functions.
 *
 ***************************************************************************/

/*
 *
 */
static inline pegnet_sock_id_t *retreive_pegnet_sock_id(flow_table_t *tcp_tbl,
							uint32_t connID)
{
        return ((pegnet_sock_id_t *)host_flowtbl_get_flow(tcp_tbl,connID));
}

/*
 * Get realdevice for given device.
 */
static inline struct net_device *nx_get_realdev(struct net_device *dev)
{
        return ((dev->priv_flags & IFF_802_1Q_VLAN) ?
		VLAN_DEV_INFO(dev)->real_dev : dev);
}

/* Check whether a connection can be offloaded.
 * For now, just check the total number of offloaded conns
 */
static inline int check_card_resources(flow_table_t *tcp_tbl)
{
	return (host_flowtbl_get_nr_flows(tcp_tbl) > (max_tcp_conn - 1));
}

/*
 * Gets the TOE adapter if the device maps to a netxen device.
 */
static inline nx_tnic_adapter_t *nx_get_dev_tnic_adapter(struct net_device *dev)
{
	int	port;

	port = nx_nic_get_device_port((dev->priv_flags & IFF_802_1Q_VLAN) ?
				      VLAN_DEV_INFO(dev)->real_dev : dev);
	if (port < 0) {
		return (NULL);
	}
        return (nx_nic_get_tnic_adapter((dev->priv_flags & IFF_802_1Q_VLAN) ?
                                        VLAN_DEV_INFO(dev)->real_dev : dev));
}
/*
 * Function to send a message to the card.
 */
static inline int unm_send_h2p_msg(struct net_device *dev, unm_msg_t *qmsg,
				   size_t size)
{
        int		rv;
#if defined(PEGNET_NIC) && (UNM_H2P_TRANSPORT==UNM_H2P_TRANSPORT_NIC)
        h2p_msg_t       *h2pmsg;
	nx_tnic_adapter_t       *tnic_adapter;

        h2pmsg = (h2p_msg_t *)qmsg;

        h2pmsg->cmd_desc.opcode = PEGNET_REQUEST;
        h2pmsg->cmd_desc.qmsg_length = (size + 7) >> 3; /* Make it number of 8
                                                         * bytes count */
        tnic_adapter = nx_get_dev_tnic_adapter(dev);
        rv = nx_nic_send_tnic_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(tnic_adapter,0), &h2pmsg->cmd_desc, 1);
        if (rv != 0) {
                printk("%s: Cannot send message to peg as cmd desc.\n",
                       __FUNCTION__);
        }

#else   /* Queue message transport */

        /* FIXME: This seems wrong. Why are we sending it to the card-to-host
         *        msg queue? (MMO)
         */

	qmsg->hdr.src_major = PEGNET_HOST_RX_MAJOR;
	qmsg->hdr.src_side = 0;
	qmsg->hdr.src_minor = PEGNET_HOST_RX_MINOR;

        /* spin_lock(&iface_lock); */
        rv = unm_qm_enqueue_tail(&qmsg->hdr, &qmsg->body, size);
        /* spin_unlock(&iface_lock); */
        if (rv != 0) {
                printk("%s: Cannot enqueue message to peg\n", __FUNCTION__);
        }
#endif  /* (UNM_H2P_MSG_TRANSPORT==UNM_H2P_MSG_TRANSPORT_NIC) */

        return (rv);
}

static inline int unm_send_h2p_msg_atomic(struct net_device *dev,
					  unm_msg_t *msg_arr, int nr_elements)
{
	h2p_msg_t	*h2pmsg;
	int		i;
	nx_tnic_adapter_t       *tnic_adapter;

	for (i = 0; i < nr_elements; i++) {

		h2pmsg = (h2p_msg_t *)&msg_arr[i];	

		h2pmsg->cmd_desc.opcode = PEGNET_REQUEST;
		/* Make it number of 8 bytes count */
		h2pmsg->cmd_desc.qmsg_length = sizeof (unm_msg_body_t) >> 3;
	}
	tnic_adapter = nx_get_dev_tnic_adapter(dev);	
	return (nx_nic_send_tnic_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(tnic_adapter,0),
				     (pegnet_cmd_desc_t *)msg_arr,
				     nr_elements));
}

#define	NR_ENQUEUE_RETRIES	3

static inline int unm_send_h2p_msg_sleep(struct net_device *dev,
					 unm_msg_t *msg_arr, int nr_elements)
{
	int 		nr_retries = 0;
	
	while (unm_send_h2p_msg_atomic(dev,msg_arr, nr_elements)) {

		if (nr_retries > NR_ENQUEUE_RETRIES) {
			nx_print("enqueue failed, elements:%d\n", nr_elements);
			/* we did our best */
			return -ENOMEM;
		}
		nx_print("Failed to queue desc, going to sleep\n");
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(HOST_TIMEOUT);
		current->state = TASK_RUNNING;
		nr_retries++;

	}

	/* guaranteed to succeed */
	return 0;
}

/* #undef	NETXEN_USE_ALLOC_PAGES_LM */
/* #undef	NETXEN_USE_ALLOC_PAGES_HM */


/*
 *
 */
static inline int netxen_page_map_pci(netxen_host_page_t *page,
				      struct pci_dev *nx_pci_dev,
				      int direction)
{
	if (page->pci_mapped) {
		return (0);
	}
#if	defined(NETXEN_USE_ALLOC_PAGES_HM) || defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
	page->phys = pci_map_page(nx_pci_dev, page->mm_page, 0, page->size,
				  direction);
#else
	page->phys = pci_map_single(nx_pci_dev, page->buffer, page->size,
				    direction);
#endif
	if (pci_dma_mapping_error(page->phys)) {
		printk("%s: pci_map_single failed\n", __FUNCTION__);
		return (-1);
	}
	page->pci_mapped = 1;
	return (0);
}

/*
 *
 */
static inline void netxen_page_unmap_pci(netxen_host_page_t *page,
					 struct pci_dev *nx_pci_dev,
					 int direction)
{
	if (page->pci_mapped) {
#if	defined(NETXEN_USE_ALLOC_PAGES_HM) || defined(NETXEN_USE_ALLOC_PAGES_LM)
#error	"Has bug in 32 bit kernel"
		pci_unmap_page(nx_pci_dev, page->phys, page->size,
			       direction);
#else
		pci_unmap_single(nx_pci_dev, page->phys, page->size,
				 direction);
#endif
		page->pci_mapped = 0;
	}
}


/*
 *
 */
static inline char* netxen_page_map_virt(netxen_host_page_t *page)
{
#if	defined(NETXEN_USE_ALLOC_PAGES_HM)
#error	"Has bug in 32 bit kernel"
	if (page->buffer == NULL) {
		page->buffer = kmap(page->mm_page);
	}
#endif	/* NETXEN_USE_ALLOC_PAGES_HM */
	return (page->buffer);
}

/*
 *
 */
static inline void netxen_page_unmap_virt(netxen_host_page_t *page)
{
#if	defined(NETXEN_USE_ALLOC_PAGES_HM)
#error	"Has bug in 32 bit kernel"
	if (page->buffer) {
		kunmap(page->mm_page);
		page->buffer = NULL;
	}
#endif	/* NETXEN_USE_ALLOC_PAGES_HM */
}

/*
 * Checks if a given device maps to a netxen device.
 */
static inline int is_netxen_device(struct net_device *dev)
{
	return (nx_nic_is_netxen_device((dev->priv_flags & IFF_802_1Q_VLAN) ?
					VLAN_DEV_INFO(dev)->real_dev : dev));
}

/*
 * Gets the port number on the card if the device maps to a netxen device.
 */
static inline int nx_get_device_port(struct net_device *dev)
{
	return (nx_nic_get_device_port((dev->priv_flags & IFF_802_1Q_VLAN) ?
				       VLAN_DEV_INFO(dev)->real_dev : dev));
}


/*
 *
 */
static inline uint16_t nx_get_ring_ctx(struct net_device *dev)
{
	int	rv;

	rv = nx_nic_get_device_ring_ctx(dev);
	if (rv < 0) {
		nx_print("Invalid device\n");
		return (0);
	}

	return ((uint16_t)rv);
}

/*
 * Register Tnic driver with Nic driver.
 *
 */
 
static inline int nx_register_tnic(struct net_device *dev,
				   nx_tnic_adapter_t *data,
				   nx_tnic_callback_t *cb)
{			 
	return (nx_nic_register_tnic(((dev->priv_flags & IFF_802_1Q_VLAN) ?
				      VLAN_DEV_INFO(dev)->real_dev : dev),
				     data, cb));
}			 


/*
 * Unregisters a known TOE adapter.
 */
static inline void nx_unregister_tnic(struct net_device *dev)
{
	nx_nic_unregister_tnic((dev->priv_flags & IFF_802_1Q_VLAN) ?
				VLAN_DEV_INFO(dev)->real_dev : dev);
	
}

/*
 * Gets the pointer to pci device if the device maps to a netxen device.
 */
static inline struct pci_dev *nx_get_pcidev(struct net_device *dev)
{
	return (nx_nic_get_pcidev((dev->priv_flags & IFF_802_1Q_VLAN) ?
				  VLAN_DEV_INFO(dev)->real_dev : dev));
}

static inline struct pci_dev * get_pci_dev_from_sid(pegnet_sock_id_t *sid)
{
    return nx_get_pcidev(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,sid->portnum));
}

/*
 * Compare adapter ids for two NetXen devices.
 */
static inline int nx_cmp_adapter_id(struct net_device *dev1,
				    struct net_device *dev2)
{
        return (nx_nic_cmp_adapter_id(((dev1->priv_flags & IFF_802_1Q_VLAN) ?
				       VLAN_DEV_INFO(dev1)->real_dev : dev1),
				      ((dev2->priv_flags & IFF_802_1Q_VLAN) ?
				       VLAN_DEV_INFO(dev2)->real_dev : dev2)));
}

/*
 * In af_unm.c
 */
int unm_create(struct net_device *netdev,struct socket *sock, int protocol,
	       int type, int zcopy);
void nx_set_route_reference(pegnet_sock_id_t *sid, uint32_t saddr,
			    uint16_t sport, uint32_t daddr, uint16_t dport);
int unm_release(struct socket *sock, int delay_sk_release);
int unm_connect(struct socket *sock, struct sockaddr *saddr, int len,
                int flags, int do_bind);
int unm_disconnect(struct socket *sock, int flags, long timeo);
int unm_listen(struct socket *sock, int backlog, int do_bind);
int unm_accept(struct socket *sock, struct socket *newsock, int flags);
int unm_getname(struct socket *sock, struct sockaddr *uaddr, int *addr_len,
                int peer);
unsigned int unm_poll(struct socket *sock);
int unm_ioctl(pegnet_sock_id_t *sid, unsigned int cmd, unsigned long arg);
int unm_shutdown(struct socket *sock, int how);
int unm_setsockopt(pegnet_sock_id_t *sid, int level, int optname, char *optval,
                   int optlen);
int unm_getsockopt(pegnet_sock_id_t *sid, int level, int optname, char *optval,
                   int *optlen);
int unm_sendmsg(struct socket *sock, struct msghdr *m, size_t size);
int unm_recvmsg(struct socket *sock, struct msghdr *m, int total_len,
                int flags);
int unm_host_is_accept_queue_empty(struct socket *sock);
int unm_host_get_next_accept_stack(struct socket *sock);
int nx_handle_offloaded_syn(struct net_device *netdev,
			    uint16_t listener_id, struct sk_buff *skb);
void unm_process_pegnet_msg(struct net_device *netdev, unm_msg_t *msg);
int nx_tnic_adapter_create_procfs(nx_tnic_adapter_t *tnic,
				  struct proc_dir_entry *adapter_dir);
void nx_tnic_adapter_destroy_procfs(nx_tnic_adapter_t *tnic,
				    struct proc_dir_entry *adapter_dir);
extern int check_pending_data(struct socket *sock);
extern void free_pegnet_sock_id(pegnet_sock_id_t *sid);
extern void nx_remove_synq(listener_group_t *grp, pegnet_sock_id_t *sid);
extern int alloc_netxen_page(netxen_host_page_t *page);
extern int free_netxen_page(netxen_host_page_t *page, struct pci_dev *unm_device_ptr,
					int direction);
extern int smart_memcpy(void *dest, void *src, int size);
void nx_init_sk_cleanup_task(nx_tnic_adapter_t *tnic_adapter);
pegnet_sock_id_t *get_sid_by_grp(nx_tnic_adapter_t *tnic_adapter,
                               listener_group_t *listener_grp);
void nx_sock_cleanup_task_internal(pegnet_sock_id_t *sid);
void nx_sk_cleanup(struct sock *sk);

/*
 * Functions in pegnet_rpc_host.c
 */
/*
 * This function processes the accept_response from the card. If successful
 * it will put the entry into accepted queue and wake up the waiting user
 * context. It also posts new accept request to the card.
 *
 * Parameters:
 *      msg     - The message from the card.
 *      sock    - The retreived socket context for the message.
 *      trans   - NULL in this case.
 *
 * Return:
 *      None.
 */
void unm_host_proc_accept_resp(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_accept_error_resp(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);

void unm_proc_connect_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans);
void unm_proc_connect_complete(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans);
void unm_proc_conn_disconnect_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans);
void unm_host_proc_listen_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_ioctl_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_setsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_getsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_iptsetsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg,
                        pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_iptgetsockopt_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg,
                        pegnet_sock_id_t *sock,pegnet_inprogress_t *trans);
void unm_host_proc_shutdown_response(nx_tnic_adapter_t *tnic_adapter,unm_msg_t *msg, 
			pegnet_sock_id_t *sid,pegnet_inprogress_t *trans);
pegnet_sock_id_t *unm_host_async_accept(listener_group_t *grp, int flags,
					int *err);
void register_card_response_handler(unm_card_response_handler_t *handler);

int nx_tnic_reset_fw(struct net_device *dev);
int host_ifconfig(struct net_device *dev, struct in_ifaddr *ifa,
		  int port, uint16_t cmd, int logical);

int host_release(pegnet_sock_id_t *sid, int delay_sk_release);
void host_release_unaccepted_child(pegnet_sock_id_t *sid);
pegnet_sock_id_t* host_create(struct net_device *netdev, int family,
			      int protocol, int *pegnet_rval);
int host_connect(pegnet_sock_id_t *sock, struct sockaddr *uaddr,
                 int addr_len, int flags, int do_bind);
int host_conn_disconnect(pegnet_sock_id_t *sid, int abortive, int tx, int rx);
int host_getname(pegnet_sock_id_t* sock, struct sockaddr* uaddr, int* addrlen,
                 int peer);
int host_ioctl(pegnet_sock_id_t* sock, int cmd, unsigned long val, int ioctl_sk);
int host_ioctl_mcast(pegnet_sock_id_t* sock, int cmd, unsigned long val);
int host_shutdown(pegnet_sock_id_t* sock, int how);
int host_setsockopt(pegnet_sock_id_t* sock, int level, int optname,
                    unsigned long optval, int optlen, unsigned char is_val);
int host_getsockopt(pegnet_sock_id_t *sock, int level, int optname,
                    uint64_t optval, int *optlen);
void nx_host_post_accept(pegnet_sock_id_t *sid, unsigned int count);
int host_listen(pegnet_sock_id_t* sock, int backlog, int do_bind);
int host_reset_stats(struct net_device *netdev, __uint32_t port);
int host_get_stats(nx_tnic_adapter_t *tnic, uint32_t port);
void nx_proc_get_stats_reply(nx_tnic_adapter_t *tnic, unm_msg_t *msg,
			     pegnet_sock_id_t *sid,
			     pegnet_inprogress_t *trans);
pegnet_sock_id_t *alloc_pegnet_sock_id(struct net_device *);

/**** SYN-queue pruning declarations *****/
#define NX_SYNQ_PRUNETYPE_BH           1
#define NX_SYNQ_PRUNETYPE_NORMAL       2
void nx_init_pruning_task(void);
void nx_prune_synq(listener_group_t *listener_grp, unsigned int type);
extern void release_accept_queues(listener_group_t *grp);

/*
 * In nx_tnic_output.c
 */
extern int is_tx_vbuf_allocatable(vbuf_type_t type, pegnet_sock_id_t *sid);

extern int host_sendmsg_fast(pegnet_sock_id_t *sid, struct msghdr *msg);
int unm_sendpage(struct socket *sock,struct page **pages, int poffset,
		 size_t psize, int flags);
extern int nx_flush_sendpage(pegnet_sock_id_t *sid);
void process_last_tx_completion(pegnet_sock_id_t *sid);
void cleanup_tx_vbuf_queues(pegnet_sock_id_t *sid);
void proc_sendmsg_control(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg,
			  pegnet_sock_id_t *sid, pegnet_inprogress_t *trans);


/*
 * In nx_tnic_input.c
 */
extern int nx_init_rxbuffers_pcb(pegnet_sock_id_t *sid);
extern int nx_init_rx_global_buffers(nx_tnic_adapter_t *tnic);
extern void nx_destroy_rx_global_buffers(nx_tnic_adapter_t *tnic);
extern void cleanup_rxbuffers_global(pegnet_sock_id_t *sid);
extern void cleanup_rxbuffers_pcb(pegnet_sock_id_t *sid);

extern int host_recvmsg_fast(pegnet_sock_id_t *sid, struct msghdr *msg,
			     int requested_len, int flags,
			     struct scm_cookie *scm);
extern int host_recv_available(pegnet_sock_id_t *sid,unsigned long arg);
extern void nx_proc_rx_urg_pointer(nx_tnic_adapter_t *tnic_adapter,
				   unm_msg_t *msg, pegnet_sock_id_t *sid,
				   pegnet_inprogress_t *trans);
extern void nx_proc_rx_buffer(struct net_device *netdev,unm_msg_t *msg);

/*
 * In nx_route_table.c
 */
extern int nx_route_tbl_create(nx_tnic_adapter_t *adapter);
extern void nx_route_tbl_free(nx_tnic_adapter_t *adapter);
extern int nx_route_add(nx_tnic_adapter_t *adapter, struct rtable *req,
			uint8_t port);
extern int nx_route_update(nx_tnic_adapter_t *adapter, struct rtable *req,
			   uint8_t port);
extern int nx_route_remove(nx_tnic_adapter_t *adapter, nx_host_rt_key_t *key);
extern void nx_proc_route_invalid(nx_tnic_adapter_t *adapter,
				  routecache_invalid_t *msg);
extern int nx_route_connection(struct rtable **rp, uint32_t saddr,
			       uint16_t sport, uint32_t daddr, uint16_t dport,
			       uint8_t tos, int oif);
struct net_device *netxen_find_dev_for_ip(u32 addr);
extern void flow_tbl_destroy_flow_free(flow_id_t id, void *flow);
extern int nx_route_tbl_init(void);
extern void nx_route_tbl_destroy(void);

/*
 * unm_pstack_init.c
 */
int pstack_init(void);
void pstack_exit(void);

#endif	/* _UNM_HOST_TYPES_H */

