/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*****************************************************************************
 * shim.c -- UNM shim layer
 *
 *
 *
 *
 *
 ****************************************************************************
 */
#include <asm/ioctls.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <linux/inetdevice.h>
#include <net/route.h>
#include <linux/if_vlan.h>
#include <linux/if.h>
#include "kernel_compatibility.h"
#include "unm_host_pnic_types.h"
#include "nx_route_table.h"
#include "pstack.h"
#include <linux/random.h>
#include <unm_version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#include <net/inet_hashtables.h>
#endif
nx_tnic_callback_t	g_tnic_callback;
extern nx_hbucket_head_t g_adapter_head;


static int nx_netdev_event(struct notifier_block *this, unsigned long event,
			   void *ptr);
static int nx_inetaddr_event(struct notifier_block *this, unsigned long event,
			     void *ptr);

extern int (*native_inet_dgrams_ioctl)(struct socket *sock, unsigned int cmd,
                                       unsigned long arg);

extern atomic_t	pegnet_init_done;

struct notifier_block	nx_netdev_cb = {
	notifier_call:	nx_netdev_event,
};

struct notifier_block	nx_inetaddr_cb = {
	notifier_call:	nx_inetaddr_event,
};

/* static unm_port_data_t	*upd = NULL; */
/* static rwlock_t		unm_offload_rwlock = RW_LOCK_UNLOCKED; */
static netxen_offload_rule_list_t       rules = {
        list : LIST_HEAD_INIT(rules.list),
        cnt : 0,
        rwlock : RW_LOCK_UNLOCKED
};
                                                 

/*
 * Function declarations.
 */
static int unm_port_manager_add(netxen_offload_rule_t *rule);
static int unm_port_manager_delete(netxen_offload_rule_t *rule);
static int unm_port_manager_list(char *user_arg, netxen_offload_ioctl_t *data);
static int nx_check_listen_offload(unsigned port, int *zcopy);

/*
 * Funtion to add, delete  or list rules for offloading.
 */
static int unm_port_manager(unsigned long arg)
{
	netxen_offload_ioctl_t  *data;
        int                     rv = 0;

	data = (netxen_offload_ioctl_t *)kmalloc(sizeof (netxen_offload_ioctl_t), GFP_KERNEL);
	if (data == NULL) {
		printk("%s: Error: Cannot allocate memory for ioctl call.\n",
                       __FUNCTION__);
		return (-ENOMEM);
	}

	if (copy_from_user(data, (void *)arg, sizeof(*data))) {
                printk("%s: copy_from_user failed\n", __FUNCTION__);
		rv = -EFAULT;
                goto done;
	}

        switch (data->hdr.cmd) {
        case NETXEN_OFFLOAD_IOCTL_GET_RULE_CNT:
                read_lock(&rules.rwlock);
                data->hdr.cnt = rules.cnt;
                data->hdr.ret_code = 0;
                read_unlock(&rules.rwlock);
                if (copy_to_user((void *)arg, data, sizeof(*data))) {
                        printk("%s: copy_to_user failed\n", __FUNCTION__);
                        rv = -EFAULT;
                }
                break;

        case NETXEN_OFFLOAD_IOCTL_GET_RULES:
                rv = unm_port_manager_list((void *)arg, data);
                break;

        case NETXEN_OFFLOAD_IOCTL_ADD_RULE:
                data->hdr.ret_code = unm_port_manager_add(&data->u.rule);
                if (data->hdr.ret_code != 0) {
                    rv = -EFAULT;
                }
                if (copy_to_user((void *)arg, data, sizeof(*data))) {
                        printk("%s: copy_to_user failed\n", __FUNCTION__);
                        rv = -EFAULT;
                }
                break;

        case NETXEN_OFFLOAD_IOCTL_REMOVE_RULE:
                data->hdr.ret_code = unm_port_manager_delete(&data->u.rule);
                if (data->hdr.ret_code != 0) {
                    rv = -EFAULT;
                }
                if (copy_to_user((void *)arg, data, sizeof(*data))) {
                        printk("%s: copy_to_user failed\n", __FUNCTION__);
                        rv = -EFAULT;
                }
                break;
        }

  done:
	kfree(data);
	return (rv);
}

/*
 * Adds a new rule for offloading.
 */
static int unm_port_manager_add(netxen_offload_rule_t *rule)
{
	netxen_offload_rule_node_t      *tmp = NULL;
        struct list_head                *node;
	netxen_offload_rule_node_t      *check_list;

	if (capable(CAP_SYS_ADMIN) == 0) {
		return (-EPERM);
	}

        if (rule->type != NETXEN_OFFLOAD_RULE_BY_TUPLE &&
            rule->type != NETXEN_OFFLOAD_RULE_BY_APP_NAME) {
                printk("%s: Error: Unknown rule type[%u]\n", __FUNCTION__,
                       rule->type);
                return (-NETXEN_IOCTL_ERR_UNSUPPORTED_RULE);
        }

	write_lock(&rules.rwlock);

	/* Before adding check whether its already there*/
        list_for_each(node, &rules.list) {
                check_list = list_entry(node, netxen_offload_rule_node_t,
                                        links);
                if (check_list->rule.type == rule->type) {
                        if (rule->type == NETXEN_OFFLOAD_RULE_BY_TUPLE) {
                                if (check_list->rule.local_port == rule->local_port &&
				    check_list->rule.remote_port == rule->remote_port) {
                                        write_unlock(&rules.rwlock);
                                        return (-NETXEN_IOCTL_ERR_RULE_EXISTS);
                                }
                        } else {
                                if (strcmp(rule->app_name,
                                           check_list->rule.app_name) == 0) {
                                        write_unlock(&rules.rwlock);
                                        return (-NETXEN_IOCTL_ERR_RULE_EXISTS);
                                }
                        }
                }
	}

	tmp = (netxen_offload_rule_node_t *)kmalloc(sizeof (netxen_offload_rule_node_t), GFP_KERNEL);
	if (tmp == NULL) {
		printk(KERN_ERR "%s: Cannot allocate memory\n", __FUNCTION__);
		write_unlock(&rules.rwlock);
		return (-ENOMEM);
	}

	tmp->rule = *rule;
        list_add_tail(&tmp->links, &rules.list);
        rules.cnt++;

	write_unlock(&rules.rwlock);

        return (0);
}

/*
 * Removes a rule from the rule list.
 */
static int unm_port_manager_delete(netxen_offload_rule_t *rule)
{
	struct list_head                *node;
        netxen_offload_rule_node_t      *check_list;
        int                             rv;

	if (capable(CAP_SYS_ADMIN) == 0) {
		return (-EPERM);
	}

        if (rule->type != NETXEN_OFFLOAD_RULE_BY_TUPLE &&
            rule->type != NETXEN_OFFLOAD_RULE_BY_APP_NAME) {
                printk("%s: Error: Unknown rule type[%u]\n", __FUNCTION__,
                       rule->type);
                return (-NETXEN_IOCTL_ERR_UNSUPPORTED_RULE);
        }

	write_lock(&rules.rwlock);

        rv = -NETXEN_IOCTL_ERR_RULE_NOT_FOUND;
        list_for_each(node, &rules.list) {
                check_list = list_entry(node, netxen_offload_rule_node_t,
                                        links);
                if (check_list->rule.type == rule->type) {
                        if (rule->type == NETXEN_OFFLOAD_RULE_BY_TUPLE) {
                                if (check_list->rule.local_port == rule->local_port &&
				    check_list->rule.remote_port == rule->remote_port) {
                                        list_del(&check_list->links);
                                        rules.cnt--;
                                        kfree(check_list);
                                        rv = 0;
                                        break;
                                }
                        } else {
                                if (strcmp(rule->app_name,
                                           check_list->rule.app_name) == 0) {

                                        list_del(&check_list->links);
                                        rules.cnt--;
                                        kfree(check_list);
                                        rv = 0;
                                        break;
                                }
                        }
                }
	}

	write_unlock(&rules.rwlock);

        return (rv);
}


/*
 *List offloaded ports and appns 
 */
static int unm_port_manager_list(char *user_arg, netxen_offload_ioctl_t *data)
{
	struct list_head                *node;
        netxen_offload_rule_node_t      *check_list;
        char                            *to;
        int                             rv;
        uint16_t                        cnt = 0;

        to = user_arg + sizeof (netxen_offload_ioctl_hdr_t);

	read_lock(&rules.rwlock);

        rv = 0;
        list_for_each(node, &rules.list) {
                check_list = list_entry(node, netxen_offload_rule_node_t,
                                        links);
                cnt++;
                if (copy_to_user(to, &check_list->rule,
                                 sizeof (netxen_offload_rule_t))) {
                        printk("%s: copy_to_user failed\n", __FUNCTION__);
                        rv = -EFAULT;
                        break;
                }
                if (cnt == data->hdr.cnt) {
                        break;
                }
                to += sizeof (netxen_offload_rule_t);
        }

        read_unlock(&rules.rwlock);

        data->hdr.ret_code = rv;
        data->hdr.cnt = cnt;
        if (copy_to_user(user_arg, &data->hdr,
                         sizeof (netxen_offload_ioctl_hdr_t))) {
                printk("%s: copy_to_user 2 failed\n", __FUNCTION__);
                rv = -EFAULT;
        }

        return (rv);
}

/*
 * Checks if a port is to be offloaded or if the calling application is to be
 * offloaded.
 */
static int nx_check_listen_offload(unsigned int port, int *zcopy)
{
	struct list_head                *node;
        netxen_offload_rule_node_t      *check_list;
        netxen_offload_rule_t		*rule = NULL;
	int				rv = 0;


        // Check whether card has started at all !!
        // XXX : should go out after multicard checkins
        if(!atomic_read(&pegnet_init_done)) {
                return 0;
        }

	read_lock(&rules.rwlock);

        list_for_each(node, &rules.list) {
                check_list = list_entry(node, netxen_offload_rule_node_t,
                                        links);
		rule = &check_list->rule;

		/* check for port number match */
		if (rule->local_port && port) {
			if (rule->local_port == port) {
				rv = 1;
				break;
			} else {
				continue;
			}
		}
	    
		/* check for app name match */
		if (strncmp(current->comm, rule->app_name, 15) == 0) {
			rv = 1;
			break;
		}
	}

	if (rv && zcopy) {
		*zcopy = rule->zero_copy;
	}

	read_unlock(&rules.rwlock);
	return (rv);
}

/*
 * Checks if a port is to be offloaded or if the calling application is to be
 * offloaded.
 */
static int nx_check_connect_offload(uint8_t family, ip_addr_t *local_ip,
				    uint16_t local_port, ip_addr_t *remote_ip,
				    uint16_t remote_port, int *zcopy)
{
	struct list_head                *node;
        netxen_offload_rule_node_t      *check_list;
        netxen_offload_rule_t		*rule = NULL;
	int				rv = 0;

        // Check whether card has started at all !!
        // XXX : should go out after multicard checkins
        if(!atomic_read(&pegnet_init_done)) {
                return 0;
        }
	read_lock(&rules.rwlock);

        list_for_each(node, &rules.list) {
                check_list = list_entry(node, netxen_offload_rule_node_t,
                                        links);
		rule = &check_list->rule;

		/* check for tuple match */
		if (rule->type == NETXEN_OFFLOAD_RULE_BY_TUPLE) {
			nx_print1("Check tuple rule: local[%u], remote[%u]: "
				  "local[%u], remote[%u]\n",
				  rule->local_port, rule->remote_port,
				  local_port, remote_port);

			if ((!rule->local_port ||
			     rule->local_port == local_port) &&
			    (!rule->remote_port ||
			     rule->remote_port == remote_port)) {
				rv = 1;
				break;
			}

		} else {
			/* check for app name match */
			if (strncmp(current->comm, rule->app_name, 15) == 0) {
				rv = 1;
				break;
			}
		}
	}

	if (rv && zcopy) {
		*zcopy = rule->zero_copy;
	}
	read_unlock(&rules.rwlock);
	return (rv);
}

/*
 * Free tnic adapter which is created at time of initialization
 *
 * Parameters:
 *	tnic_adapter - Pointer to tnic_adapter which is going to be freed.
 *
 * Returns:
 *	None
 */
static void nx_tnic_adapter_free(nx_tnic_adapter_t *tnic)
{
	nx_destroy_rx_global_buffers(tnic);

	nx_tnic_adapter_destroy_procfs(tnic, NULL);
	pci_free_consistent(nx_get_pcidev(tnic->nx_dev),
			    sizeof (netxen_pstats_t), (void *)tnic->stats.data,
			    tnic->stats.phys);

        host_flowtbl_destroy(tnic->p_ftable, flow_tbl_destroy_flow_free);
        kfree(tnic->p_ftable);
        tnic->p_ftable = NULL;

        nx_route_tbl_free(tnic);
        kfree(tnic);
}

/*
 * Cleanup per adapter structure.
 */
static void nx_tnic_cleanup(nx_tnic_adapter_t *tnic)
{
	struct in_ifaddr	*ifa;
	int			port;
	struct in_device	*in_dev;
	int			rv;

	/*
	 * Send the firmware a reset command.
	 */
	rv = nx_tnic_reset_fw(tnic->nx_dev);
	if (rv) {
		nx_print("Failed sending reset command to card\n");
	}

	/*
	 * Bring the interfaces down.
	 */
	for (port = 0; port < NX_TNIC_MAX_PORTS; ++port) {
		if (tnic->nx_port[port] != NULL) {
			in_dev = in_dev_get(tnic->nx_port[port]);
			if (in_dev != NULL) {
				ifa = in_dev->ifa_list;
				in_dev_put(in_dev);

				host_ifconfig(tnic->nx_port[port], ifa, port,
					      NX_NETDEV_DOWN, 0);
			}
			nx_unregister_tnic(tnic->nx_port[port]);
			tnic->nx_port[port] = NULL;
		}
	}

	nx_tnic_adapter_free(tnic);
}

/*
 *
 */
void nx_cleanup_all_tnics(void)
{
	struct list_head	*node1;
	struct list_head	*node2;
	nx_tnic_adapter_t	*tnic_adapter;

	/*
	 * Don't use spin_lock_bh(). This causes vfree() for host table to fail
	 * Anyway the g_adapter_head should never be accessed in the bottom
	 * half.
	 */
	spin_lock(&g_adapter_head.lock);

	list_for_each_safe(node1, node2, &g_adapter_head.head) {
		tnic_adapter = list_entry(node1, nx_tnic_adapter_t, list);
		list_del_init(&tnic_adapter->list);
		nx_tnic_cleanup(tnic_adapter);
	}

	spin_unlock(&g_adapter_head.lock);
}

/*
 * Creation and Initialization of per adapter data structure and then
 * register tnic data and callback functions to nic.
 *
 * Parameters:
 *	dev	- Pointer to device.
 *	port	- port number.
 *
 * Returns:
 *	 0	- On Success.
 */
static int nx_tnic_adapter_init(struct net_device *dev, int port)
{
	nx_tnic_adapter_t	*tnic_adapter = NULL;
	struct list_head	*node;
	int			rv = 0;
	nic_version_t		version;
	int			mismatch = 0;

	nx_print("dev %s \n", dev->name);

	spin_lock_bh(&g_adapter_head.lock);

	list_for_each(node, &g_adapter_head.head) {
		tnic_adapter = list_entry(node, nx_tnic_adapter_t, list);
		if (0 == nx_cmp_adapter_id(tnic_adapter->nx_dev, dev)) {
			rv = nx_register_tnic(dev, tnic_adapter,
					      &g_tnic_callback);
			if (rv) {
				if (rv == -1) {
					nx_print("Adapter is not set\n");
				} else {
					nx_print("Already registered\n");
					rv = 0;
				}

			} else	if (tnic_adapter->nx_port[port] == NULL) {
				tnic_adapter->nx_port[port] =
					nx_get_realdev(dev);
			}
			spin_unlock_bh(&g_adapter_head.lock);
			return (rv);
		}
	}
	spin_unlock_bh(&g_adapter_head.lock);

	/* 
	 * Check version number before preceeding. 
	 */
	nx_nic_get_tnic_version_number(dev, &version);
	if (version.major != UNM_TCP_FW_VERSION_MAJOR) {
		mismatch = 1;
	}
	if (version.minor != UNM_TCP_FW_VERSION_MINOR) {
		mismatch = 1;
	}
	if (mismatch) {
		printk(KERN_ERR
		    "%s:Version mismatch between tnic driver and tnic fw:\n"
		    "  Driver major version: %d, Firmware major version: %d\n"
		    "  Driver minor version: %d, Firmware minor version: %d\n",
		    __FUNCTION__, 
		    UNM_TCP_FW_VERSION_MAJOR, version.major, 
		    UNM_TCP_FW_VERSION_MINOR, version.minor );
		return (1);
	}
	nx_nic_get_nic_version_number(dev, &version);
	if (version.major != _UNM_NIC_LINUX_MAJOR) {
		mismatch = 1;
	}
	if (version.minor != _UNM_NIC_LINUX_MINOR) {
		mismatch = 1;
	}
	if (mismatch) {
		printk(KERN_ERR
		    "%s:Version mismatch between nic driver and nic fw:\n"
		    "  Driver major version: %d, Firmware major version: %d\n"
		    "  Driver minor version: %d, Firmware minor version: %d\n",
		    __FUNCTION__, 
		    _UNM_NIC_LINUX_MAJOR, version.major, 
		    _UNM_NIC_LINUX_MINOR, version.minor );
		return (1);
	}

	/*
	 * Not created yet so create an adapter.
	 */
	tnic_adapter = (nx_tnic_adapter_t *)kmalloc(sizeof (nx_tnic_adapter_t),
						    GFP_KERNEL);
	if (NULL == tnic_adapter) {
		nx_print("Tnic adapter structure allocation failed\n");
		return	-ENOMEM;
	}

	memset(tnic_adapter, 0, sizeof (nx_tnic_adapter_t));

	/*
	 * First initialize the lock.
	 */
	spin_lock_init(&tnic_adapter->lock);

	/*
	 * Initialize the routing table.
	 */
	tnic_adapter->rt_sync_scheduled = 0;
	INIT_LIST_HEAD(&tnic_adapter->rt_sync_node);
	if ((rv = nx_route_tbl_create(tnic_adapter))) {
		nx_print("Routing table init failed\n");
		kfree(tnic_adapter);
		return rv;
	}

	/*
	 * Init sk cleanup task.
	 */
	nx_init_sk_cleanup_task(tnic_adapter);

	/*
	 * Initialize the flow table.
	 */
	tnic_adapter->p_ftable = (flow_table_t *)kmalloc(sizeof (flow_table_t),
							 GFP_KERNEL);
	if(NULL == tnic_adapter->p_ftable) {
		nx_print("unm_init: Flow table allocation failed\n");
		nx_route_tbl_free(tnic_adapter);
		kfree(tnic_adapter);
		return	-ENOMEM;
	}

	memset(tnic_adapter->p_ftable, 0, sizeof (flow_table_t));

	/*
	 * Now allocate space for statistics.
	 */
	init_waitqueue_head(&tnic_adapter->stats.wq);
	tnic_adapter->stats.data =
		pci_alloc_consistent(nx_get_pcidev(dev),
				     sizeof (netxen_pstats_t),
				     &tnic_adapter->stats.phys);

	if (tnic_adapter->stats.data == NULL) {
		nx_print("ERROR: Could not allocate space for statistics\n");
		kfree(tnic_adapter->p_ftable);
		nx_route_tbl_free(tnic_adapter);
		kfree(tnic_adapter);
		return (-ENOMEM);
	}

	/*
	 * Set up the procfs space.
	 */
	rv = nx_tnic_adapter_create_procfs(tnic_adapter, NULL);
	if (rv) {
		nx_print("Error[%d]: Procfs init FAILED.\n", rv);
		pci_free_consistent(nx_get_pcidev(dev),
				    sizeof (netxen_pstats_t),
				    (void *)tnic_adapter->stats.data,
				    tnic_adapter->stats.phys);
		kfree(tnic_adapter->p_ftable);
		nx_route_tbl_free(tnic_adapter);
		kfree(tnic_adapter);
	}

	/*
	 * Set the correct net device
	 */
	tnic_adapter->nx_port[port] = nx_get_realdev(dev);
	tnic_adapter->nx_dev = nx_get_realdev(dev);
	tnic_adapter->ring_ctx = nx_get_ring_ctx(dev);

	atomic_set(&tnic_adapter->high_perf_cnt, NUM_HIGH_PERF_FLOWS);
	atomic_set(&tnic_adapter->num_tx_active_flows, 0);
	atomic_set(&tnic_adapter->tx_bytes_out, 0);
	rv = host_flowtbl_init(tnic_adapter->p_ftable);
	if (rv != 0) {
		nx_print("Flow table init failed. [%d]\n", rv);
		nx_tnic_adapter_destroy_procfs(tnic_adapter, NULL);
		pci_free_consistent(nx_get_pcidev(dev),
				    sizeof (netxen_pstats_t),
				    (void *)tnic_adapter->stats.data,
				    tnic_adapter->stats.phys);
		kfree(tnic_adapter->p_ftable);
		nx_route_tbl_free(tnic_adapter);
		kfree(tnic_adapter);
		return (rv);
	}

	/*
	 * Now register with the NIC for callbacks.
	 */
	rv = nx_register_tnic(dev, tnic_adapter, &g_tnic_callback);
	if (rv) {
		if (rv == -1) {
			nx_print("Port or Adapter is not set");
		} else {
			nx_print("Already registered");
		}
		nx_tnic_adapter_free(tnic_adapter);
		return rv;
	}

	/*
	 * First send a reset message to the card to clean up any residual
	 * information.
	 */
	rv = nx_tnic_reset_fw(dev);
	if (rv) {
		nx_print("Failed sending reset command to card\n");
	}

	/*
	 * Allocate and post global RX buffers.
	 */
	rv = nx_init_rx_global_buffers(tnic_adapter);
	if (rv) {
		nx_print("Global RX buffer posting failed\n");
		nx_tnic_cleanup(tnic_adapter);
		return (rv);
	}

	/*
	 * Add to the global list of known netxen adapters.
	 */
	nx_hbucket_add_tail(&tnic_adapter->list, &g_adapter_head);

	return 0;
}


/*
 * Check array is empty or not
 *
 * Parameters:
 *	tnic_adapter	- Pointer to tnic_adapter which contains array of 
 * 			  pointer to device.
 *
 * Returns:
 *	 0	- If array is empty.
 *	 1	- If array is not empty.
 */
static int nx_is_dev_registered(nx_tnic_adapter_t *tnic_adapter)
{
        int	i;

        for (i = 0; i < NX_TNIC_MAX_PORTS; i++) {
                if (tnic_adapter->nx_port[i]) {
			return (1);
                }
        }
        return (0);
}


/*
 * Netlink notifier callback.
 */
static int nx_netdev_event(struct notifier_block *this, unsigned long event,
			   void *ptr)
{
	struct net_device	*dev = (struct net_device *)ptr;
	struct in_ifaddr	*ifa;
	int			port;
        struct in_device        *in_dev;
	nx_tnic_adapter_t       *tnic_adapter;

	port = nx_get_device_port(dev);
	if (port < 0) {
		nx_print("%s: Not NetXen card\n", dev->name);
		return (NOTIFY_DONE);
	}
//	nx_print("NetXen Device: Dev[%s], Real dev[%s], vlan_id = %u\n",
//		 dev->name, realdev->name, vlan_id);

	switch (event) {
	case NETDEV_UP:
		/* Get the IP address */
		in_dev = in_dev_get(dev);
		if (in_dev == NULL) {
			nx_print("Could not get the in_device for %s.\n",
				 dev->name);
			break;
		}

		ifa = in_dev->ifa_list;
		in_dev_put(in_dev);
		if (nx_tnic_adapter_init(dev, port)) {
			nx_print("ifconfig %s UP, port = %d --> FAILED.\n",
				 ifa->ifa_label, port);
			return (NOTIFY_DONE);
		}
		nx_print("ifconfig %s UP, port = %d.\n", ifa->ifa_label, port);
		host_ifconfig(dev, ifa, port, NX_NETDEV_UP, 0);
		break;

	case NETDEV_GOING_DOWN:
		/* Get the IP address */
		in_dev = in_dev_get(dev);
		if (in_dev == NULL) {
			nx_print("Could not get the in_device for %s.\n",
				 dev->name);
			break;
		}

		ifa = in_dev->ifa_list;
		in_dev_put(in_dev);

		nx_print("ifconfig %s DOWN, port = %d.\n", dev->name, port);
		host_ifconfig(dev, ifa, port, NX_NETDEV_DOWN, 0);
		break;

	case NETDEV_REGISTER:
		if (dev->priv_flags & IFF_802_1Q_VLAN) {
			nx_print("Add VLAN device %s\n", dev->name);
                        tnic_adapter = nx_get_dev_tnic_adapter(dev);
                        if (tnic_adapter == NULL) {
                                nx_print("Could not get tnic_adapter for %s\n",
					 dev->name);
                                break;
                        }
			host_ifconfig(dev, NULL, port, NX_NETDEV_ADD, 0);
		}
		break;

	case NETDEV_UNREGISTER:
		nx_print("Inside NETDEV_UNREGISTER\n");
		tnic_adapter = nx_get_dev_tnic_adapter(dev);
		if (tnic_adapter == NULL) {
			nx_print("Could not get tnic_adapter for %s\n",
				 dev->name);
			break;
		}

		if (dev->priv_flags & IFF_802_1Q_VLAN) {
			nx_print("Remove VLAN device %s\n", dev->name);
			host_ifconfig(dev, NULL, port, NX_NETDEV_REMOVE, 0);
		} else {
			if (tnic_adapter->nx_port[port] != dev) {
				nx_print("Not the device that we saved\n");
			}
			tnic_adapter->nx_port[port] = NULL;
			nx_unregister_tnic(dev);
			if (!nx_is_dev_registered(tnic_adapter)) {
				nx_hbucket_del_init(&tnic_adapter->list,
						    &g_adapter_head);
				nx_tnic_adapter_free(tnic_adapter);
			}
		}
		break;

	case NETDEV_DOWN:
		/* nx_print("NETDEV_DOWN: %s\n", dev->name); */
		break;

	case NETDEV_REBOOT:
		nx_print("NETDEV_REBOOT: %s\n", dev->name);
		break;

	case NETDEV_CHANGE:
		/* nx_print("NETDEV_CHANGE: %s\n", dev->name); */
		break;

	case NETDEV_CHANGEMTU:
		nx_print("NETDEV_CHANGEMTU: %s\n", dev->name);
		break;

	case NETDEV_CHANGEADDR:
		nx_print("NETDEV_CHANGEADDR: %s\n", dev->name);
		break;

	case NETDEV_CHANGENAME:
		nx_print("NETDEV_CHANGENAME: %s\n", dev->name);
		break;

	default:
		nx_print("DEFAULT");
		break;
	}

	return (NOTIFY_DONE);
}




/*
 * Netlink notifier callback.
 */
static int nx_inetaddr_event(struct notifier_block *this, unsigned long event,
			     void *ptr)
{
	struct in_ifaddr	*ifa = (struct in_ifaddr *)ptr;
	struct net_device	*dev;
	int			port;
	unsigned int            aliasing = 1;
	nx_tnic_adapter_t       *tnic_adapter = NULL;

	if (strchr(ifa->ifa_label, ':') == NULL) {
		/*
		nx_print("Ip address change, device %s, address %x\n",
				ifa->ifa_label, ifa->ifa_address);
		 */
		aliasing = 0;
	}

	dev = ifa->ifa_dev ? ifa->ifa_dev->dev : NULL;
	if (dev == NULL) {
		nx_print("%s: Device NULL returning.\n", ifa->ifa_label);
		return (NOTIFY_DONE);
	}

	tnic_adapter = nx_get_dev_tnic_adapter(dev);
	if (NULL == tnic_adapter) {
		nx_print("Ignoring early callback event %lu on %s\n",
				event, ifa->ifa_label);
		return (NOTIFY_DONE);
	}

	port = nx_get_device_port(dev);
	if (port < 0) {
		nx_print("%s: Not NetXen card\n", dev->name);
		return (NOTIFY_DONE);
	}

	switch (event) {
	case NETDEV_UP:
		if (aliasing) {
			nx_print("ifconfig (alias address change) %s 0x%x UP\n",
					ifa->ifa_label, ifa->ifa_address);
		}
		else {
			nx_print("ifconfig (non-alias addr change) %s 0x%x UP\n",
					ifa->ifa_label, ifa->ifa_address);
		}

		host_ifconfig(dev, ifa, port, NX_NETDEV_UP, aliasing);
		break;

	case NETDEV_DOWN:
		if (!aliasing) {
			/* Ignored. If the device really goes down,
			   then netdev event will deal with it. */
			nx_print("Ignoring event %lu on %s\n",
					event, ifa->ifa_label);
			break;
		}

		nx_print("ifconfig %s DOWN\n", ifa->ifa_label);
		host_ifconfig(dev, ifa, port, NX_NETDEV_DOWN, aliasing);
		break;

	default:
		nx_print("DEFAULT");
		break;
	}

	return (NOTIFY_DONE);
}

void unm_register_netdev_notifiers(void)
{
	UNM_INFO(KERN_INFO"unm_register_netdev_notfiers\n");
	register_netdevice_notifier(&nx_netdev_cb);
	register_inetaddr_notifier(&nx_inetaddr_cb);
}

void unm_unregister_netdev_notifiers(void)
{
	UNM_INFO(KERN_INFO"unm_unregister_netdev_notfiers\n");
	unregister_netdevice_notifier(&nx_netdev_cb);
	unregister_inetaddr_notifier(&nx_inetaddr_cb);
}

/*
 *
 */
static inline int create_unm_socket(struct net_device *netdev,
				    struct socket *sock, int type, int zcopy)
{
        int	err = 0;

        /*
         * create pegnet_sock_id_t that will be stashed into
         * host_sk->SK_PROTINFO
         */
//         printk("%s: Entering sock=%p sk=%p\n", __FUNCTION__,sock,sock->sk);
        err = unm_create(netdev, sock, IPPROTO_TCP, type, zcopy);
        if (err) {
                printk(KERN_ERR "%s: Create failed:%d\n", __FUNCTION__, err);
        }
        return (err);
}

int unm_inet_dgrams_ioctl(struct socket *sock, unsigned int cmd,
                          unsigned long arg)
{
        int     ret;

        /* special case hack */
        if (cmd == SIOC_NETXEN_OFFLOAD) {
                return unm_port_manager(arg);
        }

        /*
         * call the native first regardless and then call ours only if it is
         * important.
         */
        ret = native_inet_dgrams_ioctl(sock, cmd, arg);

        switch (cmd) {
        case SIOCDARP:
        case SIOCGARP:
        case SIOCSARP:
                nx_print("ARP changes: cmd = 0x%x\n", cmd);
//                unm_hw_dgram_ioctl(sock, cmd, arg);
                //unm_hw_inet_dgram_ops.ioctl(WRAP_SOCK_AS_SOCKET(sock), cmd, arg);
                break;
        case SIOCSIFNETMASK:
        case SIOCSIFDSTADDR:
		/* nx_print("IF changes: cmd = 0x%x\n", cmd); */
//                unm_hw_dgram_ioctl(sock, cmd, arg);
                break;

        default:
		/* nx_print("ioctl: cmd = 0x%x\n", cmd); */
                break;
        }
        return ret;
}


/*
 * unm_sock_def_wakeup: gets unm_sk and access the native sk
 */
void
unm_sock_def_wakeup(struct sock *host_sk, int error)
{
        UNM_LOG2("UNM: unm_sock_def_wakeup: In: sock %p\n", host_sk);
        if (host_sk == 0) {
                printk("unm_sock_def_wakeup: host_sk is NULL\n");
        }

        host_sk->SK_ERR = error;
        host_sk->SK_STATE_CHANGE(host_sk);
        UNM_LOG2("UNM: unm_sock_def_wakeup: Out: sock %p\n", host_sk);
}

/*
 * Find the net device for local IP address.
 */
struct net_device *netxen_find_dev_for_ip(uint32_t addr)
{
/* #if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)) */
#if	1
	struct net_device	*dev;

	read_lock(&dev_base_lock);
	for (dev = dev_base; dev; dev = dev->next) {
		struct in_device	*indev = dev->ip_ptr;

		if (indev) {
			for_ifa(indev) {
				if (ifa->ifa_address == addr &&
				    netif_running(dev)) {
					dev_hold(dev);
					read_unlock(&dev_base_lock);
					return (dev);
				}
			} endfor_ifa(indev);
		}
	}
	read_unlock(&dev_base_lock);
	return (NULL);
#else
	return (ip_dev_find(addr));
#endif
}

/*
 *
 */
int unm_inet_release(struct socket *sock)
{
	int	ret = 0;

	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
                ret = unm_release(sock, 1);
                if(sock->sk == NULL) {
                        goto done;
                }
                if (sock->sk->SK_DESTRUCT) {
                        ret = host_inet_stream_ops.release(sock);
                } else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
                        inet_put_port(&tcp_hashinfo, sock->sk);
#else
                        tcp_put_port(sock->sk);
#endif
                        sk_free(sock->sk);
                        sock->sk = NULL;
                        UNM_HOST_PNIC_STATS_DEC(DBG_CNTR_INUSE_SK);
                }

	} else {
		ret = host_inet_stream_ops.release(sock);
	}
done:
	return (ret);
}

/*
 *   socket bind handler. 
 */
int unm_inet_bind(struct socket *sock, struct sockaddr *uaddr, int addr_len)
{
	return (host_inet_stream_ops.bind(sock, uaddr, addr_len));
}

static int nx_listen_init(listener_group_t *grp)
{
        int rv = 0;
        int i;

	/* initialize the syn-q hash table */
        grp->accept_passive.lopt = kmalloc(sizeof(nx_lopt_t), GFP_KERNEL);
        if (grp->accept_passive.lopt == NULL) {
                nx_print("group [%p] unable to allocate lopt\n", grp);
                rv = -ENOMEM;
                goto done;
        }

        for (i = 0; i < NX_SYNQ_HSIZE; i++) {
                INIT_LIST_HEAD(&grp->accept_passive.lopt->syn_table[i]);
        }
        grp->accept_passive.lopt->qlen = 0;
        INIT_LIST_HEAD(&grp->accept_passive.lopt->youngsters);
        get_random_bytes(&grp->accept_passive.lopt->hash_rnd, 4);

	/* initialize the accept queue */
        INIT_LIST_HEAD(&grp->accept_passive.accept_queue);

  done:
        return rv;

}
/*
 *
 */
int unm_inet_listen(struct socket *sock, int backlog)
{
        int             rv = 0;
        uint32_t        addr;
        unsigned short  port;
	int		zcopy = 0;

        rv = host_inet_stream_ops.listen(sock, backlog);
        if (rv) {
                return (rv);
        }

        if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
                rv = unm_listen(sock, backlog, 0);
                if (rv != 0) {
                        nx_print("Offload Listen Error: %d\n", rv);
                        unm_release(sock, 0);
                }
                goto done;
        }

        addr = SK_INET(sock->sk)->saddr;
        port = SPORT(sock->sk);

        if (port && nx_check_listen_offload(port, &zcopy)) {

                if (addr != INADDR_ANY) {
                        struct net_device	*dev;
                        struct net_device	*real_dev = NULL;

                        dev = netxen_find_dev_for_ip(addr);
                        if (dev) {
                                if (is_netxen_device(dev)) {
					real_dev = nx_get_realdev(dev);
				}
                                dev_put(dev);
                        }

                        if (real_dev == NULL) {
                                goto done;
                        }

                        rv = create_unm_socket(real_dev, sock, LISTENER_SOCK,
					       zcopy);
                        if (rv != 0) {
                                nx_print("create_unm_socket FAILED: err:%d, "
					 "not offloading listener\n", rv);
                        }
                } else {
                        struct list_head        *node;
                        nx_tnic_adapter_t       *tnic;

                        spin_lock_bh(&g_adapter_head.lock);

                        if (list_empty(&g_adapter_head.head)) {
                                nx_print("No Netxen Device is Registered\n");
                                spin_unlock_bh(&g_adapter_head.lock);
                                goto done;
                        }
                        spin_unlock_bh(&g_adapter_head.lock);

                        list_for_each(node, &g_adapter_head.head) {
                                tnic = list_entry(node, nx_tnic_adapter_t,
						  list);

                                rv = create_unm_socket(tnic->nx_dev, sock,
						       LISTENER_SOCK, zcopy);
                                if (rv != 0) {
                                        nx_print("create_unm_socket FAILED: "
						 "err:%d, not offloading "
						 "listener\n", rv);
                                }
                        }
                }

                if (UNM_SOCK_IS_SOCK_OFFLOADED(sock) == 0) {
                        nx_print("Create failed on all the cards \n");
                        goto done;
                }

                if ((rv = unm_copy_sockopts(sock)) != 0) {
                        unm_release(sock, 0);
                        goto done;
                }

                rv = nx_listen_init(NX_GET_LISTENER_GROUP_FROM_SOCK(sock));
                if (rv != 0) {
                        nx_print("Listen initialization failed rv=%d\n", rv);
                        unm_release(sock, 0);
                        goto done;
                }

                rv = unm_listen(sock, backlog, 1);
                if (rv != 0) {
                        unm_release(sock, 0);
                }
        }
  done:
        return (0);
}

/*
 *
 */
static inline int nx_ip_route_connect(struct rtable **rp, uint32_t dst,
				      uint32_t src, uint32_t tos, int oif,
				      uint8_t protocol, uint16_t sport,
				      uint16_t dport)
{
        struct flowi	fl = { .oif = oif,
			       .nl_u = { .ip4_u = { .daddr = dst,
						    .saddr = src,
						    .tos   = tos } },
			       .proto = protocol,
			       .uli_u = { .ports =
					  { .sport = sport,
					    .dport = dport } } };
        int		err;

        if (!dst || !src) {
                err = ip_route_output_key(rp, &fl);
                if (err) {
                        return err;
		}
                fl.fl4_dst = (*rp)->rt_dst;
                fl.fl4_src = (*rp)->rt_src;
                ip_rt_put(*rp);
                *rp = NULL;
        }
        return ip_route_output_key(rp, &fl);
}

/*
 *
 */
static int nx_connect_try_offload(struct socket *sock, struct sockaddr *uaddr,
				  int addr_len, int flags, int zcopy)
{
	struct sockaddr_in	*usin = (struct sockaddr_in *)uaddr;
	struct rtable		*rt;
	uint32_t		daddr;
	uint32_t		saddr;
	uint16_t		dport;
	uint16_t		sport;
	int			err;
	struct sock		*sk = sock->sk;
	uint16_t		local_port;
	int			ifport;
	nx_tnic_adapter_t 	*tnic_adapter;

	if (addr_len < sizeof(struct sockaddr_in)) {
		return (1);
	}

	if (usin->sin_family != AF_INET) {
		return (1);
	}

	if (usin->sin_addr.s_addr == 0 || usin->sin_port == 0) {
		return (1);
	}

	if (SK_INET(sk)->opt) {
		return (1);
	}

	daddr = usin->sin_addr.s_addr;
	dport = usin->sin_port;
	saddr = SK_INET(sk)->saddr;
	sport = htons(SPORT(sock->sk));
	err = nx_ip_route_connect(&rt, daddr, saddr, RT_CONN_FLAGS(sk),
				  sk->sk_bound_dev_if, IPPROTO_TCP,
				  sport, dport);
	if (err < 0) {
		return (1);
	}

	if (rt->rt_flags & (RTCF_MULTICAST | RTCF_BROADCAST)) {
		ip_rt_put(rt);
		return (1);
	}

	if (!saddr) {
		saddr = rt->rt_src;
	}
	if (!saddr) {
		nx_print("Source address is zero\n");
	}

	ip_rt_put(rt);

	if (nx_route_connection(&rt, saddr, sport, daddr, dport,
				RT_CONN_FLAGS(sk), sk->sk_bound_dev_if)) {
		return (1);
	}

	ifport = nx_get_device_port(NX_KC_ROUTE_DEVICE(rt));
	if (ifport < 0) {
		nx_print("Not netxen device\n");
		err = 1;
		goto free_route;
	}

	err = create_unm_socket(NX_KC_ROUTE_DEVICE(rt), sock, CONNECTION_SOCK,
				zcopy);
	if (err != 0) {
		nx_print("Error[%d]: Using native stack\n", err);
		err = 1;
		goto free_route;
	}
	
	err = unm_copy_sockopts(sock);
	if (err) {
		nx_print("Error[%d]: sockopts failed\n", err);
		err = 1;
		goto free_sock;
	}

	local_port = sport;
	if (local_port == 0) {
		if ((!BIND_HASH(sock->sk)) &&
		    sock->sk->SK_PROT->get_port(sock->sk, 0) != 0) {

			nx_print("get_port error for ephemeral port\n");
			err = -EADDRINUSE;
			goto free_sock;
		} else {
			sport = htons(SPORT(sock->sk));
		}
	}
	tnic_adapter = nx_get_dev_tnic_adapter(NX_KC_ROUTE_DEVICE(rt));

	/*
	 * Add the route if necessary.
	 */
	err = nx_route_add(tnic_adapter, rt, ifport);
	if (err) {
		nx_print("Error[%d]: Route add failed\n", err);
		err = 1;
		goto free_port;
	}
	nx_set_route_reference(UNM_SOCK_GET_PEGNET_SOCK_ID(sock), saddr,
			       sport, daddr, dport);

	/*
	 * Try to connect.
	 */
	err = unm_connect(sock, uaddr, addr_len, flags, 1);

	if (err != 0 && err != -EINPROGRESS && err != -EALREADY &&
	    err != -EISCONN) {
		/* Synchronous connect call failed ... */
		/* or, the failure is not because of 'good' errors */
		nx_print("Connect FAILED err:%d\n", err);
		goto free_port;
	}

	goto free_route;

  free_port:
	if (local_port == 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
		inet_put_port(&tcp_hashinfo, sock->sk);
#else
		tcp_put_port(sock->sk);
#endif
	}

  free_sock:
	unm_release(sock, 0);

  free_route:
	ip_rt_put(rt);
	return (err);
}

/*
 *
 */
int unm_inet_connect(struct socket *sock, struct sockaddr *uaddr,
		     int addr_len, int flags)
{
	int			err;
	struct sockaddr_in	*s = (struct sockaddr_in *)uaddr;
        unsigned short		local_port = SPORT(sock->sk);
	int			zcopy = 0;


	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		/*
		 * This can happen only if LISTEN or CONNECT had been called
		 * before.
		 */
		err = unm_connect(sock, uaddr, addr_len, flags, 0);

		if (err != 0 && err != -EINPROGRESS && err != -EALREADY &&
                    err != -EISCONN) {
                        /* Synchronous connect call failed ... */
                        /* or, the failure is not because of 'good' errors */
                        nx_print("Connect FAILED err:%d\n", err);
                        if (local_port == 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
                                inet_put_port(&tcp_hashinfo, sock->sk);
#else
                                tcp_put_port(sock->sk);
#endif
                                unm_release(sock, 0);
                        }
                }

		return (err);
	}

	/*
	 * Only offload if the connection is to a non local unicast address.
	 */
	if (s->sin_addr.s_addr != INADDR_ANY &&
            inet_addr_type(s->sin_addr.s_addr) == RTN_UNICAST &&
	    (sock->sk && sock->sk->SK_FAMILY == AF_INET) &&
	    nx_check_connect_offload(sock->sk->SK_FAMILY, NULL, local_port,
				     NULL, ntohs(s->sin_port), &zcopy)) {

		nx_print1("Dest IP: %d.%d.%d.%d Port %d on UNM Stack\n",
			 (int)((ip >> 24) & 0xFF),
			 (int)((ip >> 16) & 0xFF),
			 (int)((ip >> 8) & 0xFF),
			 (int)(ip & 0xFF),
			 (int)local_port);

		err = nx_connect_try_offload(sock, uaddr, addr_len, flags,
					     zcopy);
		if (err == 1) {
			goto native;
		}

	} else {
  native:
		err = host_inet_stream_ops.connect(sock, uaddr,
						   addr_len, flags);
	}

	UNM_LOG3("unm_inet_connect: out: err 0x%x(%d)\n", err, err);

	return (err);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
int unm_inet_recvmsg(struct kiocb *iocb, struct socket *sock,
		     struct msghdr *msg, size_t size, int flags)
#else
int unm_inet_recvmsg(struct socket *sock, struct msghdr *msg, size_t size,
		     int flags, struct scm_cookie *scm)
#endif
{
	int	err;

	UNM_LOG3("unm_inet_recvmsg: In: sock %p size %d\n", sock->sk, size);

	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {

                if (sock->sk->SK_ERR) {
                        err = sock_error(sock->sk);
                        printk("%s:%d: sock err %d\n", __FUNCTION__, __LINE__, err);
                        return (err);
                }
                err = unm_recvmsg(sock, msg, size, flags);

	} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		err = host_inet_stream_ops.recvmsg(iocb, sock, msg, size,
						   flags);
#else
		err = host_inet_stream_ops.recvmsg(sock, msg, size, flags,
						   scm);
#endif
	}

	UNM_LOG3("unm_inet_recvmsg: out: sock %p size 0x%x\n", sock->sk, err);

	return (err);
}

/*
 * Checks if this socket is offloaded or not and calls the appropriate sendmsg
 * function.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
int unm_inet_sendmsg(struct kiocb *iocb, struct socket *sock,
		     struct msghdr *msg, size_t size)
#else
int unm_inet_sendmsg(struct socket *sock, struct msghdr *msg, size_t size,
		     struct scm_cookie *scm)
#endif
{
	int	err;

	UNM_LOG3("unm_inet_sendmsg: In: size %d sock %p\n", size, sock->sk);

	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		err = unm_sendmsg(sock, msg, size);

                if (err < 0) {

                        if (err == -EPIPE) {
                                err = sock_error(sock->sk) ? : -EPIPE;
                        }
                        if (err == -EPIPE &&
                            !(msg->msg_flags & MSG_NOSIGNAL)) {
                                send_sig(SIGPIPE, current, 0);
                        }
                }

		//      if (err == 0 ) {
		UNM_LOG3("inet_send: bytes sent %d size %d\n", err, size);
		UNM_LOG4("iovlen %u iovec %p iov_len %u\n",
			 msg->msg_iovlen, msg->msg_iov, msg->msg_iov->iov_len);
		//      }

	} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		err = host_inet_stream_ops.sendmsg(iocb, sock, msg, size);
#else
		err = host_inet_stream_ops.sendmsg(sock, msg, size, scm);
#endif
	}

	UNM_LOG3("unm_inet_sendmsg: out: sock %p: bytes sent %d\n",
		 sock->sk, err);
	return (err);
}

/*
 *
 */
static int unm_inet_accept_internal(struct socket *sock,
                                    struct socket *newsock, int flags)
{
	int                     err;
	struct sock             *sk;

	/* Allocate & fill in host side sk */
	sk = SK_ALLOC(PF_INET, GFP_KERNEL, sock->sk->SK_PROT,2);
	if (sk == NULL) {
		return (-ENOBUFS);
	}

	sk->SK_FLAGS = ~0;
        sock_init_data(newsock, sk);
        sk->SK_DESTRUCT = NULL;         /* Has to be NULL else Linux stack
                                           can free up the stashed data in
                                           sk_protinfo. We also check this
                                           in our code. */

        /*
         * TODO: Socket options that were set on the parent will have to be
         * copied here.
         */

        err = unm_accept(sock, newsock, flags);
	if (err) {
                newsock->sk = NULL;
		sk->SK_PROTINFO = NULL;
                sk_free(sk);
		return (err);
	}

        sk->SK_FLAGS = sock->sk->SK_FLAGS;
        sk->SK_RCVTIMEO = sock->sk->SK_RCVTIMEO;
        sk->SK_SNDTIMEO = sock->sk->SK_SNDTIMEO;
        sk->SK_LINGERTIME = sock->sk->SK_LINGERTIME;
        sk->SK_PROT = sock->sk->SK_PROT;
        SPORT(sk) = SPORT(sock->sk);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
        inet_inherit_port(&tcp_hashinfo, sock->sk, sk);
#else
        tcp_inherit_port(sock->sk, sk);
#endif
        sk->SK_REUSE = sock->sk->SK_REUSE;
	sk->SK_RCVBUF = sock->sk->SK_RCVBUF;
	(UNM_SOCK_GET_PEGNET_SOCK_ID(newsock))->reported_rcvbuf =
		(UNM_SOCK_GET_PEGNET_SOCK_ID(sock))->reported_rcvbuf;

        lock_sock(sk);
        newsock->state = SS_CONNECTED;
        newsock->sk->SK_STATE = TCP_ESTABLISHED;
        release_sock(sk);

        UNM_HOST_PNIC_STATS_INC(DBG_CNTR_INUSE_SK);

	return (err);
}

/*
 *
 */
static int unm_check_accept(struct socket *sock, struct socket *newsock)
{
	int     rv;

	if (unm_host_get_next_accept_stack(sock)) {

		rv = host_inet_stream_ops.accept(sock, newsock, O_NONBLOCK);
		if (newsock->sk) {
			newsock->sk->SK_PROTINFO = NULL;
		}
		if (rv == 0) {
			/* UNM_INFO("unm_check_accept: NATIVE stack\n"); */
			return (rv);
		}

		if (!unm_host_is_accept_queue_empty(sock)) {
			rv = unm_inet_accept_internal(sock, newsock,
                                                      O_NONBLOCK);
/* 			if (rv == 0) { */
/* 				UNM_INFO("unm_check_accept: Offload stack\n"); */
/* 			} */
		}

	} else {

		if (!unm_host_is_accept_queue_empty(sock)) {
			rv = unm_inet_accept_internal(sock, newsock,
						      O_NONBLOCK);
			if (rv == 0) {
				/* UNM_INFO("unm_check_accept: Offload stack\n"); */
				return (rv);
			}
		}

		rv = host_inet_stream_ops.accept(sock, newsock, O_NONBLOCK);
		if (newsock->sk) {
			newsock->sk->SK_PROTINFO = NULL;
		}
/* 		if (rv == 0) { */
/* 			UNM_INFO("unm_check_accept: NATIVE stack\n"); */
/* 		} */
	}
	return (rv);
}

/*
 *
 */
int unm_inet_accept(struct socket *sock, struct socket *newsock, int flags)
{
	int		ret;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	DEFINE_WAIT(wait);
#else
        DECLARE_WAITQUEUE(wait, current);
#endif
	long		timeo;

	UNM_LOG("unm_inet_accept: In\n");

	if (sock->sk && sock->sk->SK_FAMILY == AF_INET &&
	    sock->sk->SK_PROTINFO && sock->sk->SK_STATE == TCP_LISTEN) {

		lock_sock(sock->sk);

		/* sock->sk->SK_PROTINFO = NULL; */
		/*
		 *
		 */
		/* timeo = (flags & O_NONBLOCK) ? 0 : sock->sk->SK_RCVTIMEO; */
		timeo = sock_rcvtimeo(sock->sk, flags & O_NONBLOCK);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
                add_wait_queue_exclusive(sock->sk->SK_SLEEP, &wait);
#endif

		for (;;) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			prepare_to_wait_exclusive(sock->sk->SK_SLEEP, &wait,
						  TASK_INTERRUPTIBLE);
#else
                        current->state = TASK_INTERRUPTIBLE;
#endif
			release_sock(sock->sk);

			ret = unm_check_accept(sock, newsock);
			if (ret == 0 || ret != -EAGAIN) {
				break;
			}

			timeo = schedule_timeout(timeo);

			ret = ((timeo == MAX_SCHEDULE_TIMEOUT) ?
			       -ERESTARTSYS : -EINTR);
			if (signal_pending(current)) {
				break;
			}
			ret = -EAGAIN;
			if (!timeo) {
				break;
			}

			lock_sock(sock->sk);
		}

		lock_sock(sock->sk);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		finish_wait(sock->sk->SK_SLEEP, &wait);
#else
                current->state = TASK_RUNNING;
                remove_wait_queue(sock->sk->SK_SLEEP, &wait);
#endif
		/* sock->sk->SK_PROTINFO = offload_sock; */

		release_sock(sock->sk);

	} else {
		UNM_LOG("unm_inet_accept: no offload\n");
		ret = host_inet_stream_ops.accept(sock, newsock, flags);
		if (newsock->sk) {
			newsock->sk->SK_PROTINFO = NULL;
		}
	}

	UNM_LOG2("unm_inet_accept: out: ret %d \n", ret);

	return (ret);
}

/*
 *
 */
static unsigned int unm_tcp_poll_hw(struct file *file, struct socket *sock,
                                    poll_table *wait)
{
        if (wait) {
                poll_wait(file, sock->sk->SK_SLEEP, wait);
        }
        return (unm_poll(sock));
}

/*
 * Need to check both the host as well as the offloaded stack for listening
 * sockets.
 */
unsigned int unm_tcp_poll(struct file *file, struct socket *sock,
                          poll_table *wait)
{
        unsigned int    ret;

        //UNM_INFO("unm_tcp_poll: In: sock = %p\n", sock);

        if (!UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
                return (host_inet_stream_ops.poll(file, sock, wait));
        }

        if (sock->sk->SK_STATE != TCP_LISTEN) {
                // UNM_INFO("unm_tcp_poll: Not a listen socket: %p\n", sock);
                return (unm_tcp_poll_hw(file, sock, wait));
        }

        // UNM_INFO("unm_tcp_poll: A listen socket: %p\n", sock);
        /*
         * First poll the card first but pass NULL for wait parameter. Then
         * call host with the wait parameter specified. The host stack will
         * add the wait to the sleep queue.
         */
        ret = unm_tcp_poll_hw(file, sock, NULL);

        /*
         * We have already found something in offload queue so don't wait on
         * the host queue.
         */
        if (ret & (POLLIN | POLLRDNORM)) {
                wait = NULL;
        }
        // UNM_INFO("unm_tcp_poll: Flags from card = %u\n", ret);

        ret |= host_inet_stream_ops.poll(file, sock, wait);

        // UNM_INFO("unm_tcp_poll: Flags as a whole = %u\n", ret);

        //UNM_LOG("unm_tcp_poll: out\n");
        return (ret);
}

/*
 *
 */
int unm_inet_getname(struct socket *sock, struct sockaddr *uaddr,
		     int *uaddr_len, int peer)
{
	int	err;

	UNM_LOG("unm_getname: In\n");
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		err = unm_getname(sock, uaddr, uaddr_len, peer);
	} else {
		err = host_inet_stream_ops.getname(sock, uaddr, uaddr_len,
						   peer);
	}
	UNM_LOG("unm_getname: out\n");

	return (err);
}

/*
 *
 */
int unm_sock_no_socketpair(struct socket *sock1, struct socket *sock2)
{
        int     err;

        UNM_LOG("unm_socketpair: In\n");
        if (UNM_SOCK_IS_SOCK_OFFLOADED(sock1)) {
                err = sock_no_socketpair(sock1, sock2);
        } else {
                err = host_inet_stream_ops.socketpair(sock1, sock2);
        }
        UNM_LOG("unm_socketpair: out\n");

        return err;
}

// ioctl vals form tcp(7)
#define IS_TCP_IOCTL(val) (val == SIOCINQ || val == SIOCATMARK || val == SIOCOUTQ)
/*
 *
 */
int unm_inet_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
        int     ret=0;
        int     tcp_ioctl=0;
        pegnet_sock_id_t *sid;


        UNM_LOG("unm_inet_ioctl: In\n");
        tcp_ioctl=IS_TCP_IOCTL(cmd);
        // socket/tcp specific ioctls on native stack should be called only if 
        // sock in not-offloaded. 
        // Other ioctls should always be invoked on native stack.
        if(tcp_ioctl == 0 || !UNM_SOCK_IS_SOCK_OFFLOADED(sock)|| 
		sock->sk->SK_STATE == TCP_LISTEN) {
            ret = host_inet_stream_ops.ioctl(sock, cmd, arg);
        }
        if (!ret && UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		if(sock->sk->SK_STATE == TCP_LISTEN){
        		struct list_head *node,*_node;
        		listener_group_t *grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
        		list_for_each_safe(node, _node,&grp->listener_list) {
                		sid = list_entry(node, pegnet_sock_id_t, listener_list);
                		if(0 != (ret = unm_ioctl(sid, cmd, arg))) {
                			nx_print("Ioctl Error for Listener grp = %p sid:%p "
					"tnic_adapter=%p dev_port=%d cmd=%d err=%d\n",
					grp,sid,sid->adapter,sid->portnum,cmd,ret);
//					nx_release_sid(sid);
				}
			}
 			if (!list_empty(&(grp->listener_list))) {
				ret = 0;
 			}
		}
		else {
			sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
                	ret = unm_ioctl(sid, cmd, arg);
		}	
        } 
        UNM_LOG("unm_inet_ioctl: out\n");
        return ret;
}

/*
 *
 */
int unm_inet_shutdown(struct socket *sock, int how)
{
        int     ret;

        UNM_LOG("unm_inet_shutdown: In\n");
        if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
                ret = unm_shutdown(sock, how);
                sock->sk->SK_STATE_CHANGE(sock->sk);
        } else {
                ret = host_inet_stream_ops.shutdown(sock, how);
        }
        UNM_LOG("unm_inet_shutdown: out\n");
        return ret;
}
int unm_copy_sockopts(struct socket *sock)
{
     return 0;
#if 0
     struct sock *sk;
     sk = sock->sk;
     int flag;
     int ret;

 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
     struct tcp_opt *tp = &(sk->tp_pinfo.af_tcp);
 #elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
     struct tcp_opt *tp = tcp_sk(sk);
 #else
     struct tcp_sock *tp = tcp_sk(sk);
 #endif

     /* TCP_NODELAY */
     flag = tp->nonagle;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_NODELAY,(char *) &flag,
                 sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

#endif
     /*
     // TCP_KEEPIDLE
     flag = tp->keepalive_time;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_KEEPIDLE,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_KEEPINTVL 
     flag = tp->keepalive_intvl;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_KEEPINTVL,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_KEEPCNT
     flag = tp->keepalive_probes;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_KEEPCNT,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_SYNCNT 
     flag = tp->syn_retries;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_SYNCNT,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_LINGER2
     flag = tp->linger2;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_LINGER2,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_DEFER_ACCEPT 
     flag = tp->defer_accept;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_DEFER_ACCEPT,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_WINDOW_CLAMP 
     flag = tp->window_clamp;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_WINDOW_CLAMP,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret;

     // TCP_QUICKACK
     flag = tp->ack.pingpong;
     ret = unm_inet_setsockopt(sock,IPPROTO_TCP,TCP_QUICKACK,(char *) &flag,
             sizeof(flag));
     if(ret!=0)
         ret = -1;
     return ret; */

}


/*
 *
 */
int unm_inet_setsockopt(struct socket *sock, int level, int optname,
                        char *optval, int optlen)
{
        int     ret = 0;
        pegnet_sock_id_t *sid;

        UNM_LOG("unm_setsockopt: In\n");
	if(sock->sk->SK_STATE == TCP_LISTEN){
                ret = host_inet_stream_ops.setsockopt(sock, level, optname,
                                              optval, optlen);
		if (!ret && UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
        		struct list_head *node,*_node;
        		listener_group_t *grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
        		list_for_each_safe(node, _node,&grp->listener_list) {
                		sid = list_entry(node, pegnet_sock_id_t, listener_list);
                		if(0 != (ret = unm_setsockopt(sid, level, optname, optval, optlen))) {
                			nx_print("Set Sockopt Error for Listener grp = %p sid:%p "
					"tnic_adapter=%p dev_port=%d optname=%d err=%d\n",
					grp,sid,sid->adapter,sid->portnum,optname,ret);
//					nx_release_sid(sid);
				}
			}
 			if (!list_empty(&(grp->listener_list))) {
				ret = 0;
 			}
		}
	}
        else {
		if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
			sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
                	ret = unm_setsockopt(sid, level, optname, optval, optlen);
        	} else {
                	ret = host_inet_stream_ops.setsockopt(sock, level, optname,
                                                      optval, optlen);
        	}
	}	
        UNM_LOG("unm_setsockopt: out\n");
        return ret;
}

/*
 *
 */
int unm_inet_getsockopt(struct socket *sock, int level, int optname,
                        char *optval, int *optlen)
{
        int     ret = 0;
        pegnet_sock_id_t *sid;

        UNM_LOG("unm_getsockopt: In\n");
	if(sock->sk->SK_STATE == TCP_LISTEN){
                ret = host_inet_stream_ops.getsockopt(sock, level, optname,
                                               optval, optlen);
		if (!ret && UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
        		struct list_head *node,*_node;
        		listener_group_t *grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
        		list_for_each_safe(node, _node,&grp->listener_list) {
                		sid = list_entry(node, pegnet_sock_id_t, listener_list);
                		if(0 != (ret = unm_getsockopt(sid, level, optname, optval, optlen))) {
                			nx_print("Get Sockopt Error for Listener grp = %p sid:%p "
					"tnic_adapter=%p dev_port=%d optname=%d err=%d\n",
					grp,sid,sid->adapter,sid->portnum,optname,ret);
//					nx_release_sid(sid);
				}
				else {
					break;
				}
			}
		}
	}
        else {
		if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
			sid = UNM_SOCK_GET_PEGNET_SOCK_ID(sock);
                	ret = unm_getsockopt(sid, level, optname, optval, optlen);
        	} else {
                	ret = host_inet_stream_ops.getsockopt(sock, level, optname,
                                                      optval, optlen);
       		 }
	}		 
        UNM_LOG("unm_getsockopt: out\n");
        return ret;
}

int
unm_sock_no_mmap(struct file *file, struct socket *sock, 
                 struct vm_area_struct *vma)
{
        int     ret;
#ifdef notdef
        UNM_LOG("unm_mmap: In\n");
        if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
                ret = sock_no_mmap(file, sock, vma);
        } else {
#endif
                ret = host_inet_stream_ops.mmap(file, sock, vma);
#ifdef notdef
        }
#endif
        UNM_LOG("unm_mmap: out\n");
        return ret;
}

/*
 *
 */
ssize_t unm_tcp_sendpage(struct socket *sock, struct page *page, int offset, 
                         size_t size, int flags)
{
        ssize_t ret;

        UNM_LOG("unm_sendpage: In\n");
        if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
                /*
                 * For now let us call net core's sock_no_sendpage
                 */
//                return sock_no_sendpage(sock, page, offset, size, flags);
		ret=unm_sendpage(sock,&page,offset,size,flags);

        } else {
                ret = host_inet_stream_ops.sendpage(sock, page, offset, size,
                                                    flags);
        }
        UNM_LOG("unm_sendpage: out\n");
        return ret;
}
