/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Main module
 */
#include <linux/kernel.h> // Needed for KERN_ALERT
#include <linux/version.h>
#include <asm/pgtable.h>  // For PAGE_KERNEL_EXEC
#include "unm_host_pnic_types.h"
#include "pstack.h"


#ifdef  UNM_CONFIG_COMPAT
#include <linux/ioctl32.h>
#define UNM_CONFIG_COMPAT
#endif

struct proto_ops		host_inet_stream_ops;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
kmem_cache_t *nx_sk_cachep;
#endif

extern nx_tnic_callback_t	g_tnic_callback;

extern int nx_handle_offloaded_syn(struct net_device *netdev,
				   uint16_t listener_id,
				   struct sk_buff *skb);
extern void unm_process_pegnet_msg(struct net_device *netdev,
				   unm_msg_t *msg);

int (*native_inet_dgrams_ioctl)(struct socket *sock, unsigned int cmd,
				unsigned long arg);
nx_hbucket_head_t g_adapter_head;
atomic_t        g_devid;
#if 0
static struct nf_hook_ops unm_nf_ipv4_input_filter = 
{ { NULL, NULL }, unm_nf_ipv4_input_handler, PF_INET, NF_IP_LOCAL_IN, NF_IP_PRI_LAST-1 };

static struct nf_hook_ops unm_nf_arp_input_filter = 
{ { NULL, NULL }, unm_nf_arp_input_handler, NF_ARP, NF_ARP_IN, 0 };


static struct nf_hook_ops unm_nf_ipv4_output_filter = 
{ { NULL, NULL }, unm_nf_ipv4_output_handler, PF_INET, NF_IP_LOCAL_OUT, NF_IP_PRI_LAST-1 };

unsigned int 
unm_nf_arp_input_handler(unsigned int hooknum,
                         struct sk_buff **pskb,
                         const struct net_device *in,
                         const struct net_device *out,
                         int (*okfn)(struct sk_buff *))

{
   struct sk_buff *sk1;

   UNM_LOG1("unm_nf_arp_input_handler\n");

   sk1 = skb_clone(*pskb, GFP_ATOMIC);
   arp_process(sk1);

   return NF_ACCEPT;
}

unsigned int 
unm_nf_ipv4_input_handler(unsigned int hooknum,
                          struct sk_buff **pskb,
                          const struct net_device *in,
                          const struct net_device *out,
                          int (*okfn)(struct sk_buff *))
{
   // return NF_DROP if it is ours or return NF_ACCEPT
   struct ethhdr *eth;
   struct iphdr *ip = NULL;
   struct tcphdr *tcp = NULL;
   struct sock *sk;
   struct sk_buff *sk1;

   eth = (*pskb)->mac.ethernet;
   ip = (*pskb)->nh.iph;

   // return non TCP by default to native stack so ping, etc can work.

   if (ip->protocol != IPPROTO_TCP) {
      return NF_ACCEPT;
   }

   tcp = (struct tcphdr *)((__u32 *)ip + ip->ihl);

   sk = __tcp_v4_lookup(ip->saddr, tcp->source,
                        ip->daddr,
                        ntohs(tcp->dest),
                        tcp_v4_iif(*pskb));
#ifdef UNM_ALLOW_NATIVE
   if (!sk) {
      return NF_ACCEPT;
   } 
#endif

  sock_put(sk);

   UNM_LOG1("nfip-in: sk %p destPort %u srcPort %u seq %u ack_seq %u syn 0x%x ack 0x%x psh 0x%x rst 0x%x fin 0x%x\n", 
            sk,
            ntohs(tcp->dest), ntohs(tcp->source), ntohl(tcp->seq), ntohl(tcp->ack_seq),
            tcp->syn, tcp->ack, tcp->psh, tcp->rst, tcp->fin);

   sk1 = skb_clone(*pskb, GFP_ATOMIC);

   ip_local_deliver_finish(sk1);

   return  NF_DROP;
}

unsigned int 
unm_nf_ipv4_output_handler(unsigned int hooknum,
                           struct sk_buff **pskb,
                           const struct net_device *in,
                           const struct net_device *out,
                           int (*okfn)(struct sk_buff *))
{
   // return NF_DROP if it is ours or return NF_ACCEPT
   struct ethhdr *eth;
   struct iphdr *ip = NULL;
   struct tcphdr *tcp = NULL;
   struct sock *sk;
   struct sk_buff *sk1;

   eth = (*pskb)->mac.ethernet;
   ip = (*pskb)->nh.iph;

#ifndef UNM_ALLOW_NATIVE
   if (ip->protocol == IPPROTO_TCP) {
      tcp = (struct tcphdr *)((__u32 *)ip + ip->ihl);
      UNM_LOG1("nfip-out: destPort %d srcPort %d syn 0x%x ack 0x%x psh 0x%x rst 0x%x\n", 
               ntohs(tcp->dest), ntohs(tcp->source), 
               tcp->syn, tcp->ack, tcp->psh, tcp->rst);
      return  NF_DROP;
   }
#endif

   return  NF_ACCEPT;
}


static void
unm_nf_init(void)
{
   nf_register_hook(&unm_nf_ipv4_output_filter);
   nf_register_hook(&unm_nf_ipv4_input_filter);
   nf_register_hook(&unm_nf_arp_input_filter);
}

#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
static int
change_perm(unsigned long ptr, unsigned int size, pgprot_t prot)
{
        unsigned long endptr = (ptr + size + PAGE_SIZE - 1) & PAGE_MASK;
        size = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;
        for (ptr &= PAGE_MASK; ptr < endptr; ptr += PAGE_SIZE) {
               if (change_page_attr(virt_to_page(ptr), 1, prot) < 0) {
                       return -1;
               }
        }
        global_flush_tlb();
        return 0;
}
#endif

struct proto_ops nx_inet_proto_ops = {
	release: unm_inet_release,
	socketpair: unm_sock_no_socketpair,
	accept: unm_inet_accept,
	getname: unm_inet_getname,
	poll: unm_tcp_poll,
	ioctl: unm_inet_ioctl,
	shutdown: unm_inet_shutdown,
	sendmsg: unm_inet_sendmsg,
	recvmsg: unm_inet_recvmsg,
	setsockopt: unm_inet_setsockopt,
	getsockopt: unm_inet_getsockopt,
	sendpage: unm_tcp_sendpage,
	listen: unm_inet_listen,
	connect: unm_inet_connect
};

extern int nx_tnic_register_ops(struct proto_ops*, struct proto_ops*);
extern void nx_tnic_unregister_ops(void);

int pstack_init(void)
{
        UNM_INFO(KERN_ALERT "pstack_module_init: \n");

#if 0
        // register netfilter inbound
        unm_nf_init();

        // initialize our stack

        inet_init();
#endif


        // Patch native inet_stream_ops

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        if (change_perm((unsigned long)&inet_stream_ops,
                         sizeof(inet_stream_ops),
                         PAGE_KERNEL_FLAG) < 0) {
                return -EIO;
        }
        if (change_perm((unsigned long)&inet_dgram_ops,
                         sizeof(inet_dgram_ops),
                         PAGE_KERNEL_FLAG) < 0) {
                change_perm((unsigned long)&inet_stream_ops,
                            sizeof(inet_stream_ops),
                            PAGE_KERNEL_RO);
                return -EIO;
        }
#endif

#if 0
        memcpy(&host_inet_stream_ops, &inet_stream_ops,
               sizeof(inet_stream_ops));

        inet_stream_ops.release = unm_inet_release;
        inet_stream_ops.bind = unm_inet_bind;
        inet_stream_ops.connect = unm_inet_connect;
        inet_stream_ops.socketpair = unm_sock_no_socketpair;
        inet_stream_ops.accept = unm_inet_accept;
        inet_stream_ops.getname = unm_inet_getname;
        inet_stream_ops.poll = unm_tcp_poll;
        inet_stream_ops.ioctl = unm_inet_ioctl;
        inet_stream_ops.listen = unm_inet_listen;
        inet_stream_ops.shutdown = unm_inet_shutdown;
        inet_stream_ops.setsockopt = unm_inet_setsockopt;
        inet_stream_ops.getsockopt = unm_inet_getsockopt;
        inet_stream_ops.sendmsg = unm_inet_sendmsg;
        inet_stream_ops.recvmsg = unm_inet_recvmsg;
        inet_stream_ops.mmap = sock_no_mmap;
        inet_stream_ops.sendpage = unm_tcp_sendpage;
#endif
	if (nx_tnic_register_ops(&nx_inet_proto_ops, &host_inet_stream_ops))
		return -EIO;

        // patch up native's inet_dgram_ops.ioctl to point to ours
        native_inet_dgrams_ioctl = inet_dgram_ops.ioctl;
        inet_dgram_ops.ioctl = unm_inet_dgrams_ioctl;
	

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
        nx_sk_cachep = kmem_cache_create("nx_sock", sizeof(struct tcp_sock), 0,
                                         SLAB_HWCACHE_ALIGN, 0, 0);
        if (!nx_sk_cachep) {
            printk(KERN_ERR "%s: Can't create protocol sock SLAB cache!\n",
                    __FUNCTION__);
        }
#endif

#ifdef  UNM_CONFIG_COMPAT
        /* Our ioctls are all "32/64bit compatible" */
        if (register_ioctl32_conversion(SIOC_NETXEN_OFFLOAD, NULL)) {
                printk(KERN_ERR "%s: Error registering ioctl32 translation.\n",
                       __FUNCTION__);
        }
#endif  /* UNM_CONFIG_COMPAT */
	INIT_NX_HBUCKET_HEAD(&g_adapter_head);
	/*
	 * Init route table and pruning task.
	 */
	nx_route_tbl_init();
        nx_init_pruning_task();
	
	g_tnic_callback.nx_syn_handler = nx_handle_offloaded_syn;
	g_tnic_callback.nx_msg_handler = unm_process_pegnet_msg;
	atomic_set(&g_devid,0);
        unm_register_netdev_notifiers();
        return 0;
}

extern void unm_unregister_netdev_notifiers(void);

void pstack_exit(void)
{
        UNM_INFO(KERN_ALERT "pstack_module_exit:\n");

	printk("%s: in\n", __FUNCTION__);
#ifdef UNM_CONFIG_COMPAT
        if (unregister_ioctl32_conversion(SIOC_NETXEN_OFFLOAD)) {
                printk(KERN_ERR "%s: Error unregistering ioctl32 "
                       "translation.\n", __FUNCTION__);
        }
#endif  /* UNM_CONFIG_COMPAT */

        unm_unregister_netdev_notifiers();
    
#if 0
        memcpy(&inet_stream_ops, &host_inet_stream_ops,
                       sizeof(inet_stream_ops));
#endif
	nx_tnic_unregister_ops();

        inet_dgram_ops.ioctl = native_inet_dgrams_ioctl;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        change_perm((unsigned long)&inet_stream_ops, sizeof(inet_stream_ops),
                    PAGE_KERNEL_RO);
        change_perm((unsigned long)&inet_dgram_ops, sizeof(inet_dgram_ops),
                    PAGE_KERNEL_RO);
#endif


	nx_route_tbl_destroy();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	kmem_cache_destroy(nx_sk_cachep);
#endif 
}

