/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef _CHIMNEY_CMN_H_
#define _CHIMNEY_CMN_H_

#define IN
#define CONST const

#ifdef WINDOWS
#if !defined(U64)
typedef __int64 U64;
typedef __int32 U32;
typedef __int16 U16;
typedef __int8 U8;
#endif
#endif


#ifndef _PEGOS_TYPES_H 
typedef unsigned long long ULONG64;
#ifndef WINDOWS
typedef unsigned int ULONG;
#endif
typedef unsigned short USHORT;
typedef unsigned char UCHAR;
#endif // _PEGOD_TYPES_H


#define STATUS_RST_ABORT		1

#define INITIATE_OFFLOAD_SUCCESS 0
#define INITIATE_OFFLOAD_FAILURE 1

#define QUERY_OFFLOAD_SUCCESS    0
#define QUERY_OFFLOAD_FAILURE    1

/* Use this to turn on/off peek-n-post mode */
#define ind_enabled(sk)     0


#define CHIMNEY_SUCCESS    0
#define CHIMNEY_FAILURE    1

// Following are the Status defines for the reply.

#define	GLOBAL_STATUS_SUCCESS	0
#define GLOBAL_STATUS_FAILURE	1

//#ifdef UNM_CHIMNEY


// TCP_OFFLOAD_EVENT_TYPE
//
// Event indications possible on an offloaded TCP connection.
//
//TCP_UPLOAD_REASON is only when we indicate TcpIndicateRetrieve
//as eventType then TCP_UPLOAD_REASON will be the subEventType...
//
typedef enum {
        /*
         * The following fields match the NDIS TCP_OFFLOAD_EVENT_TYPE.
         */
        UnmTcpIndicateSendBacklogChange = 1,
        UnmTcpIndicateDisconnect,
        UnmTcpIndicateRetrieve,
        UnmTcpIndicateAbort,

        /*
         * The following fields are UNM specific.
         */
        UnmTcpRequestRxBuffers
} UNM_OFFLOAD_EVENT_TYPE, *PUNM_OFFLOAD_EVENT_TYPE;

#ifndef WINDOWS
//
// TCP_OFFLOAD_CONNECTION_STATE
//
// TCP connection state [RFC 793].
//
typedef enum {
    TcpConnectionClosed,
    TcpConnectionListen,
    TcpConnectionSynSent,
    TcpConnectionSynRcvd,
    TcpConnectionEstablished,
    TcpConnectionFinWait1,
    TcpConnectionFinWait2,
    TcpConnectionCloseWait,
    TcpConnectionClosing,
    TcpConnectionLastAck,
    TcpConnectionTimeWait,
    TcpConnectionMaxState
} TCP_CHIMNEY_CONNECTION_STATE;

//
// Qualifies a disconnect event or a disconnect request on a TCP connection.
//

//


// The following are defined in the NDIS, so define it
// for the offload chimney stack..


#define TCP_DISCONNECT_GRACEFUL_CLOSE   0x01
#define TCP_DISCONNECT_ABORTIVE_CLOSE   0x02
//
//
// TCP_UPLOAD_REASON
//
// Reason for uploading an offloaded TCP connection back to the host.
//
typedef enum {
    LowActivity = 1,
    HighDropRate,
    SmallIO,
    NoBufferProposting,
    ReceivedUrgentData,
    HighFragmentation,
    HighOutOfOrderPackets,
    TimeoutExpiration,
    InvalidState,
    UploadRequested,
    MaxUploadReason
} UNM_CHIMNEY_REASON;

#endif
//
// TCP_OFFLOAD_STATS
//
// Statistics definitions in TCP on a per interface basis
//
typedef struct _CHIMNEY_TCP_OFFLOAD_STATS {
    ULONG64 InSegments;
    ULONG64 OutSegments;
    ULONG CurrentlyEstablished;
    ULONG ResetEstablished;
    ULONG RetransmittedSegments;
    ULONG InErrors;
    ULONG OutResets;
} CHIMNEY_TCP_OFFLOAD_STATS, *PCHIMNEY_TCP_OFFLOAD_STATS;

//
// IP_OFFLOAD_STATS
//
// Stats definitions in IPV4/IPV6 on a per interface basis
//
typedef struct _CHIMNEY_IP_OFFLOAD_STATS {
    ULONG64 InReceives;
    ULONG64 InDelivers;
    ULONG64 OutRequests;
    ULONG InHeaderErrors;
    ULONG InDiscards;
    ULONG OutDiscards;
    ULONG OutNoRoutes;
} CHIMNEY_IP_OFFLOAD_STATS, *PCHIMNEY_IP_OFFLOAD_STATS;





typedef	struct _GLOBAL_TCP_CONNECTION_OFFLOAD {
    IN UCHAR TcpAckFrequency;		//Max segments rcvd before sending ACK
    IN UCHAR TcpDelayedAckTicks; 	// ticks to wait before sending ack
    IN UCHAR TcpMaximumRetransmissions; //maximum retransmission before abort
    IN ULONG TcpDuplicateAckThreshold; 	//number of duplicate acks before doing
					// fast retransmission...
    IN ULONG TcpPushTicks;		// Push timer value for receive..
}GLOBAL_TCP_CONNECTION_OFFLOAD;

/* Bits for ConstFlags in UNM_TCP_CONST */
#define TCP_FLAG_TIMESTAMP_ENABLED      0x01 
#define TCP_FLAG_SACK_ENABLED           0x02 
#define TCP_FLAG_WINDOW_SCALING_ENABLED 0x04 

typedef	struct _UNM_TCP_CONST {
    USHORT ConstFlags; // timestamp, SACK, Window scaling enable bits
    USHORT RemotePort;
    USHORT LocalPort;
    UCHAR SndWindScale: 4;
    UCHAR RcvWindScale: 4;
    USHORT RemoteMss;
}UNM_TCP_CONST;


typedef enum
{
    TcpCachedFlags,
    InitialRcvWnd,
    RcvIndicationSize,
    KaProbeCount,
    KaTimeout,
    KaInterval,
    MaxRT,
    FlowLabel,
    TtlOrHopLimit,
    TosOrTrafficClass,
    UserPriority,
    Mtu, // IP cached
    NextHopAddr	
}UNM_UPDATE_CACHED;


typedef	struct _UNM_TCP_CACHE {
    USHORT CacheFlags;	//keepalive, NAGLE, KEEPALIVE Restart, MAX_RT_RESTART
    ULONG InitialRcvWnd;
    ULONG RcvIndicationSize;
    UCHAR KaProbeCount;
    ULONG KaTimeout;
    ULONG KaInterval;
    ULONG MaxRT;
    struct {
        UCHAR Tos;
        UCHAR Ttl;
    } Ipv4;

}UNM_TCP_CACHE;


typedef	struct _UNM_TCP_DELEGATED {

#ifndef WINDOWS
   TCP_CHIMNEY_CONNECTION_STATE State;
#else
   // This is defined in NDIS.H
   TCP_OFFLOAD_CONNECTION_STATE State;
#endif
		// one of the states (connectionSynSent, SynRcvd, Connection
		// Established, FinWait1...

    USHORT DelegatedFlags;	//not used, MS saves it for future...
    ULONG RcvNxt;
    ULONG RcvWnd;
    ULONG SndUna;
    ULONG SndNxt;
    ULONG SndMax;
    ULONG SndWnd;
    ULONG MaxSndWnd;
    ULONG SendWL1;
    ULONG SendWL2;
    ULONG CWnd;
    ULONG SsThresh;
    USHORT SRtt;
    USHORT RttVar;
    ULONG TsRecent;
    ULONG TsRecentAge;
    ULONG TsTime;
    ULONG TotalRT;
    UCHAR DupAckCount;
    UCHAR SndWndProbeCount;
  
    struct {
            UCHAR Count;
            ULONG TimeoutDelta;
    } Retransmit;
} UNM_TCP_DELEGATED;


typedef struct _TCP_STATE {
	UNM_TCP_CONST	constState;
	UNM_TCP_CACHE	cacheState;
	UNM_TCP_DELEGATED delegatedState;
}TCP_STATE;


typedef struct _IP_STATE {
   U32      IpSrcAddr;
   U32      IpDstAddr;
   ULONG    PathMtu;
}IP_STATE;


typedef struct _DL_STATE {
	UCHAR MacSrcAddr[6];
	UCHAR MacDstAddr[6];
}DL_STATE;


typedef struct _CONN_OFFLOAD {
	__uint32_t				connID;
	__uint16_t				port;	/* Need only 3 bits */
	__uint16_t				ring_ctx;
	GLOBAL_TCP_CONNECTION_OFFLOAD		globalSettings;
	TCP_STATE				tcpState;	
	IP_STATE				ipState;
	DL_STATE				dlState;

}CONN_OFFLOAD, *PCONN_OFFLOAD;

//#endif /* UNM_CHIMNEY */

/******************************************************************************/

// Following are the q messages for host <=> phantom communications. 
// the structures ending with _h are produced by host and consumed
// by the Phantom. the ones ending with _t are produced by phantom and
// consumed by host.


// UNM_SET_IOVEC macro sets up a 64 bit value to have 24 bit for length and
// 40 bits of physical address.. This is used by the host to setup the addr_h
// of the host genereated q messages.



// The following struct also is same as iov_t{}. any changes in that struct will
// effect the following...

typedef struct _addr_h {
	union {
		U64	iovec;
		struct {
		U64	phyaddr:40,
			bufferLength:24;
		};
	};
} addr_h, *paddr_h;
		
#define	UNM_PHYADDR_MASK	0xffffffffff

// With the following design for xmit, the xmit peg will use the opcode
// to see if it is chimney msg, if yes, then it will 0 out the opcode and
// the length and whole msg is already setup for the right queue...


typedef struct _unm_chim_msg_hdr {
	unm_msgword_t
		dst_minor:18,	
		dst_subq:1,
		dst_major:4,	/* Major queue number (0-15) */
		opcode:6,	/*xmit peg loop  if chimney msg to be CHIMNEY_CMD define */ 
		length:3,		/*  body length number of 64 bits*/
		msgInfo:32;
} unm_chim_msg_hdr_t;



// For all the q messages between host and peg, the first U64 is the msg hdr. Only
// 28 bits are available andout of which 4 bits (rsvd1) are not available due to
// hardware issues but may be available later on b0 boards...

typedef struct _update_offload_h {

	union {
		struct {
		U64	dstQInfo:28,
			rsvd1:4,
			updateParam:8,
			unused:16,
			rsvd:2,
			type:6;		
		};

		unm_msg_hdr_t msgHdr;
		unm_chim_msg_hdr_t chimMsgHdr;
	};

	union {
		U32		value;
		U8		macDstAddr [6];	
		U64		body;
	} ;
} update_offload_h, *pupdate_offload_h;


typedef struct _initiate_offload_h {
	union {
		struct {
		U64	dstQInfo:28,
			rsvd1:4,
			port:2,	// it can be 0-3
			unused:2,
			connID:20,	
			rsvd:2,
			type:6;	
		};
		unm_msg_hdr_t msgHdr;
		unm_chim_msg_hdr_t chimMsgHdr;
	};

	addr_h	addr;
} initiate_offload_h, *pinitiate_offload_h;



typedef struct _query_offload_h {
	union {
		struct {
		U64	dstQInfo:28,
			rsvd1:4,
			unused:24,
			rsvd:2,
			type:6;	
		};
		unm_msg_hdr_t msgHdr;
		unm_chim_msg_hdr_t chimMsgHdr;
	};
		
	addr_h	addr;
} query_offload_h, *pquery_offload_h;

typedef struct _terminate_offload_h {
	union {
		struct {
		U64	dstQInfo:28,
			rsvd1:4,
			unused:24,
			rsvd:2,
			type:6;	
		};
		unm_msg_hdr_t msgHdr;
		unm_chim_msg_hdr_t chimMsgHdr;
	};
	addr_h	addr;
} terminate_offload_h, *pterminate_offload_h;

typedef struct _invalidate_offload_h {
	union {
		struct {
		U64	dstQInfo:28,
			rsvd1:4,
			unused:24,
			rsvd:2,
			type:6;	
		};
		unm_msg_hdr_t msgHdr;
		unm_chim_msg_hdr_t chimMsgHdr;
	};
} invalidate_offload_h, *pinvalidate_offload_h;


// Following defines are for the host request to indicate the statistics 
// that we need.
// 

#define	UNM_TCP_STATS	1
#define	UNM_IP_STATS	2

// The floowing struct will be used for following commands
// Xmit,pcb, forward

// Following define the bits for sending the commands to indicate start of
// cmd and end of command set in the flgs field...

#define		START_OF_CMD	0x1
#define		END_OF_CMD	0x2
#define		REFERENCE_HANDLE_MASK	0xffff	// 16 bits
// We should make the referneceHandle to 18 bits if possible..
// This structure is same as the pegnet_types.h for txbuf_msg_t{}
// and rxbuf_msg_t .. Those structure changes will effect the
// following structure...

//    This is the descriptor which is sent from host to the phantom...

typedef struct _chimney_hostDesc_h {
	union {
		struct _invalidate_offload_h invalidate;
		struct _terminate_offload_h terminate;
		struct _query_offload_h query;
		struct _initiate_offload_h initiate;
		struct _update_offload_h update;
	};
}chimney_hostDesc_h, *pchimney_hostDesc_h;


/*
 * This is same as rxbuff_flags_t as defined in pegnet_types.h  which is part of 
 * every receive completion and is rcvflags in the chimney_statusDesc. 
 */
typedef struct {
    union {
        struct {
            U8 rcv_shutdown: 1,    /* We rcv'd a fin from the peer.    */
                urg: 1,             /* Buffer contains urgent data.     */
	        surrender: 1,       /* Unused buffer. (card to host)    */
                rsvd0: 5 ;
        } ;
        U8   word ;
    } ;
} chimney_rxbuff_flags_t ;


typedef struct chimney_statusDesc{
	union{
		struct {
			U64     flagseventType:4,// Can be used either as flags or
						// eventType as before...
				subEventType:4,
				referenceHandle:16,
				rcvflags:8,
				status:2,
				unused:19,
                                desc_cnt:3,     /* Match NIC path desc count */
                                owner:2,        /* Match NIC path owner */
				opcode:6;

			U64	connID:20,
				urgDataPtr:20,
				totalLength:20,
				notUsed:4;
		};

		U64 body[2];
	};
} chimney_statusDesc_t;

/*
 * This typedef should be in hostInterface.h.
 * The used fields are:
 *      connID
 *      flagseventType
 *      totalLength
 */
typedef struct  _chimney_initiate_offload_reply_t{
	union {
		struct {
			U64	unused:8,
				src_minor:18, /* The offload conn's q*/
				unused1:6,
				src_subq:1,	
				unused2:3,
				src_major:4, 
				unused3:18,
				opcode:6;

			U64 	status:8, 	// if successful offload or not...
				connID:20,         
				unused4:4,
				pcb_minor:18, /* Offload conn's pcb queue */
				unused5:6,
				pcb_subq:1,
				unused6:3,
				pcb_major:4; 
			};
		U64 body[2];
	};
} chimney_initiate_offload_reply_t;



typedef struct _chimney_p2p_initiate_offload_reply_t{
	unm_msg_hdr_t msgHdr;
	chimney_initiate_offload_reply_t hostStatus;
}chimney_p2p_initiate_offload_reply_t;


#endif
