/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */

/*
 * $Id: hostInterface.h,v 1.67 2007/01/10 09:57:19 krishna Exp $
 *
 */

#if !defined(_HOST_INTERFACE_H_)
#define _HOST_INTERFACE_H_

#include <message.h>
#include <nic_cmn.h>
#include <chimney_cmn.h>

/*
 * Header of message sent between host/casper and pegasus.
 */
typedef union {
        struct {
                __uint64_t      ctx : 32,
                                opcode : 8,
                                op_id : 8,
                                rsvd : 5,
                                desc_cnt : 3,   /* Card -> Host.
                                                   Used by NIC path */
                                owner : 2,      /* Card -> Host.
                                                   Used by NIC path */
                                qmsg_type : 6;  /* Card -> Host.
                                                   Used by NIC path */
        };
        __uint64_t              word;
} host_peg_msg_hdr_t;

/*
 * HOST --> CARD
 *
 * The following are the opcodes for messages from host -> card. These messages
 * have queue message type UNM_MSGTYPE_HOST_REQUEST.
 */
enum {
        UNM_H2P_OPCODE_START = 0,
        UNM_H2P_OPCODE_IFCONFIG_REQUEST,
        UNM_H2P_OPCODE_IOCTL_REQUEST,
        UNM_H2P_OPCODE_LISTEN_REQUEST,
        UNM_H2P_OPCODE_SETSOCKOPT_REQUEST,
        UNM_H2P_OPCODE_GETSOCKOPT_REQUEST,
        UNM_H2P_OPCODE_IPTSETSOCKOPT_REQUEST,
        UNM_H2P_OPCODE_IPTGETSOCKOPT_REQUEST,
        UNM_H2P_OPCODE_CONN_SHUTDOWN_REQUEST,
        UNM_H2P_OPCODE_GET_STATS_REQUEST,
        UNM_H2P_OPCODE_RESET_STATS_REQUEST,
	UNM_H2P_OPCODE_TCP_CHILD_KILL,
	UNM_H2P_OPCODE_FORCE_3WAY_HSHAKE_COMPL,
        UNM_H2P_OPCODE_LAST
};

/*
 * CARD --> HOST
 *
 * The opcodes used in the card response messages.
 * The range can be from 1 to 255. Don't use 0 for opcode to avoid bad code.
 * UNM_P2H_OPCODE_LAST is always the last one.
 */
enum {
        UNM_P2H_OPCODE_START = 0,       /* An invalid one. */
        UNM_P2H_OPCODE_CREATE_RESPONSE, /* Socket Create completion */
        UNM_P2H_OPCODE_RELEASE_RESPONSE,/* To convey a release completion */
        UNM_P2H_OPCODE_CONNECT_RESPONSE,        /* Connect reply */
        UNM_P2H_OPCODE_CONN_DISCONNECT_RESPONSE,/* Disconnect reply */
        UNM_P2H_OPCODE_CONN_SHUTDOWN_RESPONSE,  /* Shutdown reply */
        UNM_P2H_OPCODE_POLL_STATUS,     /* To convey the status of the poll */
        UNM_P2H_OPCODE_RX_URGENT_POINTER, /* Notify the host of the received
                                             urgent pointer */
        UNM_P2H_OPCODE_CONN_SIGNAL,     /* This should be killed in the long
                                           run */
        UNM_P2H_OPCODE_SENDMSG_CONTROL, /* Sendmsg failure/recovery notice */

        UNM_P2H_OPCODE_LISTEN_RESPONSE,  /* Response to listen request */
        UNM_P2H_OPCODE_IOCTL_RESPONSE,  /* Response to ioctl request */
        UNM_P2H_OPCODE_SETSOCKOPT_RESPONSE,  /* Reply to setsockopt request */
        UNM_P2H_OPCODE_GETSOCKOPT_RESPONSE,  /* Reply to getsockopt request */
        UNM_P2H_OPCODE_IPTSETSOCKOPT_RESPONSE,/* Reply to setsockopt request */
        UNM_P2H_OPCODE_IPTGETSOCKOPT_RESPONSE,/* Reply to getsockopt request */

        UNM_P2H_OPCODE_IFCONFIG_REPLY,	/* Reply to ifconfig request */
        UNM_P2H_OPCODE_RDMA_MESG,       /* RDMA card-to-host message. */
        UNM_P2H_OPCODE_GET_STATS_REPLY, /* Response to get stats request */
        UNM_P2H_OPCODE_SWITCH_RXBUF_COMPLETE,   /* Switch from local<==>global
                                                   done */
        UNM_P2H_OPCODE_TCP_EVENT,       /* TCP events */
	UNM_P2H_OPCODE_CONNECT_COMPLETE,/* fate of connection setup attempt */
	UNM_P2H_OPCODE_RESOURCE_NOTICE, /* Notice sent to host when resource
                                           thresholds are crossed */
        NX_P2H_OPCODE_SOCK_CLEANUP,	/* Cleanup host socket on last sock put at card */
	UNM_P2H_OPCODE_3WAY_HANDSHAKE_COMPLETE,	/* Passive connection sends the
						   message when a connection is
						   established. */
	UNM_P2H_OPCODE_RX_PCB_REQUEST,	/* Request the host to post receive
					   PCB. Chimney does it at upload
					   time. */
	UNM_P2H_OPCODE_PRUNABLE_SK,	/* child sk can be pruned if needed */
	UNM_P2H_OPCODE_ROUTECACHE_INVALID,	/* Message to request route
						   lookup */
	NX_P2H_OPCODE_UPDATE_CONNID,   /* update connID table at host */
        UNM_P2H_OPCODE_LAST             /* Always the last. */
};

/*
 * The IP address field used in messages.
 */
typedef	union {
	__uint32_t	v4;
	__uint32_t	v6[4];
} ip_addr_t;

/*
 * Card --> Host
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
	__uint32_t              flags;
        __uint32_t              rsvd_0;         /* For Alignment */
} poll_status_t;

/*
 * Card --> Host
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
	__uint32_t              byte_0 : 8,
                                byte_1 : 8,
                                valid : 2,      /* 0x1 - byte_0, 0x2 - byte_1,
                                                   0x3 - both */
                                rsvd_0 : 14;
        __uint32_t              position;       /* Position of urg pointer */
} rcv_urg_pointer_t;

/*
 * Card --> Host
 * This message should be killed in the long run.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
	__uint32_t              how;
        __uint32_t              band;
} conn_signal_t;

/*
 * Card --> Host
 * Response sent to the release.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
	__uint32_t              ret_code;
        __uint8_t		active_close; /* == 0, default,
						 == 1, if active close */
        __uint8_t		rsvd[3];      /* For 64 bit alignment */
} release_response_t;

/*
 * Card --> Host
 * Sent ot celanup time_wait socket in host.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
} snd_twkill_t;

typedef struct {
        host_peg_msg_hdr_t      hdr;
	__uint32_t              ret_code;
        __uint32_t              rcvBuf_qid;
} connect_response_t;

typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              child_ctx;
        __uint32_t              ret_code;
} accept_error_response_t;

#define NX_QID_2_CONNID(QID) (get_ordinal_number((QID)))

typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              ret_code;
        __uint32_t              qid;
        __uint32_t              connId;
        __uint32_t              new_ctx;
} create_response_t;

/*
 * CARD --> HOST
 *
 * Response to connection flow shutdown.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        /*
         * Following fields specify what the request was.
         */
        __uint32_t              abortive : 1,
                                rx_shutdown : 1,
                                tx_shutdown : 1,
                                rsvd_0 : 29;
        __uint32_t              ret_code;
} conn_disconnect_response_t;

/*
 * CARD --> HOST
 *
 * This message is sent when there is any TCP event to notify the host. Used
 * to indicate to the host RST received, RST sent, FIN being acked, etc.
 */
typedef union {
        struct {
                __uint64_t      type : 8,       /* types are below */
                                rsvd_0 : 24,    /* reserved */
                                reason : 32;    /* errno or subevent */
        };
        __uint64_t      word;
} unm_tcp_event_param_t;

typedef struct {
        host_peg_msg_hdr_t      hdr;
        unm_tcp_event_param_t   param;
} unm_tcp_event_msg_t;

typedef enum {
        UNM_TCP_EVENT_START = 0,        /* Just a place holder */
        UNM_TCP_EVENT_FIN_ACKED,        /* When the FIN is acked */
        UNM_TCP_EVENT_RST_RECEIVED,     /* When a RST is rxed */
        UNM_TCP_EVENT_ABORT_ON_SYN,     /* when a RST is sent in response to
                                           receiving a SYN in the middle of a
                                           connection. */
        UNM_TCP_EVENT_ABORT_ON_DATA,    /* when a RST is sent when data is
                                           received in an invalid state. */
        UNM_TCP_EVENT_ABORT_ON_CLOSE,   /* If RST is sent on tcp_close() */
        UNM_TCP_EVENT_ABORT_ON_MEMORY,
        UNM_TCP_EVENT_ABORT_ON_LINGER,  /* Exit due to linger2 */
        UNM_TCP_EVENT_ABORT_DISCONNECT,
        UNM_TCP_EVENT_KEEPALIVE_EXPIRED,/* When a keepalive try count reaches
                                           maximum. Depending on the type of
                                           offload it could reset the
                                           connection. Linux it resets the
                                           connection. */
        UNM_TCP_EVENT_RETRANSMIT_TIMEOUT,/* Sent when final retransmit timer
                                            happens and things are to be
                                            cleaned up. */
        UNM_TCP_EVENT_WND_PROBE_TIMEOUT,/* Sent when window probe fails */
        UNM_TCP_EVENT_LAST
} unm_tcp_event_types_t;

/*
 * CARD --> HOST
 *
 * Response to switch from using local to global RX buffers or from global to
 * local RX buffers. Currently used only for chimney. Currently only sent when
 * switching from global to local.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              direction:2,    /* Same as the request. */
                                rsvd:30;
        __uint32_t              ret_code;       /* 0 is successful switch */
} switch_rxbuf_queue_response_t;

/*
 * CARD --> HOST
 *
 * Response to switch from using local to global RX buffers or from global to
 * local RX buffers. Currently used only for chimney. Currently only sent when
 * switching from global to local.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              length;		/* PCB request in bytes. */
        __uint32_t              rsvd;		/* For 8 byte alignment */
} rx_pcb_request_t;

/*
 * HOST --> CARD
 *
 * Chimney initiate offload request.
 */
typedef struct {
	unm_msg_hdr_t	hdr;
	__uint64_t	connID	: 32,
			ring_ctx : 16,
			target_peg : 3,
			rsvd : 13;
	__uint64_t	addr;
} initiate_offload_t;

/*
 * CARD --> HOST
 *
 * Response for ioctl.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              ret_code;       /* return code */
	__uint32_t		rsvd;		/* For 8 byte alignment */
} ioctl_response_t;

/*
 * CARD --> HOST
 *
 * Response for setsockopt.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              ret_code;       /* return code */
	__uint32_t		rsvd;		/* For 8 byte alignment */
} setsockopt_response_t;

/*
 * CARD --> HOST
 *
 * Response for getsockopt.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              ret_code;       /* return code */
        __uint32_t              optlen;
} getsockopt_response_t;

/*
 * CARD --> HOST
 *
 * Response for iptsetsockopt.
 */
typedef setsockopt_response_t   iptsetsockopt_response_t;

/*
 * CARD --> HOST
 *
 * Response for iptsetsockopt.
 */
typedef getsockopt_response_t   iptgetsockopt_response_t;

/*
 * CARD --> HOST
 *
 * Response for socket listen.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              ret_code;       /* return code */
	__uint32_t		free_accept_qid; 
} listen_response_t;





/*
 * HOST --> CARD
 *
 * Message to card to clean up all TCP and above related information. This
 * resets all resources.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;    /* Func */
        __uint64_t              rsvd;	/* For later use */
} tnic_reset_fw_t;

/*
 * HOST --> CARD
 *
 * Message to card to do equivalent of ifconfig.
 */
#define UNM_IFCONFIG_IFNAMSIZ   16      /* Matches IFNAMSIZ in net/if.h */
#define UNM_IFCONFIG_ADDR_V4    0x1
#define UNM_IFCONFIG_ADDR_V6    0x2
typedef struct {
        host_peg_msg_hdr_t      hdr;    /* Func */
        char                    ifname[UNM_IFCONFIG_IFNAMSIZ];
        __uint64_t              port : 3,
                                family : 2,
				logical : 1,
				rsvd : 2,
                                ip_mask_prefixlen : 7,
				vlan_set : 1,
                                vlanid : 16,
                                cmd : 16,
                                ring_ctx : 16;

        __uint64_t              hwaddr;
        ip_addr_t               ipaddr;
} ifconfig_request_t;

enum {
	NX_NETDEV_ADD,
	NX_NETDEV_REMOVE,
	NX_NETDEV_UP,
	NX_NETDEV_DOWN,
};

/*
 * CARD --> HOST
 *
 * Use the same message as the request. This way when the reply is received it
 * is easy to match the request. The ret_code is the only change.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;    /* Func */
        char                    ifname[UNM_IFCONFIG_IFNAMSIZ];
        __uint64_t              port : 3,
                                family : 2,
				logical : 1,
				rsvd : 2,
                                ip_mask_prefixlen : 7,
				vlan_set : 1,
                                vlanid : 16,
                                cmd : 16,
                                ret_code : 16;

        __uint64_t              hwaddr;
        ip_addr_t               ipaddr;
} ifconfig_response_t;

/*
 * CARD --> CARD
 *
 * This is a compressed SYN packet sent from the listener peg to the child peg.
 * It triggers the SYN-ACK to be sent out by the target peg.
 */
typedef struct {
        union {
                struct {
                        __uint64_t      dst_minor:18,	/* [17:00] */
                                        dst_subq:1,	/* [18:18] */
                                        dst_major:4,	/* [22:19] */
                                        dst_side:1,	/* [23:23] */
                                        dst_c2cport:4,	/* [27:24] */
                                        hw_rsvd_1:4,	/* [31:28] */
                                        sw_rsvd_1:21,	/* [52:32] */
                                        target_peg:3,	/* [55:54] */
                                        hw_rsvd_2:2,	/* [57:56] */
                                        type:6;		/* [63:58] */
                };
                unm_msg_hdr_t		hdr;
        };
        union {
		struct {
			ip_addr_t	local_ip;	/* + 16 = 16 bytes */
			ip_addr_t	remote_ip;	/* + 16 = 32 bytes */
			__uint16_t	local_port;	/* +  2 = 34 bytes */
			__uint16_t	remote_port;	/* +  2 = 36 bytes */
			__uint32_t	seq_num;	/* +  4 = 40 bytes */
			__uint16_t	len_n_flags;	/* +  2 = 42 bytes */
			__uint16_t	window;		/* +  2 = 44 bytes */
			__uint32_t	time_stamp;	/* +  4 = 48 bytes */
			__uint16_t	mss;		/* +  2 = 50 bytes */
			__uint8_t	wscale;		/* +  1 = 51 bytes */
			__uint8_t	ts_ok : 1,
					wscale_set : 1,
					mss_set : 1,
					sack_ok : 1,
					retxed : 1,
					rsvd : 3;	/* +  1 = 52 bytes */
		};
                unm_msg_body_t		body;
        };
} compressed_syn_offload_request_t;

/* Possible values for ret_code in tcp_handshake_complete_t */
#define PEGNET_SYN_OFFLOAD_RESP_OK		0
#define PEGNET_SYN_OFFLOAD_RESP_LISTENER_FAIL	1 /* SYN dropped by listener.
						     No Child was allocated. */
#define PEGNET_SYN_OFFLOAD_RESP_CHILD_FAIL	2 /* Card has not been able to
						     process/respond. Host can
						     be offered this SYN, or
						     it can be dropped */
#define PEGNET_SYN_OFFLOAD_RESP_CHILD_FORCED   3  /* Card has been forced to
						     send 3-way handshake
						     complete */
/*
 * CARD --> HOST
 *
 * Sent by the peg to the host It will update ConnID table at host.
 * The context in the hdr is the listeners connection ID.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              child_connID;
        __uint32_t              child_ctx;
        __uint32_t              local_ip;
        __uint32_t              remote_ip;
        __uint16_t              local_port;
        __uint16_t              remote_port;
} nx_update_connID_table_t;


/*
 * CARD --> HOST
 *
 * Sent by the peg to the host when the connection 3 way handshake is complete.
 * The context in the hdr is the listeners connection ID.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              qid;
        __uint32_t              rcvBuf_qid;
	__uint32_t		child_connID;
	__uint32_t		ret_code;
	__uint32_t		child_ctx;
	__uint32_t		local_ip;
	__uint32_t		remote_ip;
	__uint16_t		local_port;
	__uint16_t		remote_port;
} tcp_handshake_complete_t;

/*
 * CARD --> HOST
 * Sent by the peg to the host when it decides that the sk meets criteria
 * for pruning
 */
typedef struct {
	host_peg_msg_hdr_t      hdr;
	__uint32_t		child_connID;
	__uint32_t		child_ctx;
	__uint32_t              child_qid;
	__uint32_t		local_ip;
	__uint32_t		remote_ip;
	__uint16_t		local_port;
	__uint16_t		remote_port;
} tcp_prunable_sk_t;

/*
 * HOST --> CARD
 *
 * Sent when a listener closes and need to kill a child still in the SYN queue.
 *
 * The context in the hdr is the listeners connection ID. This is sent to the
 * listener queue.
 */
typedef struct {
	host_peg_msg_hdr_t      hdr;
	__uint32_t		child_connID;
	__uint32_t		ip_version;
	ip_addr_t		local_ip;
	ip_addr_t		remote_ip;
	__uint16_t		local_port;
	__uint16_t		remote_port;
	__uint32_t		rsvd;
} tcp_child_kill_t;

/*
 * HOST --> CARD (listener), or CARD (listener) --> CARD (child)
 * Child MUST send 3-Way handshake response if it has
 * not already done so.
 */
typedef struct {
	host_peg_msg_hdr_t      hdr;
	ip_addr_t               local_ip;
	ip_addr_t               remote_ip;
	__uint16_t              local_port;
	__uint16_t              remote_port;
} nx_send_immediate_conn_compl_t;


/*
 * CARD --> CARD (listener to child)
 *
 * Message to a child socket to trigger cleanup
 */
typedef struct {
	unm_msg_hdr_t		hdr;
	__uint8_t		inform_host;
} tcp_child_release_t;

/*
 * HOST --> CARD
 *
 * Request for connection flow shutdown.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              abortive : 1,
                                rx_shutdown : 1,
                                tx_shutdown : 1,
                                rsvd_0 : 29;
        __uint32_t              rsvd_1;         /* For 64 bit alignment */
} conn_disconnect_request_t;

/*
 * HOST --> CARD
 *
 * Request for ioctl.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint64_t              arg;    /* ptr to ioctl argument */
        __uint32_t              cmd;    /* ioctl command */
        __uint32_t              ctx;    /* encoded cookie */
} ioctl_request_t;

/*
 * HOST --> CARD
 *
 * Request for setsockopt.
 */
#define UNM_SETSOCKOPT_MAX_OPLEN        12
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              level;          /* protocol layer affected */
        __uint32_t              optname;        /* name of option to be set */
        __uint64_t              optval_ptr;     /* the ptr to the option */
        __uint32_t              optlen;         /* length of previous field */
	__uint8_t               is_val;         /* == 0, for pass by reference,
						   != 0, for pass by value */
} setsockopt_request_t;

/*
 * HOST --> CARD
 *
 * Set connection options.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
	__uint16_t		keepalive_set	: 1,
				priority_set	: 1,
				rcvbuf_set	: 1,
				nagle_set	: 1,
				quickack_set	: 1,
				max_adv_wnd_set	: 1,
				finwait2_to_set : 1,
				rsvd0		: 9;

	__uint16_t		keepalive	: 1,
				nagle		: 1,
				quick_ack	: 1,
				rsvd1		:13;

	__uint32_t		priority;		/* The priority for
							   this connection */
        __uint32_t              rcvbuf;			/* Receive buffer */
	__uint32_t		max_adv_wnd;		/* Max advertised
							   receive window */
	__uint16_t		keepalive_idle;		/* Seconds of
							   inactivity after
							   which keepalive is
							   triggered */
	__uint16_t		keepalive_interval;	/* Number of seconds
							   between tries */
	__uint16_t		keepalive_cnt;		/* Number of tries of
							   keepalive */
	__uint16_t		finwait2_timeout;	/* Time in FIN_WAIT_2
							   state of an orphan
							   before cleaning the
							   connection */
} set_conn_options_request_t;

/*
 * HOST --> CARD
 *
 * Request for getsockopt.
 */
typedef setsockopt_request_t    getsockopt_request_t;

/*
 * HOST --> CARD
 *
 * Request for setsockopt.
 */
typedef setsockopt_request_t    iptsetsockopt_request_t;

/*
 * HOST --> CARD
 *
 * Request for getsockopt.
 */
typedef getsockopt_request_t    iptgetsockopt_request_t;

/*
 * HOST --> CARD
 *
 * Request for socket listen.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        ip_addr_t               lip_addr;
        __uint16_t              lport;
        __uint16_t              backlog;
	__uint32_t              rcvbuf; /* ignored if zero */
	__uint16_t		listener_id;
	__uint8_t		reuseflg;
	__uint8_t		do_bind;
	__uint32_t		rsvd_1;
} listen_request_t;

/*
 * Connect request from host -> card.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        ip_addr_t               rip_addr;       /* Address to connect to. */
        ip_addr_t               lip_addr;       /* Address to connect to. */
        __uint16_t              rport;
        __uint16_t              lport;
        __uint16_t              family;		/* address family */
	__uint8_t		reuseflg : 1,	/* Reuse of address */
				do_bind : 1,	/* If bind is required before
						   connect */
				rsvd0 : 6;
	__uint8_t		syncnt;
        __uint32_t              flags;          /* other bits */
	__uint32_t              rcvbuf;         /* ignored if zero */
} connect_request_t;

/*
 * Socket create from host -> card.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint16_t              family;
        __uint16_t              protocol;
        __uint16_t              ring_ctx;
        __uint16_t              rsvd_1;         /* For 64 bit alignment */
} create_request_t;

/*
 * Release request from host -> card.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint64_t              linger : 1,
                                process_exiting : 1,
                                tx_bufs_pending : 1,
                                force : 1,
                                reset : 1,
                                rsvd_0 : 27,
                                last_read_seq_num : 32;
        __uint32_t              lingertime;
        __uint32_t              rsvd_1;
} release_request_t;

/*
 * Host --> Card
 *
 * Request to switch from using local to global RX buffers or from global to
 * local RX buffers. Currently used only for chimney.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
#define UNM_SWITCH_RXBUF_QUEUE_LOC2GLOB         1
#define UNM_SWITCH_RXBUF_QUEUE_GLOB2LOC         2
        __uint32_t              direction:2,
                                rsvd_0:30;
        __uint32_t              rsvd_1;         /* For 64 bit alignment */
} switch_rxbuf_queue_request_t;

/*
 * Used in the posting of receive and transmit buffers.
 */
#define UNM_MSG_IOV_ADDRESS_BITS        40
#define UNM_MSG_IOV_ADDRESS_LEN_BITS    24
#define	UNM_MSG_IOV_PHYADDR_MASK        0xffffffffff

#define UNM_SET_IOVEC(IOVEC, LENGTH, HIGH_ADDRESS, LOW_ADDRESS) \
        (IOVEC) = (((__uint64_t)(LENGTH) << 40 ) |              \
                   (__uint64_t)(HIGH_ADDRESS) << 32 |           \
                   (__uint64_t)(LOW_ADDRESS))
typedef struct {
        union {
                struct {
                        __uint64_t      addr:40,
                                        len:24;
                };
                __uint64_t	iov;
        };
} iov_t;

#define NETXEN_RXBUF_COUNTER_BITS       4

/* Used to post empty receive buffers (host to card) */
#define QMSG_RX_PAGES           7   /* Max pages in rxbuf_msg_body_t qmsg. */
typedef struct {
        union {
                struct {
                        __uint64_t      dst_minor:18,       /* [17:00] */
                                        dst_subq:1,         /* [18:18] */
                                        dst_major:4,        /* [22:19] */
                                        dst_side:1,         /* [23:23] */
                                        dst_c2cport:4,      /* [27:24] */
                                        rsvd_0:4,           /* [31:28] */
                                        cookie:10,          /* [41:32] */
                                        counter:NETXEN_RXBUF_COUNTER_BITS,/* [45:42] */
                                        rsvd_1:2,           /* [46:47] */
                                        bufferStart:1,      /* [48:48] */
                                        bufferComplete:1,   /* [49:49] */
                                        pushSet:1,          /* [50:50] */
                                        pushTimer:1,        /* [51:51] */
                                        numBufs:4,          /* [55:52] */
                                        rsvd_2:2,           /* [57:56] */
                                        type:6;             /* [63:58] */
                };
                unm_msg_hdr_t hdr;
        };
        union {
                iov_t           iov[QMSG_RX_PAGES];
                unm_msg_body_t  body;
        };
} rxbuf_msg_t;



#define HOST_PAGE_SIZE 4096
#define HOST_PAGE_MASK (~(HOST_PAGE_SIZE-1))

#define QMSG_TX_COMPRESSED_PAGES 10

typedef struct {
	__uint32_t       addr32[QMSG_TX_COMPRESSED_PAGES];
	__uint32_t       end_len;
	__uint8_t        addr_hi[QMSG_TX_COMPRESSED_PAGES];
} iov40_t;

#define QMSG_TX_PAGES   7


/* Used to post filled send buffers (host to card) */
typedef struct {
        union {
                struct {
                        __uint64_t      dst_minor:18,       /* [17:00] */
                                        dst_subq:1,         /* [18:18] */
                                        dst_major:4,        /* [22:19] */
                                        dst_side:1,         /* [23:23] */
                                        dst_c2cport:4,      /* [27:24] */
                                        rsvd0:4,            /* [31:28] */
                                        cookie:16,          /* [47:32] */
					bufferStart:1,      /* [48:48] */
					bufferComplete:1,   /* [49:49] */
					compressed_iov:1,   /* [50:50] */
                                        urgFlag:1,          /* [51:51] */
                                        numBufs:4,          /* [55:52] */
                                        rsvd1:2,            /* [57:56] */
                                        type:6;             /* [63:58] */
                };
                unm_msg_hdr_t           hdr;
        };
        union {
                iov_t           iov[QMSG_TX_PAGES];
        	iov40_t         iov40;
                unm_msg_body_t  body;
        };
} txbuf_msg_t;

/* Return filled receive buffers.  Card to host */
typedef struct {
        union {
                struct {
                        __uint8_t       fin_rxed: 1,    /* Rcv'd a fin from
                                                           the peer. */
                                        urg: 1,         /* Buffer contains
                                                           urgent data. */
                                        surrender: 1,   /* Unused buffer. (card
                                                           to host) */
                                        global_buffer: 1,
                                        push_complete: 1,
                                        rsvd_0: 3;
                };
                __uint8_t       word;
        };
} rxbuff_flags_t;

/* Used to post filled send buffers (host to card) */
#define	QMSG_MAX_TX_SWITCH_IOV		3
typedef struct {
        union {
                struct {
                        __uint64_t      dst_minor:18,       /* [17:00] */
                                        dst_subq:1,         /* [18:18] */
                                        dst_major:4,        /* [22:19] */
                                        dst_side:1,         /* [23:23] */
                                        dst_c2cport:4,      /* [27:24] */
                                        rsvd0:4,            /* [31:28] */
                                        cookie:16,          /* [47:32] */
					bufferStart:1,      /* [48:48] */
					bufferComplete:1,   /* [49:49] */
					rsvd1:2,	    /* [51:50] */
                                        numBufs:4,          /* [55:52] */
                                        rsvd2:2,            /* [57:56] */
                                        type:6;             /* [63:58] */
                };
                unm_msg_hdr_t           hdr;
        };
        union {
		struct {
			iov_t           old[QMSG_MAX_TX_SWITCH_IOV];
			iov_t           new[QMSG_MAX_TX_SWITCH_IOV];
			__uint64_t	rsvd3;
		};
                unm_msg_body_t  body;
        };
} txbuf_switch_t;

/*
 * CARD --> HOST
 */
typedef struct {
	union {
		struct {
			__uint64_t      ctx:32,
                                        cookie:16,
                                        rsvd_0:5,
                                        desc_cnt:3,     /* Set in the NIC path
                                                           before DMA to host*/
                                        owner:2,        /* Set in the NIC path
                                                           before DMA to host*/
                                        qmsg_type:6;       /* Qmsg or NIC rcv
                                                           desc type */

			__uint64_t      length:24,
                                        rsvd_1:8,
                                        urgPtr:24,
                                        flags:8;
		};

		__uint64_t      body[2];
	};
} rxbuff_completion_t;

/*
 *
 */
typedef struct {
	union {
		struct {
			__uint64_t      ctx:32,
                                        cookie:16,
                                        rsvd_0:5,
                                        desc_cnt:3,     /* Set in the NIC path
                                                           before DMA to host*/
                                        owner:2,        /* Set in the NIC path
                                                           before DMA to host*/
                                        qmsg_type:6;    /* Qmsg or NIC rcv
                                                           desc type */

			__uint64_t      length:24,
                                        comp_type:8,    /* */
                                        rsvd_1:32;
		};

		__uint64_t      body[2];
	};
} txbuff_completion_t;

/*
 * Different TX buffer completion types.
 */
#define	NX_TX_COMP_DATA			0x01
#define	NX_TX_COMP_SHUTDOWN		0x02
#define	NX_TX_COMP_DATA_N_SHUTDOWN	(NX_TX_COMP_DATA | NX_TX_COMP_SHUTDOWN)
#define	NX_TX_COMP_STATE_ERROR		0x40	/* Not Implemented */
#define	NX_TX_COMP_TX_FAILED		0x80	/* Not Implemented */


typedef struct {
	union {
		struct {
			__uint64_t      ctx:32,
                                        cookie:16,
                                        rsvd_0:5,
                                        desc_cnt:3,     /* Set in the NIC path
                                                           before DMA to host*/
                                        owner:2,        /* Set in the NIC path
                                                           before DMA to host*/
                                        qmsg_type:6;    /* Qmsg or NIC rcv
                                                           desc type */

			__uint64_t      rsvd_1;
		};

		__uint64_t      body[2];
	};
} txbuf_switch_completion_t;

/*
 * CARD  --> HOST
 *
 * Message sent from pegasus to host/casper to notify of send failure and
 * request retransmission of the send buffers from a given point.
 */
typedef struct {
	host_peg_msg_hdr_t	hdr;
	__uint64_t		pointer;	/* The pointer from where
						   resending of data is
						   requested */
} peg_sendmsg_control_t;

/*
 * HOST --> CARD
 *
 * Do we need setting of RTT and window clamp.
 */
typedef struct {
        union {
                struct {
                        __uint64_t      dst_minor	:18,	/* [17:00] */
                                        dst_subq	: 1,	/* [18:18] */
                                        dst_major	: 4,	/* [22:19] */
                                        dst_side	: 1,	/* [23:23] */
                                        dst_c2cport	: 4,	/* [27:24] */
                                        rsvd0		:20,	/* [47:28] */
					req_type	: 8,	/* [55:48] */
                                        rsvd2		: 2,	/* [57:56] */
                                        type		: 6;	/* [63:58] */
                };
                unm_msg_hdr_t           hdr;
        };
        union {
		struct {
			ip_addr_t		src;
			ip_addr_t		dest;
			ip_addr_t		gateway;
			__uint64_t		advmss : 16,
						pmtu : 16,
						vlanid : 16,
						tos : 8,
						port :3,
						vlan_set : 1,
						qos : 3,
						ip_v6 : 1;
		};
                unm_msg_body_t  body;
        };
} routecache_request_t;

enum {
	NX_ROUTECACHE_ADD,
	NX_ROUTECACHE_MODIFY,
	NX_ROUTECACHE_INVALIDATE,
	NX_ROUTECACHE_REMOVE
};

/*
 * CARD --> HOST
 *
 * Sent when the card needs a update of the cache when there is a redirect or
 * a transmit timeout.
 */
#define	NX_ROUTECACHE_INVALID_REDIRECT		1
#define	NX_ROUTECACHE_INVALID_TIMEOUT		2
typedef struct {
        host_peg_msg_hdr_t      hdr;
	ip_addr_t		src;
	ip_addr_t		dest;
	__uint16_t		sport;
	__uint16_t		dport;
	__uint8_t		protocol;
	__uint8_t		family;
	__uint8_t		reason;
	__uint8_t		tos;
} routecache_invalid_t;

/*
 * CARD --> HOST
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint64_t              type    :32,
                                param   :32;
} netxen_resource_notice_t;

/*
 * Enumeration for the different notification types in peg_resource_notice_t.
 */
typedef enum {
        NETXEN_RESOURCE_NOTICE_START = 0,
        NETXEN_TX_MEM_MS_LOW_WATER,
        NETXEN_TX_MEM_MS_HIGH_WATER,
        NETXEN_TX_MEM_HS_LOW_WATER,
        NETXEN_TX_MEM_HS_HIGH_WATER,
        NETXEN_RESOURCE_NOTICE_LAST
} netxen_resource_notice_types_t;

/*
 * CARD --> HOST
 * Message from card to host to inform about fate of connection
 */
typedef struct {
	host_peg_msg_hdr_t      hdr;
	__uint32_t              card_error;	/* zero means no error */
        __uint32_t              rsvd;           /* 64 bit alignment */
} connect_complete_t;

/*
 * HOST --> CARD
 *
 * REquest to reset the statistics collection on the card.
 */
typedef struct {
	host_peg_msg_hdr_t      hdr;
	__uint32_t              port;
} reset_stats_request_t;

/*
 * HOST --> CARD
 *
 * Request to DMA the statistics to a certain location.
 */
typedef struct {
	host_peg_msg_hdr_t	hdr;
	__uint64_t              dma_to_addr;
	__uint32_t              dma_size;
        __uint16_t              ring_ctx;
        __uint16_t              rsvd;           /* 64 bit alignment */
} get_stats_request_t;

/*
 * CARD --> HOST
 *
 * Sent in response to get_stats_request.
 */
typedef struct {
        host_peg_msg_hdr_t      hdr;
        __uint32_t              ret_code;
        __uint32_t              rsvd;           /* 64 bit alignment */
} get_stats_response_t;


#ifdef  PEGNET_NIC
/*
 * This is the data structure used to send a message from the host to the
 * peg/card.
 */
typedef union PREALIGN(64) {
        unm_msg_t               qmsg;
        pegnet_cmd_desc_t       cmd_desc;
        rxbuf_msg_t             rxbuf_msg;
        txbuf_msg_t             txbuf_msg;
} POSTALIGN(64) h2p_msg_t;

#endif  /* PEGNET_NIC */

/*
 * Accessor macros. Use these macros to exchange data between the host
 * and the PEGs. Even though the host & the PEG share the same
 * endianness, it is preferable to go thru these macros, so that one
 * day, it may be easier to migrate to powerpc or some other processor.
 *
 * The conventions are as follows:
 *
 * Each name consists of 5 parts. Each of these parts is separated by
 * an underscore. These parts are:
 *	1.  where is this macro to be used. This can be either peg or host.
 *	2.  what to do - encode, get, or assign. encode is used to
 *	    conver the bits to the target representation.
 *	    get is used to extract the bits from the sender to the
 *	    receiver's representation. assign not only does a get, but
 *	    also places the value in the given variable.
 *	3.  type - which type is it for. It may be either a builtin type,
 *	    or an user defined type.
 *	4.  from / for - Just for decriptive purposes. Helps clarify the
 *	    macro name.
 *	5.  the target of (4) above. Again, this is mostly for descriptive
 *	    purposes.
 *
 * Feel free to modify this convention as necessary. If you do so, please
 * update the comment and explain the modifications.
 *
 */


#define unm_p2h_uint16(VALUE)           ((__uint16_t)VALUE)
#define unm_p2h_uint32(VALUE)           ((__uint32_t)VALUE)
#define unm_p2h_uint64(VALUE)           ((__uint64_t)VALUE)

#define nx_h2p_uint16(VALUE)		((__uint16_t)VALUE)
#define nx_h2p_uint32(VALUE)		((__uint32_t)VALUE)
#define nx_h2p_uint64(VALUE)		((__uint64_t)VALUE)


#define peg_encode_qid_for_host(field, value) \
		(*((__uint32_t *)&(field))) = (*((__uint32_t *)&(value)))

#define host_assign_qid_from_peg(field, value) \
		(*((__uint32_t *)&(field))) = (*((__uint32_t *)&(value)))

#define host_encode_uint32_for_peg(value) \
		((__uint32_t)(value))

#define peg_encode_uint16_for_host(value) \
		((__uint16_t)(value))

#define peg_encode_uint32_for_host(value) \
		((__uint32_t)(value))

#define peg_encode_uint64_for_host(value) \
		((__uint64_t)(value))

#define host_get_uint32_from_peg(value) \
		((__uint32_t)(value))

#define peg_get_uint32_from_host(value) \
		((__uint32_t)(value))

#define peg_get_uint64_from_host(value) \
        ((__uint64_t)(value))

#define host_encode_pointer_for_peg(ptr) \
		((__uint32_t)(ptr))

#define peg_encode_pointer_for_host(ptr) \
		((__uint32_t)((__uint64_t)(ptr)))

#define host_encode_uint16_for_peg(val) \
                ((__uint16_t)(val))

/* Macros to pass host physical addresses between the host and peg */
/* Probably not necessary for P2 */
#define peg_get_hostPtr_from_host(ptr)  \
		(((__uint64_t)(ptr))|0x8000000000000000ULL)

#define peg_encode_hostPtr_for_host(ptr)  \
		(((__uint64_t)(ptr))&(~0x8000000000000000ULL))

#define host_encode_uint64_for_peg(ptr) \
		(((__uint64_t)(ptr)))

#endif /* HOST_INTERFACE */
