/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef __KERNEL_COMPATIBILITY_H__
#define __KERNEL_COMPATIBILITY_H__

#include <linux/version.h>
#include <net/sock.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
#define	NX_KC_ROUTE_DEVICE(RT)	((RT)->u.dst.dev)
#else
#define	NX_KC_ROUTE_DEVICE(RT)	(((RT)->idev)->dev)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
extern kmem_cache_t *nx_sk_cachep;
#endif

/* 2.4 - 2.6 compatibility macros*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

#define SK_ERR 		sk_err
#define SK_ERR_SOFT     sk_err_soft
#define SK_STATE_CHANGE sk_state_change
#define SK_PROT		sk_prot
#define SK_PROTOCOL	sk_protocol
#define SK_PROTINFO	sk_protinfo
#define SK_DESTRUCT	sk_destruct
#define SK_FAMILY	sk_family
#define SK_FLAGS	sk_flags
#define SK_DATA_READY   sk_data_ready
#define SK_RCVTIMEO     sk_rcvtimeo
#define SK_SNDTIMEO     sk_sndtimeo
#define SK_STATE        sk_state
#define SK_SLEEP        sk_sleep
#define SPORT(s)	(inet_sk(s))->num
#define SK_LINGER_FLAG(SK)      sock_flag((SK), SOCK_LINGER)
#define SK_LINGERTIME   sk_lingertime
#define SK_REUSE	sk_reuse
#define SK_SOCKET       sk_socket
#define SK_RCVBUF      sk_rcvbuf
#define SK_INET(s)      (inet_sk(s))

#else   /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */

#define SK_ERR 		err
#define SK_ERR_SOFT     err_soft
#define SK_STATE_CHANGE state_change
#define SK_PROT		prot
#define SK_PROTOCOL	protocol
#define SK_PROTINFO	user_data
#define SK_DESTRUCT	destruct
#define SK_FAMILY	family
#define SK_FLAGS	dead
#define SK_DATA_READY   data_ready
#define SK_RCVTIMEO     rcvtimeo
#define SK_SNDTIMEO     sndtimeo
#define SK_STATE        state
#define SK_SLEEP        sleep
#define SPORT(s)	s->num
#define SK_LINGER_FLAG(SK)      ((SK)->linger)
#define SK_LINGERTIME   lingertime
#define SK_REUSE	reuse
#define SK_SOCKET       socket
#define SK_RCVBUF      rcvbuf
#define SK_INET(s)      s

#endif  /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */

#define NOFC 0
#define FC1  1
#define FC2  2
#define FC3  3
#define FC4  4
#define FC5  5

/*
 * sk_alloc - varieties taken care here.
 */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)) || ((LINUX_VERSION_CODE == KERNEL_VERSION(2,6,11)) && (FEDORA == FC4))

#define	SK_ALLOC(FAMILY, PRIORITY, PROTO, ZERO_IT)		\
	sk_alloc((FAMILY), (PRIORITY), (PROTO), (ZERO_IT))

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)

#define	SK_ALLOC(FAMILY, PRIORITY, PROTO, ZERO_IT)			\
	sk_alloc((FAMILY), (PRIORITY),					\
		 ((ZERO_IT) > 1) ? (PROTO)->slab_obj_size : (ZERO_IT),	\
		 (PROTO)->slab)

#else

#define	SK_ALLOC(FAMILY, PRIORITY, PROTO, ZERO_IT)	\
 	sk_alloc((FAMILY), (PRIORITY), sizeof(struct tcp_sock), nx_sk_cachep)

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
#define BIND_HASH(SK)   tcp_sk(SK)->bind_hash
#else
#define BIND_HASH(SK)   inet_csk(SK)->icsk_bind_hash
#endif

#if defined(CONFIG_X86_64)
#define PAGE_KERNEL_FLAG  PAGE_KERNEL_EXEC
#else 
#define PAGE_KERNEL_FLAG  PAGE_KERNEL
#endif


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)) || ((LINUX_VERSION_CODE == KERNEL_VERSION(2,6,11)) && (FEDORA == FC4))
#define	NX_KC_DST_MTU(DST)	dst_mtu((DST))
#else
#define	NX_KC_DST_MTU(DST)	dst_pmtu((DST))
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

typedef struct work_struct	nx_kc_work_queue_t;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
#define	NX_KC_INIT_WORK(TASK, CB_FUNC, OBJ)	INIT_WORK((TASK), (CB_FUNC))
#else
#define	NX_KC_INIT_WORK(TASK, CB_FUNC, OBJ)	\
	INIT_WORK((TASK), (CB_FUNC), (OBJ))
#endif
#define	NX_KC_SCHEDULE_WORK(TASK)		schedule_work((TASK))
#define	NX_KC_SCHEDULE_DELAYED_WORK(TASK, DELAY)	\
	schedule_delayed_work((TASK), (DELAY))

#else

typedef struct tq_struct	nx_kc_work_queue_t;
#define	NX_KC_INIT_WORK(TASK, CB_FUNC, OBJ)	\
	INIT_TQUEUE((TASK), (CB_FUNC), (OBJ))
#define	NX_KC_SCHEDULE_WORK(TASK)	schedule_task((TASK))

#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#define PM_MESSAGE_T pm_message_t
#else
#define PM_MESSAGE_T u32
#endif
#endif  /* __KERNEL_COMPATIBILITY_H__ */
