/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Message structures for communication to/from, and within, our chip.
 */
#ifndef __MESSAGE_H
#define __MESSAGE_H

/*
 * A single word in a message.
 */
typedef __uint64_t	unm_msgword_t;

/*
 * We can use raw format (as the hardware defines it) or unpacked
 * (normal C struct's) for processing at the endpoints of a message,
 * but all transmission and/or reception must be done in the raw format.
 */
typedef union {
    struct {
	unm_msgword_t
		dst_minor:18,	/* Minor queue number (0-256K) */
		dst_subq:1,	/* Use minor number (1) or not (0) */
		dst_major:4,	/* Major queue number (0-15) */
		dst_side:1,	/* Which side of chip: net(0)/md(1) */
		dst_c2cport:4,	/* Which C2C port to use (0 for local) */
		src_minor:18,	/* Minor queue number (0-256K) */
		src_subq:1,	/* Use minor number (1) or not (0) */
		src_major:4,	/* Major queue number (0-15) */
		src_side:1,	/* Which side of chip: net(0)/md(1) */
		src_c2cport:4,	/* Which C2C port to use (0 for local) */
		rsvd:2,
		type:6;		/* Overall type of message (0-64) */
    } ;
    __uint64_t  word ;
} unm_msg_hdr_t;
typedef struct {
	int	type;		/* Overall type of message (0-64) */
	int	src_c2cport;	/* Which C2C port to use (0 for local) */
	int	src_side;	/* Which side of chip: net(0)/md(1) */
	int	src_major;	/* Major queue number (0-15) */
	int	src_subq;	/* Use minor number (1) or not (0) */
	int	src_minor;	/* Minor queue number (0-256K) */
	int	dst_c2cport;	/* Which C2C port to use (0 for local) */
	int	dst_side;	/* Which side of chip: net(0)/md(1) */
	int	dst_major;	/* Major queue number (0-15) */
	int	dst_subq;	/* Use minor number (1) or not (0) */
	int	dst_minor;	/* Minor queue number (0-256K) */
} unm_msg_hdr_easy_t;

/*
 * A convenience structure for the body of a message.
 */
typedef struct {
	unm_msgword_t values[7];/* just the payload of the message */
} unm_msg_body_t;

/*
 * A convenience structure a message as a whole (in hardware native format).
 */
typedef struct {
	unm_msg_hdr_t	hdr;
	unm_msg_body_t	body;
} unm_msg_t;

typedef struct {
        unsigned int    major:4,
                        subq:1,
                        minor:18;
} nx_qaddr_t;

#define	UNM_MESSAGE_SIZE	64	/* number bytes in a message */

/*
 * Convenience macro to fill in the src/dst of a message.
 */
#define	UNM_MSGQ_ADDR(HDR,WHICH,S,M,U,N)	\
	((HDR).WHICH##_c2cport = 0,	(HDR).WHICH##_side = (S),	\
	 (HDR).WHICH##_major = (M),	(HDR).WHICH##_subq = (U),	\
	 (HDR).WHICH##_minor = (N))

#endif /* __MESSAGE_H */
