/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * (C) Copyright Universal Network Machines 2004 - 2006
 */

#ifndef	_NETXEN_LINUX_IOCTL_H_
#define	_NETXEN_LINUX_IOCTL_H_

#include <hostInterface.h>

#define SIOC_NETXEN_OFFLOAD          (SIOCPROTOPRIVATE + 6)

/*
 * Information on the TCP offload rule to add or remove.
 */
#define NETXEN_MAX_APP_NAME_LENGTH      50
#define NETXEN_MAX_DEV_NAME_LENGTH      10
typedef struct
{
        __uint16_t      type;
        __uint8_t	zero_copy;  /* 1 means zero_copy enabled;
				     * 0  means disabled */
	__uint8_t	family;	    /* 4 for IPv4 and 6 for IPv6 */
        __uint16_t      local_port;
	__uint16_t	remote_port;
	ip_addr_t	local_ip;
	ip_addr_t	remote_ip;
        char            app_name[NETXEN_MAX_APP_NAME_LENGTH];
} netxen_offload_rule_t;

typedef struct
{
        __uint16_t      reset_flag;
        char		dev_name[NETXEN_MAX_DEV_NAME_LENGTH];
} netxen_offload_dev_info_t;

/*
 * The list of valid rule types are.
 */
enum netxen_offload_rule_type_s
{
        NETXEN_OFFLOAD_RULE_INVALID = 0,
        NETXEN_OFFLOAD_RULE_BY_TUPLE,
        NETXEN_OFFLOAD_RULE_BY_APP_NAME,
        NETXEN_OFFLOAD_RULE_LAST
};

/*
 * Ioctl hdr part of every ioctl used to pass commands to the driver and
 * get data from the driver.
 */
typedef struct
{
        __uint64_t      cmd : 16,
                        cnt : 16,
                        ret_code : 32;
} netxen_offload_ioctl_hdr_t;

/*
 * Ioctl used to pass commands to the driver and get data from the driver.
 */
typedef struct
{
        netxen_offload_ioctl_hdr_t              hdr;
        union {
                netxen_offload_rule_t           rule;
                netxen_offload_rule_t           rulelist[1];
                netxen_offload_dev_info_t	dev;
        } u;
} netxen_offload_ioctl_t;

/*
 * The list of valid commands.
 */
enum netxen_offload_ioctl_cmds_s
{
        NETXEN_OFFLOAD_IOCTL_GET_RULE_CNT = 0,
        NETXEN_OFFLOAD_IOCTL_GET_RULES,
        NETXEN_OFFLOAD_IOCTL_ADD_RULE,
        NETXEN_OFFLOAD_IOCTL_REMOVE_RULE,
        NETXEN_OFFLOAD_IOCTL_LAST
};


#define NETXEN_IOCTL_ERR_RULE_EXISTS            1001
#define NETXEN_IOCTL_ERR_RULE_NOT_FOUND         1002
#define NETXEN_IOCTL_ERR_UNSUPPORTED_RULE       1003

#endif  /* _NETXEN_LINUX_IOCTL_H_ */
