/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/* Header file with definitions for both host and Phantom */

/* $Id: nic_cmn.h,v 1.97 2007/02/06 04:18:48 datton Exp $ */

#ifndef UNM_NIC_CMN_H
#define UNM_NIC_CMN_H

#include "unm_nic_config.h"
#include "unm_compiler_defs.h"

#define IP_ALIGNMENT_BYTES                2  /* make ip aligned on 16byteaddr */
#define MAX_RX_BUFFER_LENGTH           1760
#ifdef __VMKERNEL_MODULE__
#define MAX_RX_JUMBO_BUFFER_LENGTH     MAX_RX_BUFFER_LENGTH
#define MAX_RX_LRO_BUFFER_LENGTH       MAX_RX_BUFFER_LENGTH
#else
#define MAX_RX_JUMBO_BUFFER_LENGTH     9046
#define MAX_RX_LRO_BUFFER_LENGTH       ((8*1024)-512)
#endif
#define RX_DMA_MAP_LEN               (MAX_RX_BUFFER_LENGTH-IP_ALIGNMENT_BYTES)
#define RX_JUMBO_DMA_MAP_LEN   (MAX_RX_JUMBO_BUFFER_LENGTH-IP_ALIGNMENT_BYTES)
#define RX_LRO_DMA_MAP_LEN       (MAX_RX_LRO_BUFFER_LENGTH-IP_ALIGNMENT_BYTES)

/* Opcodes to be used with the commands */
#define    TX_ETHER_PKT    0x01
/* The following opcodes are for IP checksum    */
#define    TX_TCP_PKT      0x02
#define    TX_UDP_PKT      0x03
#define    TX_IP_PKT       0x04
#define    TX_TCP_LSO      0x05
#define    TX_IPSEC       0x06
#define    TX_IPSEC_CMD       0x07

/* The following opcodes are for internal consumption. */
#define UNM_CONTROL_OP  0x10
#define PEGNET_REQUEST  0x11

#ifdef UNM_RSS
#define       RSS_CNTRL_CMD       0x20
#endif
#ifdef __VMKERNEL_MODULE__
#define    MAX_NUM_CARDS   1
#define    MAX_NUM_PORTS   1          /* Deprecated. donot use this */
#else
#define    MAX_NUM_CARDS   4
#define    MAX_NUM_PORTS   4          /* Deprecated. donot use this */
#endif
#define    PORT1           0
#define    PORT2           1
#define    PORT3           2
#define    PORT4           3

#define    DESC_CHAIN      0xFF    /* descriptor command continuation */

#define MAX_BUFFERS_PER_CMD 32

/*
 * Following are the states of the Phantom. Phantom will set them and
 * Host will read to check if the fields are correct.
 */
#define PHAN_INITIALIZE_START         0xff00
#define PHAN_INITIALIZE_FAILED        0xffff
#define PHAN_INITIALIZE_COMPLETE      0xff01

/* Host writes the following to notify that it has done the init-handshake */
#define PHAN_INITIALIZE_ACK           0xf00f

/* Following defines will be used in the status descriptor */
#define TX_ETHER_PKT_COMPLETE  0xB  /* same for both commands */

#define NUM_RCV_DESC_RINGS     3 /* No of Rcv Descriptor contexts */

/* descriptor types */
#define RCV_DESC_NORMAL        0x01
#define RCV_DESC_JUMBO         0x02
#define RCV_DESC_LRO           0x04
#define RCV_DESC_NORMAL_CTXID     0
#define RCV_DESC_JUMBO_CTXID      1
#define RCV_DESC_LRO_CTXID        2

#define RCV_DESC_TYPE(ID) \
       ((ID == RCV_DESC_JUMBO_CTXID) ? RCV_DESC_JUMBO :  \
         ((ID == RCV_DESC_LRO_CTXID) ? RCV_DESC_LRO : (RCV_DESC_NORMAL)))

#define RCV_DESC_TYPE_NAME(ID) \
       ((ID == RCV_DESC_JUMBO_CTXID)  ? "Jumbo"  :  \
        (ID == RCV_DESC_LRO_CTXID)    ? "LRO"    :  \
        (ID == RCV_DESC_NORMAL_CTXID) ? "Normal" : "Unknown")

#define MAX_CMD_DESCRIPTORS       4096
#define MAX_CMD_DESCRIPTORS_HOST  (MAX_CMD_DESCRIPTORS / 4)
#define MAX_RCV_DESCRIPTORS       32768
#define MAX_RCV_DESCRIPTORS_1G    (MAX_RCV_DESCRIPTORS / 4)
#define MAX_JUMBO_RCV_DESCRIPTORS 1024
#define MAX_LRO_RCV_DESCRIPTORS   64 /*  XXX */
#define MAX_RCVSTATUS_DESCRIPTORS MAX_RCV_DESCRIPTORS
#define MAX_JUMBO_RCV_DESC        MAX_JUMBO_RCV_DESCRIPTORS
#define MAX_LRO_RCV_DESC          MAX_LRO_RCV_DESCRIPTORS
#define MAX_RCV_DESC              MAX_RCV_DESCRIPTORS
#define MAX_RCVSTATUS_DESC        MAX_RCV_DESCRIPTORS
#define NUM_RCV_DESC              (MAX_RCV_DESC+                  \
                                   MAX_JUMBO_RCV_DESCRIPTORS+     \
                                   MAX_LRO_RCV_DESCRIPTORS)

#define MAX_EPG_DESCRIPTORS       (MAX_CMD_DESCRIPTORS  * 8)

#define MIN_TX_COUNT              4096
#define MIN_RX_COUNT              4096

#define MAX_FRAME_SIZE            0x10000    /* 64K MAX size  for LSO */

#define PHAN_PEG_RCV_INITIALIZED      0xff01
#define PHAN_PEG_RCV_START_INITIALIZE 0xff00

#define get_next_index(index,length)  ((((index)  + 1) == length)?0:(index) +1)


#define get_index_range(index,length,count)  ((((index) + (count)) >= length)? \
                       (((index)  + (count))-(length)):((index) + (count)))

/*
 * Set the time bases for timer sets.
 */
#define UNM_FLOW_TICKS_PER_SEC    2048
#define UNM_FLOW_TO_TV_SHIFT_SEC  11
#define UNM_FLOW_TO_TV_SHIFT_USEC 9
#define UNM_FLOW_TICK_USEC   (1000000ULL/UNM_FLOW_TICKS_PER_SEC)
#define UNM_GLOBAL_TICKS_PER_SEC  (4*UNM_FLOW_TICKS_PER_SEC)
#define UNM_GLOBAL_TICK_USEC (1000000ULL/UNM_GLOBAL_TICKS_PER_SEC)


/*
 * Following data structures describe the descriptors that will be used.
 * Added fileds of tcpHdrSize and ipHdrSize, The driver needs to do it only when
 * we are doing LSO (above the 1500 size packet) only.
 * This is an overhead but we need it. Let me know if you have questions.
 */

/* the size of reference handle been changed to 16 bits to pass the MSS fields
   for the LSO packet */

#define FLAGS_CHECKSUM_ENABLED       0x01
#define FLAGS_LSO_ENABLED            0x02
#define FLAGS_IPSEC_SA_ADD           0x04
#define FLAGS_IPSEC_SA_DELETE        0x08
#define FLAGS_VLAN_TAGGED            0x10

#if UNM_CONF_PROCESSOR==UNM_CONF_X86

#ifndef U64
typedef uint64_t U64;
typedef uint32_t U32;
typedef uint16_t U16;
typedef uint8_t  U8;
#endif

#endif

#define NUM_SUPPORTED_RINGSETS 4

#define       MAX_RING_CTX 4
#define UNM_CTX_SIGNATURE       0xdee0
#define UNM_CTX_RESET           0xbad0
#define UNM_CTX_D3_RESET        0xacc0

//define opcode for ctx_msg
#define  RX_PRODUCER 0
#define  RX_PRODUCER_JUMBO 1
#define  RX_PRODUCER_LRO 2
#define  TX_PRODUCER 3
#define  UPDATE_STATUS_CONSUMER 4
#define  RESET_CTX  5

#define  NUM_DB_CODE 16

#define UNM_RCV_PRODUCER(ringid)        (ringid)
#define UNM_CMD_PRODUCER                TX_PRODUCER
#define UNM_RCV_STATUS_CONSUMER         UPDATE_STATUS_CONSUMER

typedef struct __msg
{
    __uint32_t  PegId:2,   // 0x2 for tx and 01 for rx.
                privId:1,  // must be 1
                Count:15,  // for doorbell
                CtxId:10,  // Ctx_id
                Opcode:4;  //opcode
}ctx_msg,CTX_MSG,*PCTX_MSG;

typedef struct __int_msg
{
    __uint32_t  Count:18, // INT
                ConsumerIdx:10,
                CtxId:4; // Ctx_id
}int_msg,INT_MSG,*PINT_MSG;

/* For use in CRB_MPORT_MODE */
#define MPORT_SINGLE_FUNCTION_MODE 0x1111
#define MPORT_MULTI_FUNCTION_MODE 0x2222

//#ifdef UNM_MPORT
typedef struct _RcvContext
{
    __uint64_t     RcvRingAddrLo:32,
                   RcvRingAddrHi:32;
    __uint32_t     RcvRingSize;
    __uint32_t     Rsrv;
}RcvContext;

typedef struct PREALIGN(64) _RingContext
{

    /* one command ring */
    __uint64_t         CMD_CONSUMER_OFFSET;
    __uint64_t         CmdRingAddrLo:32,
                       CmdRingAddrHi:32;
    __uint32_t         CmdRingSize;
    __uint32_t         Rsrv;

    /* three receive rings */
    RcvContext         RcvContext[3];

    /* one status ring */
    __uint64_t         StsRingAddrLo:32,
                       StsRingAddrHi:32;
    __uint32_t         StsRingSize;

    __uint32_t         CtxId;

    __uint64_t         D3_STATE_REGISTER;
    __uint64_t         DummyDmaAddrLo:32,
                       DummyDmaAddrHi:32;

} POSTALIGN(64) RingContext,RING_CTX,*PRING_CTX;

/*
typedef struct _RING_CONTEXT
{
    RING_CTX    RingContext;
    __uint64_t  GlobalProducerIndex;
    //__uint64_t  cmdConsumerIndex;
    //__uint64_t  ConsumerIndex;
    __uint64_t  ConsumerCount;
    __uint64_t  cmdProducerIndex;
}RING_CONTEXT,*PRING_CONTEXT,ring_context_t;
*/

#ifdef UNM_RSS
/*
 * RSS_SreInfo{} has the information for SRE to calculate the hash value
 * Will be passed by the host=> as part of comd descriptor...
 */

#if UNM_CONF_PROCESSOR==UNM_CONF_X86
typedef struct _RSS_SreInfo {
    U32     HashKeySize;
    U32    HashInformation;
    char key[40];
}RSS_SreInfo;
#endif


// The following Descriptor is used to send RSS commands to the
// PEG.... to be do the SRE registers..
//
typedef struct PREALIGN(64) _rssCmdDesc
{

    // To keep the opcode at the same location as
    // the cmdDescType0, we will have to breakup the key into
    // 2 areas.... Dont like it but for now will do... FSL
    //

#if UNM_CONF_PROCESSOR==UNM_CONF_X86
    U8              Key0[16];

    U64     HashMethod:32 ,
            HashKeySize:8,
        Unused:   16,
            opcode:8;

    U8          Key1[24];
    U64          Unused1;
    U64          Unused2;
#else

    unm_msgword_t    Key0[2];
    unm_msgword_t     HashMethod:32 ,
                HashKeySize:8,
            Unused:   16,
                opcode:8;

    unm_msgword_t    Key1[3];
    unm_msgword_t    Unused1;
    unm_msgword_t    Unused2;

#endif

} POSTALIGN(64) rssCmdDesc_t;


#endif // UNM_RSS


typedef struct PREALIGN(64) cmdDescType0
{
   union {
	struct {
   	    __uint64_t  tcpHdrOffset:8,     // For LSO only
        		ipHdrOffset:8,      // For LSO only
        		flags:7,            // as defined above
        		opcode:6,           // This location/size must not change...
        		Unused:3,           //
        		numOfBuffers:8,     // total number of segments (buffers
                        		    // for this packet. (could be more than 4)

        		totalLength:24;     // Total size of the packet
	};
	__uint64_t	word0;
    };

    union {
        struct {
            __uint32_t AddrLowPart2;
            __uint32_t AddrHighPart2;
        };
        __uint64_t AddrBuffer2;
	__uint64_t	word1;
    };

   union {
	struct {
    	     __uint64_t referenceHandle:16, // changed to U16 to add mss
        		mss:16,             // passed by NDIS_PACKET for LSO
        		port:4,
        		//rsvd1:4,
        		ctx_id:4,
        		totalHdrLength:8,   // LSO only : MAC+IP+TCP Hdr size
        		connID:16;          // IPSec offoad only
	};
	__uint64_t	word2;
    };

    union {
        struct {
            __uint32_t AddrLowPart3;
            __uint32_t AddrHighPart3;
        };
        __uint64_t AddrBuffer3;
	__uint64_t	word3;
    };

    union {
        struct {
            __uint32_t AddrLowPart1;
            __uint32_t AddrHighPart1;
        };
        __uint64_t AddrBuffer1;
	__uint64_t	word4;
    };

    union {
        struct {
    	    __uint64_t buffer1Length:16,
        		buffer2Length:16,
        		buffer3Length:16,
        		buffer4Length:16;
        };
	__uint64_t	word5;
    };

    union {
        struct {
            __uint32_t AddrLowPart4;
            __uint32_t AddrHighPart4;
        };
        __uint64_t AddrBuffer4;
	__uint64_t	word6;
    };

    __uint64_t unused;

} POSTALIGN(64) cmdDescType0_t;

#if     defined(PEGNET_NIC)
/*
 * This is the command descriptor used for moving qmessages from the host to
 * the NIC TX path and then as queue messages to the queue as specified by
 * the destination fields.
 */
typedef struct PREALIGN(64) pegnet_cmd_desc
{
    union {
        struct {
            __uint64_t      dst_minor:18,   /* The pegnet queue Minor this
                                                 * message is destined to. */
                            dst_subq:1,     /* The destination subq */
                            dst_major:4,    /* The destination Major */
                            opcode:6,       /* The opcode (PEGNET_REQUEST).
                                             * It has to be in this
                                             * location and cannot be moved. */
                            qmsg_length:3,  /* The length of the queue msg
                                             * body */
                            msginfo:24,     /* This is not used */
                            rsvd:2,
                            qmsg_type:6;    /* The actual qmsg type this
                                             * represents. This will be set in
                                             * the qmsg to the destination. */
        };
        __uint64_t      hdr;
    };

    __uint64_t      body[7];

} POSTALIGN(64) pegnet_cmd_desc_t;

#endif  /* PEGNET_NIC */

#define RCV_DESC_HOST_FREE      0x01
#define RCV_DESC_HOST_POSTED    0x00
#define RCV_DESC_HOST_PEG_USED  0x04

/* Note: sizeof(rcvDesc) should always be a mutliple of 2 */
typedef struct rcvDesc
{
    __uint64_t referenceHandle:16,
               flags:16,
               bufferLength:32;        // allocated buffer length (usually 2K)
    __uint64_t AddrBuffer;
}  rcvDesc_t;


/* opcode field in statusDesc_t */
#if     !defined(PEGNET_NIC) && !defined(UNM_RSS)
#define RCV_NIC_PKT                                     (0xA)
#endif

/* for status field in statusDesc_t */
#define STATUS_NEED_CKSUM                                 (1)
#define STATUS_CKSUM_OK                                   (2)

/* owner bits of statusDesc_t */
#define STATUS_OWNER_HOST                              (1ULL)
#define STATUS_OWNER_PHANTOM                           (2ULL)
#define HOST_STATUS_DESC          ((STATUS_OWNER_HOST) << 48)
#define PHANTOM_STATUS_DESC    ((STATUS_OWNER_PHANTOM) << 48)

#define UNM_PROT_IP                                       (1)
#define UNM_PROT_UNKNOWN                                  (0)

/* LRO specific bits of statusDesc_t */
#define LRO_LAST_FRAG                                     (1)
#define LRO_NORMAL_FRAG                                   (0)
#define LRO_LAST_FRAG_DESC              ((LRO_LAST_FRAG)<<63)
#define LRO_NORMAL_FRAG_DESC          ((LRO_NORMAL_FRAG)<<63)


#if !defined(UNM_RSS) && !defined(PEGNET_NIC)

/* Note: sizeof(statusDesc) should always be a mutliple of 2 */
typedef struct PREALIGN(8) statusDesc
{
    __uint64_t port:4,                 /* initially to be used but noe now */
               status:4,               /* completion status may not have use */
               type:4,                 /* type/index of descriptor ring */
               opcode:4,               /* Completion of receive or transmit */
               totalLength:16,         /* NIC mode...no use yet */
               referenceHandle:16,     /* handle for the associated packet */
               owner:2,                /* Owner of the descriptor */
               prot:4,                 /* protocol type for skb (Linux) */
               reserved:2,
               lro_count:7,            /* if lro, indicates frag sequence */
               last_lro_frag:1;        /* if lro, indicates last fragment */
} POSTALIGN(8) statusDesc_t;

#else // UNM_RSS
// for RSS the status desc. is 16 bytes and not 8 bytes...

#define UNM_QUEUE_MSG_DESC_COUNT        5
typedef struct PREALIGN(16) statusDesc
{
    union {
        struct {
                __uint64_t      port:4,         /* initially to be used
                                                         * but noe now */
                                        status:4,       /* completion status
                                                         * may not have use */
                                        type:4,         /* type/index of
                                                         * descriptor ring */
                                        totalLength:16, /* NIC mode...no use
                                                         * yet */
                                        referenceHandle:16, /* handle for the
                                                             * associated
                                                             * packet */
                                        prot:4,         /* Pkt protocol */
                                        UnUsed:5,
                                        descCnt:3,      /* This indicates the
                                                         * num of descriptors
                                                         * part of this
                                                         * descriptor chain. */
                                        owner:2,
                                        opcode:6;

                        __uint32_t    HashValue;
                        __uint16_t  HashType:8,
                                     MsgType:8;
            union {
                struct {
                    __uint16_t    UnUsed1:8,
                            lro_count:7, /* if lro,
                                indicates frag
                                sequence */
                            last_lro_frag:1;  /* if lro,
                                indicates last
                                frag */
                };
                /*
                 * Used in sending SYNs to the offloaded
                 * listener.
                 */
                __uint16_t listener_id;
            };
        };
        __uint64_t    body[2];
    };

} POSTALIGN(16) statusDesc_t;

#endif // UNM_RSS

#define STATUS_OWNER_NAME(sd) \
        (((sd)->owner == STATUS_OWNER_HOST) ? "Host" : "Phantom")

#ifdef UNM_IPSECOFFLOAD

#define MAX_IPSEC_SAS            1024
#define RECEIVE_IPSEC_SA_BASE           0x8000

/*
* IPSEC related structures and defines
*/

// Values for DIrFlag in the ipsec_sa_t structure below:
#define UNM_IPSEC_SA_DIR_INBOUND    1
#define UNM_IPSEC_SA_DIR_OUTBOUND    2

// Values for Operation Field below:
#define UNM_IPSEC_SA_AUTHENTICATE    1
#define UNM_IPSEC_SA_ENDECRYPT        2

// COnfidential Algorithm Types:
#define UNM_IPSEC_CONF_NONE    0    // NULL encryption?
#define UNM_IPSEC_CONF_DES    1
#define UNM_IPSEC_CONF_RESERVED    2
#define UNM_IPSEC_CONF_3DES    3

// Integrity algorithm (AH) types:
#define UNM_IPSEC_INTEG_NONE    0
#define UNM_IPSEC_INTEG_MD5    1
#define UNM_IPSEC_INTEG_SHA1    2

#define UNM_PROTOCOL_OFFSET    0x9    // from ip header begin, in bytes
#define UNM_PKT_TYPE_AH        0x33
#define UNM_PKT_TYPE_ESP    0x32


#define UNM_AHOUTPUT_LENGTH     12      // 96 bits of output for MD5/SHA1 algorithms
#define UNM_DES_ICV_LENGTH      8       // 8 bytes (64 bits) of ICV value for each block of DES_CBC at the begin of ESP payload

#if UNM_CONF_PROCESSOR==UNM_CONF_X86

typedef struct PREALIGN(512) s_ipsec_sa {
    U32    SrcAddr;
    U32 SrcMask;
    U32 DestAddr;
    U32 DestMask;
    U32    Protocol:8,
        DirFlag:4,
        IntegCtxInit:2,
        ConfCtxInit:2,
        No_of_keys:8,
        Operation:8;
    U32    IntegAlg:8,
        IntegKeyLen:8,
        ConfAlg:8,
        ConfAlgKeyLen:8;
    U32    SAIndex;
    U32    SPI_Id;
    U64    Key1[124];
} POSTALIGN(512) unm_ipsec_sa_t;

#else

typedef struct PREALIGN(512) s_ipsec_sa {
    unm_msgword_t    SrcAddr:32,
                    SrcMask:32;
    unm_msgword_t    DestAddr:32,
                    DestMask:32;
    unm_msgword_t    Protocol:8,
                    DirFlag:4,
                    IntegCtxInit:2,
                    ConfCtxInit:2,
                    No_of_keys:8,
                    Operation:8,
                    IntegAlg:8,
                    IntegKeyLen:8,
                    ConfAlg:8,
                    ConfAlgKeyLen:8;
    unm_msgword_t    SAIndex:32,
                    SPI_Id:32;
    unm_msgword_t    Key1[124];        // to round up to 1K of structure
} POSTALIGN(512) unm_ipsec_sa_t;

#endif //NOT-X86

// Other common header formats that may be needed

typedef struct _unm_ip_header_s {
    U32    HdrVer:8,
        diffser:8,
        TotalLength:16;
    U32     ipId:16,
        flagfrag:16;
    U32    TTL:8,
        Protocol:8,
        Chksum:16;
    U32    srcaddr;
    U32    destaddr;
} unm_ip_header_t;

typedef struct _unm_ah_header_s {
    U32    NextProto:8,
        length:8,
        reserved:16;
    U32    SPI;
    U32    seqno;
    U16    ICV;
    U16    ICV1;
    U16    ICV2;
    U16    ICV3;
    U16    ICV4;
    U16    ICV5;
} unm_ah_header_t;

typedef struct _unm_esp_hdr_s {
    U32 SPI;
    U32 seqno;
} unm_esp_hdr_t;

#endif //UNM_IPSECOFFLOAD

/*
 * Defines for various loop counts. These determine the behaviour of the
 * system. The classic tradeoff between latency and throughput.
 */

/*
 * MAX_DMA_LOOPCOUNT : After how many interations do we start the dma for
 * the status descriptors.
 */
#define MAX_DMA_LOOPCOUNT    (32)

/*
 * MAX_TX_DMA_LOOP_COUNT : After how many interations do we start the dma for
 * the command descriptors.
 */
#define MAX_TX_DMA_LOOP_COUNT    1000

/*
 * MAX_RCV_BUFS : Max number Rx packets that can be buffered before DMA/INT
 */
#define MAX_RCV_BUFS           (4096)

/*
* XXX;shouldnt be exposed in nic_cmn.h
 * DMA_MAX_RCV_BUFS : Max number Rx packets that can be buffered before DMA
 */
#define DMA_MAX_RCV_BUFS        (4096)

/*
* XXX;shouldnt be exposed in nic_cmn.h
 * MAX_DMA_ENTRIES : Max number Rx dma entries can be in dma list
 */
#define MAX_DMA_ENTRIES         (4096)


/*
 * MAX_INTR_LOOPCOUNT : After how many iterations do we interrupt the
 * host ?
 */
#define MAX_INTR_LOOPCOUNT   (1024)

/*
 * XMIT_LOOP_THRESHOLD : How many times do we spin before we process the
 * transmit buffers.
 */
#define XMIT_LOOP_THRESHOLD        0x20

/*
 * XMIT_DESC_THRESHOLD : How many descriptors pending before we process
 * the descriptors.
 */
#define XMIT_DESC_THRESHOLD    0x4

/*
 * TX_DMA_THRESHOLD : When do we start the dma of the command descriptors.
 * We need these number of command descriptors, or we need to exceed the
 * loop count.   P1 only.
 */
#define TX_DMA_THRESHOLD        16

/*
 * TASKLET_DELAY : Ratio of interrupts to be ignored before scheduling
 * the tasklet. We want to process the receive interrupts quickly, but can
 * delay processing the tx interrupts.
 */
#define TASKLET_DELAY            4

#if defined(UNM_IP_FILTER)
/*
 * Commands. Must match the definitions in nic/Linux/include/unm_nic_ioctl.h
 */
enum {
    UNM_IP_FILTER_CLEAR = 1,
    UNM_IP_FILTER_ADD,
    UNM_IP_FILTER_DEL,
    UNM_IP_FILTER_SHOW
};

#define MAX_FILTER_ENTRIES        16

typedef struct {
    __int32_t count;
    __uint32_t ip_addr[15];
} unm_ip_filter_t;
#endif /* UNM_IP_FILTER */

enum {
    UNM_RCV_PEG_0 = 0,
    UNM_RCV_PEG_1
};
#endif
