/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef NIC_PHAN_REG_H
#define NIC_PHAN_REG_H

#include "unm_compiler_defs.h"

#define NIC_CRB_BASE               UNM_CAM_RAM(0x200)
#define UNM_NIC_REG(X)             (NIC_CRB_BASE+(X))

#define CRB_PHAN_CNTRL_LO_OFFSET    UNM_NIC_REG(0x00)
#define CRB_PHAN_CNTRL_HI_OFFSET    UNM_NIC_REG(0x04)
#define CRB_CMD_PRODUCER_OFFSET     UNM_NIC_REG(0x08)
#define CRB_CMD_CONSUMER_OFFSET     UNM_NIC_REG(0x0c)
#define CRB_PAUSE_ADDR_LO           UNM_NIC_REG(0x10) /* C0 EPG BUG  */
#define CRB_PAUSE_ADDR_HI           UNM_NIC_REG(0x14)
#define CRB_HOST_CMD_ADDR_HI        UNM_NIC_REG(0x18) /* host add:cmd ring */
#define CRB_HOST_CMD_ADDR_LO        UNM_NIC_REG(0x1c)
#define CRB_CMD_INTR_LOOP           UNM_NIC_REG(0x20) /* 4 regs for perf */
#define CRB_CMD_DMA_LOOP            UNM_NIC_REG(0x24)
#define CRB_RCV_INTR_LOOP           UNM_NIC_REG(0x28)
#define CRB_RCV_DMA_LOOP            UNM_NIC_REG(0x2c)
#define CRB_ENABLE_TX_INTR          UNM_NIC_REG(0x30) /* phantom init status */
#define CRB_MMAP_ADDR_3             UNM_NIC_REG(0x34)
#define CRB_CMDPEG_CMDRING          UNM_NIC_REG(0x38)
#define CRB_HOST_DUMMY_BUF_ADDR_HI  UNM_NIC_REG(0x3c)
#define CRB_HOST_DUMMY_BUF_ADDR_LO  UNM_NIC_REG(0x40)
#define CRB_MMAP_ADDR_0             UNM_NIC_REG(0x44)
#define CRB_MMAP_ADDR_1             UNM_NIC_REG(0x48)
#define CRB_MMAP_ADDR_2             UNM_NIC_REG(0x4c)
#define CRB_CMDPEG_STATE            UNM_NIC_REG(0x50)
#define CRB_MMAP_SIZE_0             UNM_NIC_REG(0x54)
#define CRB_MMAP_SIZE_1             UNM_NIC_REG(0x58)
#define CRB_MMAP_SIZE_2             UNM_NIC_REG(0x5c)
#define CRB_MMAP_SIZE_3             UNM_NIC_REG(0x60)
#define CRB_GLOBAL_INT_COAL         UNM_NIC_REG(0x64) /* interrupt coalescing */
#define CRB_INT_COAL_MODE           UNM_NIC_REG(0x68)
#define CRB_MAX_RCV_BUFS            UNM_NIC_REG(0x6c)
#define CRB_TX_INT_THRESHOLD        UNM_NIC_REG(0x70)
#define CRB_RX_PKT_TIMER            UNM_NIC_REG(0x74)
#define CRB_TX_PKT_TIMER            UNM_NIC_REG(0x78)
#define CRB_RX_PKT_CNT              UNM_NIC_REG(0x7c)
#define CRB_RX_TMR_CNT              UNM_NIC_REG(0x80)
#define CRB_RX_LRO_TIMER            UNM_NIC_REG(0x84)
#define CRB_RX_LRO_MID_TIMER        UNM_NIC_REG(0x88)
#define CRB_DMA_MAX_RCV_BUFS        UNM_NIC_REG(0x8c)
#define CRB_MAX_DMA_ENTRIES         UNM_NIC_REG(0x90)
#define CRB_XG_STATE                UNM_NIC_REG(0x94) /* XG Link status */
#define CRB_AGENT_GO                UNM_NIC_REG(0x98) /* NIC pkt gen agent */
#define CRB_AGENT_TX_SIZE           UNM_NIC_REG(0x9c)
#define CRB_AGENT_TX_TYPE           UNM_NIC_REG(0xa0)
#define CRB_AGENT_TX_ADDR           UNM_NIC_REG(0xa4)
#define CRB_AGENT_TX_MSS            UNM_NIC_REG(0xa8)
#define CRB_TX_STATE                UNM_NIC_REG(0xac) /* Debug -performance */
#define CRB_TX_COUNT                UNM_NIC_REG(0xb0)
#define CRB_RX_STATE                UNM_NIC_REG(0xb4)
#define CRB_RX_PERF_DEBUG_1         UNM_NIC_REG(0xb8)
#define CRB_RX_LRO_CONTROL          UNM_NIC_REG(0xbc) /* LRO On/OFF */
#define CRB_RX_LRO_START_NUM        UNM_NIC_REG(0xc0)
#define CRB_MPORT_MODE              UNM_NIC_REG(0xc4) /* Multiport Mode */
#define CRB_CMD_RING_SIZE           UNM_NIC_REG(0xc8)
#define CRB_INT_VECTOR              UNM_NIC_REG(0xd4)
#define CRB_CTX_RESET               UNM_NIC_REG(0xd8)
#define CRB_HOST_STS_PROD           UNM_NIC_REG(0xdc)
#define CRB_HOST_STS_CONS           UNM_NIC_REG(0xe0)
#define CRB_PEG_CMD_PROD            UNM_NIC_REG(0xe4)
#define CRB_PEG_CMD_CONS            UNM_NIC_REG(0xe8)
#define CRB_HOST_BUFFER_PROD        UNM_NIC_REG(0xec)
#define CRB_HOST_BUFFER_CONS        UNM_NIC_REG(0xf0)
#define CRB_JUMBO_BUFFER_PROD       UNM_NIC_REG(0xf4)
#define CRB_JUMBO_BUFFER_CONS       UNM_NIC_REG(0xf8)

/* used for ethtool tests */
#define CRB_SCRATCHPAD_TEST         UNM_NIC_REG(0x280)

/*
#define CRB_RCV_PRODUCER_OFFSET     UNM_NIC_REG(0x100)
#define CRB_RCV_CONSUMER_OFFSET     UNM_NIC_REG(0x104)
#define CRB_GLOBALRCV_RING          UNM_NIC_REG(0x108)
#define CRB_RCV_RING_SIZE           UNM_NIC_REG(0x10c)
#define CRB_RCV_PRODUCER_OFFSET     UNM_NIC_REG(0x110)
#define CRB_RCV_CONSUMER_OFFSET     UNM_NIC_REG(0x114)
#define CRB_GLOBALRCV_RING          UNM_NIC_REG(0x118)
#define CRB_RCV_RING_SIZE           UNM_NIC_REG(0x11c)
#define CRB_RCV_PRODUCER_OFFSET     UNM_NIC_REG(0x120)
#define CRB_RCV_CONSUMER_OFFSET     UNM_NIC_REG(0x124)
#define CRB_GLOBALRCV_RING          UNM_NIC_REG(0x128)
#define CRB_RCV_RING_SIZE           UNM_NIC_REG(0x12c)
#define CRB_RCV_PRODUCER_OFFSET_1   UNM_NIC_REG(0x130)
#define CRB_RCV_CONSUMER_OFFSET_1   UNM_NIC_REG(0x134)
#define CRB_GLOBALRCV_RING_1        UNM_NIC_REG(0x138)
#define CRB_RCV_RING_SIZE_1         UNM_NIC_REG(0x13c)
#define CRB_RCV_PRODUCER_OFFSET_1   UNM_NIC_REG(0x140)
#define CRB_RCV_CONSUMER_OFFSET_1   UNM_NIC_REG(0x144)
#define CRB_GLOBALRCV_RING_1        UNM_NIC_REG(0x148)
#define CRB_RCV_RING_SIZE_1         UNM_NIC_REG(0x14c)
#define CRB_RCV_PRODUCER_OFFSET_1   UNM_NIC_REG(0x150)
#define CRB_RCV_CONSUMER_OFFSET_1   UNM_NIC_REG(0x154)
#define CRB_GLOBALRCV_RING_1        UNM_NIC_REG(0x158)
#define CRB_RCV_RING_SIZE_1         UNM_NIC_REG(0x15c)
#define CRB_RCVSTATUS_RING          UNM_NIC_REG(0x160)
#define CRB_RCV_STATUS_PRODUCER     UNM_NIC_REG(0x164)
#define CRB_RCV_STATUS_CONSUMER     UNM_NIC_REG(0x168)
#define CRB_RCVPEG_STATE            UNM_NIC_REG(0x16c)
#define CRB_STATUS_RING_SIZE        UNM_NIC_REG(0x170)
#define CRB_MPORT_SIG_0_0           UNM_NIC_REG(0x188)
#define CRB_MPORT_SIG_0_1           UNM_NIC_REG(0x18c)
#define CRB_MPORT_SIG_1_0           UNM_NIC_REG(0x190)
#define CRB_MPORT_SIG_1_1           UNM_NIC_REG(0x194)
#define CRB_MPORT_SIG_2_0           UNM_NIC_REG(0x198)
#define CRB_MPORT_SIG_2_1           UNM_NIC_REG(0x1a0)
#define CRB_MPORT_SIG_3_0           UNM_NIC_REG(0x1a4)
#define CRB_MPORT_SIG_3_1           UNM_NIC_REG(0x1a8)
*/
#define CRB_CMD_PRODUCER_OFFSET_1   UNM_NIC_REG(0x1ac)
#define CRB_CMD_CONSUMER_OFFSET_1   UNM_NIC_REG(0x1b0)
#define CRB_TEMP_STATE              UNM_NIC_REG(0x1b4)
#define CRB_CMD_PRODUCER_OFFSET_2   UNM_NIC_REG(0x1b8)
#define CRB_CMD_CONSUMER_OFFSET_2   UNM_NIC_REG(0x1bc)

// 1c0 to 1cc used for signature reg
#define CRB_CMD_PRODUCER_OFFSET_3   UNM_NIC_REG(0x1d0)
#define CRB_CMD_CONSUMER_OFFSET_3   UNM_NIC_REG(0x1d4)
#define CRB_EPG_QUEUE_BUSY_COUNT    UNM_NIC_REG(0x200)

#define CRB_V2P_0                   UNM_NIC_REG(0x290)
#define CRB_V2P_1                   UNM_NIC_REG(0x294)
#define CRB_V2P_2                   UNM_NIC_REG(0x298)
#define CRB_V2P_3                   UNM_NIC_REG(0x29c)
#define CRB_V2P(port)               (CRB_V2P_0+((port)*4))
#define CRB_DRIVER_VERSION          UNM_NIC_REG(0x2a0)
	/*
	 * Driver must set the version number register as follows:
	 *	(major << 16) | (minor << 8) | (subminor)
	 */

/* last -> 0x2a0 */

/* Upper 16 bits of CRB_TEMP_STATE:temperature value. Lower 16 bits: state */
#define nx_get_temp_val(x)                  ((x) >> 16)
#define nx_get_temp_state(x)                ((x) & 0xffff)
#define nx_encode_temp(val, state)          (((val) << 16) | (state))

/*
 * Alternative method of getting to the MMAP addr & size.
 */
#define CRB_MMAP_ADDR(idx)    (CRB_MMAP_ADDR_0 + ((idx) << 2))
#define CRB_MMAP_SIZE(idx)    (CRB_MMAP_SIZE_0 + ((idx) << 2))

/* CRB registers per Rcv Descriptor ring */
typedef struct {
    U32 PREALIGN(512) CRB_RCV_PRODUCER_OFFSET POSTALIGN(512);
    U32 CRB_RCV_CONSUMER_OFFSET;
    U32 CRB_GLOBALRCV_RING;
    U32 CRB_RCV_RING_SIZE;
} unm_rcv_desc_crb_t;

/*
 * CRB registers used by the receive peg logic.
 */

typedef struct {
    unm_rcv_desc_crb_t rcv_desc_crb[NUM_RCV_DESC_RINGS];

    U32 CRB_RCVSTATUS_RING;
    U32 CRB_RCV_STATUS_PRODUCER;
    U32 CRB_RCV_STATUS_CONSUMER;
    U32 CRB_RCVPEG_STATE;
    U32 CRB_STATUS_RING_SIZE;
} unm_recv_crb_t;

#if defined(DEFINE_GLOBAL_RECV_CRB)
unm_recv_crb_t recv_crb_registers[] = {
     {
        //rcv_desc_crb:
        {
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x100),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x104),
            //CRB_GLOBALRCV_RING:
            UNM_NIC_REG(0x108),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x10c),
            },
            /* Jumbo frames */
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x110),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x114),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x118),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x11c),
            },
            /* LRO */
            {
            //CRB_RCV_PRODUCER_OFFSET:
			UNM_NIC_REG(0x120),
            //CRB_RCV_CONSUMER_OFFSET:
			UNM_NIC_REG(0x124),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x128),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x12c),
            }
        },
        //CRB_RCVSTATUS_RING:
        UNM_NIC_REG(0x130),
        //CRB_RCV_STATUS_PRODUCER:
        UNM_NIC_REG(0x134),
        //CRB_RCV_STATUS_CONSUMER:
        UNM_NIC_REG(0x138),
        //CRB_RCVPEG_STATE:
        UNM_NIC_REG(0x13c),
        //CRB_STATUS_RING_SIZE
        UNM_NIC_REG(0x140),
    },
    {
        {
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x144),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x148),
            //CRB_GLOBALRCV_RING:
            UNM_NIC_REG(0x14c),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x150),
            },
            /* Jumbo frames */
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x154),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x158),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x15c),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x160),
            },
            /* LRO */
            {
            //CRB_RCV_PRODUCER_OFFSET:
			UNM_NIC_REG(0x164),
            //CRB_RCV_CONSUMER_OFFSET:
			UNM_NIC_REG(0x168),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x16c),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x170),
            }
        },
        //CRB_RCVSTATUS_RING:
        UNM_NIC_REG(0x174),
        //CRB_RCV_STATUS_PRODUCER:
        UNM_NIC_REG(0x178),
        //CRB_RCV_STATUS_CONSUMER:
        UNM_NIC_REG(0x17c),
        //CRB_RCVPEG_STATE:
        UNM_NIC_REG(0x180),
        //CRB_STATUS_RING_SIZE
        UNM_NIC_REG(0x184),
    },

	{
		//instance 3
        {
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x1d8),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x1dc),
            //CRB_GLOBALRCV_RING:
            UNM_NIC_REG(0x1f0),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x1f4),
            },
            /* Jumbo frames */
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x1f8),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x1fc),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x200),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x204),
            },
            /* LRO */
            {
            //CRB_RCV_PRODUCER_OFFSET:
			UNM_NIC_REG(0x208),
            //CRB_RCV_CONSUMER_OFFSET:
			UNM_NIC_REG(0x20c),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x210),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x214),
            }
        },
        //CRB_RCVSTATUS_RING:
        UNM_NIC_REG(0x218),
        //CRB_RCV_STATUS_PRODUCER:
        UNM_NIC_REG(0x21c),
        //CRB_RCV_STATUS_CONSUMER:
        UNM_NIC_REG(0x220),
        //CRB_RCVPEG_STATE:
        UNM_NIC_REG(0x224),
        //CRB_STATUS_RING_SIZE
        UNM_NIC_REG(0x228),
    },

	{
		//instance 4
        {
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x22c),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x230),
            //CRB_GLOBALRCV_RING:
            UNM_NIC_REG(0x234),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x238),
            },
            /* Jumbo frames */
            {
            //CRB_RCV_PRODUCER_OFFSET:
            UNM_NIC_REG(0x23c),
            //CRB_RCV_CONSUMER_OFFSET:
            UNM_NIC_REG(0x240),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x244),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x248),
            },
            /* LRO */
            {
            //CRB_RCV_PRODUCER_OFFSET:
			UNM_NIC_REG(0x24c),
            //CRB_RCV_CONSUMER_OFFSET:
			UNM_NIC_REG(0x250),
            //CRB_GLOBALRCV_RING:
			UNM_NIC_REG(0x254),
            //CRB_RCV_RING_SIZE:
            UNM_NIC_REG(0x258),
            }
        },
        //CRB_RCVSTATUS_RING:
        UNM_NIC_REG(0x25c),
        //CRB_RCV_STATUS_PRODUCER:
        UNM_NIC_REG(0x260),
        //CRB_RCV_STATUS_CONSUMER:
        UNM_NIC_REG(0x264),
        //CRB_RCVPEG_STATE:
        UNM_NIC_REG(0x268),
        //CRB_STATUS_RING_SIZE
        UNM_NIC_REG(0x26c),
    },
};

#define NUM_RCV_INSTANCES \
                     (sizeof(recv_crb_registers)/sizeof(recv_crb_registers[0]))

U64 ctx_addr_sig_regs[][3] = {
    {UNM_NIC_REG(0x188), UNM_NIC_REG(0x18c), UNM_NIC_REG(0x1c0)},
    {UNM_NIC_REG(0x190), UNM_NIC_REG(0x194), UNM_NIC_REG(0x1c4)},
    {UNM_NIC_REG(0x198), UNM_NIC_REG(0x19c), UNM_NIC_REG(0x1c8)},
    {UNM_NIC_REG(0x1a0), UNM_NIC_REG(0x1a4), UNM_NIC_REG(0x1cc)}
};
#define CRB_CTX_ADDR_REG_LO(FUNC_ID)         (ctx_addr_sig_regs[FUNC_ID][0])
#define CRB_CTX_ADDR_REG_HI(FUNC_ID)         (ctx_addr_sig_regs[FUNC_ID][2])
#define CRB_CTX_SIGNATURE_REG(FUNC_ID)       (ctx_addr_sig_regs[FUNC_ID][1])
#else

extern unm_recv_crb_t recv_crb_registers[];
extern U64 ctx_addr_sig_regs[][3];

#endif /* DEFINE_GLOBAL_RECEIVE_CRB */

#define lower32(x)             ( (__uint32_t)((x) & 0xffffffff) )
#define upper32(x)             ( (__uint32_t)(((unsigned long long)(x) >> 32) & 0xffffffff))

/*
 * Temperature control.
 */
enum {
    NX_TEMP_NORMAL = 0x1,	/* Normal operating range */
    NX_TEMP_WARN,		/* Sound alert, temperature getting high */
    NX_TEMP_PANIC		/* Fatal error, hardware has shut down. */
};

#define D3_CRB_REG_FUN2         (UNM_PCIX_PS_REG(0x2084))
#define D3_CRB_REG_FUN3         (UNM_PCIX_PS_REG(0x3084))
#endif /* NIC_PHAN_REG_H */
