/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * The Linux API exported by the NetXen NIC driver to the TOE driver.
 */
#ifndef NX_NIC_LINUX_TNIC_API_H
#define NX_NIC_LINUX_TNIC_API_H

#include <linux/skbuff.h>
#include "nic_cmn.h"
#include "message.h"

typedef struct nx_tnic_adapter_s		nx_tnic_adapter_t;

typedef struct {
	int	(*nx_syn_handler)(struct net_device * , __uint16_t,
				  struct sk_buff *);
	void	(*nx_msg_handler)(struct net_device *, unm_msg_t *);
} nx_tnic_callback_t;

/*
 * Checks if the specified device is a NetXen device.
 */
extern int nx_nic_is_netxen_device(struct net_device *netdev);

/*
 * Register Tnic driver with Nic driver.
 *
 * Return:
 *	0	- If registered successfully.
 *	1	- If already registered.
 *	-1	- If port or adapter is not set.
 */
extern int nx_nic_register_tnic(struct net_device *netdev,
				nx_tnic_adapter_t *data,
				nx_tnic_callback_t *cb);

/*
 * Unregisters a known TOE adapter.
 */
extern void nx_nic_unregister_tnic(struct net_device *netdev);

/*
 * Return the registered TOE adapter structure.
 *
 * Returns:
 *	The Toe adapter if already registered else returns NULL.
 */
extern nx_tnic_adapter_t *nx_nic_get_tnic_adapter(struct net_device *netdev);

/*
 * Gets the port number of the device on a netxen adaptor if it is a netxen
 * device.
 *
 * Parameters:
 *	dev	- The device for which the port is requested.
 *
 * Returns:
 *	-1	- If not netxen device.
 *	port number on the adapter if it is netxen device.
 */
extern int nx_nic_get_device_port(struct net_device *netdev);

/*
 * Gets the ring context used by the device to talk to the card.
 *
 * Parameters:
 *	dev	- The device for which the ring context is requested.
 *
 * Returns:
 *	-EINVAL	- If not netxen device.
 *	Ring context number if it is netxen device.
 */
extern int nx_nic_get_device_ring_ctx(struct net_device *netdev);

/*
 * Send a TOE message to the card.
 */
extern int nx_nic_send_tnic_msg(struct net_device *netdev,
			 pegnet_cmd_desc_t *cmd_desc, int nr_elements);

/*
 * Gets the pointer to pci device if the device maps to a netxen device.
 */
extern struct pci_dev *nx_nic_get_pcidev(struct net_device *dev);

/*
 * Compare adapter ids for two NetXen devices.
 */
extern int nx_nic_cmp_adapter_id(struct net_device *dev1,
				 struct net_device *dev2);

/* 
 * Gets the tnic version number from the card. 
 */
typedef struct {
        __uint32_t major;
        __uint32_t minor;
        __uint32_t sub;
} nic_version_t;
extern void nx_nic_get_tnic_version_number(struct net_device *netdev,
                        nic_version_t *tnic_version);
extern void nx_nic_get_nic_version_number(struct net_device *netdev,
                        nic_version_t *nic_version);

#endif	/* NX_NIC_LINUX_TNIC_API_H */
