/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
** File: pegnet_inc.h
** Description:
**  This header contains definitions used by the pegnet code.
*/

#ifndef __PEGNET_INC_H
#define __PEGNET_INC_H

#include <linux/socket.h> /* struct sockaddr */
#include <linux/net.h> /* struct socket */
#include <linux/in.h> /* struct sockaddr_in */
#include <linux/skbuff.h> /*struct sk_buff */


#include <pegnet_config.h>

/*
** Messages from the host to the card's network stack go through
** the following [major, minor] pegos queue.
*/
#define PEGNET_H2C_MAJOR        UNM_MSGQ_SQM_TYPE1
#define PEGNET_H2C_MINOR        PEGOS_PEGNET_CORE

#if !defined(PEGNET_QUEUE_ADDR_T)
#define PEGNET_QUEUE_ADDR_T
/*
** Identity (address) of the queue.
** NB: See also linux/net.h
*/
typedef union pegnet_queue_addr_s {
    struct {
        __uint32_t
#if defined(__LITTLE_ENDIAN_BITFIELD)
            major: 4,      /* Major queue number (0-15) */
            minor: 18,      /* Minor queue number (0-256K) */
            side:1;
#elif defined(__BIG_ENDIAN_BITFIELD)
            side:1,
            minor: 18,      /* Minor queue number (0-256K) */
            major: 4;      /* Major queue number (0-15) */
#else
#error  "Please fix <asm/byteorder.h>"
#endif
    };
    __uint32_t word;
} pegnet_queue_addr_t;
#endif

#if defined(__LITTLE_ENDIAN_BITFIELD)
#define null_queue(qid) (((*((__uint32_t *)&(qid))) & 0x7fffff) == 0)
#elif defined(__BIG_ENDIAN_BITFIELD)
#define null_queue(qid) (((*((__uint32_t *)&(qid))) & 0xfffffe00) == 0)
#else
#error  "Please fix <asm/byteorder.h>"
#endif

#define PEGNET_HOST_ADDR (1ULL<<63)

#if (UNM_X_HARDWARE==UNM_X_ASIC)
/* Native build */
#define PEGNET_XPTR_ADDR(A) ((void*)(((__uint64_t)(A)) & ~(1ULL<<63)))
#define PEGNET_IS_NONLOCAL_ADDR(A) (((__uint64_t)(A)) & (1ULL<<63))
#define PEGNET_MARK_NONLOCAL(A) (((__uint64_t)(A)) | (1ULL<<63))
#elif (UNM_X_HARDWARE==UNM_X_FPGA)
/* x86 Sim */
#define PEGNET_IS_NONLOCAL_ADDR(A) ((((__uint32_t)(A)) & (1<<31)))
#define PEGNET_XPTR_ADDR(A) ((void*)(((__uint32_t)(A)) & ~(1<<31)))
#define PEGNET_MARK_NONLOCAL(A) (((__uint32_t)(A)) | (1ULL<<31))
#elif (UNM_X_HARDWARE==UNM_X_SIMULATED)
/* pure simulation */
#define PEGNET_IS_NONLOCAL_ADDR(A) 0
#define PEGNET_XPTR_ADDR(A) (A)
#define PEGNET_MARK_NONLOCAL(A) (A)
#else
#error Unknown hardware platform
#endif

/*
 * Well - known object queues
 */
#define PEGOS_PEGNET_CORE	(UNM_MSGQ_TYPE1_FIRST_AVAIL+3)

#define PEGOS_LD_Q0		16 /*  64k -> queue 0 */
#define PEGOS_PAGE_SIZE		(1 << PEGOS_LD_Q0)

#endif  /* __PEGNET_INC_H */
