/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef __PEGNET_MSGTYPES_H_
#define __PEGNET_MSGTYPES_H_

/*
 * ====================== MESSAGE TYPE DEFINITIONS ======================
 * Define types for all messages used in the system.  Each h/w block will
 * be given a different "type" to use in messages it generates so that we
 * know just from the message header what we have just dequeued.
 * ====================== MESSAGE TYPE DEFINITIONS ======================
 */
#define UNM_MSGTYPE_UNKNOWN             0x00    /* message type not valid */
#define UNM_MSGTYPE_POINTER             0x01    /* generic head/tail/len linklist fmt */
#define UNM_MSGTYPE_SRE_FREE            0x02    /* unused pkt buffer (linklist fmt) */
#define UNM_MSGTYPE_SRE_ARP             0x03    /* an ARP packet (linklist fmt) */
#define UNM_MSGTYPE_SRE_ICMP            0x04    /* an ICMP packet (linklist fmt) */
#define UNM_MSGTYPE_SRE_IPQ             0x05    /* a TCP/IP packet (linklist fmt) */
#define UNM_MSGTYPE_SRE_EXP             0x06    /* an alien packet (linklist fmt) */
#define UNM_MSGTYPE_SRE_FLOW            0x07    /* a generic packet (linklist fmt) */
#define UNM_MSGTYPE_LOOPBACK            0x08    /* another generic packet (linklist fmt) */
#define UNM_MSGTYPE_TIMER_SETUP         0x09    /* set/change a timer fmt */
#define UNM_MSGTYPE_TIMER_EVENT         0x0a    /* which timer expired fmt */
#define UNM_MSGTYPE_EPG_WORK            0x0b    /* EPG packet xmit request fmt */

#define UNM_MSGTYPE_CLEAR_TUPLE		0x0c    /* req/resp to clear CAM tuple */

#define UNM_MSGTYPE_PEGOS_MEM           0x0d    /* PegOS memory mgmt */
#define UNM_MSGTYPE_PEGOS_FUNC          0x0e    /* PegOS msg. passing */

#define UNM_MSGTYPE_CONNECT_REQUEST     0x0f    /* Connect request */
#define UNM_MSGTYPE_RELEASE_REQUEST     0x10    /* Release request from host */
#define UNM_MSGTYPE_NIC_LRO             0x11    /* NIC's LRO packets */
#define UNM_MSGTYPE_SRE_FRAG            0x12    /* IP fragements go here */
#define UNM_MSGTYPE_ASYNC_DMA_COMPLETE	0x13   	/* Statistics, etc */
#define UNM_MSGTYPE_I2Q_INTERRUPT       0x14    /* interrupt from I2Q */
#define UNM_MSGTYPE_NIC_FREE_SREBUF     0x15    /* NIC: Free SRE buffers*/
#define UNM_MSGTYPE_SINGLE_DMA_REQUEST  0x16    /* Single request to DMA object */
#define UNM_MSGTYPE_TNIC_RESET_FW	0x17    /* Reset firmware related
						   resources */
#define UNM_MSGTYPE_EPG_PROXY           0x18    /* robs workaround */
#define UNM_MSGTYPE_MCAST_RPC           0x19    /* Multicast Request */
#define UNM_MSGTYPE_INITIATE_OFFLOAD    0x1a    /* CHIMNEY */
#define UNM_MSGTYPE_RESET_STATISTICS    0x1b    /* Reset network statistics */
#define UNM_MSGTYPE_GET_STATISTICS 	0x1c    /* DMA network statistics to host buffer */
#define UNM_MSGTYPE_INVALIDATE_OFFLOAD  0x1d    /* CHIMNEY */
#define UNM_MSGTYPE_QUERY_OFFLOAD       0x1e    /* CHIMNEY */
#define UNM_MSGTYPE_NIC_LRO_DUMMY       0x1f    /* NIC LRO delete-tuple msg */
#define UNM_MSGTYPE_TCP_DMA_COMP        0x20    /* SRE DMA completion */
#define UNM_MSGTYPE_DISCONNECT_OFFLOAD  0x21    /* CHIMNEY */
#define UNM_MSGTYPE_ROUTE_REQUEST	0x22    /* Add, Delete, Modify route */
#define UNM_MSGTYPE_DO_3WAY_HANDSHAKE   0x23    /* Sent from the listener PEG
					 	   to the target PEG. This  
						   triggers the SYN-ACK packet
						   to be sent.*/
#define UNM_MSGTYPE_CHILD_RELEASE       0x24    /* Trigger cleanup on child */
#define UNM_MSGTYPE_INSTANCE_SOCKET     0x25    /* invocation for instance socket routine */
#define UNM_MSGTYPE_FORWARD_OFFLOAD     0x26   /* CHIMNEY */
#define UNM_MSGTYPE_EPG_AGGR_COMP       0x27    /* Invocation of aggregate epg. */

#define UNM_MSGTYPE_CHIMNEY_SWITCH_RXBUF        0x28    /* Switch rxbuf message
                                                           local->global or
                                                           global->local. Need
                                                           to make it a
                                                           HOST_REQUEST message
                                                           once changes for WSD
                                                           are checked in. */


#define UNM_MSGTYPE_HOST_REQUEST        0x29    /* Request from host to card */
#define UNM_MSGTYPE_CARD_RESPONSE       0x2a    /* Response from card to host*/
#define UNM_MSGTYPE_RESOURCE_ALARM      0x2b	/* Free to be used */

#ifdef UNM_HWBUG_DMA_HANG 
#define UNM_MSGTYPE_DMA_WD_DMA_COMP     0x2c
#endif

#define UNM_MSGTYPE_RXED_SYN_OFFLOAD	0x2d	/* This is sent from the HOST
						   to the listener socket on
						   the CARD for starting the
						   offload. */
#define UNM_MSGTYPE_DO_NOT_USE          0x2e    /* Do not use this msg type */
#define UNM_MSGTYPE_TERMINATE_OFFLOAD   0x2f   /* CHIMNEY */
#define UNM_MSGTYPE_UPDATE_OFFLOAD      0x30   /* CHIMNEY */
#define UNM_MSGTYPE_CREATE_REQUEST      0x31    /* Create a new socket. */
#define UNM_MSGTYPE_SERVICE_ERROR       0x32	/* unmapped packet */
#define UNM_MSGTYPE_DATA_ERROR          0x33	/* unmapped packet */
#define UNM_MSGTYPE_RXBUFFER            0x34	/* Post empty rx data buffers
                                                   to card and send full rx
                                                   data buffers to host. */
#define UNM_MSGTYPE_TXBUFFER_SWITCH	0x35    /* Used in zero copy. To switch
						   a buffer in the TCB from
						   zero copy to regular copy
						   buffer. */
#define UNM_MSGTYPE_RXBUFFER_DMA_RDY    0x36    /* Receive dma to host is
						   complete.  */
#define UNM_MSGTYPE_GLOBAL_RXBUFF       0x37    /* Post empty global rx data
						   buffers to card. */
#define UNM_MSGTYPE_SENDMSG             0x38    /* Post filled send buffer to
						   card. */
#define UNM_MSGTYPE_SRE_L1R_ERROR       0x39    /* checksum error */
#define UNM_MSGTYPE_DMA_UNIT            0x3a    /* DMA completions in the NIC
						   path. */

#define UNM_MSGTYPE_DMA_FORWARD_COMPLETE           0x3b   /* CHIMNEY */


#define UNM_MSGTYPE_SREMPA              0x3c    /* Incoming MPA Packets */
#define UNM_MSGTYPE_CONNMOVE            0x3d    /* Connection migration */

#define UNM_MSGTYPE_NIC_SYN_OFFLOAD	0x3e    /* This is a SYN packet that
						   matched a Listener socket
						   and is a candidate for
						   offload. Sent from the
						   card to the host. */
#define UNM_MSGTYPE_NIC_RXPKT_DESC      0x3f    /* This is not a real queue
                                                   message. It is an opcode to
                                                   the host to specify that
                                                   this is a true RX
                                                   descriptor. Queue messages
                                                   to host from card arrive as
                                                   descriptors. The host uses
                                                   this value to figure out if
                                                   it is a queue message or a
                                                   RX descriptor. */

/* Card-to-host msg queue. (currently unused) */
#define PEGNET_HOST_RX_MAJOR    0
#define PEGNET_HOST_RX_SUBQ     1
#define PEGNET_HOST_RX_MINOR    4
#define PEGNET_HOST_RX_Q_LEN    SQG_TYPE0_UNUSED_Q_LEN


#if defined(UNM_REDIRECTOR)
#define PEGNET_C2H_MAJOR        PEGNET_HOST_RX_MAJOR
#define PEGNET_C2H_SUBQ         PEGNET_HOST_RX_SUBQ
#define PEGNET_C2H_MINOR        PEGNET_HOST_RX_MINOR
#define PEGNET_C2H_SIDE         0
#endif

#define PEGNET_TCP_CONN_MAJOR   12
#define PEGNET_TCP_CONN_MINOR   pegos_qid_requests()

#define UNM_MSGQ_HDR_LO32(HDR)          ((void *)&(((__uint32_t *)HDR)[0]))
#define UNM_MSGQ_HDR_HI32(HDR)          ((void *)&(((__uint32_t *)HDR)[1]))

/* A generic macro to create a  queue header */
#if UNM_CONF_PROCESSOR == UNM_CONF_X86  /* we're on an x86 host */
#define UNM_QHDR_CREATE(HDR,MAJOR,MINOR,SUBQ,TYPE) \
        (memset((char *)&(HDR), 0, sizeof(HDR)),        \
         UNM_MSGQ_ADDR(HDR,dst,0,MAJOR,SUBQ,MINOR),        \
         (HDR).type=UNM_MSGTYPE_##TYPE)

#else

/* let the preprocessor do the dirty work for us */
#define UNM_QHDR_CREATE(HDR, MAJOR, MINOR, SUBQ, TYPE)		\
do {								\
	(HDR).word = ((__uint64_t)(MAJOR) << 19 ) |		\
		     (__uint64_t)(SUBQ) << 18     |		\
		     (__uint64_t)(MINOR)          |		\
		     ((__uint64_t)(UNM_MSGTYPE_##TYPE) << 58);	\
} while (0)
#endif

/* Legacy header creation */
#define UNM_MSGQ_HDR(HDR,MAJOR,MINOR,TYPE) \
    UNM_QHDR_CREATE((HDR),(MAJOR),(MINOR),1,TYPE)

/* Receive global buffer queue.  Currently unused. */
#define UNM_MSGQ_GLOBAL_RXBUFF_MAJ      0
#define UNM_MSGQ_GLOBAL_RXBUFF_MIN      13
#define UNM_MSGQ_GLOBAL_RXBUFF_Q_LEN    512
#define UNM_MSGQ_GLOBAL_RXBUFF_HDR(HDR) UNM_MSGQ_HDR(HDR,                   \
                                                UNM_MSGQ_GLOBAL_RXBUFF_MAJ, \
                                                UNM_MSGQ_GLOBAL_RXBUFF_MIN, \
                                                GLOBAL_RXBUFF)

/* We will reserve 10 type 1 queues for HW blocks */
#define UNM_MSGQ_TYPE1_FIRST_AVAIL 10

/* This is the major queue number that type 1 queues hide behind */
#define UNM_MSGQ_SQM_TYPE1   12



#endif /* __PEGNET_MSGTYPES_H_ */
