/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * (C) Copyright Universal Network Machines 2004 - 2005
 * This code is distributed under FreeBSD license
 *
 * This file contains data structure definitions for Verb interfaces
 * provided by the RNIC.
 */

#ifndef	_RDMA_STRUCTS_H_
#define	_RDMA_STRUCTS_H_

#include <rdma_types.h>

/*
 * Number of Pegs that service RDMA. Peg Id's run from 0 to NUMPEGS - 1.
 */
#define	VERB_CPU_ID		0
#define	NUMPEGS			4

/*
 * Total number of constructs of different types.
 */
#define	CQS_PERPEG_SHIFT	12
#define	CQS_PERPEG		(1 << CQS_PERPEG_SHIFT)
#define	NUM_CQS			(CQS_PERPEG * NUMPEGS)
#define	QPS_PERPEG_SHIFT	10
#define	QPS_PERPEG		(1 << QPS_PERPEG_SHIFT)
#define	NUM_QPS			(QPS_PERPEG * NUMPEGS)
#define	STAGS_PERPEG_SHIFT	14
#define	STAGS_PERPEG		(1 << STAGS_PERPEG_SHIFT)
#define	MAX_STAGS		(STAGS_PERPEG * NUMPEGS)

/*
 * Limits on various types of inputs.
 */
#define	TXMAXSGL		4
#define	RECVMAXSGL		4
#define MAXPBLSIZE		8189

/*
 * Door bell page size.
 */
#define	UNM_DBELL_SHIFT		12

/*
 * Size of each DMA/PCI address.
 */
#define	RDMA_ADDR_SHIFT	3

/*
 * Location of registers used for host <-> rnic communication.
 */
#define	NIC_CRB_BASE_PORT1	UNM_CAM_RAM(0x200)
#define	UNM_RDMA_REG(X, peg)	(NIC_CRB_BASE_PORT1+(X)+(peg << 4))
#define	CRB_DMA_REQ(peg)	UNM_RDMA_REG(0x300, peg)

#define	CDB_SHIFT	9
#define	RWQE_SHIFT	6
#define	SWQE_SHIFT	8
#define	CQE_SHIFT	4

/*
 * Number of bits in doorbell message to hold number of WQE's posted.
 */
#define	DBELL_WQE_BITS	6

/*
 * Number of per QP slots to hold latency payload on host.
 */
#define	NUM_LATENCY_SLOTS	4
#define	MAX_LATENCY_LENGTH	484

/*
 * Verb opcodes.
 */
#define UNM_NOP			0x0
#define UNM_MR_REGISTER		0x1
#define	UNM_DEALLOC_STAG	0x2
#define	UNM_CQ_ALLOC		0x3
#define	UNM_CQ_FREE		0x4
#define	UNM_QP_ALLOC		0x5
#define	UNM_QP_FREE		0x6
#define UNM_RNIC_Q		0x7
#define	UNM_QP_MOD		0x8
#define	UNM_ALLOC_STAG		0x9
#define	UNM_ALLOC_MW		0xA

/*
 * Priviledged command posting format.
 */
typedef struct priv_db_post {
	u32			db_pegid:2,
				db_privid:1,
				db_data:29;
} priv_db_post_t;

/*
 * Verb command posting format.
 */
typedef struct rdma_verb_post {
	u32			verb_pegid:2,
				verb_privid:1,
				verb_opcode:5,
				verb_cmdlen:24;
} rdma_verb_post_t;

/*
 * Allocate NonShared MR Stag.
 */
typedef struct rdma_stag_attr {
	rdma_pdid_t		stag_pdid;
	rdma_stgid_t		stag_id;
	rdma_pbl_size_t		stag_reqpblsize;
	rdma_bool_t		stag_remacc;
} rdma_stag_attr_t;

typedef struct rdma_stag_result {
	rdma_pbl_size_t		stag_pblsize;
} rdma_stag_res_t;

#define	STAG0		0

/*
 * Register NonShared MR.
 */
typedef struct rdma_reg_attr {
	rdma_pdid_t		reg_pdid;		/* 4 bytes */
	rdma_stgid_t		reg_stgid;		/* 4 bytes */
	u8			reg_remacc:1,
				reg_allowmw:1,
				reg_vabasedto:1,
				reg_indirect:1,
				reg_unused:4;		/* 1 byte */		
	rdma_stag_key_t		reg_stagkey;		/* 1 byte */
	u8			reg_accbits;		/* 1 byte */
	u8			reg_pageshift;		/* 1 byte */	
	rdma_fbo_t		reg_fbo;		/* 4 bytes */
	rdma_len_t		reg_regionlen;		/* 8 bytes */
	rdma_tag_offset_t	reg_vabase;		/* 8 bytes */
	rdma_bufsize_t		reg_pbentryshift;	/* 4 bytes */
	rdma_pbl_size_t		reg_alistlen;		/* 4 bytes */
} rdma_reg_attr_t;

/*
 * Reregister NonShared MR.
 */
typedef struct rdma_rereg_attr {
	rdma_pdid_t		rereg_pdid;		/* 4 bytes */
	rdma_bool_t		rereg_remacc;		/* 1 byte */
	rdma_stag_t		rereg_stagidx;		/* 4 bytes */
	rdma_stag_key_t		rereg_stagkey;		/* 1 byte */
	rdma_bool_t		rereg_allowmw;		/* 1 byte */
	u8			rereg_accbits;		/* 1 byte */
	rdma_len_t		rereg_regionlen;	/* 8 bytes */
	rdma_fbo_t		rereg_fbo;		/* 4 bytes */
	rdma_bool_t		rereg_vabasedto;	/* 1 byte */
	rdma_tag_offset_t	rereg_vabase;		/* 8 bytes */
	rdma_bufsize_t		rereg_pbentryshift;	/* 4 bytes */
	rdma_pbl_size_t		rereg_alistlen;		/* 4 bytes */
	rdma_dmaaddr_t		rereg_alist[1];		/* 8 bytes TODO */
} rdma_rereg_attr_t;

/*
 * Register Shared MR.
 */
typedef struct rdma_regshm_attr {
	rdma_pdid_t		regshm_pdid;
	rdma_stgid_t		regshm_stgid;
	rdma_bool_t		regshm_remacc;
	rdma_stag_t		regshm_stagidx;
	rdma_stag_key_t		regshm_stagkey;
	rdma_bool_t		regshm_allowmw;
	u8			regshm_accbits;
	rdma_bool_t		regshm_vabasedto;
	rdma_tag_offset_t	regshm_vabase;
} rdma_regshm_attr_t;

/*
 * Query MR.
 */
typedef struct rdma_qmr_res {
	rdma_stag_key_t		qmr_stagkey;
	rdma_bool_t		qmr_remacc;
	rdma_pdid_t		qmr_pdid;
	rdma_bool_t		qmr_valid;
	rdma_bool_t		qmr_shared;
	rdma_pbl_size_t		qmr_pblsize;
	rdma_bool_t		qmr_allowmw;
	u8			qmr_accbits;
} rdma_qmr_res_t;

/*
 * Allocate MW.
 */
typedef struct rdma_almw_attr {
	rdma_pdid_t		almw_pdid;
	rdma_stgid_t		almw_stgid;
} rdma_almw_attr_t;

/*
 * Query MW.
 */
typedef struct rdma_qmw_res {
	rdma_stag_key_t		qmw_stagkey;
	rdma_bool_t		qmw_valid;
	rdma_pdid_t		qmw_pdid;
	u8			qmw_accbits;
} rdma_qmw_res_t;

/*
 * MR and MW access right bit position definition.
 */
#define	RDMA_REMOTE_READ	0x1
#define	RDMA_REMOTE_WRITE	0x2
#define	RDMA_LOCAL_READ		0x4
#define	RDMA_LOCAL_WRITE	0x8

/*
 * Structure of each CQE reported by RNIC to host.
 */
typedef struct rdma_cqe {
	unsigned int		cqe_hqpidx;		/* 4 bytes */
	u16			cqe_wqeindex;		/* 2 bytes */
	rdma_wcstatus_t		cqe_wcstatus;		/* 1 byte */
	u8			cqe_owner;		/* 1 byte */
	rdma_stag_t		cqe_invstag;		/* 4 bytes */
	int			cqe_numbytes;		/* 4 bytes */
} rdma_cqe_t;

/*
 * Lowest 4 bits in cqe_owner indicate WQE type (RDMA_SEND* etc). The next
 * bit indicates whether notification was generated or not. The next two
 * bits indicate how many latency slots are free for host to use.
 */
#define	CQE_NFLAG		0x10

/*
 * Host/RNIC interface for Create CQ.
 */
#define	RDMA_CQ_MAX_PAGES	16
typedef struct rdma_crcq_in {
	rdma_cqsize_t		crcq_size;			/*Byte 0-1*/
	rdma_cqid_t		crcq_cqid;			/*Byte 4-7*/
	u32			crcq_hidx;			/*Byte 8-11*/
	u32			crcq_pad;			/*Byte 12-15*/
	u64			crcq_dbaddr;			/*Byte 16-23*/
	u64			crcq_addr[RDMA_CQ_MAX_PAGES];	/*Byte 24-151*/
} rdma_crcq_in_t;

#define	RDMA_QP_MAX_PAGES	16
/*
 * Host RNIC interface for Create QP.
 */
typedef struct rdma_qp_attr {
	rdma_cqid_t	qp_sendcq;
	rdma_cqid_t	qp_recvcq;
	int		qp_sendmax;
	int		qp_recvmax;
	rdma_bool_t	qp_rdmaget;
	rdma_bool_t	qp_rdmaput;
	rdma_bool_t	qp_bindmw;
	u8		qp_scqcpu;
	u8		qp_rcqcpu;
	rdma_qpid_t	qp_qpid;
	int		qp_sendmaxsgl;
	int		qp_recvmaxsgl;
	int		qp_rdmamaxsgl;
	u8		qp_oreadmax;
	u8		qp_ireadmax;
	rdma_pdid_t	qp_pdid;
	rdma_bool_t	qp_stag0;
	rdma_bool_t	qp_freg;
	u32		qp_hevidx;
	u32		qp_hcompidx;
	u64		qp_dbaddr;
	u64		qp_latdma;
	u64		qp_rdma[RDMA_QP_MAX_PAGES];
	u64		qp_sdma[RDMA_QP_MAX_PAGES];
} rdma_qp_attr_t;

typedef struct rdma_qp_res {
	u32		qp_ident;
	int		qp_sendmax;
	int		qp_recvmax;
	int		qp_sendmaxsgl;
	int		qp_recvmaxsgl;
	int		qp_rdmamaxsgl;
	int		qp_oreadmax;
	int		qp_ireadmax;
} rdma_qp_res_t;

/*
 * Host RNIC interface for Modify QP.
 */
typedef struct rdma_qp_mattr {
	u32		qp_ident;		/* 4 bytes */
	u32		qp_llpctx;		/* 4 bytes */
	rdma_qpsize_t	qp_newsendmax;		/* 2 bytes */
	rdma_qpsize_t	qp_newrecvmax;		/* 2 bytes */
	rdma_qpsize_t	qp_neworeadmax;		/* 2 bytes */
	rdma_qpsize_t	qp_newireadmax;		/* 2 bytes */
	u32		qp_sockqid;		/* 4 bytes */
	u8		qp_newstate;		/* 1 byte  */
	rdma_bool_t     qp_rcvmrkr:1,		/* 1 byte  */
			qp_xmtmrkr:1,
			qp_crc:1,
			qp_unused:5;
} rdma_qp_mattr_t;

typedef struct rdma_qp_mres {
	rdma_qpsize_t	qp_sendmax;		/* 2 bytes */
	rdma_qpsize_t	qp_recvmax;		/* 2 bytes */
	rdma_qpsize_t	qp_oreadmax;		/* 2 bytes */
	rdma_qpsize_t	qp_ireadmax;		/* 2 bytes */
	u32		qp_sockqid;		/* 4 bytes */
	u32		qp_llpctx;		/* 4 bytes */
} rdma_qp_mres_t;

/*
 * Host RNIC interface for Destroy QP.
 */
typedef struct rdma_qp_dest {
	u32		qp_ident;		/* 4 bytes */
} rdma_qp_dest_t;

/*
 * QP states.
 */
#define	STATE_IDLE	1
#define	STATE_RTS	2
#define	STATE_CLOSING	3
#define	STATE_TERM	4
#define	STATE_ERROR	5

/*
 * Host RNIC interface for Recv WQE.
 */
typedef struct rdma_sge {
	u32		sge_tofflo;		/* 4 bytes */
	u32		sge_toffhi;		/* 4 bytes */
	rdma_stag_t	sge_stag;		/* 4 bytes */
	u32		sge_len;		/* 4 bytes */
} rdma_sge_t;

/*
 * The recv wqe has space for four sge's; if the wqe uses three or less
 * sge's, the first unused slot will have zero sgelen.
 */
typedef struct rdma_rwqe {
	rdma_sge_t	rwqe_sgel[RECVMAXSGL];	/* 64 bytes */
} rdma_rwqe_t;

/*
 * RDMA opcodes in the RDMAP control field.
 */
#define RDMA_OPCODE_WRITE	0x00
#define RDMA_OPCODE_READ_RQ	0x01
#define RDMA_OPCODE_READ_RSP	0x02
#define RDMA_OPCODE_SEND	0x03
#define RDMA_OPCODE_SEND_INV	0x04
#define RDMA_OPCODE_SEND_SE	0x05
#define RDMA_OPCODE_SEND_INVSE	0x06
#define RDMA_OPCODE_TERMINATE	0x07
#define RDMA_OPCODE_INVALID     0x0F

/*
 * Send*.
 */
typedef struct rdma_send_attr {
	u64
				swqe_optype:4,			/* 1st byte */
				swqe_signal:1,
				swqe_unused:2,
				snd_rfence:1,

				swqe_txop:8,			/* 2nd byte */

				swqe_txutqn:8,			/* 3rd byte */

				snd_nsge:8,			/* 4th byte */

				snd_invstagkidx:32;		/* 4 bytes */
	rdma_sge_t		snd_sgel[TXMAXSGL];		/* 64 bytes */
	rdma_tag_offset_t	snd_unused;			/* 8 bytes */
} rdma_send_attr_t;

/*
 * RDMA Read.
 */
typedef struct rdma_read_attr {
	u64
				swqe_optype:4,			/* 1st byte */
				swqe_signal:1,
				swqe_unused:3,

				swqe_txop:8,			/* 2nd byte */

				swqe_txutqn:8,			/* 3rd byte */

				rdr_unused4:8,			/* 4th byte */

				rdr_remstagkidx:32;		/* 4 bytes */
	rdma_sge_t		rdr_fwsge;			/* 16 bytes */
	rdma_sge_t		rdr_sge;			/* 16 bytes */
	rdma_sge_t		rdr_unused[TXMAXSGL - 2];	/* 32 bytes */
	rdma_tag_offset_t	rdr_remto;			/* 8 bytes */
} rdma_read_attr_t;

/*
 * RDMA Write.
 */
typedef struct rdma_write_attr {
	u64
				swqe_optype:4,			/* 1st byte */
				swqe_signal:1,
				swqe_unused:2,
				rdw_rfence:1,

				swqe_txop:8,			/* 2nd byte */

				rdw_unused3:8,			/* 3rd byte */

				rdw_nsge:8,			/* 4th byte */

				rdw_remstagkidx:32;		/* 4 bytes */
	rdma_sge_t		rdw_sgel[TXMAXSGL];		/* 64 bytes */
	rdma_tag_offset_t	rdw_remto;			/* 8 bytes */
} rdma_write_attr_t;

/*
 * Invalidate STAG; total 8 bytes.
 */
typedef struct rdma_inv_attr {
	u64
				swqe_optype:4,			/* 1st byte */
				swqe_signal:1,
				swqe_unused:2,
				inv_lfence:1,

				inv_unused2:8,			/* 2nd byte */

				inv_unused3:8,			/* 3rd byte */

				inv_unused4:8,			/* 4th byte */

				inv_stagkidx:32;		/* 4 bytes */
} rdma_inv_attr_t;

/*
 * Bind MW WQE; total 32 bytes.
 */
typedef struct rdma_bind_attr {
	u64
				swqe_optype:4,			/* 1st byte */
				swqe_signal:1,
				swqe_unused:3,

				bind_remaccbits:4,		/* 2nd byte */
				bind_vabasedto:1,
				bind_unused2:3,

				bind_unused3:8,			/* 3rd byte */

				bind_unused4:8,			/* 4th byte */

				bind_mrstagkidx:32;		/* 4 bytes */
	rdma_stag_t		bind_mwstagkidx;		/* 4 bytes */
	u32			bind_unused;			/* 4 bytes */
	rdma_tag_offset_t	bind_va;			/* 8 bytes */
	rdma_len_t		bind_mwlen;			/* 8 bytes */
} rdma_bind_attr_t;

/*
 * Fast Register NonShared MR WQE; total 160 bytes.
 */
#define	MAX_PBLLEN	16
typedef struct rdma_freg_attr {
	u64
				swqe_optype:4,			/* 1st byte */
				swqe_signal:1,
				swqe_unused:3,

				freg_accbits:4,			/* 2nd byte */
				freg_allowmw:1,
				freg_vabasedto:1,
				freg_unused2:2,

				freg_alistlen:8,		/* 3rd byte */

				freg_unused4:8,			/* 4th byte */

				freg_stagkidx:32;		/* 4 bytes */

	rdma_fbo_t		freg_fbo;			/* 4 bytes */
	rdma_bufsize_t		freg_pbentryshift;		/* 4 bytes */
	rdma_len_t		freg_regionlen;			/* 8 bytes */
	rdma_tag_offset_t	freg_vabase;			/* 8 bytes */
	rdma_dmaaddr_t		freg_alist[MAX_PBLLEN];		/* 8x16 bytes */
} rdma_freg_attr_t;

/*
 * Structure of a generic send WQE, sized at 256 bytes.
 */
typedef union rdma_swqe {
	rdma_freg_attr_t	swqe_freg;
	rdma_bind_attr_t	swqe_bind;
	rdma_inv_attr_t		swqe_inv;
	rdma_write_attr_t	swqe_rdmaw;
	rdma_read_attr_t	swqe_rdmar;
	rdma_send_attr_t	swqe_send;
	u8			swqe_pad[256];
} rdma_swqe_t;

/*
 * Send WQE operation types.
 */
#define	RDMA_SEND		1
#define	RDMA_SENDSE		2
#define RDMA_SENDINV		3
#define RDMA_SENDSEINV		4
#define RDMA_RDMAW		5
#define RDMA_RDMAR		6
#define RDMA_RDMARINV		7
#define RDMA_BINDMW		8
#define RDMA_FASTREG		9
#define RDMA_INV		10

/*
 * Recv WQE operation type.
 */
#define	RDMA_RECV		11

/*
 * Query RNIC
 */
#define UNM_VENDOR_INFO_SIZE  	16
#define UNM_VENDOR_INFO_TXT 	"UNM,Inc. Phantom"

typedef struct rdma_query_rnic {
	u8  		vendor_info[UNM_VENDOR_INFO_SIZE];	/* 16 bytes */
	rdma_maxnum_t	max_qps;				/* 4 bytes */
        rdma_qpsize_t	max_swr;				/* 2 bytes */
        rdma_qpsize_t	max_rwr;				/* 2 bytes */
	rdma_maxnum_t	max_cqs;				/* 4 bytes */
	rdma_cqsize_t	max_cq_entries;				/* 2 bytes */
        u8		max_wr_s_rq;
	u8		max_txsgl;
	u8		max_cq_eh;
	u8		max_mem_regs;
	u8		max_prot_dom;
	u8		max_in_rdma_rrq;
	u8		max_out_rdma_rrq;
	u8		max_in_rdma_rrq_msgs;
	u8		max_out_rdma_rrq_msgs;
	rdma_bool_t	mod_ird_flag;
	rdma_bool_t	mod_ord_flag;
	u8		max_mem_windows;
	rdma_bool_t	mod_owr_per_qp;
 	rdma_bool_t	block_list_mode;
	u8		list_sizes;
	rdma_bool_t	shared_rq_flag;
	rdma_bool_t	overflow_flag;
	u8		max_shared_rqs;
	u8		dequeuing_model;
	u8		fw_mem_pshift;
} rdma_query_rnic_t;

/*
 * Format of messages DMA'ed to host from RNIC; refer to host_peg_msg_hdr_t in
 * hostInterface.h.
 */
typedef struct rdma_hmesg {
	u32		ctx;
	u32		code;
	u32		value;
} rdma_hmesg_t;

/*
 * Interrupt type values, filled in the (lowest + 1) byte of "code".
 */
#define	IRQ_VERB	0
#define	IRQ_CQ		1
#define	IRQ_QP		2
#define	IRQ_AERR	3

/*
 * Immediate Status codes for Verbs.
 */
#define	RDMA_SUCCESS		0	/* Success */
#define	RDMA_INSUF_RESOURCES	1	/* Insufficient resources */
#define	RDMA_INVAL_PDID		6	/* Invalid PD */
#define	RDMA_INVAL_QPSTATE	23	/* Invalid QP state specified */
#define	RDMA_INVAL_LLPHDL	24	/* Invalid LLP handle specified */
#define	RDMA_FLUSHING_WQES	25	/* RI flushing WQEs */
#define	RDMA_INVAL_VA		27	/* Invalid address or offset */
#define	RDMA_INVAL_LEN		28	/* Size too big */
#define	RDMA_INVAL_PROT		29	/* Invalid access rights */
#define	RDMA_INVAL_PBL		30	/* Invalid PBL entry */
#define	RDMA_INVAL_SIZE		31	/* Requested buffer size unsupported */
#define	RDMA_INVAL_STAGIDX	32	/* Invalid Stag Index */
#define	RDMA_INVAL_FBO		33	/* Invalid FBO */
#define	RDMA_MWASSOC_MR		34	/* MW(s) associated with MR */

/*
 * Completion Status codes for Work Requests.
 */
#define	WC_SUCCESS		0	/* Success */
#define	WC_FLUSHED		1	/* WR not completed */
#define	WC_INVAL_WQE		2	/* WR format error */
#define	WC_LOCALQP_ERR		3	/* Catastrophic local QP error */
#define	WC_REMOTE_ERR		4	/* Remote termination error */
#define	WC_INVAL_STAG		5	/* Invalid STAG in SGE */
#define	WC_INVAL_BOUNDS		6	/* SGE includes out-of-bounds area */
#define	WC_PROT			7	/* Access right violation */
#define	WC_INVAL_PDID		8	/* Invalid PD ID */
#define	WC_WRAP_ERR		9	/* Machine specific address wrap */
#define	WC_PROT_STAG		10	/* Stag to invalidate has errors */
#define	WC_NORDMA_READ		11	/* 0 RDMA Read resources */
#define	WC_QP_UNPRIV		12	/* QP can not do priviledged WR */
#define	WC_STAG_UNINVAL		13	/* STAG not in Invalid state */
#define	WC_INVAL_PGSIZE		14	/* Requested page size unsupported */
#define	WC_INVAL_SIZE		15	/* Requested buffer size unsupported */
#define	WC_INVAL_PBL		16	/* Invalid PBL entry */
#define	WC_INVAL_FBO		17	/* Invalid FBO */
#define	WC_INVAL_LEN		18	/* MR Length too big for PBL */
#define	WC_INVAL_PROT		19	/* Invalid access rights */
#define	WC_INVAL_PBLSZ		20	/* PBL too long */
#define	WC_INVAL_VA		21	/* Inconsistent VA and FBO */
#define	WC_INVAL_MR		22	/* Bind MR STAG invalid */
#define	WC_INVAL_MW		23	/* Bind MW STAG invalid */
#define	WC_INVAL_LENGTH		24	/* Data size too large for RNIC */

/*
 * Resource Indicators for Async Event Records.
 */
typedef enum {
	RDMA_RES_IND_QP,
	RDMA_RES_IND_CQ,
	RDMA_RES_IND_RNIC,
	RDMA_RES_IND_S_RQ
} rdma_res_ind_t;

/*
 * Event identifiers.
 */
typedef enum {
	RDMA_LLP_CLOSE_COMPL,
	RDMA_TERM_MSG_RCVED,
	RDMA_LLP_CON_RESET,
	RDMA_LLP_CON_LOST,
        RDMA_LLP_IERR_SEG_SIZE,
	RDMA_LLP_IERR_CRC,
	RDMA_LLP_IERR_BAD_FPDU,
	RDMA_REM_INV_DDP_VER,
	RDMA_REM_INV_RDMA_VER,
	RDMA_REM_UNEXP_OPCODE,
	RDMA_REM_INV_DDP_Q_NUM,
	RDMA_REM_INV_READ_REQ_NE,
	RDMA_REM_INV_WRITE_MSG,
	RDMA_REM_INV_READ_RQ,
	RDMA_REM_INV_NO_L,
	RDMA_PROT_INV_STAG,
	RDMA_PROT_BASE_BOUNDS,
	RDMA_PROT_TAG_ACCESS,
	RDMA_PROT_TAG_INV_PD,
	RDMA_PROT_WRAP,
	RDMA_BAD_CLOSE,
	RDMA_BAD_LLP_CLOSE,
	RDMA_RQ_PROT_INVAL_MSN,
	RDMA_RQ_PROT_GAP_MSN,
	RDMA_IRRQ_PROT_MANY_MSN,
	RDMA_IRRQ_PROT_GAP_MSN,
	RDMA_IRRQ_PROT_RANGE_MSN,
	RDMA_IRRQ_INVAL_STAG,
	RDMA_IRRQ_PROT_BOUNDS,
	RDMA_IRRQ_PROT_ACCESS,
	RDMA_IRRQ_PROT_INVAL_PDID,
	RDMA_IRRQ_PROT_WRAP,
	RDMA_CQ_ERROR_SQ,
	RDMA_CQ_ERROR_RQ,
	RDMA_SRQ_ERROR_QP,
	RDMA_LOCAL_QP_CAT,
	RDMA_CQ_OVF_ERROR,
	RDMA_CQ_OP_ERROR,
	RDMA_SRQ_LIMIT,
	RDMA_QP_RQ_LIMIT,
	RDMA_SRQ_ERROR_CAT,
	RDMA_RNIC_ERROR_CAT
} rdma_event_id_t;

/*
 * MPA length and DDP/RDMAP control word format.
 */
typedef struct proto_len_ctl {
	u16	  ulpdu_len;
	u8      ddp_ctl;
	u8      rdmap_ctl;
} len_ctl_t;

/*
 * Untagged DDP header format.
 */
typedef struct ddp_hdr_untagged {
	u32     rsvd_istag;
	u32     QN;
	u32     MSN;
	u32     MO;
} ddp_hdr_ut_t;

/*
 * Tagged DDP header format.
 */
typedef struct ddp_hdr_tagged {
	u32     STag;
	u32     TO_hi;
	u32     TO_lo;
} ddp_hdr_t_t;

/*
 * Bits in the DDP control field.
 */
#define DDP_CTL_TAGGED   0x80
#define DDP_CTL_LAST     0x40
#define DDP_CTL_DV       0x00

#endif	/* _RDMA_STRUCTS_H_ */
