/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef _UNM_H
#define _UNM_H 1

    /*
     * $Id: unm.h,v 1.18 2006/06/16 17:31:03 ranjit Exp $
     * Header file defining programmatic interface to UNM card
     *
     */

#ifndef __KERNEL__
#include <asm/types.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* #define	SIOCUNMSSL		(SIOCPROTOPRIVATE + 0) */

/* PROTOPRIVATE+1 to PROROPRIVATE+4 are used by proc code */
#define SIOCSFASYNC		(SIOCPROTOPRIVATE + 5)


#define BUF_LEN 1 << 12
/*Info callback is called around 11 times if things are fine. There can also be
 alert callbacks. If there is an alert mid-way then this will be less than 11.
 Providing margin allocate space for abt MAX_INFO_CALL_STACK such calls.
 Anyway this will be freed when the user retrives it as part of an SSL_accept call.
*/
#define MAX_INFO_CALL_STACK	20

#define UNMSSL_CTX_NEW				1
    /*
     * Corresponds to SSL_CTX_new()
     *
     */
#define UNMSSL_CTX_FREE				2
    /*
     * Corresponds to SSL_CTX_free()
     *
     */
#define UNMSSL_CTX_SET_CIPHER_LIST		3
    /*
     * Corresponds to SSL_CTX_set_cipher_list()
     *
     */
#define UNMSSL_CTX_CTRL				4
    /*
     * Corresponds to SSL_CTX_ctrl()
     *
     */
#define UNMSSL_CTX_USE_CERTIFICATE		5
    /*
     * Corresponds to SSL_CTX_use_certificate()
     *
     */
#define UNMSSL_CTX_USE_RSAPRIVATEKEY		6
    /*
     * Corresponds to SSL_CTX_use_RSAPrivateKey()
     *
     */
#define UNMSSL_CTX_USE_DSAPRIVATEKEY		7
    /*
     * Corresponds to SSL_CTX_use_PrivateKey()
     *
     */
#define UNMSSL_CTX_USE_DHPRIVATEKEY		8
    /*
     * Corresponds to SSL_CTX_use_RSAPrivateKey()
     *
     */
#define UNMSSL_CTX_SET_SERVER_CERTIFICATE_CHAIN	9
    /*
     * Corresponds to SSL_CTX_use_certificate_chain()
     *
     */
#define UNMSSL_CTX_SET_TMP_RSA			10
    /*
     * Corresponds to SSL_CTX_set_tmp_rsa()
     *
     */
#define UNMSSL_NEW				11
    /*
     * Corresponds to SSL_new()
     *
     */

#define UNMSSL_GET_CURRENT_CIPHER      		12 

#define UNMSSL_GET_ERROR                        13
                                                                                                                             
#define UNMSSL_GET_PEER_CERTIFICATE             14
                                                                                                                             
#define UNMSSL_GET_STATE                        15
                                                                                                                             
#define UNMSSL_GET_VERIFY_RESULT                16
                                                                                                                             
#define UNMSSL_GET_VERSION                      17


#define UNMSSL_DO_HANDSHAKE			18


#define UNMSSL_RENEGOTIATE			19


#define UNMSSL_CTX_SET_CLIENT_CA_LIST		20


#define UNMSSL_GET_CIPHERS			21

#define UNMSSL_CALLBACK                         22
#define UNMSSL_CALLBACK_DONE                    23

#define UNMSSL_SET_VERIFY                       24

#define UNMSSL_COMMUNICATE_CB_DATA              25

#define UNMSSL_CALLBACK_VERIFY                  26
#define UNMSSL_CALLBACK_INFO                    27

#define UNMSSL_CALLBACK_NONE                    28
#define UNMSSL_CTX_SET_INFO_CALLBACK            29
#define UNMSSL_SET_INFO_CALLBACK                30
#define UNMSSL_CTX_SET_VERIFY                   31
#define UNMSSL_DO_ACCEPT                        32

#define UNMSSL_GET_CERT_DATA                    33

#define UNMSSL_DISABLE                          34


    /* Look at openssl/ssl/ssl.h to make sure no conflicts with SSL_CTRL_*
     * literal values for the following definitions
     */
#define UNMSSL_CTRL_SET_SHUTDOWN_MODE		60
#define UNMSSL_CTRL_SET_VERIFY_DEPTH		61

    /*
     * Identifiers for SSL methods
     *
     */

    /*
     * Corresponds to SSLv2 methods 
     *
     */
#define UNMSSL_SSLV2_METHOD                     1
#define UNMSSL_SSLV2_SERVER_METHOD              2
#define UNMSSL_SSLV2_CLIENT_METHOD              3
    /*
     * Corresponds to SSLv3 methods
     *
     */
#define UNMSSL_SSLV3_METHOD                     4
#define UNMSSL_SSLV3_SERVER_METHOD              5
#define UNMSSL_SSLV3_CLIENT_METHOD              6
    /*
     * Corresponds to TLSv1 methods
     *
     */
#define UNMSSL_TLSV1_METHOD                     7
#define UNMSSL_TLSV1_SERVER_METHOD              8
#define UNMSSL_TLSV1_CLIENT_METHOD              9
    /*
     * Corresponds to SSLv23 methods
     *
     */
#define UNMSSL_SSLV23_METHOD                    10
#define UNMSSL_SSLV23_SERVER_METHOD             11
#define UNMSSL_SSLV23_CLIENT_METHOD             12


typedef struct
{
	__u32		len;
	__u32		pad;
	__u64		buf;
} unm_ssl_op_t;

typedef struct
{
	__u32		cmd;
	__u32		larg;
	__u64		parg;
} unm_ssl_ctrl_t;

typedef union
{
	unm_ssl_op_t	op;
	unm_ssl_ctrl_t	ctrl;
} unm_ssl_args_t;

typedef struct
{
	__u64		ctx_handle;
	__u32		cmd;
	__s32		retval;
	unm_ssl_args_t	args;
} unm_ssl_ioctl_t;

typedef unm_ssl_ioctl_t *	unm_ssl_ioctl_ptr;

typedef struct
{
	int32_t		where;
	int32_t		ret;
	int32_t		sess_state;	
} unm_ssl_callback_stack_t;

typedef unm_ssl_callback_stack_t *	unm_ssl_callback_stack_ptr;

/* callback data gathered during session setup phase */
typedef struct
{
	int32_t				no_of_info_cb;
	unm_ssl_callback_stack_t	info_call_stack[MAX_INFO_CALL_STACK];

	int32_t				tmp_rsa_key_req; /* default is 0 */

	/* these are valid iff tmp_rsa_key_req == 1 */ 
	int32_t				nExport;
	int32_t				keylength;

	int32_t				enc_len;
	__u8				buf[BUF_LEN]; 
} unm_ssl_session_setup_data_t;

typedef unm_ssl_session_setup_data_t * 	unm_ssl_session_setup_data_ptr;

typedef struct
{
	/* used in info_callback */
	__u8		cbType;
	/* communicate cert verification result to card */
	__u8		verification_result;

	/* session read-ahead left or not communicated by card to host */
	/* set after the server reneg is done and used in SSL_read */
	__u8            read_ahead; /* 0 or 1 */ 

	/* used in info_callback */
	__s32		where;
	__s32		ret;

	/* used to trigger the s->state to 
	   SSL_ST_ACCEPT on card */
	__s32		sess_state;

	/* cert verification information length */
	__u32		enc_len;

	/* cert verification information */
	__u8		buf[4096];
} unm_callback_info_t, *unm_callback_info_ptr;

#ifdef __cplusplus
}
#endif
#ifndef __KERNEL__

#include <openssl/ssl.h>

#ifdef __cplusplus
extern "C" {
#endif

int unmSSL_CTX_new(SSL_CTX *ctx, SSL_METHOD *meth); 
int unmSSL_CTX_free(SSL_CTX *ctx); 
int unmSSL_CTX_set_cipher_list(SSL_CTX *ctx, const char * str);
int unmSSL_CTX_ctrl(SSL_CTX *ctx, int cmd, long larg, void *parg);
int unmSSL_CTX_set_quiet_shutdown(SSL_CTX *ctx, int mode);
int unmSSL_CTX_set_verify_depth(SSL_CTX *ctx, int depth);
int unmSSL_CTX_use_certificate(SSL_CTX *ctx, X509 *x);
int unmSSL_CTX_use_RSAPrivateKey(SSL_CTX *ctx, RSA *rsa);
int unmSSL_CTX_use_PrivateKey(SSL_CTX *ctx, EVP_PKEY *pkey);
int unmSSL_new(SSL * session);
int unmSSL_enable_ssl(int socket_fd, SSL_CTX * ssl_ctx);
int unmSSL_disable_ssl(int socket_fd);
SSL_CIPHER * unmSSL_get_current_cipher(SSL *ssl);
int unmSSL_get_error(SSL *ssl,int ret);
X509 *unmSSL_get_peer_certificate(SSL *ssl);
long unmSSL_get_verify_result(SSL *ssl);
const char *unmSSL_get_version(SSL *ssl);
STACK_OF(SSL_CIPHER) *unmSSL_get_ciphers(SSL *ssl);
int unmSSL_get_state(SSL *ssl);
int unmSSL_do_handshake(SSL *ssl, int state);
int unmSSL_renegotiate(SSL *ssl);
int unmSSL_shutdown(SSL *ssl);
void unmSSL_free(SSL *ssl);
void unmSSL_set_accept_state(SSL *ssl);
void unmSSL_CTX_set_client_CA_list(SSL_CTX *ctx, STACK_OF(X509_NAME) *list);
void unmSSL_set_verify(SSL *s,int mode);
int unmSSL_communicate_callback_data (SSL *s, unm_ssl_ioctl_ptr params, void *data, int size);
void unmSSL_set_info_callback (SSL *s, int mode);
void unmSSL_CTX_set_info_callback (SSL_CTX *ctx, int mode);
void unmSSL_CTX_set_verify(SSL_CTX *ctx,int mode);
int unmSSL_accept (SSL *s, int verify_mode);

    /*
     * SSL routine definitions (were static in original)
     *
     */
SSL_METHOD *ssl2_get_method(int ver);
SSL_METHOD *ssl2_get_server_method(int ver);
SSL_METHOD *ssl2_get_client_method(int ver);
SSL_METHOD *ssl3_get_method(int ver);
SSL_METHOD *ssl3_get_server_method(int ver);
SSL_METHOD *ssl3_get_client_method(int ver);
SSL_METHOD *ssl23_get_method(int ver);
SSL_METHOD *ssl23_get_server_method(int ver);
SSL_METHOD *ssl23_get_client_method(int ver);
SSL_METHOD *tls1_get_method(int ver);
SSL_METHOD *tls1_get_server_method(int ver);
SSL_METHOD *tls1_get_client_method(int ver);

#ifdef __cplusplus
}
#endif

#endif

#endif	/* unm.h */
