/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/******************************************************************************
*    unm_bdinfo.h - Phantom board information stored in flash.
*
*******************************************************************************
*/
#ifndef __UNM_BRDINFO_H
#define __UNM_BRDINFO_H

/* The version of the main data structure */
#define   UNM_BDINFO_VERSION 1

/* Magic number to let user know flash is programmed */
#define   UNM_BDINFO_MAGIC 0x12345678

/* Max number of Gig ports on a Phantom board */
#define UNM_MAX_PORTS 4

/* MN */
#define UNM_MEM_MAX_SLOTS 2

typedef enum {
    UNM_BRDTYPE_P1_BD   = 0x0000,
    UNM_BRDTYPE_P1_SB   = 0x0001,
    UNM_BRDTYPE_P1_SMAX = 0x0002,
    UNM_BRDTYPE_P1_SOCK = 0x0003,

    UNM_BRDTYPE_P2_SOCK_31  =  0x0008,
    UNM_BRDTYPE_P2_SOCK_35  =  0x0009,
    UNM_BRDTYPE_P2_SB35_4G  =  0x000a,
    UNM_BRDTYPE_P2_SB31_10G =  0x000b,
    UNM_BRDTYPE_P2_SB31_2G  =  0x000c,

    UNM_BRDTYPE_P2_SB31_10G_IMEZ =  0x000d,
    UNM_BRDTYPE_P2_SB31_10G_HMEZ =  0x000e,
    UNM_BRDTYPE_P2_SB31_10G_CX4  =  0x000f

} unm_brdtype_t;

typedef struct {
    long    is_10G:1;           /* = 1 for 10G type cards */
} nx_card_flags_t;

/* board type specific information */
typedef struct {
    unm_brdtype_t brdtype;                    /* type of board */
    long          ports;                      /* max no of physical ports */
    char          *short_name;
} unm_brdinfo_t;

#include "driver_info.h"

#define NUM_SUPPORTED_BOARDS (sizeof(unm_boards)/sizeof(unm_brdinfo_t))

#define GET_BRD_PORTS_NAME_BY_TYPE(type, ports, name)   \
{                                                       \
    int i, found = 0;                                   \
    for (i = 0; i < NUM_SUPPORTED_BOARDS; ++i) {        \
        if (unm_boards[i].brdtype == type) {            \
            ports = unm_boards[i].ports;                \
            name = unm_boards[i].short_name;            \
            found = 1;                                  \
            break;                                      \
        }                                               \
    }                                                   \
    if (!found) {                                       \
        ports = 0;                                      \
        name = "Unknown";                               \
    }                                                   \
}

#define GET_BRD_PORTS_BY_TYPE(type, ports)          \
{                                                   \
    int i, found = 0;                               \
    for (i = 0; i < NUM_SUPPORTED_BOARDS; ++i) {    \
        if (unm_boards[i].brdtype == type) {        \
            ports = unm_boards[i].ports;            \
            found = 1;                              \
            break;                                  \
        }                                           \
    }                                               \
    if (!found)                                     \
        ports = 0;                                  \
}

#define GET_BRD_NAME_BY_TYPE(type, name)            \
{                                                   \
    int i, found = 0;                               \
    for (i = 0; i < NUM_SUPPORTED_BOARDS; ++i) {    \
        if (unm_boards[i].brdtype == type) {        \
            name = unm_boards[i].short_name;        \
            found = 1;                              \
            break;                                  \
        }                                           \
    }                                               \
    if (!found)                                   \
        name = "Unknown";                           \
}

typedef enum {
    UNM_BRDMFG_INVENTEC=1
} unm_brdmfg;

typedef struct {
    union {
        struct {
            __uint16_t brdnum;
            unm_brdtype_t  brdtype;
        };
        __uint32_t brdid;
    };
} unm_brdid_t;

#define CHIP_LOT_FAST 1
#define CHIP_LOT_TYPICAL 2
#define CHIP_LOT_SLOW 3

#define CHIP_PKG_31x31 1
#define CHIP_PKG_35x35 2
typedef struct {
    __uint32_t  id:20,
         minor:4,
         major:2,
         lot:3,
         pkg:3;
} unm_chipid_t;

typedef enum {
    MEM_ORG_128Mbx4   =0x0, /* DDR1 only */
    MEM_ORG_128Mbx8   =0x1, /* DDR1 only */
    MEM_ORG_128Mbx16  =0x2, /* DDR1 only */
    MEM_ORG_256Mbx4   =0x3,
    MEM_ORG_256Mbx8   =0x4,
    MEM_ORG_256Mbx16  =0x5,
    MEM_ORG_512Mbx4   =0x6,
    MEM_ORG_512Mbx8   =0x7,
    MEM_ORG_512Mbx16  =0x8,
    MEM_ORG_1Gbx4     =0x9,
    MEM_ORG_1Gbx8     =0xa,
    MEM_ORG_1Gbx16    =0xb,
    MEM_ORG_2Gbx4     =0xc,
    MEM_ORG_2Gbx8     =0xd,
    MEM_ORG_2Gbx16    =0xe,
    MEM_ORG_128Mbx32  =0x10002, /* GDDR only */
    MEM_ORG_256Mbx32  =0x10005 /* GDDR only */
} unm_mn_mem_org_t;

typedef enum {
    MEM_ORG_512Kx36   =0x0,
    MEM_ORG_1Mx36     =0x1,
    MEM_ORG_2Mx36     =0x2
} unm_sn_mem_org_t;

typedef enum {
    MEM_DEPTH_4MB   = 0x1,
    MEM_DEPTH_8MB   = 0x2,
    MEM_DEPTH_16MB  = 0x3,
    MEM_DEPTH_32MB  = 0x4,
    MEM_DEPTH_64MB  = 0x5,
    MEM_DEPTH_128MB = 0x6,
    MEM_DEPTH_256MB = 0x7,
    MEM_DEPTH_512MB = 0x8,
    MEM_DEPTH_1GB   = 0x9,
    MEM_DEPTH_2GB   = 0xa,
    MEM_DEPTH_4GB   = 0xb,
    MEM_DEPTH_8GB   = 0xc,
    MEM_DEPTH_16GB  = 0xd,
    MEM_DEPTH_32GB  = 0xe
} unm_mem_depth_t;

/******************************************************************************
*
*
*
*******************************************************************************
*/
typedef struct {
    __uint32_t header_version;

    __uint32_t board_mfg;
    //unm_brdid_t board_id;
    __uint32_t board_type;
    __uint32_t board_num;
    //unm_chipid_t chip_id;
    __uint32_t chip_id;
    __uint32_t chip_minor;
    __uint32_t chip_major;
    __uint32_t chip_pkg;
    __uint32_t chip_lot;


    __uint32_t port_mask;       /* available niu ports */
    __uint32_t peg_mask;        /* available pegs */
    __uint32_t icache_ok;       /* can we run with icache? */
    __uint32_t dcache_ok;       /* can we run with dcache? */
    __uint32_t casper_ok;

    //unm_eth_addr_t  mac_address[MAX_PORTS];  /*  */
    __uint32_t mac_addr_lo_0;
    __uint32_t mac_addr_lo_1;
    __uint32_t mac_addr_lo_2;
    __uint32_t mac_addr_lo_3;

    /* MN-related config */
    __uint32_t mn_sync_mode;    /* enable/ sync shift cclk/ sync shift mclk */
    __uint32_t mn_sync_shift_cclk;
    __uint32_t mn_sync_shift_mclk;
    __uint32_t mn_wb_en;
    __uint32_t mn_crystal_freq; /* in MHz */
    __uint32_t mn_speed;        /* in MHz */
    __uint32_t mn_org;
    __uint32_t mn_depth;
    __uint32_t mn_ranks_0;        /* ranks per slot */
    __uint32_t mn_ranks_1;        /* ranks per slot */
    __uint32_t mn_rd_latency_0;
    __uint32_t mn_rd_latency_1;
    __uint32_t mn_rd_latency_2;
    __uint32_t mn_rd_latency_3;
    __uint32_t mn_rd_latency_4;
    __uint32_t mn_rd_latency_5;
    __uint32_t mn_rd_latency_6;
    __uint32_t mn_rd_latency_7;
    __uint32_t mn_rd_latency_8;
    __uint32_t mn_dll_val[18];
    __uint32_t mn_mode_reg;       /* See MIU DDR Mode Register */
    __uint32_t mn_ext_mode_reg;   /* See MIU DDR Extended Mode Register */
    __uint32_t mn_timing_0;       /* See MIU Memory Control Timing Rgister */
    __uint32_t mn_timing_1;       /* See MIU Extended Memory Ctrl Timing Register */
    __uint32_t mn_timing_2;       /* See MIU Extended Memory Ctrl Timing2 Register */

    /* SN-related config */
    __uint32_t sn_sync_mode;    /* enable/ sync shift cclk / sync shift mclk */
    __uint32_t sn_pt_mode;    /* pass through mode */
    __uint32_t sn_ecc_en;
    __uint32_t sn_wb_en;
    __uint32_t sn_crystal_freq;
    __uint32_t sn_speed;
    __uint32_t sn_org;
    __uint32_t sn_depth;
    __uint32_t sn_dll_tap;
    __uint32_t sn_rd_latency;

    __uint32_t mac_addr_hi_0;
    __uint32_t mac_addr_hi_1;
    __uint32_t mac_addr_hi_2;
    __uint32_t mac_addr_hi_3;

    __uint32_t magic;          /* indicates flash has been initialized */

    __uint32_t mn_rdimm;
    __uint32_t mn_dll_override;
#if 0
    __uint32_t  boot_type;        /* is board booting from flash? */

    /* following stuff might be overkill */
    __uint32_t fw_id;              /* what code is loaded here? */
    __uint32_t fw_version;     /* version of flash */
    __uint32_t fw_checksum;
    __uint32_t crb_init_version;   /* info about flash's crb_init values */
    __uint32_t crb_init_checksum;  /* verification for crb_init values */
    __uint32_t pci_ph_type;     /* strap values for host PCI */
    __uint32_t pci_ps_type;     /* strap values for secondary PCI */
#endif

}  unm_board_info_t;

#define FLASH_SECTOR_START                                  (1)
#define NUM_FLASH_SECTORS                                  (64)
#define FLASH_SECTOR_SIZE                             (64*1024)
#define FLASH_TOTAL_SIZE  (NUM_FLASH_SECTORS*FLASH_SECTOR_SIZE)
#define FLASH_NUM_PORTS                                     (4)

typedef struct {
    __uint32_t flash_addr[32];
} unm_flash_mac_addr_t;

/* flash user area */
typedef struct {
    __uint8_t  flash_md5[16];
    __uint8_t  crbinit_md5[16];
    __uint8_t  brdcfg_md5[16];
    /* bootloader */
    __uint32_t bootld_version;
    __uint32_t bootld_size;
    __uint8_t  bootld_md5[16];
    /* image */
    __uint32_t image_version;
    __uint32_t image_size;
    __uint8_t  image_md5[16];
    /* primary image status */
    __uint32_t primary_status;
    __uint32_t secondary_present;

    /* MAC address , 4 ports */
    unm_flash_mac_addr_t mac_addr[FLASH_NUM_PORTS];

    /* Any user defined data */
} unm_old_user_info_t;

#define FLASH_NUM_MAC_PER_PORT             32
typedef struct {
    __uint8_t  flash_md5[16 * 64];
    // __uint8_t  crbinit_md5[16];
    // __uint8_t  brdcfg_md5[16];
    /* bootloader */
    __uint32_t bootld_version;
    __uint32_t bootld_size;
    // __uint8_t  bootld_md5[16];
    /* image */
    __uint32_t image_version;
    __uint32_t image_size;
    // U8  image_md5[16];
    /* primary image status */
    __uint32_t primary_status;
    __uint32_t secondary_present;

    /* MAC address , 4 ports, 32 address per port */
    __uint64_t mac_addr[FLASH_NUM_PORTS * FLASH_NUM_MAC_PER_PORT];
    __uint32_t sub_sys_id;
    __uint8_t  serial_num[32];
	__uint32_t bios_version;

    /* Any user defined data */
} unm_user_info_t;

typedef struct {
   __uint8_t  major;
   __uint8_t  minor;
   __uint16_t sub;
} unm_fw_ver_t;

#define UNM_VPD_LEN         1024
#define MAX_VPD_STING_SIZE    40
#define VPD_KEYWORD_SIZE       2

typedef struct vpd_tab_s {
    __uint8_t      product_id;
    __uint8_t      length[2]; //16 bit integer at odd address boundary !!!
    __uint8_t      data[40];
    __uint8_t      vpd_r_tag;
    __uint8_t      ro_length[VPD_KEYWORD_SIZE];
    __uint8_t      ro_pn_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      pn_length;
    __uint8_t      part_num[11];
    __uint8_t      ro_ec_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      ec_length;
    __uint8_t      ec_data[6];
    __uint8_t      ro_sn_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      sn_length;
    __uint8_t      sn_data[12];
    __uint8_t      ro_v0_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      v0_length;
    __uint8_t      v0_data[19];
    __uint8_t      ro_v2_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      v2_length;
    __uint8_t      v2_data[4];
    __uint8_t      ro_cs_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      ro_cs_length;
    __uint8_t      ro_checksum;
    __uint8_t      ro_reserved;
    __uint8_t      vpd_rw_tag;
    __uint8_t      vpd_rw_length[2]; //16 bit integer at odd address boundary !!!
    __uint8_t      rw_v1_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      v1_length;
    __uint8_t      v1_data[4];
    __uint8_t      rw_v3_keyword[VPD_KEYWORD_SIZE];
    __uint8_t      v3_length;
    __uint8_t      v3_data[12];
    __uint8_t      end_tag;
} unm_vpd_tab_t;

typedef struct vpd_s {
    union {
        unm_vpd_tab_t   vpd_tab;
        __uint32_t             vpd_data[UNM_VPD_LEN/sizeof(__uint32_t)];
    } u;
} unm_vpd_t;

#define VPD_PRODUCT_ID      0x82

#define SECONDARY_IMAGE_PRESENT 0xb3b4b5b6
#define SECONDARY_IMAGE_ABSENT  0xffffffff
#define PRIMARY_IMAGE_GOOD      0x5a5a5a5a
#define PRIMARY_IMAGE_BAD       0xffffffff

/* Flash memory map */
typedef enum {
    CRBINIT_START   = 0,          /* Crbinit section */
    BRDCFG_START    = 0x4000,     /* board config */
    INITCODE_START  = 0x6000,     /* pegtune code */
    BOOTLD_START    = 0x10000,    /* bootld */
    IMAGE_START     = 0x43000,    /* compressed image */
    SECONDARY_START = 0x200000,   /* backup images */
    PXE_START       = 0x3D0000,   /* PXE image area */
    USER_START      = 0x3E8000,   /* User defined region for new boards */
    VPD_START       = 0x3E8C00,   /* Vendor private data */
    FIXED_START     = 0x3F0000    /* backup of crbinit */
} unm_flash_map_t;

#define USER_START_OLD     PXE_START /* for backward compatibility */

#define FLASH_START                                             (CRBINIT_START)
#define INIT_SECTOR                                                         (0)
#define INIT_BCKUP_SECTOR                                                  (63)
#define PXE_SECTOR                                (PXE_START/FLASH_SECTOR_SIZE)
#define USER_SECTOR                              (USER_START/FLASH_SECTOR_SIZE)
#define PRIMARY_START                                            (BOOTLD_START)
#define FLASH_CRBINIT_SIZE                                             (0x4000)
#define FLASH_BRDCFG_SIZE                            (sizeof(unm_board_info_t))
#define FLASH_USER_SIZE              (sizeof(unm_user_info_t)/sizeof(uint32_t))
#define FLASH_VPD_SIZE                           (UNM_VPD_LEN/sizeof(uint32_t))
#define FLASH_SECONDARY_SIZE                       (USER_START-SECONDARY_START)
#define NUM_PRIMARY_SECTORS                                              (0x20)
#define NUM_CONFIG_SECTORS                                                  (1)

#endif  // __UNM_BRDINFO_H
