/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef __UNM_IOCTL_H__
#define __UNM_IOCTL_H__
#include <linux/version.h>

/* UNM ioctl interface for the Linux kernel HAL transport layer */

/* ioctl's dealing with PCI read/writes */
#define UNM_XPORT_CMD 1
typedef enum {
        unm_cmd_none = 0,
        unm_cmd_pci_read,
        unm_cmd_pci_write,
        unm_cmd_pci_mem_read,
        unm_cmd_pci_mem_write,
        unm_cmd_pci_config_read,
        unm_cmd_pci_config_write,
        unm_cmd_pci_db_read,
        unm_cmd_pci_db_write,
        unm_cmd_pci_io_read,
        unm_cmd_pci_io_write
} unm_xport_cmd_t;

typedef struct {
        unsigned long long off;
        unm_xport_cmd_t cmd;
        int size;
        int rv;
        char u[64];

} unm_xport_data_t;

/* ioctl's dealing w. interrupt counts */
#define UNM_INTR_CMD 3
typedef enum {
        unm_intr_cmd_none = 0,
        unm_intr_cmd_clear,
        unm_intr_cmd_get
} unm_intr_cmd_t;

typedef enum {
        unm_intr_src_unspecified = -1,
        unm_intr_src_min = 0,
        unm_intr_src_dma_0 = 0,
        unm_intr_src_dma_1 = 1,
        unm_intr_src_i2q = 2,
        unm_intr_src_dma_2 = 3,
        unm_intr_src_dma_3 = 4,
        unm_intr_src_max = 5
} unm_intr_src_t;

typedef struct {
        unm_intr_cmd_t cmd;
        unm_intr_src_t int_src;
        int data;
} unm_intr_data_t;

/* sleep until timeout or until we get a packet on a type1 queue */
#define UNM_WAIT4WORK_CMD 4

#define UNM_GET_DMA_BASE_CMD 5
typedef struct {
        unsigned long base_addr;
        unsigned long long phys_base_addr;
} unm_dma_base_addr_t;


#define UNM_GETIFCONF 6

#endif
