/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef __UNM_NIC_IOCTL_H__
#define __UNM_NIC_IOCTL_H__

#include <linux/sockios.h>

/* ioctl's dealing with PCI read/writes */
#define UNM_CMD_START SIOCDEVPRIVATE
#define UNM_NIC_CMD  (UNM_CMD_START + 1)
#define UNM_NIC_NAME (UNM_CMD_START + 2)
#define UNM_NIC_SEND_TEST	(UNM_CMD_START + 3)
#define UNM_NIC_IP_FILTER	(UNM_CMD_START + 4)
#define UNM_NIC_IRQ_TEST	(UNM_CMD_START + 5)
#define UNM_NIC_ILB_TEST	(UNM_CMD_START + 6)
#define UNM_NIC_ELB_TEST	(UNM_CMD_START + 7)
#define UNM_NIC_LINK_TEST	(UNM_CMD_START + 8)
#define UNM_NIC_HW_TEST		(UNM_CMD_START + 9)
#define UNM_NIC_CIS_TEST	(UNM_CMD_START + 10)
#define UNM_NIC_CR_TEST		(UNM_CMD_START + 11)
#define UNM_NIC_PROTSTATS	(UNM_CMD_START + 12)

#define UNM_NIC_NAME_LEN    16
#define UNM_NIC_NAME_RSP    "NETXEN"

/* Error codes - loopback test */
enum {
	LB_TEST_OK,
	LB_UCOPY_PARAM_ERR,
	LB_UCOPY_DATA_ERR,
	LB_NOMEM_ERR,
	LB_TX_NOSKB_ERR,
	LB_SHORT_DATA_ERR,
	LB_SEQUENCE_ERR,
	LB_DATA_ERR,
	LB_ERRCNT,
	LB_NOT_SUPPORTED
};

/* Error codes - CR test */
enum {
	CR_TEST_OK,
    CR_ERROR,
	CR_ERRCNT
};

/* Error codes - CIS test */
enum {
	CIS_TEST_OK,
	CIS_WMARK,
	CIS_ERRCNT
};

/* Error codes - HW test */
enum {
	HW_TEST_OK,
	HW_DMA_BZ_0,
	HW_DMA_BZ_1,
	HW_DMA_BZ_2,
	HW_DMA_BZ_3,
	HW_SRE_PBI_HALT,
	HW_SRE_L1IPQ,
	HW_SRE_L2IFQ,
	HW_SRE_FREEBUF,
	HW_IPQ,
	HW_PQ_W_PAUSE,
	HW_PQ_W_FULL,
	HW_IFQ_W_PAUSE,
	HW_IFQ_W_FULL,
	HW_MEN_BP_TOUT,
	HW_DOWN_BP_TOUT,
	HW_FBUFF_POOL_WM,
	HW_PBUF_ERR,
	HW_FM_MSG_HDR,
	HW_FM_MSG,
	HW_EPG_CTRL_Q,
	HW_EPG_MSG_BUF,
	HW_EPG_QREAD_TOUT,
	HW_EPG_QWRITE_TOUT,
	HW_EPG_CQ_W_FULL,
	HW_EPG_MSG_CHKSM,
   	HW_EPG_MTLQ_TOUT,
	HW_PEG0,
	HW_PEG1,
	HW_PEG2,
	HW_PEG3,
	HW_ERRCNT
};


struct sections
{

   long long is_bss_flag;
   long long file_offset;
   long long size;
   long long load_offset;
};
struct driverimg_header
{
   long long no_of_sections;
   struct sections sec[1];
};

typedef enum {
        unm_nic_cmd_none = 0,
        unm_nic_cmd_pci_read,
        unm_nic_cmd_pci_write,
        unm_nic_cmd_pci_mem_read,
        unm_nic_cmd_pci_mem_write,
        unm_nic_cmd_pci_config_read,
        unm_nic_cmd_pci_config_write,
        unm_nic_cmd_get_stats,
        unm_nic_cmd_clear_stats,
        unm_nic_cmd_get_version,

        unm_nic_cmd_flash_read = 50,
        unm_nic_cmd_flash_write,
        unm_nic_cmd_flash_se
} unm_nic_ioctl_cmd_t;

#define UNM_FLASH_READ_SIZE     4
#define UNM_FLASH_WRITE_SIZE    4

typedef struct {
        __uint32_t cmd;
        __uint32_t unused1;
        __uint64_t off;
        __uint32_t size;
        __uint32_t rv;
        char u[64];
	void *ptr;
} unm_nic_ioctl_data_t;

/* FIXME vijo,  not used for now */
/* ioctl's dealing w. interrupt counts */
#define UNM_INTR_CMD 3
typedef enum {
        unm_intr_cmd_none = 0,
        unm_intr_cmd_clear,
        unm_intr_cmd_get
} unm_intr_cmd_t;

typedef enum {
        unm_intr_src_unspecified = -1,
        unm_intr_src_min = 0,
        unm_intr_src_dma_0 = 0,
        unm_intr_src_dma_1 = 1,
        unm_intr_src_i2q = 2,
        unm_intr_src_max = 3
} unm_intr_src_t;

typedef struct {
        unm_intr_cmd_t cmd;
        unm_intr_src_t int_src;
        int data;
} unm_intr_data_t;

typedef enum {
	UNM_TX_START = 1,
	UNM_TX_STOP,
	UNM_TX_SET_PARAM,
	UNM_TX_SET_PACKET,
	UNM_LOOPBACK_START,
	UNM_LOOPBACK_STOP
} unm_send_test_cmd_t;

typedef struct {
	unm_send_test_cmd_t cmd;
	__uint64_t   ifg;
	unsigned int count;
} unm_send_test_t;

    /*
     * Commands to be used for the IP Filter.
     */
enum {
	UNM_IP_ADDR_CLEAR = 1,
	UNM_IP_ADDR_ADD,
	UNM_IP_ADDR_DEL,
	UNM_IP_ADDR_SHOW
};

#define UNM_IP_ADDR_MAX_COUNT		15
typedef struct {
	__int32_t count;
	__uint32_t ip_addr[UNM_IP_ADDR_MAX_COUNT];
} unm_ip_addr_filter_t;

typedef struct {
	int cmd;	/* must be one of the commands above */
	unm_ip_addr_filter_t request;
} unm_ip_filter_request_t;

struct unm_statistics {
	__uint64_t rx_packets;
	__uint64_t tx_packets;
	__uint64_t rx_bytes;
	__uint64_t rx_errors;
	__uint64_t tx_bytes;
	__uint64_t tx_errors;
	__uint64_t rx_CRC_errors;
	__uint64_t rx_short_length_error;
	__uint64_t rx_long_length_error;
	__uint64_t rx_MAC_errors;
};

#endif
