/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Definitions relating to access/control of the Secondary Queue Manager
 * h/w block.
 */
#ifndef __HAL_PCI_H
#define __HAL_PCI_H

#include <linux/config.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#include <asm/softirq.h>
#endif

/* these initialized when the module gets loaded */
extern unsigned g_myirq;
extern spinlock_t unm_pcilock;

/* make sure a crashed user-space app releases its locks... */
void unm_xport_unlock_all(void);

#ifndef __iomem
#define __iomem
#endif

#ifndef readq
static inline __uint64_t readq(void __iomem *addr)
{
        return readl(addr) |
               (((__uint64_t)readl(addr+4)) << 32LL);
}
#endif

#ifndef writeq
static inline void writeq(__uint64_t val, void __iomem *addr)
{
        writel(((__uint32_t)(val)), (addr));
        writel(((__uint32_t)(val >> 32)), (addr + 4));
}
#endif

#define UNM_PCI_DB_WINDOW_SIZE   0x100000

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#ifdef MODULE
/************************************************************************/
/* This is in the kernel, but infortunately,the symbols are not exported */
static inline void __save_init_fpu( struct task_struct *tsk )
{
    if ( cpu_has_fxsr ) {
        asm volatile( "fxsave %0 ; fnclex"
                  : "=m" (tsk->thread.i387.fxsave) );
    } else {
        asm volatile( "fnsave %0 ; fwait"
                  : "=m" (tsk->thread.i387.fsave) );
    }
    tsk->flags &= ~PF_USEDFPU;
}

static inline void kernel_fpu_begin(void)
{
    struct task_struct *tsk = current;

    if (tsk->flags & PF_USEDFPU) {
        __save_init_fpu(tsk);
        return;
    }
    clts();
}

#define kernel_fpu_end() stts()
/************************************************************************/
#endif
#else
#include <asm/i387.h>	/* in 2.6.x kernel these functions are available through
			   this header file on x86 and x86_64 architecture*/
#endif /* KERNEL_VERSION < 2.6.0 */

#define	UNM_PCI_READ_8(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 readb(unm_pci_base + (ADDR)))
#define	UNM_PCI_READ_16(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 readw(unm_pci_base + (ADDR)))
#define	UNM_PCI_READ_32(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 readl(unm_pci_base + (ADDR)))
#define	UNM_PCI_WRITE_8(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 writeb(DATA, unm_pci_base + (ADDR)))
#define	UNM_PCI_WRITE_16(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 writew(DATA, unm_pci_base + (ADDR)))
#define	UNM_PCI_WRITE_32(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 writel(DATA, unm_pci_base + (ADDR)))

#define	UNM_PCI_DB_READ_8(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 readb(unm_pci_db_base + (ADDR)))
#define	UNM_PCI_DB_READ_16(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 readw(unm_pci_db_base + (ADDR)))
#define	UNM_PCI_DB_READ_32(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 readl(unm_pci_db_base + (ADDR)))
#define	UNM_PCI_DB_WRITE_8(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 writeb(DATA, unm_pci_db_base + (ADDR)))
#define	UNM_PCI_DB_WRITE_16(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 writew(DATA, unm_pci_db_base + (ADDR)))
#define	UNM_PCI_DB_WRITE_32(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 writel(DATA, unm_pci_db_base + (ADDR)))

#define	UNM_PCI_IO_READ_8(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 inb(unm_pci_io_base + (ADDR)))
#define	UNM_PCI_IO_READ_16(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 inw(unm_pci_io_base + (ADDR)))
#define	UNM_PCI_IO_READ_32(ADDR)				\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 inl(unm_pci_io_base + (ADDR)))
#define	UNM_PCI_IO_WRITE_8(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 outb(DATA, unm_pci_io_base + (ADDR)))
#define	UNM_PCI_IO_WRITE_16(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 outw(DATA, unm_pci_io_base + (ADDR)))
#define	UNM_PCI_IO_WRITE_32(DATA, ADDR)			\
	(unm_debug_delay ? DELAY(UNM_DELAY_HW) : 0,	\
	 outl(DATA, unm_pci_io_base + (ADDR)))

#define UNM_PCI_READ_64(ADDR)	                        \
        (readq(unm_pci_base + (ADDR)))
#define UNM_PCI_BLOCK_READ_64(ADDR, DATA_PTR, NUM_WORDS)   	\
        do {                                       		\
            int i;                                 		\
            u64 *a = (u64 *) (DATA_PTR);    			\
            u64 *b = (u64 *) (unm_pci_base + (ADDR));       	\
            for (i=0; i< (NUM_WORDS); i++) {                   	\
                writeq(readq(b), a);                           	\
                b++;                                           	\
                a++;                                           	\
            }                                                  	\
        } while (0)
#define UNM_PCI_DB_READ_64(ADDR)                                \
        (readq(unm_pci_db_base + (ADDR)))

#define UNM_PCI_WRITE_64(DATA, ADDR)                    \
        (writeq(DATA, (unm_pci_base + (ADDR))))
#define UNM_PCI_BLOCK_WRITE_64(ADDR, DATA_PTR, NUM_WORDS)	\
        do {                                       		\
            int i;                                 		\
            u64 *a = (u64 *) (DATA_PTR);    			\
            u64 *b = (u64 *) (unm_pci_base + (ADDR));          	\
            for (i=0; i< (NUM_WORDS); i++) {                    \
                writeq(readq(a), b);                            \
                b++;                                            \
                a++;                                            \
            }                                                   \
        } while (0)
#define UNM_PCI_DB_WRITE_64(DATA, ADDR)                         \
        (writeq(DATA, (unm_pci_db_base + (ADDR))))


extern char *unm_pci_base;		/* base address of our PCI board */
extern char *unm_pci_io_base;		/* base address of our PCI IO */
extern char *unm_pci_db_base;		/* base address of our doorbell PCI */

#endif /* __HAL_PCI_H */
