/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
#ifndef _UNM_PSTATS_H_
#define _UNM_PSTATS_H_

#include "unm_compiler_defs.h"

/*
 * We use all unsigned longs. Linux will soon be so reliable that even these
 * will rapidly get too small 8-). Seriously consider the IpInReceives count
 * on the 20Gb/s + networks people expect in a few years time!
 */

#define NETXEN_NUM_PORTS        4
#define NETXEN_NUM_PEGS         4

typedef struct {
	__uint64_t IpInReceives;
	__uint64_t IpInHdrErrors;
	__uint64_t IpInAddrErrors;
	__uint64_t IpForwDatagrams;
	__uint64_t IpInUnknownProtos;
	__uint64_t IpInDiscards;
	__uint64_t IpInDelivers;
	__uint64_t IpOutRequests;
	__uint64_t IpOutDiscards;
	__uint64_t IpOutNoRoutes;
	__uint64_t IpReasmTimeout;
	__uint64_t IpReasmReqds;
	__uint64_t IpReasmOKs;
	__uint64_t IpReasmFails;
	__uint64_t IpFragOKs;
	__uint64_t IpFragFails;
	__uint64_t IpFragCreates;
} nx_ip_mib_t;

typedef struct {
	__uint64_t PREALIGN(512) IcmpInMsgs POSTALIGN(512);
	__uint64_t IcmpInErrors;
	__uint64_t IcmpInDestUnreachs;
	__uint64_t IcmpInTimeExcds;
	__uint64_t IcmpInParmProbs;
	__uint64_t IcmpInSrcQuenchs;
	__uint64_t IcmpInRedirects;
	__uint64_t IcmpInTimestamps;
	__uint64_t IcmpInTimestampReps;
	__uint64_t IcmpInAddrMasks;
	__uint64_t IcmpInAddrMaskReps;
	__uint64_t IcmpOutMsgs;
	__uint64_t IcmpOutErrors;
	__uint64_t IcmpOutDestUnreachs;
	__uint64_t IcmpOutTimeExcds;
	__uint64_t IcmpOutParmProbs;
	__uint64_t IcmpOutSrcQuenchs;
	__uint64_t IcmpOutRedirects;
	__uint64_t IcmpOutTimestamps;
	__uint64_t IcmpOutTimestampReps;
	__uint64_t IcmpOutAddrMasks;
	__uint64_t IcmpOutAddrMaskReps;
	__uint64_t dummy;
} nx_icmp_mib_t;

typedef struct {
	__uint64_t TcpRtoAlgorithm;
	__uint64_t TcpRtoMin;
	__uint64_t TcpRtoMax;
	__uint64_t TcpMaxConn;
	__uint64_t TcpActiveOpens;
	__uint64_t TcpPassiveOpens;
	__uint64_t TcpAttemptFails;
	__uint64_t TcpEstabResets;
	__uint64_t TcpCurrEstab;
	__uint64_t TcpInSegs;
	__uint64_t TcpOutSegs;
	__uint64_t TcpRetransSegs;
	__uint64_t TcpInErrs;
	__uint64_t TcpOutRsts;
} nx_tcp_mib_t;

typedef struct {
	__uint64_t PREALIGN(512) SyncookiesSent POSTALIGN(512);
	__uint64_t SyncookiesRecv;
	__uint64_t SyncookiesFailed;
	__uint64_t EmbryonicRsts;
	__uint64_t PruneCalled;
	__uint64_t RcvPruned;
	__uint64_t OfoPruned;
	__uint64_t OutOfWindowIcmps;
	__uint64_t LockDroppedIcmps;
	__uint64_t ArpFilter;
	__uint64_t TimeWaited;
	__uint64_t TimeWaitRecycled;
	__uint64_t TimeWaitKilled;
	__uint64_t PAWSPassiveRejected;
	__uint64_t PAWSActiveRejected;
	__uint64_t PAWSEstabRejected;
	__uint64_t DelayedACKs;
	__uint64_t DelayedACKLocked;
	__uint64_t DelayedACKLost;
	__uint64_t ListenOverflows;
	__uint64_t ListenDrops;
	__uint64_t TCPPrequeued;
	__uint64_t TCPDirectCopyFromBacklog;
	__uint64_t TCPDirectCopyFromPrequeue;
	__uint64_t TCPPrequeueDropped;
	__uint64_t TCPHPHits;
	__uint64_t TCPPureAcks;
	__uint64_t TCPHPAcks;
	__uint64_t TCPRenoRecovery;
	__uint64_t TCPSackRecovery;
	__uint64_t TCPSACKReneging;
	__uint64_t TCPFACKReorder;
	__uint64_t TCPSACKReorder;
	__uint64_t TCPRenoReorder;
	__uint64_t TCPTSReorder;
	__uint64_t TCPFullUndo;
	__uint64_t TCPPartialUndo;
	__uint64_t TCPDSACKUndo;
	__uint64_t TCPLossUndo;
	__uint64_t TCPLoss;
	__uint64_t TCPLostRetransmit;
	__uint64_t TCPRenoFailures;
	__uint64_t TCPSackFailures;
	__uint64_t TCPLossFailures;
	__uint64_t TCPFastRetrans;
	__uint64_t TCPForwardRetrans;
	__uint64_t TCPSlowStartRetrans;
	__uint64_t TCPTimeouts;
	__uint64_t TCPRenoRecoveryFail;
	__uint64_t TCPSackRecoveryFail;
	__uint64_t TCPSchedulerFailed;
	__uint64_t TCPRcvCollapsed;
	__uint64_t TCPDSACKOldSent;
	__uint64_t TCPDSACKOfoSent;
	__uint64_t TCPDSACKRecv;
	__uint64_t TCPDSACKOfoRecv;
	__uint64_t TCPAbortOnSyn;
	__uint64_t TCPAbortOnData;
	__uint64_t TCPAbortOnClose;
	__uint64_t TCPAbortOnMemory;
	__uint64_t TCPAbortOnTimeout;
	__uint64_t TCPAbortOnLinger;
	__uint64_t TCPAbortFailed;
	__uint64_t TCPMemoryPressures;
} nx_misc_mib_t;


typedef struct {
	nx_ip_mib_t	ip_statistics;
	nx_tcp_mib_t	tcp_statistics;
/*
	icmp_mib_t		icmp_stats[NR_PORTS];
*/
} netxen_port_stats_t;

typedef struct {
	netxen_port_stats_t	PREALIGN(512) port[NETXEN_NUM_PORTS]	\
						POSTALIGN(512);
} netxen_peg_stats_t;

typedef struct {
        netxen_peg_stats_t      peg[NETXEN_NUM_PEGS];
} netxen_pstats_t;

#endif /* _UNM_PSTATS_H_ */

